/** @file
 *  @details : Private header file for timer wheel implementation
 */

/*
 * Copyright (c) 2008 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __TIMER_WHEEL_PRIVATE_H
#define __TIMER_WHEEL_PRIVATE_H

#include <swi.h>
#include <timer_wheel.h>

#define DEFINE_TIMER_CONFIG(num_bits...) \
		 __PERCPU static uint32_t bits_each_level[] = {num_bits};

DEFINE_TIMER_CONFIG(15, 12, 4);	/*Sum of the args should be less than 31 */

#define MAX_WHEEL_LEVEL	(sizeof(bits_each_level)/sizeof(typeof(bits_each_level[0])))

#define CAS	compare_and_swap
#define MAX_NUM_LWE_INSTANCES	8

#define MAX_BITS_PERIOD_TICK	31	/*2G ticks */
#define MAX_PERIOD_TICK		(0x1 << MAX_BITS_PERIOD_TICK)

#define MAX_NUM_TIMER_OBJECTS	(0x1000000)	/*16M per core */
#define MAX_RECLAIM_QUEUE_LEN	(0x10000)	/*64K */

#define NUM_BITS_FOR_LWE_ID	8
#define NUM_BITS_FOR_TIMER_INDEX (32 - NUM_BITS_FOR_LWE_ID)
#define TIMER_INDEX_MASK	(0xFFFFFFFF >> NUM_BITS_FOR_LWE_ID)
#define TIMER_LWE_ID_MASK	(0xFFFFFFFF << NUM_BITS_FOR_TIMER_INDEX)

#define GET_LWE_ID_FROM_TIMER_ID(timer_id) (((timer_id) & TIMER_LWE_ID_MASK) >> NUM_BITS_FOR_TIMER_INDEX)
#define GET_TIMER_INDEX_FROM_TIMER_ID(timer_id) (((timer_id) & TIMER_INDEX_MASK))

#define TIMER_STATE_ID_MASK			(0x7)
#define TIMER_INFLIGHT_STATUS_MASK		(0x8)
#define TIMER_STATE_NUM_MASK			(0xFFFFFFF0)

#define GET_TIMER_STATE_ID(state)		((state) & TIMER_STATE_ID_MASK)
#define GET_TIMER_STATE_NUM(state) 		(((state) & TIMER_STATE_NUM_MASK) >> 4)
#define GET_TIMER_INFLIGHT_STATUS(state)	(((state) & TIMER_INFLIGHT_STATUS_MASK) >> 3)
#define GET_COMBINED_TIMER_STATE(id, inflight, num)	((id) | ((inflight) << 3) | ((num) << 4))

#define GET_TICK_FROM_MS(ms)	(((ms)/TICK_SIZE_MS) + ((ms % TICK_SIZE_MS)? 1:0))

enum timer_state_t {
	TIMER_STATE_FREE,
	TIMER_STATE_STOPPED,
	TIMER_STATE_SCHEDULED,
	TIMER_STATE_RESET_PENDING,
	TIMER_STATE_EXPIRED
};

/* The size of timer object is 36 bytes.
 * Good if we can reduce it to 32 bytes so that 2 timer objects may get aligned 
 * to boundary of 64 bytes (size of cache line in P4080)
 */
struct timer_object_t {
	node_t ll_header;

	union {
		uint32_t timer_id;
		struct {
			uint32_t lwe_id:8; /*LWE ID which owns this object*/
			uint32_t timer_index:24;
		};
	};

	union {
		timer_callbk_t callbk;
		struct qman_fq *expiry_fq;
	};

	void *p_data;	/*User specified cookie per timer object*/

	struct {
		uint32_t wheel_level:2;	/*Used for cascaded wheel level */
		uint32_t wheel_id:4;
		uint32_t swi_priority:3;
		uint32_t is_periodic:1;
	};

	volatile enum timer_state_t state;
	volatile uint32_t period_tick;
	volatile uint32_t expiry_tick;
};

typedef list_t bucket_array_t[1];	/*Any arbit dimension */
typedef struct timer_object_t timer_array_t[MAX_NUM_TIMER_OBJECTS];

struct reclaim_queue_t {
	uint32_t read_index;
	uint32_t write_index;
	struct timer_object_t *queue[];
};

struct wheel_level_t {
	uint32_t curr_bucket;
	uint32_t num_buckets;
	uint32_t mask;
	uint32_t shift;
	bucket_array_t *buckets;	/*Strechable array of buckets*/
};

struct timer_wheel_t {
	struct wheel_level_t wheel_level[MAX_WHEEL_LEVEL];
	timer_array_t *timer_objects;

	volatile struct timer_object_t *free_timer_db; /*Free timer pool*/
	struct reclaim_queue_t *reclaim_queues[MAX_NUM_LWE_INSTANCES];

	volatile uint32_t curr_tick;	/*Value of ticks elapsed since start*/
	uint32_t num_levels; 		/*Number of wheel levels*/
	uint32_t max_timeout_tick;	/*Maximum period supported in ticks*/
	uint32_t reclaim_queue_len;	/*Length of fast reclamation queue*/
};

struct timer_framework_t {
	struct timer_wheel_t wheel[MAX_NUM_LWE_INSTANCES];	/*There is one wheel per core */
};

int timer_framework_init(uint32_t max_timeout_ms,
			 uint32_t num_timer_objects,
			 uint32_t reclaim_queue_len);


#endif	/* __TIMER_WHEEL_PRIVATE_H */
