
/** @file barrier.h
 * This file contains APIs for implementing barriers.
 * Barrier is a synchronization construct, which causes cores to wait till
 * a set of cores have reached the barrier. The number of cores	 required
 * is set when the barrier is initialized.
 *
 * Barrier can be initalized or destroyed using barrier_init() and 
 * barrier_destroy()
 *
 * barrier_wait() allows cores to wait for barrier. 
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef BARRIER_H
#define BARRIER_H

#ifdef __cplusplus
    extern "C" {
#endif

/**Barrier Structure
 */
struct lwe_barrier {
	int status;
	volatile uint32_t mask;
	uint32_t threshold;
	volatile uint32_t cond_check;
	uint32_t lock;
};

enum BARRIER_STATUS {
	BARRIER_AVAILABLE = 0,
	BARRIER_VALID
};

/**
 * @brief- Initialize barrier
 * @details Initializes the barrier for the coremask provided.
 * 
 * @param[in] *barrier -- address of barrier shared by all cores
 * @param[in] uImask -- Bitmask of all the cores that should wait on the barrier
 * 			mask [0 - 31 bits] - LSB(Bit 32), MSB (Bit 0)
 *			Bit 31 - VCPU 0 ... Bit 30  VCPU 1 and so on
 *
 * @return - 0 on success -ve on failure
 * @error     -EINVAL if barrier is already in use 
 *
 * @note- Can be called from any or all of the lwe cores but only one core 
 * will get it initialized
*/
int barrier_init(struct lwe_barrier *barrier, uint32_t uImask);

/**
 * @brief- Synchronize at the barrier
 * @details - The barrier_sync() function shall synchronize participating cores
 * at the barrier refernced by barrier parameter passed.The calling core shall 
 * block till the required number of cores have called the barrier_sync() 
 * specifying the barrier. When the required number of cores have called 
 * the barrier_sync , 1 shall be returned to the last core entering 
 * barrier_sync and 0 to others. At this point, barrier shall be reset to the 
 * state it had as a result of barrier_init() function that referenced it.
 *
 * @param[in] *barrier -- address of barrier shared  by all cores
 *
 * @return - >=0 on success -ve on failure
 * @error	-EINVAL - The barrier passed is not an initialized barrier object.
 *		 or if the barrier_sync is called by core not initialized to do so.
 *
 * @note- The results are undefined if it is called form an uninitialzed barrier
*/
int barrier_sync(struct lwe_barrier *barrier);

/**
 * @brief- Destroy barrier
 * @details - barrier_destroy() shall destroy the barrier reference by barrier.
 * To use this barrier again it should be re-initialized.
 *
 * @param[in] *barrier -- address of barrier shared by all cores
 *
 * @return - 0 on success -ve on failure 
 *
 * @note- May be called from all the lwe cores but will be destroyed by only
 * 	  one core
*/
int barrier_destroy(struct lwe_barrier *barrier);

#ifdef __cplusplus
    }
#endif
#endif
