/**
\file
\brief	crc support
*/

/*
 * Copyright (c) 2008, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __CHECKSUMS_H
#define __CHECKSUMS_H

#ifdef __cplusplus
    extern "C" {
#endif

#include <stdint.h>

/**
\brief		Calculates the ones complement checksum over any length of
		input data. The calling function can initialize the sum to
		a non-zero value. This can be useful when the checksum is
		to be calculated over a pseudo-header as the case is
		for UDPoIP. 

\param[in] 	addr - Pointer to the start of data for which checksum
		is to be calculated

\param[in] 	len - Number of bytes to be read for checksum

\param[in] 	sum - Input checksum (0 or partial checksum)

\return	Ones complement checksum
*/
uint16_t checksum(uint8_t * addr, uint32_t len, unsigned long sum);

/**
\brief		Generates a CRC64 (ECMA) over the input data. This is the same
		CRC as the one that is used by the FMan. This implementation
		here is a byte-wise CRC, which is 'fast' for small data sizes.
		Given that the most likely use case for this test is key
		generation for LUTs, that assumption would be correct

\param[in]	ptr - Pointer to the start of the input data

\param[in]	length - Length of the input data

\return   	Generated CRC64-ECMA over the input data
*/
uint64_t crc64(uint8_t * bytePtr, uint32_t bytes);

/**
\brief		Calculates IEEE FCS / IEEE802.16E CRC (CRC32) over a given
		input. Uses a word-wise CRC for increasing performance.
		This CRC is also known as CRC32_bzip2.

\param[in]	bytePtr - Pointer to input data over which CRC is to be
		calculated

\param[in]	length - Length of input data

\return		Calculated IEEE FCS/ 802.16e (Mobile WiMAX) CRC
*/
uint32_t ieeeFCS(uint8_t * bytePtr, uint32_t bytes);

/**
\brief		Calculates the WiMAX/802.16d (OFDM) CRC over a given input.
		Uses a word-wise CRC for increasing performance. Current
		implementation is known incorrect
 
\param[in]	bytePtr - Pointer to input data over which CRC is to be
		calculated

\param[in]	length - Length of input data

\return   	Calculated 802.16d (Fixed WiMAX) CRC
*/
uint32_t ieee80216Dcrc(uint8_t * bytePtr, uint32_t bytes);

/**
\brief		Calculates the CRC16-IBM over a given input. Uses word-wise CRC for
		increasing performance.

\param[in] 	bytePtr - Pointer to the input data over which checksum is to be
		calculated

\param[in]	length - Length of the input data

\return   	CRC16-IBM
*/
uint16_t crc16IBM(uint8_t * bytePtr, uint32_t bytes);

/**
\brief		Calculates the CRC16-CCITT over a given input. Uses word-wise
		CRC for increased performance.

\param[in]	bytePtr - Pointer to the input data for which the checksum to
		be calculated

\param[in]	length - Length of the input data

\return   	CRC16-CCITT
*/
uint16_t crc16CCITT(uint8_t * bytePtr, uint32_t bytes);

/**
\brief		Calculates a CRC8 over a given input. The CRC is calculated
		byte-wise. This is not the fastest method if the input data
		size is very large, but efficient for small input sizes
		(which is more likely anyway)

\param[in]	bytePtr - Pointer to input

\param[in]	length - Length of input

\return   	CRC8
*/
uint8_t crc8(uint8_t * bytePtr, uint32_t bytes);

#ifdef __cplusplus
    }
#endif
#endif				/* __CHECKSUMS_H */
