/**
\file	cpu_usage.h
\brief	Core Usage calculator
	This is a header file. All the functions are declared inline
*/
/*
 * Copyright (C) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _CORE_USAGE_H_
#define _CORE_USAGE_H_

#include <timebase.h>
#include <io.h>
#include <init.h>
#include "misc.h"
#include <stdint.h>

#define SCALING_FACTOR (UINT64_MAX/(cpu_usage_data.integ_interval*4)+1)
#define SCALING_FACTOR_SHIFT (62 - cpu_usage_data.integ_interval_shift)

enum state_e {
	IDLE,
	BUSY
};

/* Store the per core data */
struct cpu_usage_data_t {
	uint64_t usage_value;		/* The cpu usage information */
	enum state_e state;		/* The state of the core */
	uint64_t prev;			/* Stores the value of timebase
					register upto which usage value
					has been calculated */
	uint64_t busy_start;		/* Stores the value of timebase
					register at which the app goes busy */
	uint32_t integ_interval;	/* Integration interval */
	uint8_t integ_interval_shift;	/* Integration Interval is in multiple
					of 2, so the shift factor for that */
	uint32_t threshold;
	uint32_t busy_accumulated;	/* The accumulated busy period */
	uint32_t busy_acc_end_time;	/* Time of last accumulation*/
	bool flag_idle;
};

__PERCPU extern volatile struct cpu_usage_data_t cpu_usage_data;

void init_cpu_usage(void);


/**
\brief:		It calculates the core usage in terms of percentage over an fixed integration period of time.
\param[in]:	void
\param[out]:	void
\return:	void
*/
static inline void calc_core_usage(void)
{
	uint64_t busy_duration;
	uint64_t idle_duration, discard_duration;
	uint64_t busy_start, prev, curr;
	uint64_t temp;
	uint64_t window_start;

	prev = cpu_usage_data.prev;
	busy_start = cpu_usage_data.busy_start;
	curr = get_timebase();

	if (BUSY == cpu_usage_data.state) {
		busy_duration = curr - busy_start;

		/* If core is busy for more than integration interval */
		if (unlikely(busy_duration >= cpu_usage_data.integ_interval)) {
			cpu_usage_data.usage_value = SCALING_FACTOR;
			goto down;
		}

		if (unlikely(curr - prev >= cpu_usage_data.integ_interval)) {
			if (unlikely(cpu_usage_data.busy_accumulated)) {
				window_start = curr -
					cpu_usage_data.integ_interval;
				busy_duration +=
					(cpu_usage_data.busy_accumulated*
					(cpu_usage_data.busy_acc_end_time -
					 window_start))/
					(cpu_usage_data.busy_acc_end_time -
					 prev);
			}
			cpu_usage_data.usage_value = (busy_duration <<
				SCALING_FACTOR_SHIFT) >>
				cpu_usage_data.integ_interval_shift;
			goto down;
		}

		busy_duration += cpu_usage_data.busy_accumulated;
	} else {		/* IDLE state case */
		if (unlikely(cpu_usage_data.busy_accumulated))
			idle_duration = curr - cpu_usage_data.busy_acc_end_time;
		else
			idle_duration = curr - prev;

		/* If core is idle for more than integration interval */
		if (unlikely(idle_duration >=
			cpu_usage_data.integ_interval)) {
			cpu_usage_data.usage_value = 0;
			goto down;
		}

		if (unlikely(curr - prev >= cpu_usage_data.integ_interval)) {
			window_start = curr - cpu_usage_data.integ_interval;
			busy_duration = (cpu_usage_data.busy_accumulated*
				(cpu_usage_data.busy_acc_end_time -
				 window_start))/
				(cpu_usage_data.busy_acc_end_time - prev);
			cpu_usage_data.usage_value = (busy_duration <<
				SCALING_FACTOR_SHIFT) >>
				cpu_usage_data.integ_interval_shift;
			goto down;
		}
		busy_duration = cpu_usage_data.busy_accumulated;
	}

	discard_duration = curr - prev;

	/*Now simply recalculate average*/
	temp = (busy_duration << SCALING_FACTOR_SHIFT) +
			cpu_usage_data.usage_value*
			(cpu_usage_data.integ_interval-discard_duration);
	cpu_usage_data.usage_value = temp >>
			cpu_usage_data.integ_interval_shift;

down:
	cpu_usage_data.prev = curr;
	cpu_usage_data.busy_start = curr;
	cpu_usage_data.busy_accumulated = 0;
}

/**
\brief:		This API is called by the LWE-APP when it goes idle from busy
\param[in]:	void
\param[out]:	void
\return:	void
*/
static inline void idle_start(void)
{
	uint64_t busy_duration;
	uint64_t busy_start, prev, curr;
	uint32_t discard_duration;
	uint64_t temp;
	uint64_t window_start;
	register_t saved;

	if (likely(IDLE == cpu_usage_data.state))
		return;

	/* Save the previous interrupt enabled/disabled state */
	saved = disable_critint_save();

	curr = get_timebase();

	cpu_usage_data.state = IDLE;

	prev = cpu_usage_data.prev;
	busy_start = cpu_usage_data.busy_start;

	busy_duration = curr - busy_start;

	if (curr - prev <= cpu_usage_data.threshold) {
		cpu_usage_data.busy_accumulated += busy_duration;
		cpu_usage_data.busy_acc_end_time = curr;
		restore_critint(saved);
		return;
	}

	/* If core was busy for more than integration interval */
	if (unlikely(busy_duration >= cpu_usage_data.integ_interval)) {
		cpu_usage_data.usage_value = SCALING_FACTOR;
	} else {
		/*Now simply recalculate average*/
		if (unlikely(curr - prev >= cpu_usage_data.integ_interval)) {
			if (cpu_usage_data.busy_accumulated) {
				window_start = curr -
				cpu_usage_data.integ_interval;
				busy_duration +=
					(cpu_usage_data.busy_accumulated*
					(cpu_usage_data.busy_acc_end_time -
					 window_start))/
					(cpu_usage_data.busy_acc_end_time -
					 prev);
			}
			cpu_usage_data.usage_value = (busy_duration <<
				SCALING_FACTOR_SHIFT) >>
				cpu_usage_data.integ_interval_shift;
		} else {
			busy_duration += cpu_usage_data.busy_accumulated;
			discard_duration = curr - prev;
			temp = (busy_duration << SCALING_FACTOR_SHIFT) +
					cpu_usage_data.usage_value*
					(cpu_usage_data.integ_interval-
					discard_duration);
			cpu_usage_data.usage_value = temp >>
				cpu_usage_data.integ_interval_shift;
		}
	}

	cpu_usage_data.prev = curr;
	cpu_usage_data.busy_accumulated = 0;

	/* restore the previous interrupt state */
	restore_critint(saved);
}

/**
\brief:		This API is called by the LWE-APP when it goes from
		idle to busy
\param[in]:	void
\param[out]:	void
\return:	void
*/
static inline void busy_start(void)
{
	/* The core is not idle */
	cpu_usage_data.flag_idle = false;

	if (unlikely(BUSY != cpu_usage_data.state)) {
		/* Update the idle-end time */
		cpu_usage_data.state = BUSY;
		cpu_usage_data.busy_start = get_timebase();
	}
}

/**
\brief:		When the core wants to get its usage percentage,
		it calls this API
\param[in]:	void
\param[out]:	void
\return:	The core usage value in percentage
*/
static inline uint8_t get_core_usage(void)
{
	/* Calculate the core usage and return the value */
	calc_core_usage();
	return 100*cpu_usage_data.usage_value/SCALING_FACTOR;
}

/**
\brief:		This funtion is called to assert that the core is idle
\param[in]:	void
\param[out]:	void
\return	:	void
*/
static inline void assert_idle(void)
{
	cpu_usage_data.flag_idle = true;
}

/**
\brief:		This funtion is called when the core is idle
\param[in]:	void
\param[out]:	void
\return	:	void
*/
static inline void set_idle(void)
{
	if (cpu_usage_data.flag_idle)
		idle_start();
}

#endif
