/**
\file           
\brief          DPA buffer allocator interface definitions
*/
/*
 * Copyright (C) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef __cplusplus
    extern "C" {
#endif
#include <stdint.h>
#include <compat.h>
#include <fsl_bman.h>

#define DPA_ALLOC_MAX_BUFF_SIZE		2048
#define MAX_NUM_BMAN_POOLS		64

struct dpa_buff_allocator;

/**
\brief          DPA allocator object creation function

\details        This function creates a core local DPA allocator object.
		Several BMAN pools can be registered in a DPA allocator.

\param[out]      ecode - Error code in case of failure

\retval         On success, dpa allocator object handle.
\retval		On failure, NULL.
*/
struct dpa_buff_allocator *dpa_allocator_create(int *ecode);

/**
\brief          DPA allocator object free function.

\details        Frees a previously allocated DPA allocator object.

\param[in]      allocator - Previously allocated DPA allocator object handle

\return         void
*/
void dpa_allocator_free(struct dpa_buff_allocator *allocator);

/**
\brief          BMAN bpool registration function

\details        Registers a BMAN pool with DPA allocator object.

\param[in]      allocator - Previously allocated DPA allocator object handle
\param[in] 	size - Size of buffers in bpool being registered
\param[in] 	bpid - BPID being registered
\param[in] 	bpool - Bpool handle

\retval         0 - Success
\retval		Negative - Failure
*/
int dpa_allocator_add_bpool(struct dpa_buff_allocator *allocator,
			    const uint32_t size,
			    const uint32_t bpid, struct bman_pool *bpool);

/**
\brief          BMAN bpool de-registration function

\details        De-registers a BMAN pool with DPA allocator object.

\param[in]      allocator - Previously allocated DPA allocator object handle
\param[in] 	bpid - BPID being de-registered

\retval         0 - Success
\retval		Negative - Failure 
*/
int dpa_allocator_del_bpool(struct dpa_buff_allocator *allocator,
			    const uint32_t bpid);

/**
\brief          Buffer allocation function

\details        Allocates a buffer from set of bpools registered with allocator
		The size returned may be greater than that requested

\param[in]      allocator - Previously allocated DPA allocator object handle
\param[in] 	size - Size of buffer required
\param[out]	buf - Allocated buffer

\retval         Non-zero-positive - Size of actual buffer allocated
\retval		Negative - Failure
*/

int dpa_allocator_get_buff(const struct dpa_buff_allocator *allocator,
			   const uint32_t size, struct bm_buffer *buf);

/**
\brief          Buffer freeing function

\details        Frees a BMAN buffer back to allocator.

\param[in]      allocator - Previously allocated DPA allocator object handle
\param[in]	buf - Previously allocated buffer

\retval          0 - SUCCESS
\retval		Negative - Failure
*/
int dpa_allocator_free_buff(const struct dpa_buff_allocator *allocator,
			    const struct bm_buffer *buf);
#ifdef __cplusplus
    }
#endif
