/** @file
 * This file provides APIs for HASH calculation on the basis of ORD32 and 
 * Static calculation algorithms.
 */

/*
 * Copyright (c) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __HASH_H
#define __HASH_H

#ifdef __cplusplus
    extern "C" {
#endif
#include <stdint.h>
#include <stdbool.h>

#undef Ord32

#undef statichash32

#undef statichash16

#define Ord32

/*
 * Here are some functions for 32-bit shifts, to make the code more
 * readable.
 */

/** Shift left 'x' by 'i' bits
*/
#define SHIFT_LEFT32(x, i) ((x)<<(i))	/* Left shift 'x' by 'i' bits */

/** Shift right 'x' by 'i' bits */
#define SHIFT_RIGHT32(x, i) ((x)>>(i))	/* Right shift 'x' by 'i' bits */

/** Rotate left 'x' by 'i' bits*/
#define ROTATE_LEFT32(x, i) (SHIFT_LEFT32(x, i) | SHIFT_RIGHT32(x, 32-i))

/**
 * Ord32 based HASH function: Given 3 words, this function returns the 32-bit hash
 * @param[in] a -- First 32bit word
 * @param[in] b -- Second 32bit word
 * @param[in] c -- Third 32bit word
 * @param[in] initval -- HASH initialization value
 * @return Ord32 based 32bit HASH
 */
static inline uint32_t
Ord32_3words(uint32_t a, uint32_t b, uint32_t c, uint32_t initval)
{
	uint32_t key = initval;
	/*
	 * Process 32 bits at a time, ROUND1
	 * First, do a combine step that combines the new
	 * input with our existing state.
	 */
	key = key ^ a;

	/*
	 * First, we perform a series of reversible
	 *left-shift-or-xor operations.
	 */
	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));

	/*
	 * Next, we perform a series of reversible
	 * right-shift-or-xor operations.
	 */
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));

	/*
	 * Next, we rotate the key, in order to ensure that high
	 * bits and low bits are mixed well.
	 */
	key = ROTATE_LEFT32(key, 8);

	/*
	 * Process 32 bits at a time, ROUND2
	 * First, do a combine step that combines the new
	 * input with our existing state.
	 */
	key = key ^ b;
	/*
	 * First, we perform a series of reversible
	 * left-shift-or-xor operations.
	 */
	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	/*
	 * Next, we perform a series of reversible
	 * right-shift-or-xor operations.
	 */
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	/*
	 * Next, we rotate the key, in order to ensure that high
	 * bits and low bits are mixed well.
	 */
	key = ROTATE_LEFT32(key, 8);

	/*
	 * Process 32 bits at a time, ROUND3
	 * First, do a combine step that combines the new
	 * input with our existing state.
	 */
	key = key ^ c;
	/*
	 * First, we perform a series of reversible
	 * left-shift-or-xor operations.
	 */
	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	/*
	 * Next, we perform a series of reversible
	 * right-shift-or-xor operations.
	 */
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	/*
	 * Next, we rotate the key, in order to ensure that high
	 * bits and low bits are mixed well.
	 */
	key = ROTATE_LEFT32(key, 8);

	/*
	 * Now we perform 3 sets of post-mixes.  These are the same as
	 * the above.
	 */
	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	key = ROTATE_LEFT32(key, 8);

	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	key = key ^ SHIFT_RIGHT32(key, 20)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	key = ROTATE_LEFT32(key, 8);

	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	key = ROTATE_LEFT32(key, 8);

	return key;
}

/**
 * Ord32 based HASH function: Given 2 words, returns the 32-bit hash
 * @param[in] a -- First 32bit word
 * @param[in] b -- Second 32bit word
 * @param[in] initval -- HASH initialization value
 * @return Ord32 based 32bit HASH
 */
static inline uint32_t
Ord32_2words(uint32_t a, uint32_t b, uint32_t initval)
{
	uint32_t key = initval;
	/*
	 * Process 32 bits at a time, ROUND1
	 * First, do a combine step that combines the new
	 * input with our existing state.
	 */
	key = key ^ a;
	/*
	 * First, we perform a series of reversible
	 * left-shift-or-xor operations.
	 */
	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	/*
	 * Next, we perform a series of reversible
	 * right-shift-or-xor operations.
	 */
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	/*
	 * Next, we rotate the key, in order to ensure that high
	 * bits and low bits are mixed well.
	 */
	key = ROTATE_LEFT32(key, 8);

	/*
	 * Process 32 bits at a time, ROUND2
	 * First, do a combine step that combines the new
	 * input with our existing state.
	 */
	key = key ^ b;
	/*
	 * First, we perform a series of reversible
	 * left-shift-or-xor operations.
	 */
	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	/*
	 * Next, we perform a series of reversible
	 * right-shift-or-xor operations.
	 */
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	/*
	 * Next, we rotate the key, in order to ensure that high
	 * bits and low bits are mixed well.
	 */
	key = ROTATE_LEFT32(key, 8);

	/*
	 * Now we perform 3 sets of post-mixes.  These are the same as
	 * the above.
	 */
	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	key = ROTATE_LEFT32(key, 8);

	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	key = key ^ SHIFT_RIGHT32(key, 20)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	key = ROTATE_LEFT32(key, 8);

	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	key = ROTATE_LEFT32(key, 8);

	return key;
}

/**
 * Ord32 based HASH function: Given 1 word, return the 32-bit hash
 * @param[in] a -- 32bit word
 * @param[in] initval -- HASH initialization value
 * @return Ord32 based 32bit HASH
 */
static inline uint32_t
Ord32_1word(uint32_t a, uint32_t initval)
{
	uint32_t key = initval;
	/*
	 * Process 32 bits at a time, ROUND1
	 * First, do a combine step that combines the new
	 * input with our existing state.
	 */
	key = key ^ a;
	/*
	 * First, we perform a series of reversible
	 * left-shift-or-xor operations.
	 */
	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	/*
	 * Next, we perform a series of reversible
	 * right-shift-or-xor operations.
	 */
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	/*
	 * Next, we rotate the key, in order to ensure that high
	 * bits and low bits are mixed well.
	 */
	key = ROTATE_LEFT32(key, 8);

	/* Now we perform 3 sets of post-mixes.  These are the
	   same as the above. */
	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	key = ROTATE_LEFT32(key, 8);

	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	key = key ^ SHIFT_RIGHT32(key, 20)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	key = ROTATE_LEFT32(key, 8);

	key = key ^ SHIFT_LEFT32(key, 6)
	    ^ (SHIFT_LEFT32(key, 5) | SHIFT_LEFT32(key, 5 + 6));
	key = key ^ SHIFT_RIGHT32(key, 10)
	    ^ (SHIFT_RIGHT32(key, 1) | SHIFT_RIGHT32(key, 1 + 10));
	key = ROTATE_LEFT32(key, 8);

	return key;
}

/**
 * Static HASH calculation function: Given 3 words, return the 32-bit HASH
 * @param[in] a -- First 32bit word
 * @param[in] b -- Second 32bit word
 * @param[in] c -- Third 32bit word
 * @return 32bit static HASH
 */
static inline uint32_t
statichash32_3words(uint32_t a, uint32_t b, uint32_t c)
{
	/*
	 * Initialize 2 keys to 0. Either one by itself will yield a 32 bit
	 * result, of which 16 bits are useful. At the end of the code, throw
	 * the 2 x 16 bit together to form a single 32 bit key that is returned.
	 */
	uint32_t key[2] = { 0, 0 };

	key[0] = (key[0] << 8) | ((key[0] >> 24) & 0xFF);	/* Rotate */
	key[0] ^= a;
	key[0] = (key[0] << 8) | ((key[0] >> 24) & 0xFF);	/* Rotate */
	key[0] ^= b;
	key[0] = (key[0] << 8) | ((key[0] >> 24) & 0xFF);	/* Rotate */
	key[0] ^= c;

	key[0] *= 0x9CCF9319;	/* discard overflow */

	key[1] = (key[1] << 8) | ((key[1] >> 24) & 0xFF);	/* Rotate */
	key[1] ^= a;
	key[1] = (key[1] << 8) | ((key[1] >> 24) & 0xFF);	/* Rotate */
	key[1] ^= b;
	key[1] = (key[1] << 8) | ((key[1] >> 24) & 0xFF);	/* Rotate */
	key[1] ^= c;

	key[1] *= 0x10DCD;	/* discard overflow */

	/* Concatenate 2 x 16 bit halfs */
	return ((key[0] >> 16) & 0xFFFF) | ((key[1]) & 0xFFFF0000);
}

/**
 * Static HASH calculation function: Given 2 words, return the 32-bit HASH
 * @param[in] a -- First 32bit word
 * @param[in] b -- Second 32bit word
 * @return 32bit static HASH
 */
static inline uint32_t
statichash32_2words(uint32_t a, uint32_t b)
{
	uint32_t key[2] = { 0, 0 };

	key[0] = (key[0] << 8) | ((key[0] >> 24) & 0xFF);	/* Rotate */
	key[0] ^= a;
	key[0] = (key[0] << 8) | ((key[0] >> 24) & 0xFF);	/* Rotate */
	key[0] ^= b;

	key[0] *= 0x9CCF9319;	/* discard overflow */

	key[1] = (key[1] << 8) | ((key[1] >> 24) & 0xFF);	/* Rotate */
	key[1] ^= a;
	key[1] = (key[1] << 8) | ((key[1] >> 24) & 0xFF);	/* Rotate */
	key[1] ^= b;

	key[1] *= 0x10DCD;	/* discard overflow */

	return ((key[0] >> 16) & 0xFFFF) | ((key[1]) & 0xFFFF0000);
}

/**
 * Static HASH calculation function: Given 1 word, return the 32-bit HASH
 * @param[in] a -- First 32bit word
 * @return 32bit static HASH
 */
static inline uint32_t
statichash32_1word(uint32_t a)
{
	uint32_t key[2] = { 0, 0 };

	key[0] = (key[0] << 8) | ((key[0] >> 24) & 0xFF);	/* Rotate */
	key[0] ^= a;

	key[0] *= 0x9CCF9319;	/* discard overflow */

	key[1] = (key[1] << 8) | ((key[1] >> 24) & 0xFF);	/* Rotate */
	key[1] ^= a;

	key[1] *= 0x10DCD;	/* discard overflow */

	return ((key[0] >> 16) & 0xFFFF) | ((key[1]) & 0xFFFF0000);
}

/**
 * Static HASH calculation function: Given 3 words, return the 16-bit HASH
 * @param[in] a -- First 32bit word
 * @param[in] b -- Second 32bit word
 * @param[in] c -- Third 32bit word
 * @return 16bit Static HASH
 */
static inline uint32_t
statichash16_3words(uint32_t a, uint32_t b, uint32_t c)
{
	uint32_t key = 0;

	key = (key << 8) | ((key >> 24) & 0xFF);	/* Rotate */
	key ^= a;
	key = (key << 8) | ((key >> 24) & 0xFF);	/* Rotate */
	key ^= b;
	key = (key << 8) | ((key >> 24) & 0xFF);	/* Rotate */
	key ^= c;

	key *= 0x9CCF9319;	/* discard overflow */

	return ((key >> 16) & 0xFFFF);
}

/**
 * Static HASH calculation function: Given 2 words, return the 16-bit HASH
 * @param[in] a -- First 32bit word
 * @param[in] b -- Second 32bit word
 * @return 16bit Static HASH
 */
static inline uint32_t
statichash16_2words(uint32_t a, uint32_t b)
{
	uint32_t key = 0;;

	key = (key << 8) | ((key >> 24) & 0xFF);	/* Rotate */
	key ^= a;
	key = (key << 8) | ((key >> 24) & 0xFF);	/* Rotate */
	key ^= b;

	key *= 0x9CCF9319;	/* discard overflow */

	return ((key >> 16) & 0xFFFF);
}

/**
 * Static HASH calculation function: Given 1 words, return the 16-bit HASH
 * @param[in] a -- First 32bit word
 * @return 16bit static HASH
 */
static inline uint32_t
statichash16_1word(uint32_t a)
{
	uint32_t key = 0;;

	key = (key << 8) | ((key >> 24) & 0xFF);	/* Rotate */
	key ^= a;

	key *= 0x9CCF9319;	/* discard overflow */

	return ((key >> 16) & 0xFFFF);
}

static inline uint32_t
hash_3words(uint32_t a, uint32_t b, uint32_t c, uint32_t initval)
{
#ifdef Ord32
	return Ord32_3words(a, b, c, initval);
#endif				/* Ord32 */

#ifdef statichash32
	/* Save a warning */
	initval++;
	return statichash32_3words(a, b, c);
#endif				/* statichash32 */

#ifdef statichash16
	/* Save a warning */
	initval++;
	return statichash16_3words(a, b, c);
#endif				/* statichash16 */
}

static inline uint32_t
hash_2words(uint32_t a, uint32_t b, uint32_t initval)
{
#ifdef Ord32
	return Ord32_2words(a, b, initval);
#endif				/* Ord32 */

#ifdef statichash32
	/* Save a warning */
	initval++;
	return statichash32_2words(a, b);
#endif				/* statichash32 */

#ifdef statichash16
	/* Save a warning */
	initval++;
	return statichash16_2words(a, b);
#endif				/* statichash16 */
}

static inline uint32_t
hash_1word(uint32_t a, uint32_t initval)
{
#ifdef Ord32
	return Ord32_1word(a, initval);
#endif				/* Ord32 */

#ifdef statichash32
	/* Save a warning */
	initval++;
	return statichash32_1word(a);
#endif				/* statichash32 */

#ifdef statichash16
	/* Save a warning */
	initval++;
	return statichash16_1word(a);
#endif				/* statichash16 */
}

#ifdef __cplusplus
    }
#endif
#endif				/* ifndef __HASH_H */
