/**
\file
\brief	Hardware Interrupts setup and control.
	This file provides apis for management of
	external input interrupts presented to interrupt mechanism 
	when MSR[EE] = 1.
	
	register_hwi() registers handlers to be called when a particular
	interrupt occurs.  	

	The enable_hwi() and disable_hwi()
	functions can be used to enable or disable these interrupts.
*/

/*
 * Copyright (c) 2008 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __HWI_EXT_H
#define __HWI_EXT_H

#ifdef __cplusplus
    extern "C" {
#endif
#include <stdint.h>
#include <init.h>

extern volatile __PERCPU int in_irq;

typedef uint32_t(*hwi_func) (int hwi_handle, void *);

/**
\brief		Register a handler for given interrupt number

\param[in]	hwi_handle - Interrupt number (from the guest device tree)
 
\param[in]	handler - Handler function to be called when interrupt is generated
 
\param[in]	arg - Argument '2' of handler function

\param[in]	priority - Specifies the interrupt priority. If it is defined
		as 0, priority is set to 15. Valid values (1 - 15)

\return -	0 on success, otherwise -ve value 
\note 		This api doesnot enable the interrupt. User should explicitly use
		enable_hwi to enable this interrupt.	
		Use of printf statement form the registered handler code is not
		supported
*/
int register_hwi(uint32_t hwi_handle, hwi_func handler, void *p,
							uint32_t priority);

/**
\brief		Enable a given interrupt

\param[in]	hwi_handle - Interrupt number (from the guest device tree)

\return		0 on success, otherwise -ve value
*/
int enable_hwi(uint32_t hwi_handle);

/**
\brief		Disable a given interrupt

\param[in]	hwi_handle - Interrupt number (from the guest device tree)

\return		0 on success, otherwise -ve value
*/
int disable_hwi(uint32_t hwi_handle);

/** 
\brief		Free handler attached with given interrupt number

\param[in]	hwi_handle - Interrupt number (from the guest device tree)

\return      0 on success, otherwise -ve value
*/
int free_hwi(uint32_t hwi_handle);

#ifdef __cplusplus
    }
#endif
#endif
