/** @file
 * This file contains functions to initialize and manage a doubly linked list.
 */

/*
 * Copyright (c) 2008 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __LISTMGR_H
#define	__LISTMGR_H

#ifdef __cplusplus
    extern "C" {
#endif
/*
* If the tail node points back to the head node, we don't have any data
 * nodes in the list. Not worthwhile to implement this as a function.
 */
/**
* Check if list is currently empty
*/

#define list_is_empty(ml) ((ml)->mlh_tailpred == (struct node *)ml)

/**
 * Initializes list header before very first use
 * @param[in] ml -- Pointer to the linked list to be created
 */
static inline void
newlist(struct list *ml)
{
	/*The list header really consists of two nodes,
	 *a head node and a tail node. These are identified
	 *by one NULL link in the node.
	 *Because they both have a NULL link, we can overlap
	 *the NULL pointers and end up with an overlapping
	 *head and tail node. Any data nodes are added in
	 *between them, forming an effective ring of data nodes
	 *linked together by the list header.
	 *Initially though, we don't have any data nodes, so
	 *the head points to the tail, and the tail points to
	 *the head.
	 */
	ml->mlh_head = (struct node *) &ml->mlh_tail;
	ml->mlh_tail = NULL;
	ml->mlh_tailpred = (struct node *) &ml->mlh_head;

}				/* newlist */

/** Adds a node to the head of list
 * @param[in] ml -- Pointer to the linked list
 * @param[in] mn -- Node to be added to the head
 */
static inline void
add_head(struct list *ml, struct node *mn)
{
	/* Note the order of modifying the fields so that concurrent
	 * list traversal of another task will not be disturbed
	 * This works both for forward and backward traversal
	 */
	mn->mln_succ = ml->mlh_head;
	mn->mln_pred = (struct node *) &ml->mlh_head;

	mn->mln_succ->mln_pred = mn;
	mn->mln_pred->mln_succ = mn;

}

/** Adds a node to the tail of list
 * @param[in] ml -- Pointer to the linked list
 * @param[in] mn -- Node to be added to the tail
 */
static inline void
add_tail(struct list *ml, struct node *mn)
{
	/* Note the order of modifying the fields so that concurrent
	 * list traversal of another task will not be disturbed
	 * This works both for forward and backward traversal
	 */
	mn->mln_succ = (struct node *) &ml->mlh_tail;
	mn->mln_pred = ml->mlh_tailpred;

	mn->mln_succ->mln_pred = mn;
	mn->mln_pred->mln_succ = mn;

}

/** Removes a node from the linked list
 * @param[in] mn -- Node to be removed
 */
static inline void
remove(struct node *mn)
{
	/* Unlink the node in both directions */
	mn->mln_succ->mln_pred = mn->mln_pred;
	mn->mln_pred->mln_succ = mn->mln_succ;

}

/** Adds a node to the linked list at specific location
 * @param[in] mn -- Node to be added
 * @param[in] mp -- Predecessor to the node to be added
 * @param[in] ms -- Successor to the node to be added
 */
static inline void
insert(struct node *mn, struct node *mp, struct node *ms)
{
	mn->mln_pred = mp;
	mn->mln_succ = ms;

	mp->mln_succ = mn;
	ms->mln_pred = mn;
}

/** Removes the head of linked list
 * @param[in] ml -- Pointer to the linked list
 */
static inline struct node *
rem_head(struct list *ml)
{
	struct node *mn;

	mn = ml->mlh_head;
	if (mn != (struct node *) &ml->mlh_tail)
		remove(mn);
	else
		mn = NULL;

	return (mn);

}

/** Removes the tail of linked list
 * @param[in] ml -- Pointer to the linked list
 */
static inline struct node *
rem_tail(struct list *ml)
{
	struct node *mn;

	mn = ml->mlh_tailpred;
	if (mn != (struct node *) &ml->mlh_head)
		remove(mn);
	else
		mn = NULL;

	return (mn);

}

/** Appends the complete content of the second list to the first
 * list and empties out the second list
 * @param[in] ml -- Pointer to the first linked list
 * @param[in] aml -- Pointer to the list to be appended behind first
 */
static inline void
append_list(struct list *ml, struct list *aml)
{
	/* Rather than looping with remove/add to move the
	 * entries, we chain the lists directly.
	 */
	if (!list_is_empty(aml)) {
		ml->mlh_tailpred->mln_succ = aml->mlh_head;
		aml->mlh_head->mln_pred = ml->mlh_tailpred;
		ml->mlh_tailpred = aml->mlh_tailpred;
		aml->mlh_tailpred->mln_succ = (struct node *) &ml->mlh_tail;

		/* We moved all the entries, empty out the 2nd list now */
		newlist(aml);
	}

}

#ifdef __cplusplus
    }
#endif
#endif
