/** @file
 * LWE logging APIs
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __LOG_H__
#define __LOG_H__

#ifdef __cplusplus
    extern "C" {
#endif
#include <types.h>
#include <fsl_hcalls.h>
#include <stdio.h>
#include <conf.h>
#include <init.h>
#include <hv_misc.h>
#include <stdarg.h>

#define LOG_BUF_SIZE 16384

enum log_type_t {
#define DEFINE_LOG_TYPE(type)	LOG_T_##type,
#include <log_def.h>
#undef DEFINE_LOG_TYPE

	LOG_T_LWE_END,
	LOG_T_APP_START = 128, /*Apps defined logs can go after this*/
	NUM_LOG_TYPES = 256
};


enum log_level_t {
	LWE_LOG_LEVEL_OFF,
	LWE_LOG_LEVEL_CRIT,
	LWE_LOG_LEVEL_ERROR,
	LWE_LOG_LEVEL_WARN,
	LWE_LOG_LEVEL_INFO,
	LWE_LOG_LEVEL_DEBUG,
	NUM_LOG_LEVELS
};


enum log_dest_t {
	LOG_DEST_INVALID,
	LOG_DEST_BUFF,		/** Dump logs to log-buffer*/
	LOG_DEST_CONSOLE,	/** Dump logs to console*/
	LOG_DEST_NETWORK,	/** Currently unimplemented*/
	NUM_LOG_DEST
};

extern __PERCPU enum log_level_t enabled_log_level[NUM_LOG_TYPES];
extern __PERCPU enum log_dest_t log_destination;

/* These isr_xxx apis can be used to print from context of an ISR */
int isr_vsnprintf(char *buf, size_t size, const char *str, va_list args);
int isr_printf(const char *str, ...) __attribute__ ((format(printf, 1, 2)));

int isr_sprintf(char *buf, const char *str, ...)
		__attribute__ ((format(printf, 2, 3)));

int isr_snprintf(char *buf, size_t size, const char *str, ...)
	 __attribute__ ((format(printf, 3, 4)));

void printlog(enum log_level_t level, enum log_type_t type,
	const char *format, ...) __attribute__ ((format(printf, 3, 4)));

void print_log_buf(void);
int logger_init(void);

/** Application callable API to set enabled log level*/
int set_log_level(enum log_type_t type, enum log_level_t level);

/** Application callable API to register a log_type*/
int register_log_type(enum log_type_t type, char *name);

/** Application callable API to set log destination*/
static inline int set_log_dest(enum log_dest_t dest)
{
	log_destination = dest;
	return 0;
}

/**Critical error logs cannot be disabled at compile time*/
#define LWE_CRIT(fmt, args...) do { \
	if (unlikely(enabled_log_level[LOG_T_UNSPEC] >= LWE_LOG_LEVEL_CRIT)) \
		printlog(LWE_LOG_LEVEL_CRIT, (enum log_type_t)0, fmt, \
	##args); } while (0)

#if (LWE_LOG_LEVEL >= LWE_LOG_LEVEL_ERROR)
#define LWE_ERROR(fmt, args...)	do { \
	if (unlikely(enabled_log_level[LOG_T_UNSPEC] >= LWE_LOG_LEVEL_ERROR)) \
		printlog(LWE_LOG_LEVEL_ERROR, (enum log_type_t)0, fmt, \
	##args); } while (0)

#else
#define LWE_ERROR(fmt, args...)
#endif

#if (LWE_LOG_LEVEL >= LWE_LOG_LEVEL_WARN)
#define LWE_WARN(fmt, args...) do { \
	if (unlikely(enabled_log_level[LOG_T_UNSPEC] >= LWE_LOG_LEVEL_WARN)) \
		printlog(LWE_LOG_LEVEL_WARN, (enum log_type_t)0, fmt, \
	##args); } while (0)
#else
#define LWE_WARN(fmt, args...)
#endif

#if (LWE_LOG_LEVEL >= LWE_LOG_LEVEL_INFO)
#define LWE_INFO(fmt, args...) do { \
	if (unlikely(enabled_log_level[LOG_T_UNSPEC] >= LWE_LOG_LEVEL_INFO)) \
		printlog(LWE_LOG_LEVEL_INFO, (enum log_type_t)0, fmt, \
	##args); } while (0)
#else
#define LWE_INFO(fmt, args...)
#endif

#if (LWE_LOG_LEVEL >= LWE_LOG_LEVEL_DEBUG)
#define LWE_DEBUG(fmt, args...) do { \
	if (unlikely(enabled_log_level[LOG_T_UNSPEC] >= LWE_LOG_LEVEL_DEBUG)) \
		printlog(LWE_LOG_LEVEL_DEBUG, (enum log_type_t)0, fmt, \
	##args); } while (0)
#else
#define LWE_DEBUG(fmt, args...)
#endif


/**Used to signal a catastrophic condition after which it not
 possible to proceed*/
#define LWE_PANIC(fmt, args...) \
do { \
	printlog(LWE_LOG_LEVEL_CRIT, (enum log_type_t)0, \
	"PANIC: "fmt, ##args); \
	while (1); \
} while(0)


#ifdef __cplusplus
    }
#endif
#endif /*__LOG_H__*/
