/** @file
 * Atomic functions 
 * This file provided APIs to atomically read, write,
 * increment, decrememt
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef HEADER_LWE_ATOMIC_H
#define HEADER_LWE_ATOMIC_H

#ifdef __cplusplus
    extern "C" {
#endif

/* Atomic stuff */
#include <bitops.h>

/** 
 * @brief	Atomically read from the memory location and 
 * 		return the read value
 *
 * @param[in] ptr -- Pointer to the memory location to read
 *
 * @return Returns value in memory location
 */
static inline int
atomic_read(unsigned long *ptr)
{
	int ret;
	asm volatile ("lwz%U1%X1 %0,%1":"=r" (ret):"m"(*ptr));

	return ret;
}

/** 
 * @brief	Update the memory location with the passed value atomically
 *
 * @param[in] ptr -- Pointer to the target memory location
 * @param[in] i -- 32bit value to be written in *ptr
 *
 * @return None
 */
static inline void
atomic_set(unsigned long *ptr, int i)
{
	asm volatile ("stw%U0%X0 %1,%0":"=m" (*ptr):"r"(i):"memory");
}

/**
 * @brief	Increments the value in the specifed memory location
 * 
 * @param[in] ptr -- Pointer to the target memory location
 *
 * @return None
 */
static inline void
atomic_inc(unsigned long *ptr)
{
	atomic_add(ptr, 1);
}

/**
 * @brief	Decrements the value in the specified memory location
 *
 * @param[in] ptr -- Pointer to the target memory location
 *
 * @return None
 */
static inline void
atomic_dec(unsigned long *ptr)
{
	atomic_add(ptr, -1);
}

/**
 * @brief 	Atomically decrements the value and tests it against zero. If value
 *		after decrement is 0, returns > 0 else returns <= 0
 *
 * @param[in] ptr -- Pointer to the target memory location
 *
 * @return +ve on SUCCESS -ve on failure
 *
 */
static inline int
atomic_dec_and_test(unsigned long *ptr)
{
	return (atomic_add(ptr, -1)) == 0;
}
#ifdef __cplusplus
    }
#endif

#endif /* !HEADER_LWE_ATOMIC_H */
