/** @file
 * memory allocation/free functions 
 * This file provided APIs for memory_allocation free from
 * private heap of each core and statistics region.
 *
 * It also priovides memalign API not available in LIBC library.
 */
/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef ALLOC_H
#define ALLOC_H
#ifdef __cplusplus
    extern "C" {
#endif
#include "malloc.h"
#include "bitops.h"

extern mspace priv_mspace;
extern mspace stats_mspace;
extern mspace sys_mspace;
extern uint32_t stats_memlock;

/**
\brief         	Allocates memory from Private Heap Area of each core

\details        Allocates size bytes from private heap area of the core
		and returns a pointer to the allocated memory.
		The memory is not cleared.

\param[in]      size : size of the memory to be allocated

\return         pointer to the allocated memory, on SUCCESS 
		or NULL on failure.
*/
static inline void *
tlmalloc(size_t size)
{
	return mspace_malloc(priv_mspace, size);
}

/**
\brief         	Allocates memory aligned to the required alignment from 
		private heap area of each core

\details        Allocates size bytes from private heap area of the core
		and returns a pointer to the allocated memory.
		The address of the allocated memory will be a multiple of align 
		which must be a power of 2. The memory is not cleared.

\param[in]      align : alignment required
\param[in]      size  : size of the memory to be allocated

\return         pointer to the allocated memory, on SUCCESS 
		or NULL on failure.
*/
static inline void *
tlmemalign(size_t align, size_t size)
{
	return mspace_memalign(priv_mspace, align, size);
}

/**
\brief         	Free memory space from Private Heap Area of each core

\details        Free the memory space pointed by ptr which must have been
		 returned by tlmalloc or tlmemalign

\param[in]      ptr : Pointer to the memory space to be freed

\return		None      
*/
static inline void
tlfree(void *ptr)
{
	mspace_free(priv_mspace, ptr);
}

/**
\brief         	Allocates memory from area reserved for statistics
		(Memory area marked wih Gaurded bit)

\details        Allocates size bytes from statistics area
		and returns a pointer to the allocated memory.
		The memory is not cleared.

\param[in]      size : size of the memory to be allocated

\return         pointer to the allocated memory, on SUCCESS 
		or NULL on failure.
*/
__attribute__ ((malloc))
static inline void *
stats_malloc(size_t size)
{
	void *mem;
	register_t lock_ret;
	lock_ret = spin_lock_intsave(&stats_memlock);
	mem = mspace_malloc(stats_mspace, size);
	spin_unlock_intsave(&stats_memlock, lock_ret);
	return mem;
}

/**
\brief         	Allocates memory aligned to the required alignment from 
		area reserved for statistics (marked with Gaurded bit)	

\details        Allocates size bytes from area reserved for statistics
		and returns a pointer to the allocated memory.
		The address of the allocated memory will be a multiple of align 
		which must be a power of 2. The memory is not cleared.

\param[in]      align : alignment required
\param[in]      size  : size of the memory to be allocated

\return         pointer to the allocated memory, on SUCCESS 
		or NULL on failure.
*/
static inline void *
stats_memalign(size_t align, size_t size)
{
	void *mem;
	register_t lock_ret;
	lock_ret = spin_lock_intsave(&stats_memlock);
	mem = mspace_memalign(stats_mspace, align, size);
	spin_unlock_intsave(&stats_memlock, lock_ret);
	return mem;
}

/**
\brief         	Free memory space from statistics Area

\details        Free the memory space pointed by ptr which must have been
		returned by stats_malloc or stats_memalign

\param[in]      ptr : Pointer to the memory space to be freed

\return		None      
*/
static inline void
stats_free(void *ptr)
{
	register_t lock_ret;
	lock_ret = spin_lock_intsave(&stats_memlock);
	mspace_free(stats_mspace, ptr);
	spin_unlock_intsave(&stats_memlock, lock_ret);
}

/**
\brief         	Allocates memory aligned to the required alignment from 
		Global heap area.

\details        Allocates size bytes from Global heap area and returns a 
		pointer to the allocated memory. The address of the
		allocated memory will be a multiple of align which
		must be a power of 2. The memory is not cleared.

\param[in]      align : alignment required
\param[in]      size  : size of the memory to be allocated

\return         pointer to the allocated memory, on SUCCESS 
		or NULL on failure.
*/
static inline void *
memalign(size_t align, size_t size)
{
	return mspace_memalign(sys_mspace, align, size);
}

#ifdef __cplusplus
    }
#endif
#endif
