/** @file
 * Contains APIs related to LWE memory map.
 * This file provides APIs to map/unmap area using TLB0 with appropriate
 * permissiions and flags set.
 *  
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __MMAP_H__
#define __MMAP_H__

#ifdef __cplusplus
    extern "C" {
#endif
#include <stdint.h>
#include <types.h>

/** 
\brief		Maps the size bytes(maximum of 2MB) starting from physical
		address passed with the required attributes at location
		which is returned back in ea.  Mapping is created in TLB0. 

\param[in] 	pa - Physical address for which the mapping has to be created
		The address passed should be 4K aligned.
\param[in]	size - Size for which mapping is to be created. 
		This should be 4K aligned. maximum size allowed is 2MB.
\param[in] 	align - Virtual address retuend shpuld be mapped to this 
		alignment.Passed in terms of mumber of bytes. The 
		minimum alignment gauranteed is 4K.				

\param[in] 	mas2flags - must be a combination of MAS2[WIMGE]
\param[in] 	mas3flags - must be a combination of MAS3[UX,SX,UW,SW,UR,SR]
\param[out]     ea - Effective address to which the physical address passed 
		is mapped.The address returned would be aligned to 4K or 
		alignment passed if the alignment id greater than 4K.

\return		0 on success and -ve on failure

\note 		TLB0 is used for mapping the physical address passe. Maximum of
		2MB mapping is supported using this function to avoid any 
		TLB Misses.	
*/
int dev_mmap(phys_addr_t pa, size_t size, size_t align, register_t mas2flags,
	     register_t mas3flags, void **ea);

/** 
\brief		Unmaps the address passed in this function

\param[in] 	va - The address which ahs to be unmapped	

\return		Success: 0; Fail: -EINVAL
*/
int dev_unmap(void *va);

/**
\brief		Maps the virtual address to physical address

\param[in] 	ptr - The virtual address which has to be mapped

\return		Physical address which the passed virtual address maps to
		-1 is returned if the passed virtual address doesn't map
		 to any physical address
*/
phys_addr_t vtop(void *ptr);

/**
\brief		Maps the physical address to virtual address

\param[in] 	ptr - The physical address which has to be mapped

\return		Virtual address which the passed Physical address maps to
		NULL is returned if the passed Physical address doesn't map
		 to any virtual address
*/
void *ptov(phys_addr_t ptr);

/**
\brief		virtual address space allocation.

\param[in]	size - specify the allocation size.

\param[in]	align - alignment of start address of the allocated virtual
		address space. It must be a non-zero value and is in multiple
		of 4k.

\return		Start address of allocated virtual address space,
		NULL is returned in case of failure.
*/
void *valloc(unsigned long size, unsigned long align);

/**
\brief		Release virtual address space allocated using valloc.

\param[in]	ptr - start address of allocated virtual addres space
		using valloc, that needs to be released.
*/
void vfree(void *ptr);

/**
\brief		Providing virtual address space segment/range to Virtual
		Address Space Manager for managing it.

\param[in]	start - start address of the virtual address segment/range.

\param[in]	size - size of virtual address segment.

\return		0 on Success, -1 on error.
*/
int valloc_add(unsigned long start, unsigned long size);

enum tlb_sel {
	TLB0 = 0,
	TLB1
};

/**
\brief		Maps the size bytes starting from physical
		address passed with the required attributes at location
		which is returned back in ea in the desired TLB.

\param[in] 	pa - Physical address for which the mapping has to be created
		The address passed should be 4K aligned.
\param[in]	size - Size for which mapping is to be created.
		This should be 4K aligned. maximum size allowed is 2MB.
\param[in] 	attr - Programed in MAS2. Allowed values
		MAS2_W, MAS2_I, MAS2_M, MAS2_G, MAS2_E, MAS2_X0, MAS2_X1
\param[in] 	perm - must be a combination of MAS3[UX,SX,UW,SW,UR,SR]
		MAS3_UR, MAS3_UW, MAS3_UX, MAS3_SX, MAS3_SW, MAS3_SR
\param[in] 	tlbnum -- TLB0 or TLB1.

\param[out]     ea - Effective address to which the physical address passed
		is mapped.The address returned would be aligned to 4K or
		alignment passed if the alignment id greater than 4K.

\return		0 on success and -ve on failure

*/
int mmap(phys_addr_t pa, size_t size, uint32_t attr, uint32_t perm,
					enum tlb_sel tlbnum, void **ea);

/**
\brief		Unmaps the address passed in this function

\param[in] 	va - The address which has to be unmapped

\return		0 - Succes
		-EFAULT	- Bad address
*/
int munmap(void *va);

#ifdef __cplusplus
    }
#endif
#endif /*__MMAP_H__*/
