/**
 \file rbtree.h
 \brief Common datatypes, hash-defines and function prototypes for
 \RED-BLACK Tree's APIs implementation
 */
/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef __cplusplus
    extern "C" {
#endif
#include <io.h>
#include <bitops.h>
#include <init.h>
#include <misc.h>
#include <errno.h>

#define RBTREE_NODE2OBJ(ptr, type, node_field) \
	(type *)((char *)ptr - offsetof(type, node_field))

struct rbt_node {
	unsigned int color;		/*color of the node: 0-Black, 1-Red*/
	struct rbt_node *left;		/*pointer of the left child*/
	struct rbt_node *right;		/*pointer of the right child*/
	struct rbt_node *parent;	/*pointer of the parent*/
};

struct red_black_tree {
	int (*compfunc) (const void *a, const void *b);	/*function for
							comparing keys*/
	struct rbt_node *root;		/*pointer to the root of the tree*/
	struct rbt_node *nil;		/*dummy nil node to be
						used as sentinel*/
};

struct red_black_tree *create_rbtree
		(int (*compare_func) (const void *, const void *));
struct rbt_node *rbtree_insert_node(struct red_black_tree *tree,
				    struct rbt_node *);
struct rbt_node *rbtree_delete_node(struct red_black_tree *tree,
				    struct rbt_node *);

#define NEW_RBTREE(compare) create_rbtree(compare);

#define IMPLEMENT_RBTREE(name, type, node_field, val_field) \
static inline int name##_push(struct red_black_tree *tree, type *obj) \
{ \
	type *obj1, *obj2, *item; \
	struct rbt_node *z, *temp, *temp1; \
	int *val, *compval, *val1; \
	z = &obj->node_field; \
	if (tree->root == tree->nil) { \
		printf("1st node\n"); \
		tree->root = z; \
		tree->root->color = 0; \
		tree->root->parent = tree->nil; \
		tree->root->left = tree->nil; \
		tree->root->right = tree->nil; \
		return 0; \
	} \
	item = RBTREE_NODE2OBJ(z, type, node_field); \
	compval = (int *)item->val_field; \
	z->left = z->right = tree->nil; \
	temp1 = NULL; \
	temp = tree->root; \
	while (temp != tree->nil) { \
		temp1 = temp; \
		obj1 = RBTREE_NODE2OBJ(temp, type, node_field); \
		val = (int *)obj1->val_field; \
		if (*val > *compval) {  \
			temp = temp->left; \
		} else {  \
			temp = temp->right; \
		} \
	} \
	z->parent = temp1; \
	obj2 = RBTREE_NODE2OBJ(temp1, type, node_field); \
	val1 = (int *)obj2->val_field; \
	if (*val1 > *compval) { \
		temp1->left = z; \
	} else { \
		temp1->right = z; \
	} \
	rbtree_insert_node(tree, z); \
	return 0; \
} \
static inline int name##_del(struct red_black_tree *tree, type *obj) \
{ \
	struct rbt_node *del, *arg1; \
	type *item1; \
	arg1 = &obj->node_field; \
	del = rbtree_delete_node(tree, arg1); \
	item1 = RBTREE_NODE2OBJ(del, type, node_field); \
	free(item1); \
	return 0; \
} \
static inline type *name##_find(const struct red_black_tree *tree, \
						const void *val) \
{ \
	struct rbt_node	*x = tree->root; \
	type *item; \
	int compval; \
	if (x == tree->nil) \
		return 0; \
	item = RBTREE_NODE2OBJ(x, type, node_field); \
	compval = tree->compfunc((void *)item->val_field, val); \
	while (0 != compval) { \
		if (1 == compval) {  \
			x = x->left; \
		} else { \
			x = x->right; \
		} \
		if (x == tree->nil) \
			return 0; \
		item = RBTREE_NODE2OBJ(x, type, node_field); \
		printf("entered place find:val:%lu\r\n", \
				*((int *)item->val_field)); \
		compval = tree->compfunc((void *)item->val_field, val); \
	} \
	return item; \
}
#ifdef __cplusplus
    }
#endif
