/**
\file
\brief	Read-write spinlocks, allowing multiple readers but exclusive writer.
*/

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __RWLOCK_H
#define __RWLOCK_H

#ifdef __cplusplus
    extern "C" {
#endif
#include <stdint.h>

/* Lock definition */
#define rwlock_t  uint32_t

/**
\brief		lock initialization, should be called only from one core
		before read or write lock operations are attempted

\param[in]	ptr - R/W lock to be initialized

\return   	None
*/
static inline void
lwe_rwlock_init(rwlock_t * ptr)
{
	*ptr = 0;
}

/**
\brief		Multiple readers can read the data in parallel but readers
		will be blocked until the writer is finished writing.

\param[in]	ptr - Acquire lock for read operation

\return   	None
*/
static inline void
lwe_rwlock_read_lock(rwlock_t * ptr)
{
	uint32_t val = 0;
	uint32_t val1 = 0;
	uint32_t val2 = 0;
	uint32_t mask = 0;

	asm volatile ("1: lwarx %0, %y3 \n" "andi. %1, %0, 0x7f \n" "andi. %2, %0, 0x7f00 \n" "rotrwi %2, %2, 8 \n" "xor. %2, %2, %1 \n" "bne 1b \n"	/* loop until all writes are complete */
		      "li %4, 1 \n" "rotlwi %4, %4, 16 \n" "add %1, %0, %4 \n" "rotlwi %4, %4, 7 \n" "and. %0, %1, %4 \n"	/* no over flows allowed */
		      "bne 1b \n" "stwcx. %1, %y3 \n" "bne 1b \n"	/* increment readers and retry if reserve fails */
		      :"+&r" (val), "+&r"(val1), "+&r"(val2), "+Z"(*ptr),
		      "+&r"(mask)
		      ::"memory", "cc");
}

/** 
\brief		Release previosuly acquired lock for read operations

\param[in]	ptr - Read lock to be released

\return   	None
*/
static inline void
lwe_rwlock_read_unlock(rwlock_t * ptr)
{
	uint32_t val = 0;
	uint32_t count = 0;

	asm volatile ("li %2, 1 \n" "rotlwi %2, %2, 16 \n" "1: lwarx %0, %y1 \n" "sub %0, %0, %2 \n" "stwcx. %0, %y1 \n" "bne 1b \n;"	/* increment readers and retry if reserve fails */
		      :"+&r" (val), "+Z"(*ptr), "+&r"(count)
		      ::"memory", "cc");
}

/** 
\brief		Exclusive access for write operations.

\param[in]	ptr - Acquire lock for write operation

\return   	None
*/
static inline void
lwe_rwlock_write_lock(rwlock_t * ptr)
{
	uint32_t val = 0;
	uint32_t val1 = 0;
	uint32_t val2 = 0;
	uint32_t mask = 0;

	asm volatile ("li %4, 0x80 \n" "nor %4, %4, %4 \n" "1: lwarx %0, %y3 \n" "addi %1, %0, 1 \n" "and %1, %1, %4 \n" "stwcx. %1, %y3 \n" "bne 1b \n"	/* increment write requests, retry if reserve fails */
		      "andi. %1, %0, 0x7f \n"	/* save old request count */
		      "2:lwarx %0, %y3 \n" "li %4, 0x7f00 \n" "rotlwi %4, %4, 8 \n" "and %2, %0, %4 \n"	/* read count */
		      "or %2, %2, %1 \n" "andi. %4, %0, 0x7f00 \n"	/* write compare */
		      "rotrwi %4, %4, 8 \n" "xor. %2, %2, %4 \n" "bne 2b \n"	/* loop we reach our write slot and no readers exist */
		      :"+&r" (val), "+&r"(val1), "+&r"(val2), "+Z"(*ptr),
		      "+&r"(mask)
		      ::"memory", "cc");
}

/**
\brief		Release previosuly acquired lock for write operations

\param[in]	ptr - Write lock to be released

\return   	None
*/
static inline void
lwe_rwlock_write_unlock(rwlock_t * ptr)
{
	uint32_t val = 0;
	uint32_t mask = 0;

	asm volatile ("li %2, 0x4000 \n" "rotlwi %2, %2, 1 \n" "nor %2, %2, %2 \n" "1: lwarx %0, %y1 \n" "addi %0, %0, 0x100 \n" "and %0, %0, %2 \n" "stwcx. %0, %y1 \n" "bne 1b \n"	/* increment write requests, retry if reserve fails */
		      :"+&r" (val), "+Z"(*ptr), "+&r"(mask)
		      ::"memory", "cc");
}

#ifdef __cplusplus
    }
#endif
#endif				/* __RWLOCK_H */
