/**
\file
\brief	create and free different frame descriptors 
*/

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __SG_API_H
#define __SG_API_H

#ifdef __cplusplus
    extern "C" {
#endif
#include <compat.h>
#include <fsl_bman.h>
#include <dpa_alloc.h>
#include <string.h>
#include <fsl_qman.h>

#define GET_ADDR36B(s)	((phys_addr_t)(s)->addr_hi << 32 | (s)->addr_lo)
#define GET_ADDR36B_BM_BUF(s)	((phys_addr_t)(s).hi << 32 | (s).lo)
#define BPID_ADDR_CPY(d, s) \
do { \
	(d)->bpid = (s)->bpid; \
	(d)->addr_hi = (s)->addr_hi; \
	(d)->addr_lo = (s)->addr_lo; \
} while (0)
#define BM_BUF_CPY(d, s) \
do { \
	(d)->bpid = (s).bpid; \
	(d)->addr_hi = (s).hi; \
	(d)->addr_lo = (s).lo; \
} while (0)

enum frame_format_t {
	SINGLE_BUFFER_SHORT = qm_fd_contig,
	SINGLE_BUFFER_LONG = qm_fd_contig_big,
	MULTI_BUFFER_SHORT = qm_fd_sg,
	MULTI_BUFFER_LONG = qm_fd_sg_big
};

struct frame_t {
	enum frame_format_t format;
	u32 addr_hi;
	u32 addr_lo;
	u32 length;
	u32 offset;
	u32 bpid;
};

struct frame_fragment_t {
	u32 bpid;
	u32 addr_hi;
	u32 addr_lo;
	u32 offset;
	u32 length;
};

typedef int (*lwe_fd_free_func_t) (const struct dpa_buff_allocator *
				   buff_allocator, const struct qm_fd * p_fd);

/**
\brief		Scatter Gather Entry

\details 	Speicifies the the Scatter Gather Format related information 
*/
struct sg_entry_t {
	uint16_t reserved_zero;
	uint16_t addr_hi;	/**< Memory Address of the start of the buffer - hi*/
	uint32_t addr_lo;	/**< Memory Address - lo*/
	unsigned int extension:1;
	unsigned int final:1;
	unsigned int length:30;	/**< Length of the data in the frame */
	uint8_t reserved_zero2;
	uint8_t bpid;		/**< Buffer Pool Id */
	unsigned int reserved_offset:3;
	unsigned int offset:13;
};

/**
\brief		Create a frame which may consist of single buffer or multiple
 		buffer linked through scatter/gather list

\param[in]	allocator - DPA allocator object handle.

\param[out]	p_frame - Frame to be constructed

\param[in]	p_frag - Array of buffer(s)
 				
\param[in]	count - Total no. of buffers. It starts with value 1.

\return		0 - On success, otherwise -ve value
*/
int lwe_prepare_frame(struct dpa_buff_allocator *buff_allocator,
		      struct frame_t *p_frame, struct frame_fragment_t *p_frag,
		      u32 count);

/**
\brief		Create a frame descriptor.

\details	An FD directly describe a single buffer, which contains a frame,
 		or it may, indirectly through scatter/gather lists,	describe a
 		frame which occupies multiple buffers or it may	potentially
 		describe multiple frames where each of these frames can in turn
 		consist of a single or multiple buffers.

\param[in]	allocator - DPA allocator object handle.

\param[out]	p_fd - FD to be constructed.

\param[in]	p_frame - Array of frame(s)

\param[in]	count - Total no. of frames. It starts with value 1.

\return		0  - On success, otherwise -ve value
*/
int lwe_prepare_fd(struct dpa_buff_allocator *buff_allocator,
		   struct qm_fd *p_fd, struct frame_t *p_frame, u32 count);

/**
\brief		Drop a frame descriptor

\details	Free memory associated with a frame descriptor.

\param[in]	allocator - DPA allocator object handle.

\param[in]	p_fd - Frame descriptor of frame to be freed.

\return		0 - Success, otherwise -ve value

*/
int lwe_free_fd(const struct dpa_buff_allocator *buff_allocator,
		const struct qm_fd *p_fd);

/**
\brief		Find total number of frame present in FD.

\param[in]	p_fd - Frame descriptor.

\return 	Number of frames(s).
*/
uint32_t get_num_frame(const struct qm_fd *p_fd);

/**
\brief		Find pointer of each frame present in FD.

\param[in]	p_fd - Frame descriptor

\param[out] 	p_frame - Array of frame(s)

\param[inout] 	count - Total no. of buffer(s) to be appended in p_frame.
		if count is less than maximum no. of frames in FD then same
		value will be returned but if it is greter than maximum no. of
		frames in FD then count will have maximum no. of frame present
		in FD.

\return 	None
*/
void get_frame_from_fd(const struct qm_fd *p_fd, struct frame_t *p_frame,
		       uint32_t *count);

/**
\brief   	Append fragment(s) to a frame and it's FD could be rebuild with
		the updated frame using lwe_prepare_fd() API.

\param[in]	allocator - DPA allocator object handle

\param[inout]	p_frame -  Frame to be updated with new appended buffer(s).

\param[in] 	p_frag - Array of buffer(s)

\param[out]	count - Total no. of frames returned in p_frame
*/
int append_frame_fragment(const struct dpa_buff_allocator *buff_allocator,
			  struct frame_t *p_frame,
			  const struct frame_fragment_t *p_frag, u32 count);

#ifdef __cplusplus
    }
#endif
#endif /* __SG_API_H */
