/** @file
 * This file provides APIs for decorated store functionality
 *
 * P4080 statistics functions using decorated stores, loads and notifies.
 * This file contains definitions for statistics manipulation functions, as
 * well as relative offsets for the primary statistics blocks.
 * The statistics area is assumed to start at the globally-defined
 * APP_STATS_BAR, and this virtual address is assumed to be readable and
 * writeable by all cores.  The total size of the statistics area is assumed
 * to be 1MB.
 *
 * This file defines several statistics manipulation functions, 
 */

/*
 * Copyright (c) 2008 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __STATISTICS_H__
#define __STATISTICS_H__

#ifdef __cplusplus
    extern "C" {
#endif
#include <stdint.h>
#include "compiler.h"

/*!
 The decorated store (write) transactions require the address operand to
 sometimes to not point to the start of the structure they are updating,
 but to the value operand-LSB-aligned accumulator portion of the statistic.
 If there is a quad-word statistic (64 bit inc, 64 bit acc) at 0x100,
 the decorated store should target address 0x10C if the value operand is a
 32 bit value.
 */
#define DEC_STORE_DEC0_ACC_64_OPER_32   4
#define DEC_STORE_DEC2_ACC_64_OPER_32   12
#define DEC_STORE_DEC2_ACC_64_OPER_64   8
#define DEC_STORE_DEC3_ACC_32_OPER_32   4
#define DECORATED_INC_ACC_ADDR_OFFSET   12

#define STAT_NAME_MAX_LENGTH            30

/* 32 bit statistics, single accumulator */
typedef volatile int32_t stat32_t;

/* 32 bit statistics, incrementor (number of packets) and accumulator
 *(number of bytes) pair.
 */
struct stat32_pair_t {
	stat32_t inc;
	stat32_t acc;
};

/* 64 bit statistics, single accumulator */
union stat64_t {
	volatile int64_t stat;
	struct {
		volatile uint32_t hi;
		volatile uint32_t lo;
	} words;
};

/* 64 bit statistics, incrementor (number of packets) and accumulator
 * (number of bytes) pair
 */
struct stat64_pair_t {
	union stat64_t inc;
	union stat64_t acc;
};

/*!
 Four decorations are allowed on Read transactions:
 - Read-clear
 - Read-set
 - Read-decrement
 - Read-increment
*/
enum LOAD_DECORATION {
	LOAD_DECORATION_CLEAR = 0,
	LOAD_DECORATION_SET = 1,
	LOAD_DECORATION_DEC = 2,
	LOAD_DECORATION_INC = 3
};

/*!
 Decorated load. Decoration = Read-clear. Returns 8 bit value
*/
/** 
\brief		8 bit Decorated load. Decoration = Read-clear

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 8 bit value	
*/
static inline uint8_t
decorated_load_clear_8(volatile void *a)
{
	uint8_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_CLEAR;
      __ASM("lbdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-set. Returns 8 bit value
*/
/** 
\brief		8 bit Decorated load. Decoration = Read-set

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 8 bit value	
*/
static inline uint8_t
decorated_load_set_8(volatile void *a)
{
	uint8_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_SET;
      __ASM("lbdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-dec. Returns 8 bit value
*/
/** 
\brief		8 bit Decorated load. Decoration = Read-dec

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 8 bit value	
*/
static inline uint8_t
decorated_load_dec_8(volatile void *a)
{
	uint8_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_DEC;
      __ASM("lbdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-inc. Returns 8 bit value
*/
/** 
\brief		8 bit Decorated load. Decoration = Read-inc

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 8 bit value	
*/
static inline uint8_t
decorated_load_inc_8(volatile void *a)
{
	uint8_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_INC;
      __ASM("lbdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-clear. Returns 16 bit value
*/
/** 
\brief		16 bit Decorated load. Decoration = Read-clear

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 16 bit value	
*/
static inline uint16_t
decorated_load_clear_16(volatile void *a)
{
	uint16_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_CLEAR;
      __ASM("lhdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-set. Returns 16 bit value
*/
/** 
\brief		16 bit Decorated load. Decoration = Read-set

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 16 bit value	
*/
static inline uint16_t
decorated_load_set_16(volatile void *a)
{
	uint16_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_SET;
      __ASM("lhdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-dec. Returns 16 bit value
*/
/** 
\brief		16 bit Decorated load. Decoration = Read-dec

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 16 bit value	
*/
static inline uint16_t
decorated_load_dec_16(volatile void *a)
{
	uint16_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_DEC;
      __ASM("lhdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-inc. Returns 16 bit value
*/
/** 
\brief		16 bit Decorated load. Decoration = Read-inc

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 16 bit value	
*/
static inline uint16_t
decorated_load_inc_16(volatile void *a)
{
	uint16_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_INC;
      __ASM("lhdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-clear. Returns 32 bit value
*/
/** 
\brief		32 bit Decorated load. Decoration = Read-clear

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 32 bit value	
*/
static inline uint32_t
decorated_load_clear_32(volatile void *a)
{
	uint32_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_CLEAR;
      __ASM("lwdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-set. Returns 32 bit value
*/
/** 
\brief		32 bit Decorated load. Decoration = Read-set

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 32 bit value	
*/
static inline uint32_t
decorated_load_set_32(volatile void *a)
{
	uint32_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_SET;
      __ASM("lwdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-dec. Returns 32 bit value
*/
/** 
\brief		32 bit Decorated load. Decoration = Read-dec

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 32 bit value	
*/
static inline uint32_t
decorated_load_dec_32(volatile void *a)
{
	uint32_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_DEC;
      __ASM("lwdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-inc. Returns 32 bit value
*/
/** 
\brief		32 bit Decorated load. Decoration = Read-inc

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 32 bit value	
*/
static inline uint32_t
decorated_load_inc_32(volatile void *a)
{
	uint32_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_INC;
      __ASM("lwdx    %0, %1, %2":"=r"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-clear. Returns 64 bit value
*/
/** 
\brief		64 bit Decorated load. Decoration = Read-clear

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 64 bit value	
*/
static inline uint64_t
decorated_load_clear_64(volatile void *a)
{
	uint64_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_CLEAR;
      __ASM("lfddx    %0, %1, %2":"=f"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-set. Returns 64 bit value
*/
/** 
\brief		64 bit Decorated load. Decoration = Read-set

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 64 bit value	
*/
static inline uint64_t
decorated_load_set_64(volatile void *a)
{
	uint64_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_SET;
      __ASM("lfddx    %0, %1, %2":"=f"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-dec. Returns 64 bit value
*/
/** 
\brief		64 bit Decorated load. Decoration = Read-dec

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 64 bit value	
*/
static inline uint64_t
decorated_load_dec_64(volatile void *a)
{
	uint64_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_DEC;
      __ASM("lfddx    %0, %1, %2":"=f"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*!
 Decorated load. Decoration = Read-inc. Returns 64 bit value
*/
/** 
\brief		64 bit Decorated load. Decoration = Read-inc

\param[in] 	a - The address from which decorated load has to be performed

\return		 Returns 64 bit value	
*/
static inline uint64_t
decorated_load_inc_64(volatile void *a)
{
	uint64_t r;
	enum LOAD_DECORATION d = LOAD_DECORATION_INC;
      __ASM("lfddx    %0, %1, %2":"=f"(r)
      :      "r"(d), "r"(a)
      :      "memory");
	return r;
}

/*
 These decorated store functions utilize the decorated storage mechanism to
 atomically increment a particular 32/64-bit counter, and optionally
 simultaneously accumulate an associated 32/64-bit accumulator.  Both the
 accumulator and the counter must reside in the same aligned double/quad-word.
 The accumulator must be in the most significant doubleword, and the counter
 is in the least significant doubleword. See the structure definitions
 elsewhere in this file.
*/

/*!
 Four decorations are allowed on Write transactions:
 - 64-bit accumulator update
 - 32-bit accumulator update
 - 64-bit accumulator increment and 64-bit accumulator upate (packet count
 and byte count)
 - 32-bit accumulator increment and 32-bit accumulator update (packet count
 and byte count)
*/
enum STORE_DECORATION {
	STORE_DECORATION_ACC_64 = 0,
	STORE_DECORATION_ACC_32 = 1,
	STORE_DECORATION_INC_ACC_64 = 2,
	STORE_DECORATION_INC_ACC_32 = 3
};

/*!
 Decorated store. Decoration = Write-accumulate. 32 bit accumulator update,
 32 bit argument
*/
/** 
\brief		Decorated store. Decoration = Write-accumulate.
		32 bit accumulator update, 32 bit argument

\param[in] 	a - The address to which decorated store has to be performed
\param[in] 	v - value to be accumulated in the passed address

\return		None	
\note		These decorated store functions utilize the decorated storage
		 mechanism to accumulate an associated 32/64-bit accumulator.
*/
static inline void
decorated_store_32_acc_32(volatile void *a, register int32_t v)
{
	volatile void *address = a;
	enum STORE_DECORATION d = STORE_DECORATION_ACC_32;
      __ASM("stwdx    %0, %1, %2":
      :"r"(v), "r"(d), "r"(address)
      :"memory");
}

/*!
 Decorated store. Decoration = Write-accumulator increment and update. 32 bit
 accumulator update, 32 bit argument
*/
/** 
\brief		Decorated store. Decoration = Write-accumulator increment and update.
		32 bit accumulator update, 32 bit argument

\param[in] 	a - The address to which decorated store has to be performed
\param[in] 	v - value to be accumulated in the passed address

\return		None	
\note		These decorated store functions utilize the decorated storage
		 mechanism to accumulate an associated 32/64-bit accumulator.
*/
static inline void
decorated_store_32_inc_acc_32(volatile void *a, register int32_t v)
{
	volatile void *address = (void *) ((uintptr_t) a + 4);
	enum STORE_DECORATION d = STORE_DECORATION_INC_ACC_32;
      __ASM("stwdx    %0, %1, %2":
      :"r"(v), "r"(d), "r"(address)
      :"memory");
}

/*!
 Decorated store. Decoration = Write-accumulate. 64 bit accumulator update,
 32 bit argument
*/
/** 
\brief		Decorated store. Decoration = Write-accumulate.
		64 bit accumulator update, 32 bit argument

\param[in] 	a - The address to which decorated store has to be performed
\param[in] 	v - value to be accumulated in the passed address

\return		None	
\note		These decorated store functions utilize the decorated storage
		 mechanism to accumulate an associated 32/64-bit accumulator.
*/
static inline void
decorated_store_64_acc_32(volatile void *a, register int32_t v)
{
	volatile void *address = (void *) ((uintptr_t) a + 4);
	enum STORE_DECORATION d = STORE_DECORATION_ACC_64;
      __ASM("stwdx    %0, %1, %2":
      :"r"(v), "r"(d), "r"(address)
      :"memory");
}

/*!
 Decorated store. Decoration = Write-accumulator increment and update. 64 bit
 accumulator update, 32 bit argument
*/
/** 
\brief		Decorated store. Decoration = Write-accumulator increment and update.
		64 bit accumulator update, 32 bit argument

\param[in] 	a - The address to which decorated store has to be performed
\param[in] 	v - value to be accumulated in the passed address

\return		None	
\note		These decorated store functions utilize the decorated storage
		 mechanism to accumulate an associated 32/64-bit accumulator.
*/
static inline void
decorated_store_64_inc_acc_32(volatile void *a, register int32_t v)
{
	volatile void *address = (void *) ((uintptr_t) a + 12);
	enum STORE_DECORATION d = STORE_DECORATION_INC_ACC_64;
      __ASM("stwdx    %0, %1, %2":
      :"r"(v), "r"(d), "r"(address)
      :"memory");
}

/*!
 Decorated store. Decoration = Write-accumulate. 64 bit accumulator update,
 64 bit argument
*/
/** 
\brief		Decorated store. Decoration = Write-accumulate
		64 bit accumulator update, 64 bit argument

\param[in] 	a - The address to which decorated store has to be performed
\param[in] 	v - value to be accumulated in the passed address

\return		None	
\note		These decorated store functions utilize the decorated storage
		 mechanism to accumulate an associated 32/64-bit accumulator.
*/
static inline void
decorated_store_64_acc_64(volatile void *a __UNUSED,
			register int64_t v __UNUSED)
{
	volatile void *address = a;
	enum STORE_DECORATION d = STORE_DECORATION_ACC_64;
      __ASM("stfddx    %0, %1, %2":
      :"f"(v), "r"(d), "r"(address)
      :"memory");
}

/*!
 Decorated store. Decoration = Write-accumulator increment and update. 64 bit
 accumulator update, 64 bit argument
*/
/** 
\brief		Decorated store. Decoration = Write-accumulator increment and update.
		64 bit accumulator update, 64 bit argument

\param[in] 	a - The address to which decorated store has to be performed
\param[in] 	v - value to be accumulated in the passed address

\return		None	
\note		These decorated store functions utilize the decorated storage
		 mechanism to accumulate an associated 32/64-bit accumulator.
*/
static inline void
decorated_store_64_inc_acc_64(volatile void *a __UNUSED,
				register int64_t v __UNUSED)
{
	volatile void *address = (void *) ((uintptr_t) a + 8);
	enum STORE_DECORATION d = STORE_DECORATION_INC_ACC_64;
      __ASM("stfddx    %0, %1, %2":
      :"f"(v), "r"(d), "r"(address)
      :"memory");
}

/*
  Decorated notify functions do increment, clear etc only
*/
/*!
 Four decorations are allowed on Notify transactions:
 - 64-bit accumulator increment
 - 32-bit accumulator increment
 - 64-bit accumulator clear
 - 32-bit accumulator clear
*/
enum NOTIFY_DECORATION {
	NOTIFY_DECORATION_INC_64 = 0,
	NOTIFY_DECORATION_INC_32 = 1,
	NOTIFY_DECORATION_CLEAR_64 = 2,
	NOTIFY_DECORATION_CLEAR_32 = 3
};

/*!
 Decorated notify. Decoration = increment. 32 bit accumulator
*/
/** 
\brief		 Decorated notify. Decoration = increment. 32 bit accumulator

\param[in] 	a - The address from which decorated notify has to be performed

\return		 None
*/
static inline void
decorated_notify_inc_32(volatile void *a)
{
	register enum STORE_DECORATION d =
		(enum STORE_DECORATION)NOTIFY_DECORATION_INC_32;
      __ASM("dsn    %0, %1":
      :"r"(d), "r"(a)
      :"memory");
}

/*!
 Decorated notify. Decoration = clear. 32 bit accumulator
*/
/** 
\brief		 Decorated notify. Decoration = clear. 32 bit accumulator

\param[in] 	a - The address from which decorated notify has to be performed

\return		 None	
*/
static inline void
decorated_notify_clear_32(volatile void *a)
{
	register enum STORE_DECORATION d =
		(enum STORE_DECORATION)NOTIFY_DECORATION_CLEAR_32;
      __ASM("dsn    %0, %1":
      :"r"(d), "r"(a)
      :"memory");
}

/*!
 Decorated notify. Decoration = increment. 64 bit accumulator
*/
/** 
\brief		Decorated notify. Decoration = increment. 64 bit accumulator

\param[in] 	a - The address from which decorated notify has to be performed

\return		 None	
*/
static inline void
decorated_notify_inc_64(volatile void *a)
{
	register enum STORE_DECORATION d =
		(enum STORE_DECORATION)NOTIFY_DECORATION_INC_64;
      __ASM("dsn    %0, %1":
      :"r"(d), "r"(a)
      :"memory");
}

/*!
 Decorated notify. Decoration = clear. 64 bit accumulator
*/
/** 
\brief		Decorated notify. Decoration = clear. 64 bit accumulator

\param[in] 	a - The address from which decorated notify has to be performed

\return		 None	
*/
static inline void
decorated_notify_clear_64(volatile void *a)
{
	register enum STORE_DECORATION d =
		(enum STORE_DECORATION)NOTIFY_DECORATION_CLEAR_64;
      __ASM("dsn    %0, %1":
      :"r"(d), "r"(a)
      :"memory");
}

#ifdef __cplusplus
    }
#endif
#endif				/* __STATISTICS_H__ */
