/** @file
 \brief Software events support
 */

/*
 * Copyright (c) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __SWI_EXT_H
#define __SWI_EXT_H
#ifdef __cplusplus
    extern "C" {
#endif

#include <hv_misc.h>

#define SWI_ID_INV 		0	/** Invalid SWI ID*/
#define SWI_ID_TIMER_EXPIRY	1	/** SWI ID used for timer expiry*/
#define SWI_ID_SHUTDOWN 	2	/** SWI ID used for partition  shutdown doorbell */

enum swi_priority_t {
	SWI_PRI_HIGH = 0,
	SWI_PRI_MED,
	SWI_PRI_LOW,
	SWI_MAX_PRIORITIES
};

__PERCPU extern volatile uint32_t g_pending_swi_count;

typedef void (*swi_callbk_hndl) (uint32_t swi_id, void *data);
uint32_t swi_consume(void **p_p_data);


/**
\brief          Activates processing of an SWI.

\details        Check if there is a SWI in the application's swi queue. 
		If any SWI is present, its callback handler is called if present
		otherwise SWI ID is delivered to caller as a return value.
		SWIs are dequeued as per priority order.

\param[out]      p_p_data - SWI data if callback handler not installed.

\return         SWI ID if callback handler not installed

\note           Should be called periodically preferrably after every frame.
*/
static inline uint32_t swi_activate(void **data)
{
	if (likely(0 == g_pending_swi_count))
		return SWI_ID_INV;

	return swi_consume(data);
}

/**
\brief          Adds an SWI.

\details        Enqueues a SWI event at a particular priority. 

\param[in] id		: An identifier for the SWI to be added
\param[in] p_data	: Cookie to be passed to callback function when 
			the given SWI is scheduled
\param[in] f_hndl	: function to be called which the given SWI is schefuled
\param[in] priority	: priority to be assigned to the SWI event.

\retval         0	 for SUCCESS 
\retval 	negative for failure.
*/
int swi_add(uint32_t id, void *p_data,
	    swi_callbk_hndl hndl, enum swi_priority_t priority);

#ifdef __cplusplus
    }
#endif
#endif				/* __SWI_EXT_H */
