/** @file
 * This file provides APIs for getting timebase, delays.
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __TIME_H__
#define __TIME_H__

#ifdef __cplusplus
    extern "C" {
#endif
#include <stdint.h>
#include "hv_misc.h"
#include "init.h"
#include <io.h>

#define NSEC_PER_SEC 1000000000
#define USEC_PER_SEC 1000000
#define USEC_PER_MSEC 1000
#define MSEC_PER_SEC 1000

/** Read per core tick value
 * @return Returns the current tick value in TBU+TBL
 */
static inline uint64_t
get_timebase(void)
{
	uint32_t hi, lo, chk;

	/*
	 * To make sure that there is no carry over
	 * between checking of TBU and TBL
	 */
	do {
		hi = mfspr(SPR_TBU);
		lo = mfspr(SPR_TBL);
		chk = mfspr(SPR_TBU);
	} while (unlikely(hi != chk));

	return (uint64_t) hi << 32 | (uint64_t) lo;
}

/** Read per core alternative ticks via alternate timebase
 * @return Returns the current tick value found in ATBU+ATBL
 */
static inline uint64_t
get_atb(void)
{
	uint32_t hi, lo, chk;

	/*
	 * To make sure that there is no carry over
	 * between checking of ATBU and ATBL
	 */
	do {
		hi = mfspr(SPR_ATBU);
		lo = mfspr(SPR_ATBL);
		chk = mfspr(SPR_ATBU);
	} while (unlikely(hi != chk));

	return (uint64_t) hi << 32 | (uint64_t) lo;
}

/** Delay based on timebase reg
 * @param[in] nsec -- Delay time in nanoseconds
 */
static inline void
ndelay(uint64_t nsec)
{
	uint64_t tb;

	tb = (get_timebase() + ((timebase_freq * nsec) / NSEC_PER_SEC));
	while (tb > get_timebase())
		;
}

/** Delay based on timebase reg
 * @param[in] usec -- Delay time in microseconds
 */
static inline void
udelay(uint32_t usec)
{
	uint64_t tb;

	tb = (get_timebase() + (timebase_freq / USEC_PER_SEC * usec));
	while (tb > get_timebase()) ;
}

/** Delay based on timebase reg
 * @param[in] msec -- Delay time in milliseconds
 */
static inline void
mdelay(uint32_t msec)
{
	udelay(msec * USEC_PER_MSEC);
}

/** Delay based on timebase reg
 * @param[in] sec -- Delay time in seconds
 */
static inline void
sdelay(uint32_t sec)
{
	udelay(sec * USEC_PER_SEC);
}

#ifdef __cplusplus
    }
#endif
#endif /*__TIME_H__*/
