/** @file
 \brief Timer Wheel support
 
\details Timer wheel provides support of soft real time timers. The total
	number of timers which can be started can be in millions and the 
	time period range can be from 10 milliseconds to days.
 */

/*
 * Copyright (c) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __TIMER_WHEEL_EXT_H
#define __TIMER_WHEEL_EXT_H

#ifdef __cplusplus
    extern "C" {
#endif
#include <conf.h>
#include <compat.h>
#include <fsl_bman.h>
#include <fsl_qman.h>
#include "swi.h"

#define INV_TIMER_ID		(0xFFFFFFFF)  /**< Max 32 bit*/
#define INV_FQ_ID		(0x0)

/**
\brief          Returns application specific data stored in timer object

\details        Checks whether the ethernet frame is allowed to be processed.

\param[in]      timer_id - Timer ID from which application data needs to be 
				extracted.

\return         Timer data (NULL is also valid value)
*/
void *get_app_data_from_timer_id(uint32_t timer_id);

typedef void (*timer_callbk_t) (uint32_t timer_id, void *p_data);

/**
\brief		Timer framework start function

\details	Starts the tick processing in timer framework on local core

\return		On success 0, on failure negative ecode.
*/
int timer_framework_start(void);

/**
\brief		Timer framework stop function

\details	Stops the tick processing in timer framework on local core

\return		On success 0, on failure negative ecode.
*/
int timer_framework_stop(void);

/**
\brief          Timer start function

\details        Starts a software timer

\param[in]      period_ms - Period for which timer to be started
\param[in]      is_periodic - Whether timer is periodic or not
\param[in]      expiry_fq - FQ ID on which to receive timer expiry 
\param[in]      swi_priority - Priority of SWI for timer expiry
\param[in]      callbk - Timer expiry callback function
\param[in]      p_data - Application data stored with timer context 

\return         On success, timer ID, on failure INV_TIMER_ID.
*/
uint32_t start_timer(uint32_t period_ms, bool is_periodic,
		     struct qman_fq *expiry_fq,
		     enum swi_priority_t swi_priority,
		     timer_callbk_t callbk, void *p_data);

/**
\brief          Stops a timer

\details        Stops a software timer.

\param[in]      timer_id - Timer ID to be stopped

\return         0 - Success
  		-1 - Failure
*/
int stop_timer(uint32_t timer_id);

/**
\brief          Reset a timer

\details        Resets a software timer.

\param[in]      timer_id - Timer ID to reset.
\param[in]      new_period_ms  - New period for which timer needs to be restarted

\return         Old or new timer ID on success.
  		INV_TIMER_ID on failure
*/
uint32_t reset_timer(uint32_t timer_id, uint32_t new_period_ms);

/**
\brief          Timer expiry process function

\details        To be called by application on receiving timer expiry event.

\param[in]      timer_id - Timer ID which has expired,

\return         void
*/
void proc_timer_expiry(uint32_t timer_id);


/**
\brief          Get the period of tick in milliseconds

\details        Get the period of tick in milliseconds

\return         Tick size in milliseconds
*/
uint32_t get_tick_size(void);

/**
\brief          Get the time elapsed in milliseconds since startup

\details        Get the time elapsed in milliseconds since startup

\return         Time in milliseconds
*/
uint32_t get_current_time(void);

/**
\brief          Get the present time tick.

\details        Get the number of ticks elapsed since startup.

\return         Present value tick counter.
*/
uint32_t get_current_tick(void);

#ifdef __cplusplus
    }
#endif
#endif				/* __TIMER_WHEEL_EXT_H */
