/** @file
 * Watchdog API support
 * This file provides APIs to program the watchdog timer available on
 * e500mc core.
 *
 * watchdog_init() function can be used to register handler to be invoked 
 * after the specified timeout. However , teh applictaion can restart 
 * the timer by using wdog_poke() api.
 *
 * To stop the timer, use wdog_stop() api.
 *
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __WDOG_H
#define __WDOG_H

#ifdef __cplusplus
    extern "C" {
#endif
#include "core-regs.h"

typedef void (*wdoghndlr_t) (void);

/** 
\brief		Start watch dog timer with the given timeout values in seconds
		Register handler to be invoked when the timeout would occur.

\param[in]	handler : Handler called when the watchdog expires.
\param[in]	wrc - Watchdog reset control
\param[in]	timeout : expiry value in milliseconds

\return		Success: 0; Fail: -EINVAL
\note		Watchdog interrupt would occur only if MSR[CE] is enabled.
		In this implementation, registered watchdog handler will 
		be invoked in time 
			timeout + delta(timeout)
		Where delta(timeout) can vary from 0 to timeout.
		Use of printf statement form theregistered handler code is not
		supported
	
*/
int wdog_init(wdoghndlr_t hndlr, uint32_t wrc, uint32_t timeout_ms);

/** 
\brief		Stop watchdog timer. System wouldnot be interrupted on watchdog

\return		None
*/
void wdog_stop(void);

/** 
\brief		Poke watchdog timer. Once poked, timer would not expire 
		for altleast time eqal to the timeout value provided during
		initialization of wdog using wdog_init. 

\return		None
*/
static inline void
wdog_poke(void)
{
	mtspr(SPR_TSR, TSR_WIS | TSR_ENW);
}

#ifdef __cplusplus
    }
#endif
#endif				/* __WDOG_H */
