/* Copyright (c) 2008-2010 Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Freescale Semiconductor nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "bman_private.h"
#include "ldt.h"

/*****************/
/* Portal driver */
/*****************/

#define PORTAL_MAX	10
#define POOL_MAX	64
#define NUM_PORTAL_ADDR	2

static struct bm_portal_config configs[PORTAL_MAX];
static u8 num_portals;

static struct bm_portal_config *__bm_portal_add(
					const struct bm_portal_config *config)
{
	struct bm_portal_config *ret;
	BUG_ON((num_portals + 1) > PORTAL_MAX);
	ret = &configs[num_portals++];
	*ret = *config;
	ret->portal = NULL;
	return ret;
}

u8 bm_portal_num(void)
{
	return num_portals;
}

const struct bm_portal_config *bm_portal_config(u8 idx)
{
	if (unlikely(idx >= num_portals))
		return NULL;

	return &configs[idx];
}

static int __init fsl_bman_portal_init(int node)
{
	struct bm_portal_config cfg, *pconfig;
	phys_addr_t portal_addr[NUM_PORTAL_ADDR];
	uint64_t portal_size[NUM_PORTAL_ADDR];
	int irq, ret;

	if (get_fdt_mem_ranges(node, portal_addr, portal_size) < 0) {
		pr_err("bm_init error, reading portal addr and size\n");
		return -1;
	}

	/* set appropriate TLB entries */
	/* Block Guide states that 16K region of BMAN portal should have 
 	 * WIMG = x01x (coherence ON to enable stashing). 
 	 * But it turns out, the 16K region should be marked as 0000
 	 * -- the LAW/PAMU needs to be marked as M=1, but if you 
	 * mark the MMU entry as M=0, it saves some snoops.  This seems icky, 
	 * but it's been approved by Gary W. and others as acceptable.
	 */
	ret = dev_mmap(portal_addr[0], portal_size[0], portal_size[0], 0x0,
			 MAS3_SR | MAS3_SW, &cfg.addr.addr_ce);
	if (ret < 0) {
		pr_err("Failed to create tlb entry for qman driver\n");
		return -1;
	}
	ret = dev_mmap(portal_addr[1], portal_size[1], portal_size[1],
			 TLB_MAS2_IO, MAS3_SR | MAS3_SW , &cfg.addr.addr_ci);
	if (ret < 0) {
		pr_err("Failed to create tlb entry for qman driver\n");
		return -1;
	}
	if (get_fdt_irq(node, &irq, NULL) < 0) {
		pr_err("Failed in getting IRQ number of");
		pr_err("bman portal %p from device tree\n", cfg.addr.addr_ce);
		return -1;
	}
	cfg.irq = irq;
	bman_depletion_fill(&cfg.mask);

	/* TODO: this is ripped from the old Bman LWE driver - we should instead
	 * be looking at the 'cpu-handle' property in the device-tree node,
	 * making sure everything is kosher, etc. */

	cfg.cpu = mfspr(SPR_PIR);

	pconfig = __bm_portal_add(&cfg);
	if (!pconfig) {
		dev_unmap(cfg.addr.addr_ce);
		dev_unmap(cfg.addr.addr_ci);
		return -ENOMEM;
	}
	pr_info("Bman portal at %p:%p (%d)\n", cfg.addr.addr_ce,
		cfg.addr.addr_ci, cfg.cpu);
	/* If the portal is affine to a cpu and that cpu has no default affine
	 * portal, auto-initialise this one for the job. */
	if (cfg.cpu == -1)
		return 0;
	if (!bman_have_affine_portal()) {
		u32 irq_sources = 0;
#ifdef CONFIG_FSL_BMAN_HAVE_IRQ
		irq_sources = BM_PIRQ_RCRI | BM_PIRQ_BSCN;
#endif
		if (bman_create_affine_portal(pconfig, irq_sources))
			pr_err("Bman portal auto-initialisation failed\n");
		else
			pr_info("Bman portal %d auto-initialised\n", cfg.cpu);
	}
	return 0;
}

int bman_init(__UNUSED int secondary)
{
	int node = fdt_node_offset_by_compatible(devtree_ptr,
						 -1, "fsl,bman-portal");
	for ( ; node != -1; node = fdt_node_offset_by_compatible(devtree_ptr,
					 node, "fsl,bman-portal")) {
		int ret;
		const uint32_t *cpu_phandle;
		const uint32_t *cpu_id;
		int cpu_node;

		cpu_phandle = fdt_getprop(devtree_ptr, node, "cpu-handle", NULL);
		if (!cpu_phandle)
			panic("fdt_getprop failed for cpu-handle");

		cpu_node = fdt_node_offset_by_phandle(devtree_ptr, *cpu_phandle);
		if (cpu_node < 0)
			panic("fdt_node_offset_by_phandle for cpu node");

		cpu_id = fdt_getprop(devtree_ptr, cpu_node, "reg", NULL);
		if (!cpu_id)
			panic("fdt_getprop failed for reg");

		if (*cpu_id != get_lwe_id())
			continue;

		ret = fsl_bman_portal_init(node);
		if (ret)
			return ret;
	}
	pr_info("Bman driver initialised (%d)\n", get_lwe_id());
	return 0;
}
