/** @file
 * This file provides cache locking and unlocking support for L1 and L2 cache.
 * L3 cache management is not supported as of now.
 */

/*
 * Copyright (c) 2008 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdint.h>
#include <types.h>
#include <libfdt.h>
#include <fsl_hcalls.h>
#include <hv_misc.h>
#include <io.h>
#include <init.h>
#include <log.h>
#include <errno.h>
#include <cache.h>
#include <ldt.h>

int
dcache_lock(void *addr_start, uint32_t size, enum cache_target_t ct)
{
	uint32_t tmp;
	uint32_t cache_blk_size;

	assert(addr_start);
	assert(size);

	if (unlikely(!guest_cache_lock_on)) {
		LWE_ERROR("%s: guest cache lock mode is disabled",
			 __func__);
		return -EINVAL;
	}

	switch (ct) {
	case CT_L1_CACHE:
		cache_blk_size = l1_data_cache_blk_size;
		break;

	case CT_L2_CACHE:
		cache_blk_size = l2_cache_blk_size;
		break;

	default:
		LWE_ERROR("%s: unsupported cache target!", __func__);
		return -EINVAL;
	}

	if (cache_blk_size == 0) {
		LWE_WARN
		("%s:could not find required cache parameter in device tree.",
			__func__);
		return -1;
	}

	for (tmp = (uint32_t) addr_start;
	     tmp <= (uint32_t) (addr_start + size); tmp += cache_blk_size) {
		if (ct == CT_L1_CACHE) {
			asm volatile ("dcbtls %0, 0, %1"::"i" (CT_L1_CACHE),
				      "r"(tmp):"memory");
		} else {
			asm volatile ("dcbtls %0, 0, %1"::"i" (CT_L2_CACHE),
				      "r"(tmp):"memory");
		}

		asm("msync");

		/* Check unable to lock and overflow condition */
		if (likely(!ct)) {
			if (mfspr(SPR_L1CSR0) & L1CSR0_DCLO) {
				LWE_ERROR
				    ("%s: lock overflow, L1 cache line "
				     "with addr=%x", __func__, tmp);
				return -1;
			} else if (mfspr(SPR_L1CSR0) & L1CSR0_DCUL) {
				LWE_ERROR
				    ("%s: unable to lock, L1 cache line"
				     " with addr=%x", __func__, tmp);
				return -1;
			}
		} else {
			if (mfspr(SPR_L2CSR0) & L2CSR0_L2LO) {
				LWE_ERROR
				    ("%s: lock overflow, L2 cache line"
				     " with addr=%x", __func__, tmp);
				return -1;
			}
		}
	}

	return 0;
}

int
dcache_unlock(void *addr_start, uint32_t size, enum cache_target_t ct)
{
	uint32_t tmp;
	uint32_t cache_blk_size;

	assert(addr_start);
	assert(size);

	if (unlikely(!guest_cache_lock_on)) {
		LWE_ERROR("%s: guest cache lock mode is disabled!",
			 __func__);
		return -EINVAL;
	}

	switch (ct) {
	case CT_L1_CACHE:
		cache_blk_size = l1_data_cache_blk_size;
		break;

	case CT_L2_CACHE:
		cache_blk_size = l2_cache_blk_size;
		break;

	default:
		LWE_ERROR("%s: unsupported cache target!", __func__);
		return -EINVAL;
	}

	if (cache_blk_size == 0) {
		LWE_WARN
		("%s:could not find required cache parameter in device tree.",
			__func__);
		return -1;
	}

	for (tmp = (uint32_t) addr_start;
	     tmp <= (uint32_t) (addr_start + size); tmp += cache_blk_size) {
		if (ct == CT_L1_CACHE) {
			asm volatile ("dcblc %0,0,%1"::"i" (CT_L1_CACHE),
				      "r"(tmp):"memory");
		} else {
			asm volatile ("dcblc %0,0,%1"::"i" (CT_L2_CACHE),
				      "r"(tmp):"memory");
		}

		asm("msync");

		/* Check for lock bit reset */
		if (likely(!ct)) {
			if (mfspr(SPR_L1CSR0) & L1CSR0_DCLFC) {
				LWE_ERROR
				    ("%s: unable to unlock, L1 cache line"
				     " with addr=%x", __func__, tmp);
				return -1;
			}
		} else if (mfspr(SPR_L2CSR0) & L2CSR0_L2LFC) {
			LWE_ERROR
			    ("%s: unable to unlock, L2 cache line with addr=%x",
			    __func__, tmp);
			return -1;
		}
	}

	return 0;
}

int
icache_lock(void *addr_start, uint32_t size, enum cache_target_t ct)
{
	uint32_t tmp;
	uint32_t cache_blk_size;

	assert(addr_start);
	assert(size);

	if (unlikely(!guest_cache_lock_on)) {
		LWE_ERROR("%s: guest cache lock mode is disabled!",
			 __func__);
		return -EINVAL;
	}

	switch (ct) {
	case CT_L1_CACHE:
		cache_blk_size = l1_inst_cache_blk_size;
		break;

	case CT_L2_CACHE:
		cache_blk_size = l2_cache_blk_size;
		break;

	default:
		LWE_ERROR("%s: unsupported cache target!", __func__);
		return -EINVAL;
	}

	if (cache_blk_size == 0) {
		LWE_WARN
		("%s:could not find required cache parameter in device tree.",
			__func__);
		return -1;
	}

	for (tmp = (uint32_t) addr_start;
	     tmp <= (uint32_t) (addr_start + size); tmp += cache_blk_size) {
		if (ct == CT_L1_CACHE) {
			asm volatile ("icbtls %0,0,%1"::"i" (CT_L1_CACHE),
				      "r"(tmp):"memory");
		} else {
			asm volatile ("icbtls %0,0,%1"::"i" (CT_L2_CACHE),
				      "r"(tmp):"memory");
		}

		asm("msync");

		/* Check unable to lock and overflow condition */
		if (mfspr(SPR_L1CSR1) & L1CSR1_ICLO) {
			LWE_ERROR
			    ("%s: lock overflow, L1 cache line with addr=%x",
			    __func__, tmp);
			return -1;
		} else if (mfspr(SPR_L1CSR1) & L1CSR1_ICUL) {
			LWE_ERROR
			    ("%s: unable to lock, L1 cache line with addr=%x",
			    __func__, tmp);
			return -1;
		} else if (mfspr(SPR_L2CSR0) & L2CSR0_L2LO) {
			LWE_ERROR
			    ("%s: lock Overflow, L2 cache line with addr=%x",
			    __func__, tmp);
			return -1;
		}
	}

	return 0;
}

int
icache_unlock(void *addr_start, uint32_t size, enum cache_target_t ct)
{
	uint32_t tmp;
	uint32_t cache_blk_size;

	assert(addr_start);
	assert(size);

	if (unlikely(!guest_cache_lock_on)) {
		LWE_ERROR("%s: guest cache lock mode is disabled!",
			 __func__);
		return -EINVAL;
	}

	switch (ct) {
	case CT_L1_CACHE:
		cache_blk_size = l1_inst_cache_blk_size;
		break;

	case CT_L2_CACHE:
		cache_blk_size = l2_cache_blk_size;
		break;

	default:
		LWE_ERROR("%s: unsupported cache target!", __func__);
		return -EINVAL;
	}

	if (cache_blk_size == 0) {
		LWE_WARN
		("%s:could not find required cache parameter in device tree.",
			__func__);
		return -1;
	}

	for (tmp = (uint32_t) addr_start;
	     tmp <= (uint32_t) (addr_start + size); tmp += cache_blk_size) {
		if (ct == CT_L1_CACHE) {
			asm volatile ("icblc %0,0,%1"::"i" (CT_L1_CACHE),
				      "r"(tmp):"memory");
		} else {
			asm volatile ("icblc %0,0,%1"::"i" (CT_L2_CACHE),
				      "r"(tmp):"memory");
		}

		asm("msync");

		/* Check for lock bit reset */
		if (mfspr(SPR_L1CSR1) & L1CSR1_ICLFC) {
			LWE_ERROR
			    ("%s: unable to unlock, L1 cache line with addr=%x",
			    __func__, tmp);
			return -1;
		} else if (mfspr(SPR_L2CSR0) & L2CSR0_L2LFC) {
			LWE_ERROR
			    ("%s: unable to unlock, L2 cache line with addr=%x",
			    __func__, tmp);
			return -1;
		}
	}

	return 0;
}
