/** @file barrier.c
 * This file contains definitions of APIs for implementing barriers
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <stdbool.h>
#include <errno.h>
#include <init.h>
#include <misc.h>
#include <bitops.h>
#include <fsl_hcalls.h>
#include <barrier.h>

int
barrier_init(struct lwe_barrier *barrier, uint32_t uImask)
{
	uint32_t coremask = sec_core_mask | 1;
	
	coremask = ~coremask;

	if (unlikely(uImask & coremask))
		return -EINVAL;

	spin_lock(&barrier->lock);
	if (barrier->status != BARRIER_AVAILABLE) {
		spin_unlock(&barrier->lock);
		return -EINVAL;
	}
	barrier->threshold = barrier->mask = uImask;
	barrier->cond_check = 0;
	barrier->status = BARRIER_VALID;
	spin_unlock(&barrier->lock);
	return 0;
}

int
barrier_sync(struct lwe_barrier *barrier)
{
	int id = 1 << get_lwe_id();
	int status = 0;
	if (barrier->status != BARRIER_VALID)
		return -EINVAL;

	if ((id & barrier->mask) == 0)
		return -EINVAL;

	while (barrier->cond_check == 1
	       && barrier->mask != barrier->threshold) ;
	barrier->cond_check = 0;
	int ret = atomic_and((unsigned long *) &barrier->mask, ~id);
	if (ret == 0) {
		barrier->cond_check = 1;
		status = 1;
	} else {
		while (barrier->cond_check == 0) ;
	}
	atomic_or((unsigned long *) &barrier->mask, id);
	return status;
}

int
barrier_destroy(struct lwe_barrier *barrier)
{
	if (barrier->status != BARRIER_VALID)
		return -EINVAL;

	if (barrier->mask != barrier->threshold)
		return -EBUSY;

	barrier->threshold = barrier->mask = 0;
	barrier->cond_check = 0;
	barrier->status = BARRIER_AVAILABLE;
	return 0;
}
