/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdbool.h>
#include <libfdt.h>
#include <UART.h>
#include <log.h>
#include <ldt.h>
#include <percpu.h>
#include <init.h>
#include <hwi.h>
#include <console.h>
#include <compiler.h>
#include <ns16550.h>
#include <epapr_hcalls.h>

struct console_t lwe_console[MAX_NUM_CONSOLES];

/*This function is executed by boot core only*/
int init_consoles(void)
{
	int ret;
	int len;
	int con_node;
	uint32_t index;
	static uint32_t nsobj_idx;

	/*For each byte channel having compatible lwe-stdout, init a console*/
	index = 0;
	con_node = fdt_node_offset_by_compatible(devtree_ptr, -1,
			"fsl,lwe-stdout");

	while ((con_node > 0) && (index < MAX_NUM_CONSOLES)) {

		const uint32_t *ch_handle;
		/*Get the type of console (UART or byte channel)*/
		ret = fdt_node_check_compatible(devtree_ptr, con_node,
						"epapr,hv-byte-channel");

		if (ret == 0) {
			lwe_console[index].driver = HV_BYTE_CHANNEL;
			ch_handle = fdt_getprop(devtree_ptr,
						con_node, "reg", &len);

			if ((NULL == ch_handle) || (len != sizeof(uint32_t)))
				LWE_PANIC("Can't get console channel handle");

			lwe_console[index].ch_handle = *ch_handle;
		} else if (fdt_node_check_compatible(devtree_ptr,
					con_node, "ns16550") == 0) {
			lwe_console[index].driver = UART;

			if (0 > uart_init(nsobj_idx, con_node))
				return -1;

			lwe_console[index].ch_handle = nsobj_idx++;
		} else {
			LWE_WARN("Invalid driver for console");
			return -1;
		}

		lwe_console[index].ch_node = con_node;
		lwe_console[index].log_lock = 0;
		lwe_console[index].console_ok = true;

		/*Stdin not attached yet*/
		lwe_console[index].stdin_handler = NULL;

		/*Find the vcpu which handles interrupts from this device*/
		const uint32_t *vcpu = fdt_getprop(devtree_ptr,
						con_node, "vcpu", &len);
		if ((NULL == vcpu) || (len != sizeof(uint32_t)))
			LWE_WARN("Can't get vcpu which handles interrupts");
		else
			lwe_console[index].stdin_vcpu = *vcpu;

		LWE_INFO("Initialised console ID: %u", index);

		index++;
		if (MAX_NUM_CONSOLES == index) {
			LWE_WARN("LWE supports max: %u consoles",
						MAX_NUM_CONSOLES);
			break;
		}

		con_node = fdt_node_offset_by_compatible(devtree_ptr,
						con_node, "fsl,lwe-stdout");
	}

	if ((index < MAX_NUM_CONSOLES) && (-FDT_ERR_NOTFOUND != con_node))
		LWE_CRIT("fdt_node_offset_by_compatible failed, ecode = %d",
							con_node);

	return index;
}

/*This function is executed by all cores to attach their stdout to a console*/
int subscribe_console(void)
{
	int node_off;
	uint32_t index;
	char alias_name[32];
	static uint32_t nsobj_idx;
	int ret;

	sprintf(alias_name, "stdout_vcpu%d", lwe_instance);

	ret = get_stdout_handle(alias_name, &node_off);
	if (ret < 0) {
		/* If search for private stdout hadnle fails search for
		 *  the stdout property */
		ret = get_stdout_handle("stdout", &node_off);
		if (ret < 0) {
			LWE_WARN("No console for vcpu: lwe_instance");
			return 0;
		}
	}

	/*Now hookup stdout with console which has matching driver handle*/
	for (index = 0; index < MAX_NUM_CONSOLES; index++) {
		if (lwe_console[index].ch_node == node_off)
			break;
	}

	if (MAX_NUM_CONSOLES == index) {
		LWE_ERROR("Could not find console for handle: %u", node_off);
		return -1;
	}

	if (lwe_console[index].driver == UART) {
		nsobj_idx = lwe_console[index].ch_handle;
		get_uart_va(&uart_obj[nsobj_idx], lwe_instance);
	}

	/*Mark the CPU console as UP*/
	LWE_INFO("Subscribing to console ID: %u", index);

	cpu->console_id = index;
	cpu->console_ok = 1;

	/*Flush log buffer*/
	print_log_buf();

	return 0;
}

/*This function is used by LWE logger modules to display log
messages on console*/
int console_log(const char *str, size_t log_size)
{
	int ret = 0;
	register_t saved;

	/*Lock the console so that logs don't get mixed*/
	saved = spin_lock_intsave(&lwe_console[cpu->console_id].log_lock);

	WriteUARTN(str, log_size);

	spin_unlock_intsave(&lwe_console[cpu->console_id].log_lock, saved);

	return ret;
}

static uint32_t console_intr_handler(int interrupt __UNUSED, void *arg __UNUSED)
{
	uint32_t console_id;
	stdin_intr_handler_t stdin_handler;

	console_id = cpu->console_id;
	stdin_handler = lwe_console[console_id].stdin_handler;

	/*Notify the shell*/
	stdin_handler();

	return 0;
}

/*This function would be used by shell to attach stdin to a console*/
int attach_console_input(stdin_intr_handler_t stdin_handler)
{
	uint32_t console_id;
	uint32_t stdin_vcpu;
	int ch_node;
	int len;
	const uint32_t *intr;

	assert(NULL != stdin_handler);

	/*Check if this core handles interrupts from its console*/
	console_id = cpu->console_id;
	stdin_vcpu = lwe_console[console_id].stdin_vcpu;

	if (lwe_instance != stdin_vcpu) {
		LWE_ERROR("Stdin of console is handled on vcpu: %u",
				stdin_vcpu);
		return -1;
	}

	if (lwe_console[console_id].stdin_handler) {
		LWE_ERROR("Stdin (console: %u) already busy", console_id);
		return -1;
	}

	ch_node = lwe_console[console_id].ch_node;
	intr = fdt_getprop(devtree_ptr, ch_node, "interrupts", &len);
	if (NULL == intr) {
		LWE_ERROR("interrupts property not found");
		return -1;
	}

	if (register_hwi(intr[0], console_intr_handler, NULL, 0) < 0) {
		LWE_ERROR("Can not register HWI for interrupt: %u", intr[0]);
		return -1;
	}

	lwe_console[console_id].stdin_handler = stdin_handler;

	if (enable_hwi(intr[0]) < 0) {
		LWE_ERROR("Could not enable interrupt: %u", intr[0]);
		return -1;
	}


	return 0;
}

/*This function is to be used by shell stdin callback to read 1 byte
when in ISR context*/
int console_gets(char *cmd_buf, uint32_t len)
{
	uint32_t count = 0;

	if (len == 0)
		return -1;

	if (HV_BYTE_CHANNEL == lwe_console[cpu->console_id].driver) {
		uint32_t ch_handle = lwe_console[cpu->console_id].ch_handle;
		uint32_t tx_avail, rx_avail = 0;

		if (ev_byte_channel_poll(ch_handle, &rx_avail, &tx_avail))
			return -1;

		if (rx_avail > len) {
			LWE_WARN("Command length exceeds the limit\n");
			return -1;
		}

		count = rx_avail;
	} else {
		count = 1;
	}

	ReadUARTN(cmd_buf, count);

	return count;
}

/*This function is to be used by shell stdin callback to write byte
string when in ISR context*/
int console_put(uint8_t *str, size_t size)
{
	if (kUARTNoError == WriteUARTN(str, size))
		return 0;
	else
		return -1;
}
