/** @file debug_cmds.c
 * Performance Monitoring command's handler
 */

/*
 * Copyright (c) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* For all the feilds refer to P4080RM_RevG.pdf */

#include "compat.h"
#include <linux/fsl_qman.h>
#include <limits.h>
#include <errno.h>
#include <fsl_hcalls.h>

extern uint32_t coremask;

int shell_cmd_qman_query_fq(int argc, char *argv[])
{
	struct qm_fqd fqd;
	struct qman_fq fq;
	uint16_t cong_mgt, tail_drop, orp, stashing, cpc_stash;
	int ret;

	if (argc != 2) {
		printf("Usage: Query Qman Frame Queue <arg>\n");
		printf("Valid argument<arg> is Frame Queue ID\n");
		return -1;
	}

	errno = 0;
	fq.fqid = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == fq.fqid)
			|| (0 != errno && 0 == fq.fqid)) {
		printf("strtoul failed converting command's argument(fqid).\n");
		return -1;
	}

	if (fq.fqid > 0x00ffffff) {
		printf("Please enter valid Frame Queue ID\n");
		return 0;
	}

	ret = qman_query_fq(&fq, &fqd);

	if (ret) {
		printf("ERROR in Querying QMAN\n");
		return 0;
	}

	printf("\n\n---------Frame Queue Descriptor Data----------\n\n");

	cong_mgt = fqd.fq_ctrl & QM_FQCTRL_CGE;
	if (cong_mgt)
		printf("(FQ_CTRL) WRED and CS tail drop congession management "
			"is enabled for this FQ\n");
	else
		printf("(FQ_CTRL) WRED and CS tail drop congession management "
			"is disabled for this FQ\n");

	if (cong_mgt)
		printf("\t(CONG_ID) Congession Group ID (to which "
			"this FQ belongs: %d\n", fqd.cgid);

	tail_drop = fqd.fq_ctrl & QM_FQCTRL_TDE;
	if (tail_drop) {
		printf("(FQ_CTRL) FQ tail drop congession management is "
			"enabled for this FQ\n");
		printf("\t(TD_MANT) Tail Drop Threshold Mantesa: %d\n",
			fqd.td.mant);
		printf("\t(TD_EXP) Tail Drop Threshold Exponent: %d\n",
			fqd.td.exp);
	} else {
		printf("(FQ_CTRL) FQ tail drop congession management is "
			"disabled for this FQ\n");
	}

	orp = fqd.fq_ctrl & QM_FQCTRL_ORP;
	if (orp) {
		printf("(FQ_CTRL) Order Restoration using this FQID as an "
			"ORP is enabled\n");

		printf("\t(ORPRWS) ORP Restoration Window Size is %d frames\n",
			0x20 << fqd.orprws);

		if (fqd.oa)
			printf("\t(OA) ORP Auto Advance NESN Window Size is "
				"same as ORP Restoration Window Size\n");
		else
			printf("\t(OA) NESN Auto Advance is Disabled\n");

		printf("\t(OLWS) ORP Acceptable Late Arrival ");
		if (0 == fqd.olws)
			printf("is Disabled\n");
		else if (1 == fqd.olws)
			printf("Window Size is 32 frames\n");
		else if (2 == fqd.olws)
			printf("Window Size is same as ORP Restoration "
				"Window Size\n");
		else
			printf("Window Size is 8192 frames\n");
	} else {
		printf("(FQ_CTRL) Order Restoration using this FQID as an "
			"ORP is disabled\n");
	}

	stashing = fqd.fq_ctrl & QM_FQCTRL_CTXASTASHING;
	if (stashing) {
		printf("(FQ_CTRL) Context A is used for stashing\n");
		printf("\tNo of Cache Lines for\n");
		printf("\t\tAnnotation: %d\n",
			fqd.context_a.stashing.annotation_cl);
		printf("\t\tData: %d\n", fqd.context_a.stashing.data_cl);
		printf("\t\tContext: %d\n", fqd.context_a.stashing.context_cl);
	} else {
		printf("(FQ_CTRL) Context A is not used for stashing\n");
		printf("(CONTEXT_A) Frame Queue Context A 0x%llx\n",
			fqd.context_a.opaque);
	}

	cpc_stash = fqd.fq_ctrl & QM_FQCTRL_CPCSTASH;
	if (cpc_stash)
		printf("(FQ_CTRL) CPC stashing is enabled\n");
	else
		printf("(FQ_CTRL) CPC stashing is disabled\n");

	printf("(FQ_CTRL) Force SFDR Allocate: %d\n",
		fqd.fq_ctrl & QM_FQCTRL_FORCESFDR);
	printf("(FQ_CTRL) Avoid Blocking: %d\n",
		fqd.fq_ctrl & QM_FQCTRL_AVOIDBLOCK);
	printf("(FQ_CTRL) Hold Active: %d\n",
		fqd.fq_ctrl & QM_FQCTRL_HOLDACTIVE);
	printf("(FQ_CTRL) Prefer in Cache: %d\n",
		fqd.fq_ctrl & QM_FQCTRL_PREFERINCACHE);

	printf("(DEST_WQ) Destination Work Queue Channel No: %d\n",
		fqd.dest.channel);
	printf("(DEST_WQ) Work Queue no. within the channel: %d\n",
		fqd.dest.wq);

	printf("(ICS_CRED) Intra-Class Scheduling Credit (No. of bites to "
		"consume from FQ): %d\n", fqd.ics_cred);

	printf("(CONTEXT_B) Frame Queue Context B: %#x\n", fqd.context_b);

	return ret;
}

int shell_cmd_qman_query_fq_np(int argc, char *argv[])
{
	struct qm_mcr_queryfq_np np;
	struct qm_fqd fqd;
	struct qman_fq fq;
	uint16_t orp;
	int ret1, ret2;

	if (argc != 2) {
		printf("usage: Query Qman Frame Queue <arg>\n");
		printf("Valid argument<arg> is Frame Queue ID\n");
		return -1;
	}

	errno = 0;
	fq.fqid = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == fq.fqid)
			|| (0 != errno && 0 == fq.fqid)) {
		printf("strtoul failed converting command's argument(fqid).\n");
		return -1;
	}

	if (fq.fqid > 0x00ffffff) {
		printf("Please enter valid Frame Queue ID\n");
		return 0;
	}

	ret1 = qman_query_fq(&fq, &fqd);
	if (ret1) {
		printf("ERROR in Querying QMAN\n");
		return -1;
	}

	ret2 = qman_query_fq_np(&fq, &np);
	if (ret2) {
		printf("ERROR in Querying QMAN\n");
		return -1;
	}

	printf("\n\n---------Non Programable Frame Queue Data----------\n\n");

	printf("(FQD_LINK) Link to the next FQD: %d\n", np.fqd_link);
	printf("(ODP_SEQ) ODP Sequence Number: %d\n", np.odp_seq);

	orp = fqd.fq_ctrl & QM_FQCTRL_ORP;
	if (orp) {
		printf("(ORP_NSEN) ORP Next Expected Sequence Number: %d\n",
			np.orp_nesn);
		printf("(ORP_EA_HSEQ) ORP Early Arrival Head Sequence Number: "
			"%d\n",	np.orp_ea_hseq);
		printf("(ORP_EA_TSEQ) ORP Early Arrival Tail Sequence Number: "
			"%d\n",	np.orp_ea_tseq);

		printf("(ORP_EA_HPTR) ORP Early Arrival Head "
			"Pointer: 0x%06x\n", np.orp_ea_hptr);
		printf("(ORP_EA_TPTR) ORP Early Arrival Tail "
			"Pointer: 0x%06x\n", np.orp_ea_tptr);
	}

	printf("(PFDR_HPTR) Packed Frame Description Record Head Pointer: "
		"0x%06x\n", np.pfdr_hptr);
	printf("(PFDR_TPTR) Packed Frame Description Record Tail Pointer: "
		"0x%06x\n", np.pfdr_tptr);

	printf("(IS) Intra Class Scheduling is ");
	if (np.is)
		printf("Surplus\n");
	else
		printf("Defict\n");

	printf("(ICS_SURP) No of bytes to consume from FQ for current "
		"Dequeue event: %d\n", np.ics_surp);
	printf("(BYTE_CNT) Byte Count: %d\n", np.byte_cnt);
	printf("(FRM_CNT) Frame Count: %d\n", np.frm_cnt);

	printf("(RA1_SFDR) SFDR Pointer for Recently Arrived frame number 1: "
		"0x%04x\n", np.ra1_sfdr);
	printf("(RA2_SFDR) SFDR Pointer for Recently Arrived frame number 2: "
		"0x%04x\n", np.ra2_sfdr);
	printf("(OD1_SFDR) SFDR Pointer for On Deck frame number 1: 0x%04x\n",
		np.od1_sfdr);
	printf("(OD2_SFDR) SFDR Pointer for On Deck frame number 2: 0x%04x\n",
		np.od2_sfdr);
	printf("(OD3_SFDR) SFDR Pointer for On Deck frame number 3: 0x%04x\n",
		np.od3_sfdr);

	return ret2;
}

char *get_state(uint32_t vcpu)
{
	int ret = 0;
	unsigned int state;
	ret = fh_get_core_state(-1, vcpu, &state);
	if (ret) {
		printf("Error in getting state of the core\n");
		return NULL;
	}

	switch (state) {
	case FH_VCPU_NAP:
		return "NAP";
	case FH_VCPU_IDLE:
		return "IDLE";
	case FH_VCPU_RUN:
		return "RUNNING";
	default:
		return NULL;
	}

	return NULL;
}

int shell_cmd_get_core_state(int argc, char *argv[])
{
	unsigned long core_num;
	const char *state;
	if (argc > 2) {
		printf("Usage: gcs [core_num]>\n\n");
		return -1;
	}

	printf("\tvcpu\t\t\tstate\n");
	printf("\t-----------------------------------\n");
	if (argc == 2) {
		core_num  = strtoul(argv[1], NULL, 0);

		if (!((1 << core_num) & coremask)) {
			printf("Error -- Core Number not valid\n\n");
			return -1;
		}

		state = get_state(core_num);
		if (!state)
			return -1;

		printf("\t  %lu   \t\t\t%s\n", core_num, state);

	} else {

		for (int i = 0; i < get_num_cores(); i++) {
			if (!((1 << i) & coremask))
				continue;
			state = get_state(i);
			if (!state)
				return -1;

			printf("\t  %d   \t\t\t%s\n", i, state);
		}
	}
	return 0;
}
