/**
\file           
\brief          DPA buffer allocator routines
*/
/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <dpa_alloc_private.h>
#include <lwe_malloc.h>
#include <hv_misc.h>
#include <compiler.h>

struct dpa_buff_allocator *
dpa_allocator_create(int *ecode __UNUSED)
{
	uint32_t index;
	struct dpa_buff_allocator *allocator;

	/*This should come from private heap only */
	allocator = tlmalloc(sizeof (struct dpa_buff_allocator));
	if (unlikely(NULL == allocator)) {
		/**ecode = -ENOMEM;*//*To be fixed */
		return NULL;
	}

	for (index = 0; index < MAX_NUM_BMAN_POOLS; index++)
		allocator->bpool[index] = NULL;

	for (index = 0; index < MAX_NUM_BMAN_POOLS; index++)
		allocator->bpid_to_size[index] = 0;

	for (index = 0; index <= DPA_ALLOC_MAX_BUFF_SIZE; index++)
		allocator->size_to_bpid[index] = INV_BPID;

	return allocator;
}

void
dpa_allocator_free(struct dpa_buff_allocator *allocator)
{
	tlfree(allocator);
}

int
dpa_allocator_add_bpool(struct dpa_buff_allocator *allocator,
			const uint32_t size,
			const uint32_t bpid, struct bman_pool *bpool)
{
	uint32_t index;
	uint32_t old_bpid;

	assert(allocator);
	assert(bpool);
	assert(bpid <= MAX_BPID);

	if (unlikely(size > DPA_ALLOC_MAX_BUFF_SIZE))
		return -EINVAL;	/*Size too big */

	if (unlikely(NULL != allocator->bpool[bpid]))
		return -EINVAL;	/*Same pool ID already registered */

	index = size;

	old_bpid = allocator->size_to_bpid[index];

	for (; index > 0; index--) {	/*Oth index not used */
		if (old_bpid != allocator->size_to_bpid[index])
			break;

		allocator->size_to_bpid[index] = bpid;
	}

	allocator->bpid_to_size[bpid] = size;
	allocator->bpool[bpid] = bpool;

	return 0;
}

int
dpa_allocator_del_bpool(struct dpa_buff_allocator *allocator,
			const uint32_t bpid)
{
	uint32_t index;
	uint32_t size;
	uint32_t next_bpid;

	assert(allocator);

	if (unlikely(NULL == allocator->bpool[bpid]))
		return -EINVAL;

	size = allocator->bpid_to_size[bpid];

	assert(size);
	assert(size <= DPA_ALLOC_MAX_BUFF_SIZE);

	index = size;

	if (unlikely(size == DPA_ALLOC_MAX_BUFF_SIZE))	/*Max possible size */
		next_bpid = INV_BPID;
	else
		next_bpid = allocator->size_to_bpid[size + 1];

	do {
		if (likely(bpid == allocator->size_to_bpid[index]))
			allocator->size_to_bpid[index] = next_bpid;
		else
			break;	/*There is some other BPID here */

		index--;
	} while (index);	/*0th index is unused */

	allocator->bpid_to_size[bpid] = 0;
	allocator->bpool[bpid] = NULL;

	return 0;
}

int
dpa_allocator_get_buff(const struct dpa_buff_allocator *allocator,
		       const uint32_t size, struct bm_buffer *buf)
{
	uint32_t bpid;
	uint32_t index;
	int ret;

	assert(buf);
	assert(allocator);

	if (unlikely(size > DPA_ALLOC_MAX_BUFF_SIZE))
		return -EINVAL;

	bpid = allocator->size_to_bpid[size];

	/*Try to acquire buffer from the BPID queried */
	while (likely(INV_BPID != bpid)) {
		assert(allocator->bpool[bpid]);

		ret = bman_acquire(allocator->bpool[bpid], buf, 1, 0);
		if (likely(1 == ret)) {
			/*bman_acquire does not fill correct bpid*/
			buf->bpid = bpid;

			/*Returns the size of buffer allocated */
			return allocator->bpid_to_size[bpid];
		}

		/*Else try next next higher size available */
		index = allocator->bpid_to_size[bpid] + 1;
		if (unlikely(index >= DPA_ALLOC_MAX_BUFF_SIZE)) {
			return -EINVAL;
		}

		bpid = allocator->size_to_bpid[index];
	}

	return -1;		/*No bpool can satisfy size requested */
}

int
dpa_allocator_free_buff(const struct dpa_buff_allocator *allocator,
			const struct bm_buffer *buf)
{
	uint32_t bpid;
	struct bman_pool *bpool;

	assert(buf);

	bpid = buf->bpid;

	assert(bpid <= MAX_BPID);

	/* Return from dpa_allocator_free_buff as the buffer is already
	* free/not managed by BMAN */
	if (unlikely((0 == buf->lo) && (0 == buf->hi)))
		return 0;

	bpool = allocator->bpool[bpid];
	if (unlikely(NULL == bpool))	/*Pool not registered */
		return -EINVAL;

	assert(allocator->bpid_to_size[bpid]);

	return bman_release(bpool, buf, 1, 0);
}
