/** @file
 * This file provides APIs for querying guest device tree. These APIs
 * read memory windows, interrupts for the hardware portals
 */

/*
 * Copyright (c) 2008 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <io.h>
#include <stdlib.h>
#include <libfdt.h>
#include <log.h>
#include <ldt.h>
#include <fsl_hcalls.h>
#include <log.h>
#include <init.h>
#include <compiler.h>

/** get_addr_format - Reads "#address-cell" and "#size-cell" property
 * of given node offset from lwe device tree 
 *
 * @param[in] fdt Guest device tree pointer as passed from the hypervisor
 * @param[in] offset Offset of the node whose property to find
 * @param[in] naddr #address-cell property value i.e no. of <u32> cells used
 *			  to encode the address field in a child node's reg property
 * @param[in] nsize #size-cell property value i.e no. of <u32> cells used
 * 			  to encode the size field in a child node's reg property
 */
static int
get_addr_format(const void *fdt, int offset, uint32_t * naddr, uint32_t * nsize)
{
	const uint32_t *prop;
	int len;

	*naddr = 2;
	*nsize = 1;

	prop = fdt_getprop(fdt, offset, "#address-cells", &len);
	if (prop) {
		if (len != sizeof (uint32_t)) {
			LWE_ERROR("Bad #addresss-cells");
			return -1;
		}

		*naddr = *prop;
		if (*naddr > MAX_ADDR_CELLS) {
			LWE_ERROR("Bad #addresss-cells");
			return -1;
		}
	}

	prop = fdt_getprop(fdt, offset, "#size-cells", &len);
	if (prop) {
		if (len != sizeof (uint32_t)) {
			LWE_ERROR("Bad #size-cells");
			return -1;
		}

		*nsize = *prop;
		if (*nsize > MAX_SIZE_CELLS) {
			LWE_ERROR("Bad #size-cells");
			return -1;
		}
	}

	return 0;
}

int
calc_ch_handle(void)
{
	int len;
	int node;

	if ((node =
	     fdt_node_offset_by_compatible(devtree_ptr, -1,
					   "epapr,hv-byte-channel")) < 0)
		return -1;

	const int *prop = fdt_getprop(devtree_ptr, node, "reg", &len);
	if (!prop)
		return -1;

	return *prop;
}

static int
get_parent_bus_addr(int node, phys_addr_t *start,
		phys_addr_t *size __UNUSED, int idx,
		uint32_t caddr, uint32_t csize)
{
	uint32_t naddr, nsize;
	int parent;
	int len;

	const uint32_t *memranges =
	    fdt_getprop(devtree_ptr, node, "ranges", &len);
	if (memranges == NULL) {
		/* Remving the error check from here as currently 
		 * root node also has compatibility property simple-bus.
		 * To correct it once we get confirmation from Hypervisor team */
#if 0
		LWE_ERROR
		    ("get_parent_bus_addr: ERROR reading ranges property\n");
		return ERR_NOTFOUND;
#endif
		return 0;
	}

	/* If ranges property is present and  it is empty, 
	 * parent and child-address are identical, no translation required */
	if (len == 0)
		return 0;

	parent = fdt_parent_offset(devtree_ptr, node);
	if (get_addr_format(devtree_ptr, parent, &naddr, &nsize) < 0) {
		LWE_ERROR("get_parent_bus_addr: Wrong address format");
		return -1;
	}

	if (naddr < 1 || naddr > 2 || nsize < 1 || nsize > 2) {
		LWE_ERROR
		    ("get_parent_bus_addr: bad address format addr = %x ,size %x\n",
		     naddr, nsize);
		return -1;
	}

	int num_of_words = len / sizeof (uint32_t);
	if (num_of_words % (caddr + naddr + csize)) {
		LWE_ERROR("get_parent_bus_addr: Wrong address format");
		return -1;
	}

	int num_of_trip = num_of_words / (caddr + naddr + csize);
	struct dev_range *range =
	    malloc(sizeof (struct dev_range) * num_of_trip);

	for (int i = 0; i < num_of_trip; i++) {
		range[i].chaddr = *memranges++;
		if (caddr == 2) {
			range[i].chaddr <<= 32;
			range[i].chaddr |= *memranges++;
		}
		range[i].phaddr = *memranges++;
		if (naddr == 2) {
			range[i].phaddr <<= 32;
			range[i].phaddr |= *memranges++;
		}
		range[i].chsize = *memranges++;
		if (csize == 2) {
			range[i].chsize <<= 32;
			range[i].chsize |= *memranges++;
		}
	}

	for (int i = 0; i < idx; i++) {
		for (int j = 0; j < num_of_trip; j++) {
			if (start[i] >= range[j].chaddr
			    && start[i] < range[j].chaddr + range[j].chsize) {
				start[i] =
				    range[j].phaddr + (start[i] -
						       range[j].chaddr);
				break;
			}
		}
	}
	free(range);

	return 0;

}

/** get_fdt_mem_ranges - Reads the (64b) base address and (64b) size of the
 *  memory window defined in the lwe device tree
 *
 * @param[in] node Handle to device-tree node
 * @param[out] start Pointer to start address of memory/register ranges
 * @param[out] size Pointer to size of memory/register ranges
 * @return Success: 0; Fail: -ive value. See errors.h for details
 */
int
get_fdt_mem_ranges(int node, phys_addr_t * start, phys_addr_t * size)
{
	uint32_t naddr, nsize;
	int parent;
	int len;
	phys_addr_t addr;
	uint64_t msize;

	parent = fdt_parent_offset(devtree_ptr, node);
	if (get_addr_format(devtree_ptr, parent, &naddr, &nsize) < 0) {
		LWE_ERROR("get_fdt_mem_ranges: Wrong address format");
		return -1;
	}

	if (naddr < 1 || naddr > 2 || nsize < 1 || nsize > 2) {
		LWE_ERROR
		    ("get_fdt_mem_ranges: bad address format addr = %x ,size %x\n",
		     naddr, nsize);
		return -1;
	}

	const uint32_t *memreg = fdt_getprop(devtree_ptr, node, "reg", &len);
	if (!memreg) {
		LWE_ERROR("get_fdt_mem_ranges: ERROR reading reg property\n");
		return -1;
	}

	uint8_t i = 0;
	const uint32_t *reg = memreg;
	while ((reg + naddr + nsize) <= (memreg + len / 4)) {
		addr = *reg++;
		if (naddr == 2) {
			addr <<= 32;
			addr |= *reg++;
		}

		msize = *reg++;
		if (nsize == 2) {
			msize <<= 32;
			msize |= *reg++;
		}

		start[i] = addr;
		size[i] = msize;
		i++;
	}

	int ret = fdt_node_check_compatible(devtree_ptr, parent, "simple-bus");
	if (ret == 0) {
		ret = get_parent_bus_addr(parent, start, size, i, naddr, nsize);
	}

	return 0;
}

/** get_fdt_irq - Reads the interrupt number and its type
 *  as defined in the LWE device tree
 *
 * @param[in] node Handle to device-tree node
 * @param[out] irq IRQ number to be read
 * @param[out] type IRQ type (polarity) to be read
 * @return Success: 0; Fail: -ive value.
 */
int
get_fdt_irq(int node, int *irq, int *type)
{
	int len;
	const int *irq_handle;

	irq_handle = fdt_getprop(devtree_ptr, node, "interrupts", &len);
	if (!irq_handle) {
		LWE_ERROR("get_fdt_irq: error reading interrupts");
		return -1;
	}

	*irq = *irq_handle++;
	if (type)
		*type = *irq_handle;
	return 0;
}

/** get_shutdn_dbell_irq - Reads interrupt number of receive shutdown
 *  doorbell in LWE device tree
 *
 *  @param[out] irq IRQ number to be read
 *  @return Success: IRQ number; Fail: -ve value.
 */
int get_shutdn_dbell_irq(void)
{

	int node = -1;
	int irq;

	node = fdt_node_offset_by_compatible(devtree_ptr, node,
			"fsl,hv-shutdown-doorbell");
	if (0 > node) {
		LWE_ERROR("offset by compatible for shutdown doorbell failed");
		return -1;
	}

	if (0 > get_fdt_irq(node, &irq, NULL)) {
		LWE_ERROR("Failed in getting IRQ number of shutdown-doorbell");
		return -1;
	}

	return irq;
}

/** get_fdt_int - Reads a 32-bit property from a node
 *
 * @param[in] node Handle to device-tree node
 * @param[in] propname Name of the device tree property to read
 * @param[out] val The value read from the property
 * @return Success: 0; Fail: -ive value.
 */
int
get_fdt_int(int node, const char *propname, int *val)
{
	int len;
	const int *ptr;

	ptr = fdt_getprop(devtree_ptr, node, propname, &len);
	if (!ptr || (len != 4))
		return -1;
	*val = *ptr;
	return 0;
}

/** get_stdout_handle - Returns the byte channel handle or UART handle
 * to be used for stdout by reading it from alias subnode in LWE device tree 
 *  
 *	@param[in] alias_name - Property to be searched in alias node for stdout
 * 	@param[out] node_off - offset of the node specified in
 * 				the alias property.
 *	@return 0 on Success and < 0 on Failure. 
 */
int
get_stdout_handle(const char *alias_name, int *node_off)
{
	void *fdt = (void *) devtree_ptr;
	int node;
	int len = 0;
	*node_off = 0;

	node = fdt_subnode_offset(fdt, 0, "aliases");
	if (node < 0) {
		return -1;
	}

	const char *output = fdt_getprop(fdt, node, alias_name, &len);
	if (!output) {
		return -1;
	}

	node = fdt_path_offset(fdt, output);
	if (node < 0) {
		return -1;
	}

	*node_off = node;
	return 0;
}

/** get_tbfreq - Reads the timebase frequency from the LWE device tree.
 *  
 *     @return 0 on Failure & timebase frequency ( > 0)on Success. 
 */
uint32_t
get_tbfreq(void)
{
	int node;
	uint32_t tb_freq = 0;
	unsigned int core_no;

	core_no = mfspr(SPR_PIR);

	/*Read timebase freq from the devtree */
	if ((node =
	     fdt_node_offset_by_prop_value(devtree_ptr, -1, "device_type",
					   "cpu", strlen("cpu") + 1)) < 0) {
		LWE_ERROR("Missing cpu node");
		return 0;
	}

	do {
		int len = 0;

		const uint32_t *reg =
		    fdt_getprop(devtree_ptr, node, "reg", &len);
		if (!reg || len != sizeof (uint32_t)) {
			LWE_ERROR("Missing reg property in cpu node");
			continue;
		}
		if (*reg != core_no) {
			continue;
		}

		const uint32_t *freq =
		    fdt_getprop(devtree_ptr, node, "timebase-frequency", &len);
		if (!freq || len != sizeof (uint32_t)) {
			LWE_ERROR("Reading timebase freq failed");
			break;
		}
		tb_freq = *freq;
		break;

	} while ((node = fdt_node_offset_by_prop_value(devtree_ptr, node,
						       "device_type", "cpu",
						       strlen("cpu") + 1)) !=
		 -FDT_ERR_NOTFOUND);
	return tb_freq;
}

/** get_sec_core_mask - Returns the secondary cores mask
 *  
 *     @return 0 on Failure & mask on Success. 
 */
uint32_t
get_sec_core_mask(void)
{
	int node = -1;
	int mask = 0;

	node = fdt_node_offset_by_prop_value(devtree_ptr, -1, "device_type",
					     "cpu", strlen("cpu") + 1);
	if (node < 0) {
		LWE_ERROR("Missing cpu node");
		return mask;
	}

	do {
		int len = 0;
		const char *str;

		str = fdt_getprop(devtree_ptr, node, "status", &len);
		if (!str || strcmp(str, "disabled") != 0)
			continue;

		str = fdt_getprop(devtree_ptr, node, "enable-method", &len);
		if (!str) {
			LWE_ERROR("BROKEN: Missing enable-method on"
				  "disabled cpu node");
			continue;
		}
		if (strcmp(str, "spin-table") != 0) {
			LWE_ERROR
			    ("BROKEN: Unknown enable-method \"%s\";not enabling\n",
			     str);
			continue;
		}
		const uint32_t *reg = fdt_getprop(devtree_ptr, node, "reg",
						  &len);
		if (!reg || len != sizeof (uint32_t)) {
			LWE_ERROR("BROKEN: Missing reg property in cpu" "node");
			continue;
		}
		mask |= 1 << *reg;
	} while ((node = fdt_node_offset_by_prop_value(devtree_ptr, node,
						       "device_type", "cpu",
						       strlen("cpu") + 1)) !=
		 -FDT_ERR_NOTFOUND);
	return mask;
}

/** get_bootargs_mask - Reads the coresmask passed in
 * bootargs property of chosen node 
 *  
 *    @param[out] mask - mask read from the bootargs property
 *     @return < 0 on Failure & 0 on Success. 
 */
int
get_bootargs_coremask(uint32_t * mask)
{
	int len = 0;
	int flag = 0;
	char str[100];
	char *result = NULL;
	int ret = -1;

	int node = fdt_subnode_offset(devtree_ptr, 0, "chosen");
	if (node < 0) {
		LWE_WARN("Missing chosen node");
		return ret;
	}

	const char *bargs = fdt_getprop(devtree_ptr, node, "bootargs", &len);
	if (bargs == NULL) {
		LWE_WARN("Missing bootargs property in chosen node");
		return ret;
	}

	strcpy(str, bargs);
	result = strtok((char *) str, " =");
	while (result != NULL) {
		if (flag) {
			*mask = (uint32_t) strtoul((result), NULL, 0);
			ret = 0;
			break;
		}
		if (strcmp(result, "coremask") == 0)
			flag = 1;

		result = strtok(NULL, " =");
	}

	return ret;
}

/** Reads L1 data/inst cache or L2 cache block size from device tree
 *
 * @param[in]	cache_type - L1_DATA_CACHE for L1-data cache,
 * 							 L1_INST_CACHE for L1-inst cache,
 * 							 L2_CACHE for L2 cache
 *
 * @return   Success: Block Size; Fail: -1
 */
int
get_cache_blk_size(enum cache_type_t cache_type)
{
	int node;
	uint32_t cache_blk_size = 0;
	const int *prop;
	unsigned int core_no;

	core_no = mfspr(SPR_PIR);

	if ((node =
	     fdt_node_offset_by_prop_value(devtree_ptr, -1, "device_type",
					   "cpu", strlen("cpu") + 1)) < 0) {
		LWE_ERROR("Missing cpu node");
		return -1;
	}

	do {
		int len = 0;

		const uint32_t *reg =
		    fdt_getprop(devtree_ptr, node, "reg", &len);
		if (!reg || len != sizeof (uint32_t)) {
			LWE_ERROR("Missing reg property in cpu node");
			continue;
		}
		if (*reg != core_no) {
			continue;
		}

		if (cache_type == L1_DATA_CACHE) {
			if ((prop =
			     fdt_getprop(devtree_ptr, node,
					 "d-cache-block-size", &len)) == NULL)
				return -1;

			cache_blk_size = *prop;
			break;
		}

		if (cache_type == L1_INST_CACHE) {
			if ((prop =
			     fdt_getprop(devtree_ptr, node,
					 "i-cache-block-size", &len)) == NULL)
				return -1;

			cache_blk_size = *prop;
			break;
		}

		if ((node =
		     fdt_subnode_offset(devtree_ptr, node, "l2-cache")) < 0)
			return -1;

		if ((prop =
		     fdt_getprop(devtree_ptr, node, "cache-block-size",
				 &len)) == NULL)
			return -1;

		cache_blk_size = *prop;
		break;

	} while ((node = fdt_node_offset_by_prop_value(devtree_ptr, node,
						       "device_type", "cpu",
						       strlen("cpu") + 1)) !=
		 -FDT_ERR_NOTFOUND);

	return cache_blk_size;
}

/** Provides the current LWE partition's name
 *  param[in] 	ecode - error code pointer, populated with error code
 *		in case of failure
 *  return	Returns the name/label of the partition in case of success
 *		NULL- In case of failure
 */
const char *get_part_name(int *ecode)
{
	const char *label;
	int len;

	label = (const char *)fdt_getprop(devtree_ptr, 0, "label", &len);

	if (!label) {
		LWE_ERROR("Label property not found.");
		*ecode = len;
		return NULL;
	}

	return label;
}
