/** @file
 * LWE logging implementation
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <stdarg.h>
#include <percpu.h>
#include <log.h>
#include <hwi.h>
#include <statistics.h>
#include <console.h>

#define LOG_BUF_WRAP_THRESHOLD	32
#define LOG_TYPE_STR_LEN	16


__PERCPU enum log_level_t enabled_log_level[NUM_LOG_TYPES];
__PERCPU char log_buf[LOG_BUF_SIZE + 1]; /*Reserve extra byte for NULL*/
__PERCPU char *log_buf_ptr;
__PERCPU enum log_dest_t log_destination;
__STATS uint32_t g_log_seq_no;

static const char *log_level_str[NUM_LOG_LEVELS] = {
			"OFF",
			"CRIT",
			"ERROR",
			"WARN",
			"INFO",
			"DEBUG"
		};

static char log_type_str[NUM_LOG_TYPES][LOG_TYPE_STR_LEN + 1];

void printlog(enum log_level_t level, enum log_type_t type,
		const char *fmt, ...)
{
	int log_size;
	int rem_size;
	int ret;
	va_list args;
	char *log_start;
	uint32_t seq_no;
	register_t saved;

	saved = disable_int_save();

	/*Get and increment the global log sequence number*/
	seq_no = decorated_load_inc_32(&g_log_seq_no);

	/*Check if log buffer needs to wrap around due to lack of space*/
	rem_size = log_buf + LOG_BUF_SIZE - log_buf_ptr;
	if (unlikely(rem_size < LOG_BUF_WRAP_THRESHOLD)) {
try_again:
		log_buf_ptr = log_buf;
		rem_size = LOG_BUF_SIZE;
	} else if (unlikely(log_buf_ptr != log_buf)) {
		log_buf_ptr--;	/*Discard previous NULL*/
	}

	log_start = log_buf_ptr;

	/*Prepend log message with cpu_id and severity*/
	if (unlikely(in_irq))
		log_size = isr_sprintf(log_buf_ptr, "[%d][%u][%s:%s] ",
					lwe_instance, seq_no, 
					log_type_str[type], log_level_str[level]);
	else
		log_size = sprintf(log_buf_ptr, "[%d][%u][%s:%s] ",
					lwe_instance, seq_no, 
					log_type_str[type], log_level_str[level]);

	log_buf_ptr = log_buf_ptr + log_size;
	rem_size = rem_size - log_size;

	/*Reserve two bytes for '\r\n' which may need to be appended*/
	rem_size = rem_size - 2;

	va_start(args, fmt);

	if (in_irq)
		ret = isr_vsnprintf(log_buf_ptr, rem_size, fmt, args);
	else
		ret = vsnprintf(log_buf_ptr, rem_size, fmt, args);

	va_end(args);

	if (unlikely(ret >= rem_size)) { 	
		/*This means that the log got clipped due to lack of space at
		the end of the log buffer. Wrap around and try again.*/

		if (unlikely(ret + log_size > LOG_BUF_SIZE)) {
			/*If the log message is too big to fit into whole
			log_buf, then try once with as much as possible*/
			if (log_start == log_buf)
				/*Already tried with whole log_buf*/
				goto proceed; 
		}

		log_start[0] = NULL; /*Terminate the last log with NULL*/
		goto try_again;
	}

proceed:
	log_buf_ptr = log_buf_ptr + ret;
	log_size = log_size + ret;

	/*Put a new line at the end if there is none*/
	if (*(log_buf_ptr - 1) != '\n') {
		log_buf_ptr[0] = '\r';
		log_buf_ptr[1] = '\n';
		log_buf_ptr = log_buf_ptr + 2;
		log_size = log_size + 2;
	} else if (*(log_buf_ptr - 2) != '\r') {
		*(log_buf_ptr - 1) = '\r'; /*Overwrite \n with \r*/
		log_buf_ptr[0] = '\n';
		log_buf_ptr = log_buf_ptr + 1;
		log_size = log_size + 1;
	}

	log_buf_ptr[0] = NULL;	/*Terminate the log string*/
	log_buf_ptr++;

	restore_int(saved);

	if ((LOG_DEST_CONSOLE == log_destination) && cpu->console_ok)
		console_log(log_start, log_size);

	return;
}

void print_log_buf()
{
	size_t size;

	/*Print in order of history*/
	size = strlen(log_buf_ptr);
	if (size)
		console_log(log_buf_ptr, size);

	size = strlen(log_buf);
	if ((log_buf_ptr != log_buf) && (0 != size))
		console_log(log_buf, size);
}

int register_log_type(enum log_type_t type, char *name)
{
	if (type >= NUM_LOG_TYPES)
		return -EINVAL;

	if (name) {
		strncpy(&log_type_str[type][0], name, LOG_TYPE_STR_LEN);
		log_type_str[type][LOG_TYPE_STR_LEN] = NULL;
	}

	return 0;
}

int logger_init()
{
	unsigned long int type;

	/*Set the global log sequence number to zero from boot core*/
	/*Variable 'g_log_seq_no' is in .stats_data which is not a loadable
	ELF section. Hence explicit initialisation is required*/
	if (0 == lwe_instance)
		g_log_seq_no = 0;

	/*Set the whole log buffer to NULL initially*/
	memset(log_buf, 0, sizeof(log_buf));

	/*Initialise log buffer pointer to start*/
	log_buf_ptr = log_buf;

	/*Turn ON all the logs which are compiled in for all log types*/
	for (type = 0; type < NUM_LOG_TYPES; type++)
		enabled_log_level[type] = LWE_LOG_LEVEL;

	/*Set default log destination to console*/
	log_destination = LOG_DEST_CONSOLE;

	/*If its boot core, register all the log types*/
	if (0 == lwe_instance) {
#define DEFINE_LOG_TYPE(type) 	register_log_type(LOG_T_##type, #type);
#include <log_def.h>
#undef DEFINE_LOG_TYPE
	}

	LWE_INFO("LWE logger initialised");

	return 0;
}

int set_log_level(enum log_type_t type, enum log_level_t level)
{
	if (type >= NUM_LOG_TYPES)
		return -EINVAL;

	if (level > LWE_LOG_LEVEL_INFO)
		return -EINVAL;

	enabled_log_level[type] = level;
	return 0;
}

/*Shell command "Set log level"*/
int shell_cmd_sll(int argc, char *argv[])
{
	enum log_level_t level;
	enum log_type_t type;

	if (argc < 2) {
		printf("%s: Missing log level\n\n", argv[0]);
		printf("Usage: sll <log_level> [log_type]\n\n");
		printf("Valid log levels\n");
		printf("----------------\n");

		for (level = 0; level < NUM_LOG_LEVELS; level++)
			printf("\t%s - %u\n", log_level_str[level], level);

		printf("\n");
		printf("Valid log types: 0 - %u\n",
				NUM_LOG_TYPES - 1);
		printf("------------------------\n");

		for (type = 0; type < NUM_LOG_TYPES; type++) {
			if (NULL != log_type_str[type][0])
				printf("\t%s - %u\n", log_type_str[type], type);
		}

		return -1;
	}

	errno = 0;
	level = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == level)
			|| (0 != errno && 0 == level)) {
		printf("strtoul failed converting log level");
		return -1;
	}

	if (level >= NUM_LOG_LEVELS) {
		printf("Invalid log level: %u\n", level);
		return -1;
	}

	if (LWE_LOG_LEVEL_OFF == level) {
		printf("Warning! Turning off logs completely.\n");
		printf("It is not advisable to turn off critical logs.\n");
	}

	if (argc < 3) {
		/*Log type not given, set log-level for all log-types*/
		for (type = 0; type < NUM_LOG_TYPES; type++)
			enabled_log_level[type] = level;

		return 0;
	}

	/*Else set log-level of requested type*/
	errno = 0;
	type = strtoul(argv[2], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == type)
			|| (0 != errno && 0 == type)) {
		printf("strtoul failed converting log type");
		return -1;
	}

	if (type >= NUM_LOG_TYPES) {
		printf("Invalid log type: %u\n", type);
		printf("Valid log types: 0 - %u\n", NUM_LOG_TYPES - 1);

		for (type = 0; type < NUM_LOG_TYPES; type++) {
			if (log_type_str[type])
				printf("%s - %u\n", log_type_str[type],
							type);
		}

		return -1;
	}

	enabled_log_level[type] = level;
	return 0;
}

/*Shell command "Print log level"*/
int shell_cmd_pll(int argc, char *argv[])
{
	enum log_level_t level;
	enum log_type_t type;

	if (argc > 1) {
		errno = 0;
		type = strtoul(argv[1], NULL, 0);
		if ((ERANGE == errno && ULONG_MAX == type)
				|| (0 != errno && 0 == type)) {
			printf("strtoul failed converting log type");
			return -1;
		}

		if (type >= NUM_LOG_TYPES) {
			printf("Invalid log type: %lu\n", type);
			printf("Valid log types: 0 - %u\n", NUM_LOG_TYPES - 1);
			return -1;
		}

		level = enabled_log_level[type];
		printf("Log_TYPE_%s (%u) - %s (%u)\n",
				log_type_str[type], type,
				log_level_str[level], level);
		return 0;
	}

	for (type = 0; type < NUM_LOG_TYPES; type++) {
		if (NULL != log_type_str[type][0]) {
			level = enabled_log_level[type];
			printf("LOG_TYPE_%s (%u) - %s (%u)\n",
				log_type_str[type], type,
				log_level_str[level], level);
		}
	}

	return 0;
}

/*Handler for command for setting log destination*/
int shell_cmd_sld(int argc, char *argv[])
{
	enum log_dest_t dest;

	if (argc < 2) {
		printf("Usage: sld <log-destination>\n\n");
		printf("%s - %u\n", "LOG_DEST_BUFF", LOG_DEST_BUFF);
		printf("%s - %u\n", "LOG_DEST_CONSOLE", LOG_DEST_CONSOLE);
		return -1;
	}

	errno = 0;
	dest = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == dest) ||
				(0 != errno && 0 == dest)) {
		printf("strtoul failed converting log destination\n");
		return -1;
	}

	if (dest >= NUM_LOG_DEST || LOG_DEST_INVALID == dest) {
		printf("Invalid log destination: %u\n", dest);
		printf("Valid log types: 1 - %u\n", NUM_LOG_DEST - 1);
		return -1;
	}

	if (LOG_DEST_NETWORK == dest) {
		printf("LOG_DEST_NETWORK not yet supported\n");
		return -1;
	}

	log_destination = dest;
	return 0;
}

/*Handler for command for printing log buffer*/
int shell_cmd_plb(int argc __UNUSED, char *argv[] __UNUSED)
{
	print_log_buf();
	return 0;
}

