/** @file mem_cmd.c
 * Memory dump commands handlers
 */

/*
 * Copyright (c) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <limits.h>
#include <compat.h>
#include <mmap.h>
#include <ldt.h>
#include <io.h>
#include <stdlib.h>
#include <libfdt.h>
#include <init.h>
#include <compiler.h>
#include <mmap_private.h>
#include <fsl-booke-tlb.h>
#include "perf_mon.h"
#define NUM_PHY_ADDR 6
#define REQD_NUM_ARGS 3
#define BYTES_PER_LINE 16
#define WORDS_PER_LINE 4

/* Function for checking the validity of
 * physical address that is whether the
 * given physical address is visible to current partition or not.*/
int phys_address_check(phys_addr_t phys_addr)
{
	uint64_t phy_addr[NUM_PHY_ADDR];
	uint64_t phy_addr_size[NUM_PHY_ADDR];
	int node = -1;
	int i = 0;

	node = fdt_node_offset_by_prop_value(devtree_ptr, -1, "device_type",
				   "memory", strlen("memory") + 1);
	 if (node < 0) {
		LWE_ERROR("Missing memory node");
		return -1;
	}

	do {
		if (get_fdt_mem_ranges
			(node, &phy_addr[i], &phy_addr_size[i]) < 0) {
			LWE_ERROR("Error in reading physical addr and size\n");
			return -1;
		}
		if (phys_addr >= phy_addr[i] &&
			phys_addr <= (phy_addr[i]+phy_addr_size[i])) {
				return 0;
		}
		i++;
	} while ((node = fdt_node_offset_by_prop_value(devtree_ptr, node,
					       "device_type", "memory",
					       strlen("memory") + 1)) !=
		 -FDT_ERR_NOTFOUND);

	printf("Physical address not in the range of partition.\n");
	return -1;

}


/* Handler for the pmem shell command that dumps
 * the memory contents for given physical address*/
int shell_cmd_pmem(int argc, char *argv[])
{
	phys_addr_t paddr, align_paddr;
	phys_addr_t offset;
	uint32_t *vaddr, *vaddr_mapped;
	uint32_t i, j, multiplier;
	int ret;
	unsigned long int num_of_bytes;

	if (argc != REQD_NUM_ARGS) {
		printf("usage: pmem <address> <number_of_bytes>\n");
		printf("Valid argument values:\n");
		printf("-----------------------\n");
		printf("      address - physical address"
					"(prefix with 0x for hex address)\n");
		printf("      number_of_bytes - bytes of data "
			"to print(upto 4k allowed),"
			"must be a multiple of 4\n");
		return -1;
	}

	errno = 0;
	paddr = strtoull(argv[1], NULL, 0);
	if ((ERANGE == errno && ULLONG_MAX == paddr)
			|| (0 != errno && 0 == paddr)) {
		printf("strtoull failed converting physical address.\n");
		return -1;
	}

	num_of_bytes = strtoul(argv[2], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == num_of_bytes)
			|| (0 != errno && 0 == num_of_bytes)) {
		printf("strtoul failed converting number of bytes.\n");
		return -1;
	}

	if ((num_of_bytes > PAGE_SIZE)
			|| ((num_of_bytes % BYTES_PER_WORD) != 0)) {
		printf("Number of bytes not in permissible limit.\n");
		return -1;
	}

	offset = paddr & (PAGE_SIZE - 1);
	align_paddr = paddr & ~(PAGE_SIZE - 1);

	if (!(phys_address_check(paddr))
			&& !(phys_address_check(paddr + num_of_bytes))) {
		if ((offset + num_of_bytes) > PAGE_SIZE)
			multiplier = 2;
		else
			multiplier = 1;
		ret = dev_mmap(align_paddr, (PAGE_SIZE * multiplier),
				PAGE_SIZE, TLB_MAS2_MEM, MAS3_SR,
				(void *)&vaddr_mapped);
		if (ret < 0) {
			printf("Failed to create tlb entry for"
						" physical address\n");
			return -1;
		}
	} else {
		printf("Physical address is not in valid range.\n");
		return -1;
	}

	vaddr = (uint32_t *)vaddr_mapped;
	vaddr = (uint32_t *)((uint8_t *)vaddr + offset);
	i = num_of_bytes % BYTES_PER_LINE;
	for (j = 0; j < (num_of_bytes / BYTES_PER_LINE); j++) {
		printf("%09llx: %08x %08x %08x %08x\n",
				paddr, *vaddr, *(vaddr + 1),
				*(vaddr + 2), *(vaddr + 3));
		vaddr = vaddr + WORDS_PER_LINE;
		paddr = paddr + BYTES_PER_LINE;
	}
	switch (i) {
	case 4:
		printf("%09llx: %08x\n", paddr, *vaddr);
		break;
	case 8:
		printf("%09llx: %08x %08x\n", paddr, *vaddr, *(vaddr + 1));
		break;
	case 12:
		printf("%09llx: %08x %08x %08x\n", paddr, *vaddr,
				*(vaddr + 1), *(vaddr + 2));
		break;
	default:
		break;
	}

	dev_unmap((void *)vaddr_mapped);
	return 0;

}


/* Handler for the vmem shell command that dumps
 * the memory contents for given virtual address*/
int shell_cmd_vmem(int argc, char *argv[])
{
	uint32_t j, i;
	bool flag = 0;
	bool tlbsx_pass = 0;
	uint32_t vaddr, align_vaddr, offset;
	uint32_t *v_addr;
	phys_addr_t paddr1, paddr2;
	register_t mas1_1, mas1_2;
	unsigned long int num_of_bytes;

	if (argc != REQD_NUM_ARGS) {
		printf("usage: vmem <address> <number_of_bytes>\n");
		printf("Valid argument values:\n");
		printf("-----------------------\n");
		printf("       address - virtual address"
					"(prefix with 0x for hex address)\n");
		printf("       number_of_bytes - bytes "
			"to print (upto 4k allowed),"
			"must be a multiple of 4\n");
		return -1;
	}
	errno = 0;
	vaddr = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == vaddr)
			|| (0 != errno && 0 == vaddr)) {
		printf("strtoul failed converting virtual address.\n");
		return -1;
	}

	num_of_bytes = strtoul(argv[2], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == num_of_bytes)
			|| (0 != errno && 0 == num_of_bytes)) {
		printf("strtoul failed converting number of bytes.\n");
		return -1;
	}

	if ((num_of_bytes > PAGE_SIZE) ||
				((num_of_bytes % BYTES_PER_WORD) != 0)) {
		printf("Number of bytes not in permissible limit.\n");
		return -1;
	}

	offset = vaddr & (PAGE_SIZE - 1);
	align_vaddr = vaddr & ~(PAGE_SIZE - 1);

	isync();
	asm volatile("tlbsx 0, %0" : : "r" (vaddr));
	isync();
	mas1_1 = mfspr(SPR_MAS1);
	if (mas1_1 & MAS1_VALID) {
		paddr1 = (((phys_addr_t)(mfspr(SPR_MAS7) & MAS7_RPN)) << 32)
				| (mfspr(SPR_MAS3) & MAS3_RPN)
					| (vaddr & ~MAS3_RPN);
		if ((phys_address_check(paddr1)) == -1) {
			printf("Corresponding Physical address not valid\n");
			return -1;
		}
	}

	if ((offset + num_of_bytes) > PAGE_SIZE) {
		flag = 1;
		isync();
		asm volatile("tlbsx 0, %0" : : "r" (align_vaddr + PAGE_SIZE));
		isync();
		mas1_2 = mfspr(SPR_MAS1);
		if (mas1_2 & MAS1_VALID) {
			paddr2 =
			(((phys_addr_t)(mfspr(SPR_MAS7) & MAS7_RPN)) << 32)
					| (mfspr(SPR_MAS3) & MAS3_RPN)
						| (vaddr & ~MAS3_RPN);
			if ((phys_address_check(paddr2)) == -1) {
				printf("Corresponding Physical address"
							" not valid\n");
				return -1;
			}
		}
	}

	if (flag)
		tlbsx_pass = (mas1_1 & MAS1_VALID) && (mas1_2 & MAS1_VALID);
	else
		tlbsx_pass = mas1_1 & MAS1_VALID;

	if (tlbsx_pass) {
		v_addr = (uint32_t *)vaddr;
		i = num_of_bytes % BYTES_PER_LINE;
		for (j = 0; j < (num_of_bytes / BYTES_PER_LINE); j++) {
			printf("%08p: %08x %08x %08x %08x\n",
					v_addr, *v_addr, *(v_addr + 1),
					*(v_addr + 2), *(v_addr + 3));
			v_addr = v_addr + WORDS_PER_LINE;
		}
		switch (i) {
		case 4:
			printf("%08p: %08x\n", v_addr, *v_addr);
			break;
		case 8:
			printf("%08p: %08x %08x\n",
				v_addr, *v_addr, *(v_addr + 1));
			break;
		case 12:
			printf("%08p: %08x %08x %08x\n",
				v_addr, *v_addr, *(v_addr + 1),
				*(v_addr + 2));
			break;
		default:
			break;
		}
	} else {
		printf("No such virtual address found.\n");
		return -1;
	}
	return 0;
}
