/** @file mmap.c
 * This file contains APIs for mapping I/O devices.
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <stdbool.h>
#include <log.h>
#include <init.h>
#include <bitops.h>
#include <errno.h>
#include <fsl-booke-tlb.h>
#include <fsl_hcalls.h>
#include <types.h>
#include <elf.h>
#include <mmap_private.h>
#include <compiler.h>
#include <string.h>
#include <mmap.h>
#include <percpu.h>
#include <lwe_malloc.h>
#include <hv_misc.h>

static __PERCPU __ALIGNED(64) struct addr_tbl vptbl;

static __PERCPU struct tlbtbl_entry tlb0_tbl[512];
static __PERCPU struct tlbtbl_entry *tlb1_tbl;
static __PERCPU int tlb0_idx;

int
dev_unmap(void *ea)
{
	unsigned long va = (unsigned long) ea;
	register_t mas0, mas1, mas2;
	int tlb_idx = (va >> 12) & 511;
	unsigned long way = 0;
	int pages = 0;

	if ((tlb0_tbl[tlb_idx].attr & TLB0_VALID_MASK) == 0)
		return -EINVAL;

	pages = tlb0_tbl[tlb_idx].attr & TLB0_SIZE_MASK;
	if (pages == 0)
		return -EINVAL;

	for (int i = tlb_idx; i < (tlb_idx + pages); i++) {
		tlb0_tbl[i].addr = 0x0;
		tlb0_tbl[i].attr = 0x0;
		way = (va >> 19) & WAY_MASK;
		mas0 =
		    MAS0_TLBSEL0 | ((way << MAS0_ESEL_SHIFT) & MAS0_ESEL_MASK);
		mas2 = (va & MAS2_EPN);
		mas1 = 0;
		mtspr(SPR_MAS0, mas0);
		mtspr(SPR_MAS1, mas1);
		mtspr(SPR_MAS2, mas2);
		asm volatile ("isync; tlbwe; isync; msync":::"memory");
		va += PAGE_SIZE;
	}

	return 0;
}

static int
get_free_tlb0_idx(int num_4k_pages, uint32_t align)
{
	int ret_tlb0_idx = -1;
	int start_idx = tlb0_idx;
	int end_idx = start_idx + num_4k_pages;
	int num_iter = 0;
	int iter = 0;
	while (num_iter < 512) {
		iter = 0;
		while ((start_idx << PAGE_SHIFT) & (align - 1)) {
			start_idx++;
			num_iter++;
			if (num_iter >= 512)
				break;
		}
		end_idx = start_idx + num_4k_pages;
		if (end_idx > 512) {
			num_iter += (512 - start_idx);
			start_idx = 0;
			end_idx = num_4k_pages;
		}
		for (int j = start_idx; j < end_idx; j++) {
			num_iter++;
			if (tlb0_tbl[j].attr & TLB0_VALID_MASK) {
				start_idx = j + 1;
				break;
			}
			iter++;
		}
		if (iter == num_4k_pages) {
			ret_tlb0_idx = start_idx;
			break;
		}
	}
	return ret_tlb0_idx;
}

static int
update_tlb0_tbl(int idx, phys_addr_t pa, int num_entries)
{
	tlb0_tbl[idx].addr =
	    TLB0_VALID_MASK | ((pa >> PAGE_SHIFT) & TLB0_RPN_MASK);
	tlb0_tbl[idx].attr = TLB0_VALID_MASK | (num_entries & TLB0_SIZE_MASK);
	tlb0_idx = idx + 1;
	if (tlb0_idx == 512)
		tlb0_idx = 0;
	return tlb0_idx;
}

int
dev_mmap(phys_addr_t pa, size_t size, size_t align, register_t mas2flags,
	 register_t mas3flags, void **ea)
{
	int tlb_idx = -1;
	uint32_t pages = 0;
	uint32_t num_entries = 0;
	unsigned int ts = 0;
	unsigned int tid = 0;
	unsigned long va = 0;
	register_t mas0, mas1, mas2, mas3, mas7;

	if (DEV_MMAP_START & (TLB0_SIZE - 1))
		LWE_PANIC("Virtual address(va_start) %x for the IO devices"
			  " is not aligned to 2MB", DEV_MMAP_START);

	if (pa & (PAGE_SIZE - 1)) {
		LWE_WARN("Physical address of the memory to be mapped is not"
			 " 4K aligned. Cannot be mapped with TLB0");
		return -EINVAL;
	}

	if (size & (PAGE_SIZE - 1)) {
		LWE_WARN("Size of the memory to be mapped is not 4K aligned."
			 "Cannot be mapped with TLB0");
		return -EINVAL;
	}

	if (size > TLB0_SIZE || align > TLB0_SIZE) {
		LWE_WARN("Maximum of 2MB region can be mapped by this call."
			 "Check the size or alignment passed");
		return -EINVAL;
	}

	if (align < PAGE_SIZE)
		align = PAGE_SIZE;

	pages = size >> PAGE_SHIFT;
	if (pages == 0)
		return -EINVAL;
	num_entries = pages;

	tlb_idx = get_free_tlb0_idx(pages, align);
	if (tlb_idx < 0)
		return -ENOMEM;

	mas1 = MAS1_VALID | ts | tid;
	va = DEV_MMAP_START + (tlb_idx << PAGE_SHIFT);
	*ea = (void *) va;
	for (uint32_t i = 0; i < pages; i++) {
		unsigned long way = (va >> 19) & WAY_MASK;
		mas0 =
		    MAS0_TLBSEL0 | ((way << MAS0_ESEL_SHIFT) & MAS0_ESEL_MASK);

		mas2 = (va & MAS2_EPN) | mas2flags;
		mas3 = (pa & MAS3_RPN) | mas3flags;
		mas7 = pa >> 32;
		mtspr(SPR_MAS0, mas0);
		mtspr(SPR_MAS1, mas1);
		mtspr(SPR_MAS2, mas2);
		mtspr(SPR_MAS3, mas3);
		mtspr(SPR_MAS7, mas7);
		asm volatile ("isync; tlbwe; isync; msync":::"memory");
		tlb_idx = update_tlb0_tbl(tlb_idx, pa, num_entries);
		num_entries = 0;
		pa += (1 << PAGE_SHIFT);
		va += (1 << PAGE_SHIFT);
	}
	return 0;
}

void initialize_vptbl(uint32_t cpu_id)
{
	uint32_t percpu_size = 0x0;
	/* 64k reserved for PHDRS */
	unsigned long _begin = (unsigned long)&_start - 0x10000;
	struct elf_header *ehdr = (struct elf_header *)_begin;
	struct program_header *ph = (struct program_header *)((uint32_t) _begin
							      + ehdr->phoff);
	int i = 1;
	int j = 0;
	memset(&vptbl, 0x0, sizeof(vptbl));

	for (i = 1; i < ehdr->phnum; i++, j++) {

		/* if ph[i].vaddr is = percore area or heapstack
		 * paddr += cpu_id * percpu_size
		 */
		if (ph[i].vaddr == (uint32_t)&_percpu_start) {
			percpu_size = (uint32_t)&_percpu_end
			    - (uint32_t)&_percpu_start;

		} else if (ph[i].vaddr ==
			   (uint32_t)&_percpu_bss_heap_stack_start) {
			percpu_size = ph[i].filesz;
			percpu_size =
			    (uint32_t)&_percpu_bss_heap_stack_end -
			    (uint32_t)&_percpu_bss_heap_stack_start;
		} else
			percpu_size = 0;

		vptbl.entry[j].vaddr = ph[i].vaddr;
		vptbl.entry[j].paddr = (phys_addr_t)ph[i].paddr +
						(cpu_id * percpu_size);
		vptbl.entry[j].memsz = percpu_size ? percpu_size :  ph[i].memsz;

	}

	/*ASSUMPTION -Due to known bug in linker script.
	 * First entry in PHDR tbl will always be text segment
	 * containing PHDR table.
	 * So moving that entry to last index in vptbl
	 * TO REMOVE - Once above limitation is fixed*/
	vptbl.entry[j].vaddr = ph[0].vaddr;
	vptbl.entry[j].paddr = ph[0].paddr;
	vptbl.entry[j].memsz = ph[0].memsz;

	vptbl.max_index = ehdr->phnum;
}

/* Function to print vvp table for debugging */
void print_vptbl(void)
{
	/* 64k reserved for PHDRS */
	unsigned long _begin = (unsigned long)&_start - 0x10000;
	struct elf_header *ehdr = (struct elf_header *)_begin;
	struct program_header *ph = (struct program_header *)((uint32_t) _begin
							      + ehdr->phoff);
	for (int i = 0; i <  ehdr->phnum; i++) {
		LWE_INFO("Phdr %d, va %x, pa %x, memsize %x\n",
			 i,  ph[i].vaddr, ph[i].paddr, ph[i].memsz);
		LWE_INFO("Entry %d, va %x, pa %llx, mesize %x\n",
			i,  vptbl.entry[i].vaddr, vptbl.entry[i].paddr,
			 vptbl.entry[i].memsz);
	}
}

phys_addr_t vtop(void *addr)
{
	uint32_t vaddr = (uint32_t) addr;

	/* For performance improvement on 4 most frequently queried translations
	 */ 
	if (likely(vaddr >= vptbl.entry[0].vaddr
		  && vaddr < (vptbl.entry[0].vaddr + vptbl.entry[0].memsz)))
		return vptbl.entry[0].paddr + (vaddr - vptbl.entry[0].vaddr);

	if (likely(vaddr >= vptbl.entry[1].vaddr
		   && vaddr < (vptbl.entry[1].vaddr + vptbl.entry[1].memsz)))
		return vptbl.entry[1].paddr + (vaddr - vptbl.entry[1].vaddr);

	if (likely(vaddr >= vptbl.entry[2].vaddr
		   && vaddr < (vptbl.entry[2].vaddr + vptbl.entry[2].memsz)))
		return vptbl.entry[2].paddr + (vaddr - vptbl.entry[2].vaddr);

	if (likely(vaddr >= vptbl.entry[3].vaddr
		   && vaddr < (vptbl.entry[3].vaddr + vptbl.entry[3].memsz)))
		return vptbl.entry[3].paddr + (vaddr - vptbl.entry[3].vaddr);

	int i = 4;
	while (vptbl.entry[i].memsz > 0) {
		if (likely(vaddr >= vptbl.entry[i].vaddr
		    && vaddr < (vptbl.entry[i].vaddr + vptbl.entry[i].memsz)))
			return (vptbl.entry[i].paddr +
				(vaddr - vptbl.entry[i].vaddr));
		i++;
	}

	return UINTMAX_MAX;
}

void *ptov(phys_addr_t paddr)
{
	/* For performance improvement on 4 most frequently queried translations
	 */ 
	if (likely(paddr >= vptbl.entry[0].paddr &&
	    paddr < (vptbl.entry[0].paddr + vptbl.entry[0].memsz)))
		return ((void *)(vptbl.entry[0].vaddr +
			 (uint32_t) (paddr - vptbl.entry[0].paddr)));

	if (likely(paddr >= vptbl.entry[1].paddr &&
	    paddr < (vptbl.entry[1].paddr + vptbl.entry[1].memsz)))
		return ((void *)(vptbl.entry[1].vaddr +
			 (uint32_t) (paddr - vptbl.entry[1].paddr)));

	if (likely(paddr >= vptbl.entry[2].paddr &&
	    paddr < (vptbl.entry[2].paddr + vptbl.entry[2].memsz)))
		return ((void *)(vptbl.entry[2].vaddr +
			 (uint32_t) (paddr - vptbl.entry[2].paddr)));

	if (likely(paddr >= vptbl.entry[3].paddr &&
	    paddr < (vptbl.entry[3].paddr + vptbl.entry[3].memsz)))
		return ((void *)(vptbl.entry[3].vaddr +
			 (uint32_t) (paddr - vptbl.entry[3].paddr)));

	int i = 4;
	while (vptbl.entry[i].memsz > 0) {
		if (likely(paddr >= vptbl.entry[i].paddr &&
		    paddr < (vptbl.entry[i].paddr + vptbl.entry[i].memsz)))
			return ((void *)(vptbl.entry[i].vaddr +
				 (uint32_t) (paddr - vptbl.entry[i].paddr)));
		i++;
	}
	return NULL;
}

static void sort_vtop_tbl(struct vtop_info *arr, int n)
{
	for (int i = 0; i < n; i++)
		for (int j = i; j < n; j++)
			if (arr[j].vaddr < arr[i].vaddr) {
				struct vtop_info tmp;
				tmp = arr[j];
				arr[j] = arr[i];
				arr[i] = tmp;
			}
}

void alloc_vspace(void)
{
	struct vtop_info *tbl;
	int idx = vptbl.max_index;
	int i;

	/* Allocate memory for tlb1tbl for each core */
	tlb1_tbl = tlmalloc(sizeof(struct tlbtbl_entry) * cpu->tlb1cnt);
	if (unlikely(!tlb1_tbl)) {
		LWE_ERROR("Memory not sufficient\n");
		return;
	}

	tbl = malloc(sizeof(struct vtop_info) *  (idx + 1));
	if (unlikely(!tbl)) {
		LWE_ERROR("Virtual space allocation failed\n");
		return;
	}

	memcpy(tbl, vptbl.entry, sizeof(struct vtop_info) *  idx);

	/* ADD the DEV_MMAP entry reserved for TLB0 also */
	tbl[idx].memsz = TLB0_SIZE;
	tbl[idx].vaddr = DEV_MMAP_START;
	idx++;

	sort_vtop_tbl(tbl, idx);
	for (i = 0; i <  idx; i++) {
		LWE_DEBUG("Entry %d, va %x, pa %llx, mesize %x\n",
			i,  tbl[i].vaddr, tbl[i].paddr, tbl[i].memsz);
	}

	/* Leaving a 4K hole in the beginning to
	   disallow addition of address 0 */
	unsigned long start_addr = 0x1000, size = 0;
	size = tbl[0].vaddr - start_addr;
	LWE_DEBUG("Adding vspace of addr - %lx size - %lx\n", start_addr, size);
	if (size > 0) {
		if (valloc_add(start_addr, size) < 0)
			LWE_WARN("Addition of segment addr%lx size%lx failed\n",
							     start_addr, size);
	}

	for (i = 0; i < idx - 1; i++) {
		start_addr = tbl[i].vaddr + tbl[i].memsz;
		size = tbl[i + 1].vaddr - start_addr;
		if (!size)
			continue;
		LWE_DEBUG("Adding vspace of addr - %lx size - %lx\n",
							 start_addr, size);
		if (valloc_add(start_addr, size) < 0)
			LWE_WARN("Addition of segment addr%lx size%lx failed\n",
							     start_addr, size);
	}

	start_addr = tbl[i].vaddr + tbl[i].memsz;
	size = UINTPTR_MAX - start_addr + 1;
	if (size > 0) {
		LWE_DEBUG("Adding vspace of addr - %lx size - %lx\n",
							 start_addr, size);
		if (valloc_add(start_addr, size) < 0)
			LWE_WARN("Addition of segment addr%lx size%lx failed\n",
							     start_addr, size);
	}

	free(tbl);
}

static void free_tlb1(uint32_t used_tlb1_idx)
{
	for (uint32_t i = 0; i < MAX_TLB1; i++)
		if (used_tlb1_idx & 1 << i)
			tlb1_clear_entry(i);
}

/* Returns 0 if match for va found else returns -1*/
static int free_tlb1_entries(unsigned long va)
{
	uint32_t i = 0;
	int loop_idx = MAX_TLB1;
	unsigned long size = 0 ;
	uint32_t tsize = 0;

	while (i++, loop_idx--) {
		if (i == MAX_TLB1)
			i = 0;
		if ((tlb1_tbl[i].attr & TLB1_VALID_MASK) == 0)
			continue;
		if (tlb1_tbl[i].addr == va) {
			tlb1_clear_entry(i);
			if (tlb1_tbl[i].attr & TLB1_LAST_ENTRY) {
				tlb1_tbl[i].attr = 0;
				return 0;
			}
			tsize = (tlb1_tbl[i].attr & TLB1_TSIZE_MASK) >>
							       TLB1_TSIZE_SHIFT;
			tlb1_tbl[i].attr = 0;
			size = tsize_to_pages(tsize) << PAGE_SHIFT;
			va = va + size;
			loop_idx = MAX_TLB1 - 1;
		}
	}
	return -1;
}

/* This function returns alignment of addr in terms of power of four */
static unsigned long align_pow_four(unsigned long addr)
{
	if (!addr)
		return 1UL << (LONG_BITS - 1);

	unsigned long zero_count = count_lsb_zeroes(addr);

	uint32_t align = zero_count % 2  != 0 ? 1UL << (zero_count - 1) :
							 1UL << zero_count;

	return align;
}

static int populate_tlb1(uintptr_t va, phys_addr_t pa, uint32_t size,
				 register_t mas2flags, register_t mas3flags)
{
	uint32_t tlb1_idx = 0xff;
	uint32_t used_tlb1_idx = 0;
	uint32_t epn = va >> PAGE_SHIFT;
	uint32_t rpn = pa >> PAGE_SHIFT;
	uint32_t pages = size >> PAGE_SHIFT;
	uint32_t tsize;
	uint32_t npages;
	uint32_t min_pages_va, min_pages_rpn;

	uint32_t tepn = epn;
	uint32_t tpages = pages;

	while (tepn < (epn + pages)) {

		/* Get alignment of rpn and epn */
		min_pages_va =  align_pow_four(tepn);
		min_pages_rpn = align_pow_four(rpn);
		npages = min(min_pages_va, min_pages_rpn);

		/* if npages is greater than pages required for the given size,
		   since entry to be created in TLB has to be 4K aligned,
		   divide npages by 4 */
		while (npages > tpages)
			npages = npages / 4;

		tsize = pages_to_tsize(npages);

		tlb1_idx = get_freetlb1_idx();
		if (tlb1_idx == 0xff) {
			/* free_allocated_tlb1 for this va */
			LWE_WARN("Number of free TLB entries not sufficient to"
				 "create mapping of given address and size\n");
			free_tlb1(used_tlb1_idx);
			return -ENOMEM;
		}

		tlb1_set_entry(tlb1_idx, tepn << PAGE_SHIFT,
				((phys_addr_t)rpn) << PAGE_SHIFT,
				tsize, mas2flags, mas3flags, 0, 0, 0);

		used_tlb1_idx |= 1 << tlb1_idx;
		if ((tlb1_tbl[tlb1_idx].attr & TLB1_VALID_MASK) == 1)
			LWE_PANIC("Unexpected error : Local TLB table and TLB1"
				  "donot match\n");

		tlb1_tbl[tlb1_idx].attr =  tlb1_tbl[tlb1_idx].addr = 0;
		tlb1_tbl[tlb1_idx].attr = TLB1_VALID_MASK |
						 (tsize << TLB1_TSIZE_SHIFT);
		tlb1_tbl[tlb1_idx].addr = tepn << 12;

		tepn = tepn + npages;
		rpn = rpn + npages;
		tpages -= npages;
	}

	tlb1_tbl[tlb1_idx].attr |= TLB1_LAST_ENTRY;

	return 0;
}

int mmap(phys_addr_t pa, size_t size, uint32_t attr,
		uint32_t perm, enum tlb_sel tlbnum, void **ea)
{
	uintptr_t va = NULL;
	unsigned long align = 0;
	register_t mas2flags = attr & MAS2_FLAGS;
	register_t mas3flags = perm & MAS3_FLAGS;

	if (tlbnum != TLB0 && tlbnum != TLB1) {
		LWE_WARN("Wrong parameter tlbnum, Can be either TLB0 or TLB1");
		return -EINVAL;
	}

	if (tlbnum == TLB0)
		return dev_mmap(pa, size, size, mas2flags, mas3flags, ea);

	if (pa & (PAGE_SIZE - 1)) {
		LWE_WARN("Physical address of the memory to be mapped is not"
			 " 4K aligned. Cannot be mapped with TLB1");
		return -EINVAL;
	}

	if (size == 0 || size & (PAGE_SIZE - 1)) {
		LWE_WARN("Size of the memory to be mapped is not 4K aligned."
			 "Cannot be mapped with TLB1");
		return -EINVAL;
	}

	/* Find alignment of the pa and get a va aligned to this pa */
	align =	align_pow_four(pa);

	while (align >= PAGE_SIZE) {
		va = (uintptr_t)valloc(size, align);
		if (va)
			break;
		align = align / 4;
	}

	if (!va) {
		LWE_WARN("Couldn't allocate virtual address for the given pa"
			 "and size");
		return -ENOMEM;
	}

	if (populate_tlb1(va, pa, size, mas2flags, mas3flags) != 0) {
		vfree((void *)va);
		*ea = NULL;
		return -EINVAL;
	}
	*ea = (void *)va;

	return 0;
}

int munmap(void *ea)
{
	uintptr_t va = (uintptr_t) ea;
	if (!va || (va & (PAGE_SIZE - 1))) {
		LWE_WARN("virtual address not correct\n");
		return -EFAULT;
	}

	/* Check if va is in range alloted for TLB0 mapping */
	if (va >= DEV_MMAP_START && va < (DEV_MMAP_START + TLB0_SIZE))
		return dev_unmap(ea);

	if (free_tlb1_entries(va) < 0) {
		LWE_WARN("Virtual address not correct\n");
		return -EFAULT;
	}

	vfree((void *)va);
	return 0;
}
