/** @file ns16550.c
 * UART implementation
 */

/*
 * Copyright (c) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <ns16550.h>
#include <libfdt.h>
#include <stdlib.h>
#include <init.h>
#include <compat.h>
#include <log.h>
#include <ldt.h>
#include <mmap_private.h>

/* Global array for UARTs */
ns16550 uart_obj[NUM_OF_UARTS];

/* Probe UART for
 * certain properties
 */
int uart_probe(ns16550 *ns, int offset)
{
	const uint32_t *prop = NULL;
	uint32_t freq = 0;

	prop = fdt_getprop(devtree_ptr, offset, "clock-frequency", NULL);
	freq = *prop;

	if (!freq) {
		LWE_ERROR("%s: bad/missing clock-frequency in %s\n",
			__func__, fdt_get_name(devtree_ptr, offset, NULL));
		return -1;
	}

	if (unlikely(0 > get_fdt_mem_ranges(offset, &ns->uart_pa, &ns->size))) {
		LWE_ERROR("%s: get_fdt_mem_ranges failed for uart node",
			  __func__);
		return -1;
	}

	ns->freq = freq;
	ns->baud = BAUD_RATE;
	ns->lock = 0; /* init */
	ns->uart_offset = ns->uart_pa & (PAGE_SIZE - 1);

	return 0;
}

/* Get the virtual address space
 * for UART
 */
int get_uart_va(ns16550 *ns, int lwe_instance)
{
	int ret = -1;

	if (ns->uart_pa & (PAGE_SIZE - 1)) {
		LWE_WARN("Physical address of the memory to be mapped is not"
			 " 4K aligned");
		ns->uart_pa = ns->uart_pa & ~(PAGE_SIZE - 1);
	}

	if (ns->size & (PAGE_SIZE - 1)) {
		LWE_WARN("Size of the memory to be mapped is not 4K aligned");
		ns->size = PAGE_SIZE;
	}

	ret =
		dev_mmap(ns->uart_pa, ns->size, ns->size, MAS2_M, TLB_MAS3_KERN,
		     &ns->uart_va[lwe_instance]);

	if (unlikely(0 > ret)) {
		LWE_ERROR("%s: failed to create tlb entry for uart!",
			  __func__);
		return -1;
	}

	return 0;
}

/* Release uart_va */
int free_uart_va(ns16550 *ns, int lwe_instance)
{
	int ret = -1;

	ret = dev_unmap(ns->uart_va[lwe_instance]);

	if (unlikely(0 > ret)) {
		LWE_ERROR("%s: failed to remove tlb entry for uart!",
			  __func__);
		return -1;
	}

	return 0;
}

/* Initialize UART */
int ns16550_init(ns16550 *ns)
{
	int divisor;

	uint8_t *uart_va = (uint8_t *)ns->uart_va[lwe_instance] +
						ns->uart_offset;
	uint32_t baud = ns->baud * 16;
	divisor = (ns->freq + baud / 2) / baud;

	/* Set divisor latch access bit in LCR */
	out8((uart_va + NS16550_LCR), NS16550_LCR_DLAB);

	/* Populate divisor most significant byte in DMB */
	out8((uint8_t *)(uart_va + NS16550_DMB), (divisor >> 8) & 0xff);

	/* Populate divisor least significant byte in DLB */
	out8((uint8_t *)(uart_va + NS16550_DLB), divisor & 0xff);

	/* Select word length as 8 bits */
	out8((uint8_t *)(uart_va + NS16550_LCR), NS16550_LCR_8BIT);

	in8((uint8_t *)(uart_va + NS16550_RBR)); /* dummy read to clear
						any pending interrupts */

	/* Assert RTS. This indicates UART is ready for
	 * sending / receiving data.
	 */
	out8((uint8_t *)(uart_va + NS16550_MCR),
		in8((uint8_t *)(uart_va + NS16550_MCR)) | NS16550_MCR_RTS);

	/* Set ERDAI bit in IER. This enables and asserts interrupts when
	 * new data character is received.
	 */
	out8((uint8_t *)(uart_va + NS16550_IER), NS16550_IER_ERDAI);

	return 0;
}

/* Dump on UART */
ssize_t ns16550_tx(ns16550 *ns, uint8_t *buf,
			int count, int flags)
{
	int ret = 0;
	int i;

	uint8_t *uart_va = (uint8_t *)ns->uart_va[lwe_instance] +
						ns->uart_offset;

	do {
		while (!(in8((uint8_t *)(uart_va + NS16550_LSR)) &
						NS16550_LSR_THRE))
			if (!(flags & CHARDEV_BLOCKING))
				return ret;

		unsigned long saved = spin_lock_intsave(&ns->lock);

		if (!(in8((uint8_t *)(uart_va + NS16550_LSR)) &
						NS16550_LSR_THRE)) {
			spin_unlock_intsave(&ns->lock, saved);
			continue;
		}

		for (i = 0; i < UART_FIFO_SIZE && ret < count; i++) {
			out8((uint8_t *)(uart_va + NS16550_THR),
							buf[ret++]);
		}

		spin_unlock_intsave(&ns->lock, saved);
	} while (ret < count);

	return ret;
}

/* Get from UART */
int ns16550_rx(ns16550 *ns, uint8_t *result)
{
	uint8_t iir;

	uint8_t *uart_va = (uint8_t *)ns->uart_va[lwe_instance] +
						ns->uart_offset;

	iir = in8((uint8_t *)(uart_va + NS16550_IIR));
	if (iir & NS16550_IIR_NOIRQ) {
		LWE_ERROR("No pending UART interrupt");

		return -1;
	}

	iir &= NS16550_IIR_IID;

	if (iir == NS16550_IIR_MSI)
		in8((uint8_t *)(uart_va + NS16550_MSR));

	/* Either receiver data available or receiver timeout, call store */
	if (iir == NS16550_IIR_RDAI || iir == NS16550_IIR_RXTIME) {
		if (in8((uint8_t *)(uart_va + NS16550_LSR)) & NS16550_LSR_DR) {
			uint8_t data = in8((uint8_t *)(uart_va + NS16550_RBR));
			*result = data;
		}
	}

	return 0;
}

/* Wrapper function for UART initialization
 * sequential functions
 */
int uart_init(uint32_t nsobj_idx, int con_node)
{
	if (0 > uart_probe(&uart_obj[nsobj_idx], con_node))
		return -1;

	if (0 > get_uart_va(&uart_obj[nsobj_idx], lwe_instance))
		return -1;

	if (0 > ns16550_init(&uart_obj[nsobj_idx]))
		return -1;

	if (0 > free_uart_va(&uart_obj[nsobj_idx], lwe_instance))
		return -1;

	return 0;
}
