/** @file perf_mon.c
 * Performance Monitoring APIs
 */

/*
 * Copyright (c) 2009 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include "types.h"
#include "log.h"
#include "io.h"
#include "bitops.h"
#include "perf_mon.h"

#define MAX_PMC_REG	3
#define MAX_PM_EVENT	180

#define PMC		0x10
#define PMC0		0x10
#define PMC1		0x11
#define PMC2		0x12
#define PMC3		0x13

#define PMLCA		0x90
#define PMLCA0      	0x90
#define PMLCA1		0x91
#define PMLCA2		0x92
#define PMLCA3		0x93

#define PMLCB		0x110
#define PMLCB0		0x110
#define PMLCB1		0x111
#define PMLCB2		0x112
#define PMLCB3		0x113

#define PMGC		0x190
#define PMGC0		0x190

#define MAX_THRESHOLD	63
#define MAX_THRESHMUL	128

#define THRESHOLD(val) (val & 0x3f)
#define THRESHMUL(val) (val & 0x700)

#define MIN_THRESHOLD_EVENT	PM_DLFB_LOAD_MISS_CYCLES
#define MAX_THRESHOLD_EVENT	\
	PM_CRITICAL_INPUT_INTERRUPT_PENDING_LATENCY_CYCLES

union pmr_lca {
	struct {
		uint8_t fc:1;
		uint8_t fcs:1;
		uint8_t fcu:1;
		uint8_t fcm1:1;
		uint8_t fcm0:1;
		uint8_t ce:1;
		uint8_t reserved1:2;
		uint8_t event;
		uint16_t reserved3;
	} pmlca;
	uint32_t reg_val;
};

union pmr_gcr {
	struct {
		uint8_t fac:1;
		uint8_t pmie:1;
		uint8_t fcece:1;
		uint8_t reserved1:5;
		uint8_t reserved2[3];
	} pmgcr;
	uint32_t reg_val;
};

union pmr_lcb {
	struct {
		uint8_t reserved1[2];
		uint8_t reserved2:5;
		uint8_t threshmul:3;
		uint8_t reserved3:2;
		uint8_t threshold:6;
	} pmlcb;
	uint32_t reg_val;
};

static int get_free_pmr(pmr_handle_t *handle, int thresh_event)
{
	union pmr_lca u;

	if (!thresh_event) {
		u.reg_val = (uint32_t) mfpmr(PMLCA + 2);
		if (u.pmlca.event == PM_NOTHING) {
			*handle = 2;
			return 0;
		}

		u.reg_val = (uint32_t) mfpmr(PMLCA + 3);
		if (u.pmlca.event == PM_NOTHING) {
			*handle = 3;
			return 0;
		}
	}

	u.reg_val = (uint32_t) mfpmr(PMLCA + 0);
	if (u.pmlca.event == PM_NOTHING) {
		*handle = 0;
		return 0;
	}

	u.reg_val = (uint32_t) mfpmr(PMLCA + 1);
	if (u.pmlca.event == PM_NOTHING) {
		*handle = 1;
		return 0;
	}

	return -1;
}

static uint32_t get_pmr(uint32_t reg, uint32_t reg_nr)
{
	switch (reg + reg_nr) {
	case PMC0:
		return mfpmr(PMC0);
	case PMC1:
		return mfpmr(PMC1);
	case PMC2:
		return mfpmr(PMC2);
	case PMC3:
		return mfpmr(PMC3);
	case PMLCA0:
		return mfpmr(PMLCA0);
	case PMLCA1:
		return mfpmr(PMLCA1);
	case PMLCA2:
		return mfpmr(PMLCA2);
	case PMLCA3:
		return mfpmr(PMLCA3);
	case PMLCB0:
		return mfpmr(PMLCB0);
	case PMLCB1:
		return mfpmr(PMLCB1);
	case PMLCB2:
		return mfpmr(PMLCB2);
	case PMLCB3:
		return mfpmr(PMLCB3);
	case PMGC0:
		return mfpmr(PMGC0);

	default:
		break;
	}

	return -1;
}

static void set_pmr(uint32_t reg, uint32_t reg_nr, uint32_t val)
{
	switch (reg + reg_nr) {
	case PMC0:
		return mtpmr(PMC0, val);
	case PMC1:
		return mtpmr(PMC1, val);
	case PMC2:
		return mtpmr(PMC2, val);
	case PMC3:
		return mtpmr(PMC3, val);
	case PMLCA0:
		return mtpmr(PMLCA0, val);
	case PMLCA1:
		return mtpmr(PMLCA1, val);
	case PMLCA2:
		return mtpmr(PMLCA2, val);
	case PMLCA3:
		return mtpmr(PMLCA3, val);
	case PMLCB0:
		return mtpmr(PMLCB0, val);
	case PMLCB1:
		return mtpmr(PMLCB1, val);
	case PMLCB2:
		return mtpmr(PMLCB2, val);
	case PMLCB3:
		return mtpmr(PMLCB3, val);
	case PMGC0:
		return mtpmr(PMGC0, val);

	default:
		break;
	}
}

void initPM(void)
{
	union pmr_lca u;
	union pmr_gcr u1;

	u.reg_val = 0;
	u.pmlca.fc = PM_DISABLE;
	set_pmr(PMLCA, 0, u.reg_val);
	set_pmr(PMLCA, 1, u.reg_val);
	set_pmr(PMLCA, 2, u.reg_val);
	set_pmr(PMLCA, 3, u.reg_val);

	u1.pmgcr.fac = 0;
	u1.pmgcr.pmie = PMI_DISABLE;
	u1.pmgcr.fcece = 0;

	set_pmr(PMGC, 0, u1.reg_val);
}

int configurePM_event(enum PM_event event, enum PME_state initial_state,
		      uint32_t threshold, uint32_t threshmul,
		      enum PMI_state pmi, pmr_handle_t *handle)
{
	union pmr_lca u;
	int thresh_event = 0;

	if (event == PM_NOTHING) {
		LWE_ERROR("%s: PM_NOTHING event cannot be used to monitor"
			  "anything\n", __func__);
		return -EINVAL;
	}

	if (event > MAX_PM_EVENT ||
	   (event > PM_COHERENT_LOOKUP_MISS && event < PM_IAC1_DETECTED) ||
	   (event > PM_IAC2_DETECTED && event < PM_DAC1_DETECTED) ||
	   (event > PM_DAC2_DETECTED && event < PM_DVT0_DETECTED) ||
	   (event > PM_CYCLES_COMPLETION_STALLED &&
			event < PM_FPU_DOUBLEPUMP) ||
	   (event > PM_FPU_INPUT_DATA_STALL && event < PM_DECORATED_LOADS)) {
		LWE_ERROR("%s: Invalid Event %d\n", __func__, event);
		return -EINVAL;
	}

	if (event >= MIN_THRESHOLD_EVENT && event <= MAX_THRESHOLD_EVENT)
		thresh_event = 1;

	if (thresh_event) {
		if (threshold > MAX_THRESHOLD) {
			LWE_ERROR("%s: Threshold value is greater than max"
				" %d value\n", __func__, MAX_THRESHOLD);
			return -EINVAL;
		}

		if ((!threshmul || threshmul > MAX_THRESHMUL ||
		    (threshmul & (threshmul - 1)))) {
			LWE_ERROR("%s: Max threshold value multiplier is"
			  " %d, and it should be 2^n where n can be 0 - 7\n",
			   __func__, MAX_THRESHMUL);
			return -EINVAL;
		}
	}

	if (get_free_pmr(handle, thresh_event) < 0) {
		LWE_ERROR("%s: Performance Monitoring counter not available\n",
			  __func__);
		return -EBUSY;
	}

	/* Initialize the counter */
	set_pmr(PMC, *handle, 0);

	if (thresh_event) {
		/* set the threshold */
		union pmr_lcb u1;
		u1.pmlcb.threshold = threshold;
		u1.pmlcb.threshmul = count_lsb_zeroes(threshmul);

		set_pmr(PMLCB, *handle, u1.reg_val);
	}

	/* set the counter's control register */
	u.reg_val = 0;
	u.pmlca.event = event;
	u.pmlca.fcs = PM_ENABLE;
	u.pmlca.fcu = PM_ENABLE;
	u.pmlca.fcm0 = PM_ENABLE;
	u.pmlca.fcm1 = PM_ENABLE;
	u.pmlca.ce = pmi;
	u.pmlca.fc = initial_state;

	set_pmr(PMLCA, *handle, u.reg_val);

	return 0;
}

int unconfigurePM_event(pmr_handle_t handle)
{
	union pmr_lca u;

	if (handle > MAX_PMC_REG) {
		LWE_ERROR("%s: invalid handle\n", __func__);
		return -EINVAL;
	}

	u.reg_val = get_pmr(PMLCA, handle);
	if (u.pmlca.event != PM_NOTHING) {
		u.pmlca.fc = PM_DISABLE;
		u.pmlca.event = PM_NOTHING;
		u.pmlca.ce = PMI_DISABLE;

		set_pmr(PMLCA, handle, u.reg_val);
		set_pmr(PMLCB, handle, 0);
		set_pmr(PMC, handle, 0);
	}

	return 0;
}

int startPM_event(pmr_handle_t handle)
{
	union pmr_lca u;

	if (handle > MAX_PMC_REG) {
		LWE_ERROR("%s: invalid handle\n", __func__);
		return -EINVAL;
	}

	u.reg_val = get_pmr(PMLCA, handle);
	if (u.pmlca.event == PM_NOTHING) {
		LWE_ERROR("%s: No Event configured\n", __func__);
		return -EINVAL;
	}

	if (u.pmlca.fc == PM_DISABLE) {
		u.pmlca.fc = PM_ENABLE;
		set_pmr(PMLCA, handle, u.reg_val);
	}

	return 0;
}

int stopPM_event(pmr_handle_t handle)
{

	union pmr_lca u;

	if (handle > MAX_PMC_REG) {
		LWE_ERROR("%s: invalid handle\n", __func__);
		return -EINVAL;
	}
	u.reg_val = get_pmr(PMLCA, handle);
	if (u.pmlca.event == PM_NOTHING) {
		LWE_ERROR("%s: No Event configured\n", __func__);
		return -EINVAL;
	}

	if (u.pmlca.fc == PM_ENABLE) {
		u.pmlca.fc = PM_DISABLE;
		set_pmr(PMLCA, handle, u.reg_val);
	}

	return 0;
}

int restartPM_event(pmr_handle_t handle)
{
	union pmr_lca u;

	if (handle > MAX_PMC_REG) {
		LWE_ERROR("%s: invalid handle\n", __func__);
		return -EINVAL;
	}
	u.reg_val = get_pmr(PMLCA, handle);
	if (u.pmlca.event == PM_NOTHING) {
		LWE_ERROR("%s: No Event configured\n", __func__);
		return -EINVAL;
	}

	if (u.pmlca.fc == PM_ENABLE) {
		u.pmlca.fc = PM_DISABLE;
		set_pmr(PMLCA, handle, u.reg_val);
	}

	set_pmr(PMC, handle, 0);
	u.pmlca.fc = PM_ENABLE;
	set_pmr(PMLCA, handle, u.reg_val);

	return 0;
}

int getPM_event_count(pmr_handle_t handle, uint32_t *val)
{
	union pmr_lca u;

	if (!val) {
		LWE_ERROR("%s: out parameter val must not be NULL\n",
			  __func__);
		return -EINVAL;
	}

	if (handle > MAX_PMC_REG) {
		LWE_ERROR("%s: invalid handle\n", __func__);
		return -EINVAL;
	}

	u.reg_val = get_pmr(PMLCA, handle);
	if ((u.pmlca.event == PM_NOTHING)) {
		LWE_ERROR("%s: No Event configured\n", __func__);
		return -EINVAL;
	}

	*val = get_pmr(PMC, handle);

	return 0;
}

int getPM_configured_events(uint32_t *event_array)
{
	int i;
	union pmr_lca u;

	if (!event_array) {
		LWE_ERROR("%s: out parameter event_array must not be NULL\n",
			  __func__);
		return -EINVAL;
	}

	for (i = 0; i < 4; i++) {
		u.reg_val = get_pmr(PMLCA, i);
		event_array[i] = u.pmlca.event;
	}

	return 0;
}

int setPM_event_threshold(pmr_handle_t handle, uint32_t threshold,
			  uint32_t threshmul)
{
	union pmr_lcb ub;
	union pmr_lca ua;

	if (handle > 1) {
		LWE_ERROR("%s: invalid handle\n", __func__);
		return -EINVAL;
	}

	ua.reg_val = get_pmr(PMLCA, handle);
	if (ua.pmlca.event < MIN_THRESHOLD_EVENT &&
	   ua.pmlca.event > MAX_THRESHOLD_EVENT) {
		LWE_ERROR("%s: The Event is not a threshold event\n", __func__);
		return -EINVAL;
	}

	if (threshold > MAX_THRESHOLD) {
		LWE_ERROR("%s: Threshold value is greater than max %d value\n",
			  __func__, MAX_THRESHOLD);
		return -EINVAL;
	}

	if (!threshmul || threshmul > MAX_THRESHMUL ||
	    (threshmul & (threshmul - 1))) {
		LWE_ERROR("%s: Max threshold value multiplier is %d, and it"
			  "should be 2^n where n can be 0 - 7\n", __func__,
			  MAX_THRESHMUL);
		return -EINVAL;
	}

	ub.pmlcb.threshold = threshold;
	ub.pmlcb.threshmul = count_lsb_zeroes(threshmul);

	set_pmr(PMLCB, handle, ub.reg_val);

	return 0;
}

int getPM_event_status(pmr_handle_t handle, struct pmes *status)
{
	union pmr_lca u;
	union pmr_gcr u1;
	uint32_t val;

	if (!status) {
		LWE_ERROR("%s: out parameter status must not be NULL\n",
			  __func__);
		return -EINVAL;
	}

	if (handle > MAX_PMC_REG) {
		LWE_ERROR("%s: invalid handle\n", __func__);
		return -EINVAL;
	}

	u.reg_val = get_pmr(PMLCA, handle);
	if (u.pmlca.event == PM_NOTHING) {
		LWE_ERROR("%s: No Event configured\n", __func__);
		return -EINVAL;
	}

	status->event = u.pmlca.event;
	status->state = u.pmlca.fc;

	if (status->event >= MIN_THRESHOLD_EVENT &&
		status->event <= MAX_THRESHOLD_EVENT) {
		val = get_pmr(PMLCB, handle);
		status->threshold = THRESHOLD(val);

		status->threshmul = 1 << THRESHMUL(val);
	}

	val = get_pmr(PMGC, 0);
	u1.reg_val = (uint32_t) val;

	status->pmi = PMI_DISABLE;
	if (u1.pmgcr.pmie == PMI_ENABLE)	/* interrupts enabled */
		status->pmi = u.pmlca.ce;

	return 0;
}

int enablePM_event_interrupt(pmr_handle_t handle)
{
	union pmr_lca u;
	union pmr_gcr u1;

	if (handle > MAX_PMC_REG) {
		LWE_ERROR("%s: invalid handle\n", __func__);
		return -EINVAL;
	}
	u.reg_val = get_pmr(PMLCA, handle);
	if (u.pmlca.event == PM_NOTHING) {
		LWE_ERROR("%s: No Event configured\n", __func__);
		return -EINVAL;
	}

	u.pmlca.ce = PMI_ENABLE;
	set_pmr(PMLCA, handle, u.reg_val);

	u.reg_val = get_pmr(PMGC, 0);
	u1.pmgcr.pmie = PMI_ENABLE;
	set_pmr(PMGC, 0, u1.reg_val);

	return 0;
}

int disablePM_event_interrupt(pmr_handle_t handle)
{
	union pmr_lca u;

	if (handle > MAX_PMC_REG) {
		LWE_ERROR("%s: invalid handle\n", __func__);
		return -EINVAL;
	}
	u.reg_val = get_pmr(PMLCA, handle);
	if (u.pmlca.event == PM_NOTHING) {
		LWE_ERROR("%s: No Event configured\n", __func__);
		return -EINVAL;
	}

	u.pmlca.ce = PMI_DISABLE;
	set_pmr(PMLCA, handle, u.reg_val);

	return 0;
}
