/** @file profile_cmds.c
 * Performance Monitoring command's handler
 */

/*
 * Copyright (c) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <limits.h>
#include "perf_mon.h"
#include <compiler.h>
#include <cpu_usage.h>

__PERCPU static pmr_handle_t ld_event_handle, st_event_handle;
__PERCPU static pmr_handle_t dcache_reloads_handle;
__PERCPU static pmr_handle_t inst_cnt_handle, icache_reloads_handle;
__PERCPU static pmr_handle_t ipc_inst_cnt_handle, cycle_cnt_handle;

/*profiling commands*/
enum profile_cmd {
	START = 0,
	STOP,
	PAUSE,
	RESTART,
	RESUME,
	DISPLAY,
	MAX_NUM_CMDS
};

/*State of profiling*/
enum profiling_state {
	STARTED,
	PAUSED,
	STOPPED
};

__PERCPU static enum profiling_state dcache_profiling_state = STOPPED;
__PERCPU static enum profiling_state icache_profiling_state = STOPPED;
__PERCPU static enum profiling_state ipc_state = STOPPED;

/*Handler for the L1 data cache profile command*/
int shell_cmd_profile_dcache(int argc, char *argv[])
{
	uint32_t val_loads;
	uint32_t val_stores;
	uint32_t val;
	int status;
	int ret = 0;

	enum profile_cmd arg1;

	if (argc != 2) {
		printf("usage: profile_dcache <arg>\n");
		printf("Valid argument values:\n");
		printf("-----------------------\n");
		printf("       0 - Start the cache profiling events\n");
		printf("       1 - Stop the cache profiling events\n");
		printf("       2 - Pause the cache profiling events\n");
		printf("       3 - Restart the cache profiling"
						 " event's counters\n");
		printf("       4 - Resume the cache profiling"
						 " event's counters\n");
		printf("       5 - Display the events counter values\n");
		return -1;
	}

	errno = 0;
	arg1 = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == arg1)
			|| (0 != errno && 0 == arg1)) {
		printf("strtoul failed converting command's argument1.\n");
		return -1;
	}

	switch (arg1) {
	case START:
		if (dcache_profiling_state == STARTED) {
			printf("Data cache profling has already"
						" been started\n");
			return -1;
		}
		/*Configuring Load micro-ops completed event*/
		ret = configurePM_event(PM_LOAD_MIRCRO_OPS_COMPLETED,
					PM_ENABLE, 1, 4, PMI_DISABLE,
							&ld_event_handle);

		if (ret < 0) {
			printf("Couldn't configure load micro-ops event\n");
			return -1;
		}

		/*Configuring Store micro-ops completed events*/
		ret = configurePM_event(PM_STORE_MIRCRO_OPS_COMPLETED,
					PM_ENABLE, 1, 4, PMI_DISABLE,
						&st_event_handle);

		if (ret < 0) {
			printf("Couldn't configure store micro-ops"
					" completed event\n");
			unconfigurePM_event(ld_event_handle);
			return -1;
		}


		/*Configuring Data L1 cache reloads event*/
		ret = configurePM_event(PM_DATA_L1_CACHE_RELOADS,
					PM_ENABLE, 1, 4, PMI_DISABLE,
					&dcache_reloads_handle);

		if (ret < 0) {
			printf("Couldn't configure data"
					 " L1 cache reloads event\n");
			unconfigurePM_event(ld_event_handle);
			unconfigurePM_event(st_event_handle);
			return -1;
		}

		/*starting Load micro-ops completed event*/
		status = startPM_event(ld_event_handle);
		if (status) {
			printf("Couldn't start load"
				 " micro-ops event counting\n");
			unconfigurePM_event(ld_event_handle);
			unconfigurePM_event(st_event_handle);
			unconfigurePM_event(dcache_reloads_handle);
			return -1;
		}

		/*starting Store micro-ops completed events*/
		status = startPM_event(st_event_handle);
		if (status) {
			printf("Couldn't start store micro-ops"
				" completed event counting\n");
			unconfigurePM_event(ld_event_handle);
			unconfigurePM_event(st_event_handle);
			unconfigurePM_event(dcache_reloads_handle);
			return -1;
		}

		/*starting Data L1 cache reloads event*/
		status = startPM_event(dcache_reloads_handle);
		if (status) {
			printf("Couldn't start Data L1 cache"
				" reloads event counting\n");
			unconfigurePM_event(ld_event_handle);
			unconfigurePM_event(st_event_handle);
			unconfigurePM_event(dcache_reloads_handle);
			return -1;
		}

		dcache_profiling_state = STARTED;
		printf("Started L1 Data cache profiling\n");
		break;

	case STOP:
		if ((dcache_profiling_state != STARTED) &&
					(dcache_profiling_state != PAUSED)) {
			printf("Data cache profling has not been started,"
						" nothing to stop\n");
			return -1;
		}

		unconfigurePM_event(ld_event_handle);
		unconfigurePM_event(st_event_handle);
		unconfigurePM_event(dcache_reloads_handle);

		dcache_profiling_state = STOPPED;
		printf("L1 Data cache profiling stopped.\n");
		break;

	case PAUSE:
		if (dcache_profiling_state != STARTED) {
			printf("Data cache profling has not been started,"
						" nothing to pause\n");
			return -1;
		}

		stopPM_event(ld_event_handle);
		stopPM_event(st_event_handle);
		stopPM_event(dcache_reloads_handle);
		dcache_profiling_state = PAUSED;
		printf("L1 Data cache profiling events paused.\n");
		break;

	case RESTART:
		if ((dcache_profiling_state != STARTED) &&
					(dcache_profiling_state != PAUSED)) {
			printf("Data cache profling has not been started,"
						" cannot restart\n");
			return -1;
		}

		status = restartPM_event(ld_event_handle);
		if (status) {
			printf("Couldn't Restart Load micro-ops"
					" completed event.\n");
			return -1;
		}

		status = restartPM_event(st_event_handle);
		if (status) {
			printf("couldn't Restart Store micro-ops"
					" completed event.\n");
			stopPM_event(ld_event_handle);
			stopPM_event(st_event_handle);
			stopPM_event(dcache_reloads_handle);
			return -1;
		}

		status = restartPM_event(dcache_reloads_handle);
		if (status) {
			printf("Couldn't Restart L1 data cache"
						" reloads event.\n");
			stopPM_event(ld_event_handle);
			stopPM_event(st_event_handle);
			stopPM_event(dcache_reloads_handle);
			return -1;
		}

		dcache_profiling_state = STARTED;
		printf("L1 Data cache profiling restarted.\n");
		break;

	case RESUME:
		if (dcache_profiling_state != PAUSED) {
			printf("Data cache profling has not been paused,"
						" so cannot resume.\n");
			return -1;
		}

		status = startPM_event(ld_event_handle);
		if (status) {
			printf("Couldn't start load micro-ops"
					" completed event counting\n");
			return -1;
		}

		status = startPM_event(st_event_handle);
		if (status) {
			printf("Couldn't start store micro-ops"
					" completed event counting\n");
			stopPM_event(ld_event_handle);
			return -1;
		}

		status = startPM_event(dcache_reloads_handle);
		if (status) {
			printf("Couldn't start Data L1 cache"
					" event counting\n");
			stopPM_event(ld_event_handle);
			stopPM_event(st_event_handle);
			return -1;
		}

		dcache_profiling_state = STARTED;
		printf("Resumed L1 Data cache profiling.\n");
		break;

	case DISPLAY:
		if (dcache_profiling_state == STOPPED) {
			printf("Data cache profiling not in running state.\n");
			return -1;
		}

		status = getPM_event_count(ld_event_handle, &val_loads);
		if (status) {
			printf("Event counting for load micro-ops"
						" completed event failed\n");
			return -1;
		} else {
			printf("Load micro-ops completed are:%u\n", val_loads);
		}

		status = getPM_event_count(st_event_handle, &val_stores);
		if (status) {
			printf("Event counting for stores completed"
						" event failed\n");
			return -1;
		} else {
			printf("Store micro-ops completed are: %u\n",
							val_stores);
		}

		status = getPM_event_count(dcache_reloads_handle, &val);
		if (status) {
			printf("Event counting for data cache event failed\n");
			return -1;
		} else {
			printf("L1 data cache reloads are: %u\n", val);
		}

		printf("L1 data cache miss rate: %u / (%u + %u)\n",
						val, val_loads, val_stores);
		break;

	default:
		printf("usage: profile_dcache <arg>\n");
		printf("Valid argument values:\n");
		printf("----------------------\n");
		printf("       0 - Start the cache profiling events\n");
		printf("       1 - Stop the cache profiling events\n");
		printf("       2 - Pause the cache profiling events\n");
		printf("       3 - Restart the cache profiling event's"
							 " counters\n");
		printf("       4 - Resume the cache profiling event's"
							 " counters\n");
		printf("       5 - Display the events counter values\n");
		break;
	}

	return 0;
}

/*Handler for the L1 Instruction cache profile command*/
int shell_cmd_profile_icache(int argc, char *argv[])
{
	uint32_t val_inst;
	uint32_t val;
	int status;
	int ret = 0;

	enum profile_cmd arg1;

	if (argc != 2) {
		printf("usage: profile_icache <arg>\n");
		printf("Valid argument values:\n");
		printf("-----------------------\n");
		printf("       0 - Start the cache profiling events\n");
		printf("       1 - Stop the cache profiling events\n");
		printf("       2 - Pause the cache profiling events\n");
		printf("       3 - Restart the cache profiling event's"
							" counters\n");
		printf("       4 - Resume the cache profiling event's"
							" counters\n");
		printf("       5 - Display the events counter values\n");
		return -1;
	}

	errno = 0;
	arg1 = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == arg1)
			|| (0 != errno && 0 == arg1)) {
		printf("strtoul failed converting command's argument1\n");
		return -1;
	}

	switch (arg1) {
	case START:
		if (icache_profiling_state != STOPPED) {
			printf("Instruction cache profling has"
					" already been started\n");
			return -1;
		}
		/*Configuring PM Instruction Completed event*/
		ret = configurePM_event(PM_INSTRUCTIONS_COMPLETED,
				PM_ENABLE, 1, 4, PMI_DISABLE, &inst_cnt_handle);

		if (ret < 0) {
			printf("Couldn't configure PM Instructions"
					 " Completed event\n");
			return -1;
		}

		/*Configuring Instruction L1 cache reloads event*/
		ret = configurePM_event
				(PM_INSTRUCTION_L1_CACHE_RELOADS_FROM_FETCH,
				PM_ENABLE, 1, 4, PMI_DISABLE,
				&icache_reloads_handle);

		if (ret < 0) {
			printf("Couldn't configure L1 Instruction"
							" cache event\n");
			unconfigurePM_event(inst_cnt_handle);
			return -1;
		}

		/*starting PM Instruction Completed event*/
		status = startPM_event(inst_cnt_handle);
		if (status) {
			printf("Couldn't start PM Instructions"
				 " Completed event counting\n");
			unconfigurePM_event(inst_cnt_handle);
			unconfigurePM_event(icache_reloads_handle);
			return -1;
		}

		/*starting Instruction L1 cache reloads event*/
		status = startPM_event(icache_reloads_handle);
		if (status) {
			printf("Couldn't start Instruction L1 cache"
						" event counting\n");
			unconfigurePM_event(inst_cnt_handle);
			unconfigurePM_event(icache_reloads_handle);
			return -1;
		}

		icache_profiling_state = STARTED;
		printf("Started L1 Instruction cache profiling.\n");
		break;

	case STOP:
		if ((icache_profiling_state != STARTED) &&
					(icache_profiling_state != PAUSED)) {
			printf("Instruction cache profling has not been"
						" started, nothing to stop\n");
			return -1;
		}

		unconfigurePM_event(inst_cnt_handle);
		unconfigurePM_event(icache_reloads_handle);

		icache_profiling_state = STOPPED;
		printf("Instruction Cache profiling stopped.\n");
		break;

	case PAUSE:
		if (icache_profiling_state != STARTED) {
			printf("Instruction cache profling has not been"
					" started, nothing to pause\n");
			return -1;
		}

		stopPM_event(inst_cnt_handle);
		stopPM_event(icache_reloads_handle);
		icache_profiling_state = PAUSED;
		printf("Instruction Cache profiling events paused.\n");
		break;

	case RESTART:
		if ((icache_profiling_state != STARTED) &&
				(icache_profiling_state != PAUSED)) {
			printf("Instruction cache profling has not been"
						" started, cannot restart\n");
			return -1;
		}

		status = restartPM_event(inst_cnt_handle);
		if (status) {
			printf("Couldn't Restart Instructions count event.\n");
			return -1;
		}

		status = restartPM_event(icache_reloads_handle);
		if (status) {
			printf("Couldn't Restart L1 Instruction cache"
							" reloads event.\n");
			stopPM_event(inst_cnt_handle);
			stopPM_event(icache_reloads_handle);
			return -1;
		}

		icache_profiling_state = STARTED;
		printf("L1 Instruction cache profiling restarted.\n");
		break;

	case RESUME:
		if (icache_profiling_state != PAUSED) {
			printf("Instruction cache profiling has not been"
					" paused, so cannot resume.\n");
			return -1;
		}

		status = startPM_event(inst_cnt_handle);
		if (status) {
			printf("Couldn't start PM Instructions"
					" Completed event counting\n");
			return -1;
		}

		status = startPM_event(icache_reloads_handle);
		if (status) {
			printf("Couldn't start Instruction L1 cache"
						" event counting\n");
			stopPM_event(inst_cnt_handle);
			return -1;
		}

		icache_profiling_state = STARTED;
		printf("Resumed L1 Instruction cache profiling.\n");
		break;

	case DISPLAY:
		if (icache_profiling_state == STOPPED) {
			printf("Instruction cache profiling has not been"
							" started.\n");
			return -1;
		}

		status = getPM_event_count(inst_cnt_handle, &val_inst);
		if (status) {
			printf("Event counting for Instructions"
					" completed event failed\n");
			return -1;
		} else {
			printf("Instructions completed are: %u\n", val_inst);
		}

		status = getPM_event_count(icache_reloads_handle, &val);
		if (status) {
			printf("Event counting for instruction cache"
						" reloads event failed\n");
			return -1;
		} else {
			printf("L1 Instruction cache reloads are: %u\n", val);
		}
		printf("L1 Instruction cache miss rate: %u / (%u)\n",
							val, val_inst);
		break;

	default:
		printf("usage: profile_icache <arg>\n");
		printf("Valid argument values:\n");
		printf("----------------------\n");
		printf("       0 - Start the cache profiling events\n");
		printf("       1 - Stop the cache profiling events\n");
		printf("       2 - Pause the cache profiling events\n");
		printf("       3 - Restart the cache profiling event's"
							" counters\n");
		printf("       4 - Resume the cache profiling event's"
							" counters\n");
		printf("       5 - Display the events counter values\n");
		break;
	}

	return 0;
}

/*Handler for the IPC(Instructions per cycle) command*/
int shell_cmd_ipc(int argc, char *argv[])
{
	uint32_t val_inst;
	uint32_t val_cycles;
	int status;
	int ret = 0;

	enum profile_cmd arg1;

	if (argc != 2) {
		printf("usage: ipc <arg>\n");
		printf("Valid argument values:\n");
		printf("-----------------------\n");
		printf("       0 - Start the ipc events\n");
		printf("       1 - Stop the ipc events\n");
		printf("       2 - Pause the ipc events\n");
		printf("       3 - Restart the ipc event's"
						" counters\n");
		printf("       4 - Resume the ipc event's"
						" counters\n");
		printf("       5 - Display IPC and related events values\n");
		return -1;
	}

	errno = 0;
	arg1 = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == arg1)
			|| (0 != errno && 0 == arg1)) {
		printf("strtoul failed converting command's argument1\n");
		return -1;
	}

	switch (arg1) {
	case START:
		if (ipc_state != STOPPED) {
			printf("Events for IPC have already been"
						" started\n");
			return -1;
		}
		/*Configuring PM Instruction Completed event*/
		ret = configurePM_event(PM_INSTRUCTIONS_COMPLETED,
					PM_ENABLE, 1, 4, PMI_DISABLE,
					&ipc_inst_cnt_handle);

		if (ret < 0) {
			printf("Couldn't configure PM Instructions"
					 " Completed event\n");
			return -1;
		}

		/*Configuring Cycle count event*/
		ret = configurePM_event(PM_CYCLE, PM_ENABLE, 1, 4,
					PMI_DISABLE, &cycle_cnt_handle);

		if (ret < 0) {
			printf("Couldn't configure cycle count event\n");
			unconfigurePM_event(ipc_inst_cnt_handle);
			return -1;
		}

		/*starting PM Instruction Completed event*/
		status = startPM_event(ipc_inst_cnt_handle);
		if (status) {
			printf("Couldn't start PM Instructions"
				 " Completed event counting\n");
			unconfigurePM_event(ipc_inst_cnt_handle);
			unconfigurePM_event(cycle_cnt_handle);
			return -1;
		}

		/*starting Cycle count event*/
		status = startPM_event(cycle_cnt_handle);
		if (status) {
			printf("Couldn't start cycle counting\n");
			unconfigurePM_event(ipc_inst_cnt_handle);
			unconfigurePM_event(cycle_cnt_handle);
			return -1;
		}

		ipc_state = STARTED;
		printf("Started events for calculating IPC.\n");
		break;

	case STOP:
		if ((ipc_state != STARTED) &&
					(ipc_state != PAUSED)) {
			printf("Events for IPC have not been"
					" started, nothing to stop\n");
			return -1;
		}

		unconfigurePM_event(ipc_inst_cnt_handle);
		unconfigurePM_event(cycle_cnt_handle);

		ipc_state = STOPPED;
		printf("IPC related events stopped.\n");
		break;

	case PAUSE:
		if (ipc_state != STARTED) {
			printf("Events for IPC have not been"
					" started, nothing to pause\n");
			return -1;
		}

		stopPM_event(ipc_inst_cnt_handle);
		stopPM_event(cycle_cnt_handle);
		ipc_state = PAUSED;
		printf("IPC related events paused.\n");
		break;

	case RESTART:
		if ((ipc_state != STARTED) &&
				(ipc_state != PAUSED)) {
			printf("Events for IPC have not been"
						" started, cannot restart\n");
			return -1;
		}

		status = restartPM_event(ipc_inst_cnt_handle);
		if (status) {
			printf("Couldn't Restart Instructions count event.\n");
			return -1;
		}

		status = restartPM_event(cycle_cnt_handle);
		if (status) {
			printf("Couldn't Restart cycle count event.\n");
			stopPM_event(ipc_inst_cnt_handle);
			stopPM_event(cycle_cnt_handle);
			return -1;
		}

		ipc_state = STARTED;
		printf("IPC related events restarted.\n");
		break;

	case RESUME:
		if (ipc_state != PAUSED) {
			printf("Events for IPC have not been"
					" paused, so cannot resume.\n");
			return -1;
		}

		status = startPM_event(ipc_inst_cnt_handle);
		if (status) {
			printf("Couldn't start PM Instructions"
					" Completed event counting\n");
			return -1;
		}

		status = startPM_event(cycle_cnt_handle);
		if (status) {
			printf("Couldn't start cycle count"
						" event counting\n");
			stopPM_event(ipc_inst_cnt_handle);
			return -1;
		}

		ipc_state = STARTED;
		printf("Events for IPC have been Resumed.\n");
		break;

	case DISPLAY:
		if (ipc_state == STOPPED) {
			printf("Instructions per cycle profiling not in"
						" running state.\n");
			return -1;
		}

		status = getPM_event_count(ipc_inst_cnt_handle, &val_inst);
		if (status) {
			printf("Event counting for Instructions"
					" completed event failed\n");
			return -1;
		} else {
			printf("Instructions completed are: %u\n", val_inst);
		}

		status = getPM_event_count(cycle_cnt_handle, &val_cycles);
		if (status) {
			printf("Event counting for cycle count failed\n");
			return -1;
		} else {
			printf("Total cycle count is: %u\n", val_cycles);
		}
		printf("Instructions executed per cycle(IPC) are: %u / (%u)\n",
							val_inst, val_cycles);
		break;

	default:
		printf("usage: ipc <arg>\n");
		printf("Valid argument values:\n");
		printf("-----------------------\n");
		printf("       0 - Start the ipc events\n");
		printf("       1 - Stop the ipc events\n");
		printf("       2 - Pause the ipc events\n");
		printf("       3 - Restart the ipc event's"
						" counters\n");
		printf("       4 - Resume the ipc event's"
						" counters\n");
		printf("       5 - Display IPC and related events values\n");
		break;
	}

	return 0;
}

int shell_cmd_top(int argc __UNUSED, char *argv[] __UNUSED)
{
	printf("The CPU Usage is: %d%\n", get_core_usage());
	return 0;
}
