/**
 \file rbtree.c
 \brief Implementation of APIs of RED-BLACK tree
 */
/*
 * Copyright (c) 2009 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "rbtree.h"

/**  function name  create_rbtree
 \   param[in]      compare_func - pointer of compare function
 \   return    	    returns the tree pointer
 \   brief	    Allocate the memory for the tree, create a sentinel node
*/
struct red_black_tree *create_rbtree
		(int (*compare_func) (const void *, const void *))
{
	struct red_black_tree *newtree;
	struct rbt_node *newnode;

	newtree =
	    (struct red_black_tree *)malloc(sizeof(struct red_black_tree));
	newtree->compfunc = compare_func;
	newnode = newtree->nil =
	    (struct rbt_node *)malloc(sizeof(struct rbt_node));
	newnode->parent = newnode->left = newnode->right = NULL;
	newnode->color = 0;
	newtree->root = newtree->nil;
	return newtree;
}

/** function name   rotate_left
 \  param[in]       struct red_black_tree * - pointer of the rbtree and
		    struct rbt_node * - pointer of the node to rotate on.
 \  return	    None
 \  brief           This function makes the right child of x,
		    to be the parent of x, previous parent of x
		    now points to the right child of x
		    and fixes other pointers accordingly.
 */
void rotate_left(struct red_black_tree *tree, struct rbt_node *x)
{
	struct rbt_node *y;

	y = x->right;
	x->right = y->left;

	if (y->left != tree->nil)
		y->left->parent = x;	/* as sentinel used here */
	/* therefore an unconditional assignment has been
			made instead of testing for nil*/
	y->parent = x->parent;

	/* instead of checking if x->parent is the root, we */
	/* rely on the root sentinel to implicitly take care of this case */
	if (x != tree->root) {
		if (x == x->parent->left)
			x->parent->left = y;
		else
			x->parent->right = y;
	} else
		tree->root = y;
	y->left = x;
	x->parent = y;
}

/** function name   rotate_right
 \  param[in]       struct red_black_tree * - pointer of the rbtree and
		    struct rbt_node * - pointer of the node to rotate on.
 \  return	    None
 \  brief           This function makes the left child of y,
		    to be the parent of y, previous parent of y
		    now points to the left child of y
		    and fixes other pointers accordingly.
 */

void rotate_right(struct red_black_tree *tree, struct rbt_node *y)
{
	struct rbt_node *x;

	x = y->left;
	y->left = x->right;

	if (tree->nil != x->right)
		x->right->parent = y;	/*as sentinel used here */
	/* therefore an unconditional assignment has been made
				instead of testing for nil*/

	/* instead of checking if y->parent is the root, we */
	/* rely on the root sentinel to implicitly take care of this case */
	x->parent = y->parent;
	if (y != tree->root) {
		if (y == y->parent->left)
			y->parent->left = x;
		else
			y->parent->right = x;
	} else
		tree->root = x;
	x->right = y;
	y->parent = x;
}

/*  Before calling Insert RBTree the node x should have its key set */

/** function name   rbtree_insert_node
 \  param[in] 	    struct red_black_tree * - pointer of the rbtree in which to insert a node and
		    struct rbt_node * - pointer of the node to insert.
 \  return	    This function returns a pointer to the newly inserted node
		    which is guarunteed to be valid until this node is deleted.
		    What this means is if another data structure stores this
		    pointer then the tree does not need to be searched when this
		    is to be deleted.
 \  brief   Creates a node node which contains the appropriate key and
	    data pointers and inserts it into the tree.
 */

struct rbt_node *rbtree_insert_node(struct red_black_tree *tree,
				    struct rbt_node *x)
{
	struct rbt_node *newNode, *y;

	newNode = x;
	x->color = 1;		/*color newnode red */
	while ((x != tree->root) && (x->parent->color)) {
			/* use sentinel instead of checking for root */
		if (x->parent == x->parent->parent->left) {
			y = x->parent->parent->right;
			if (y == tree->nil) {
				if (x == x->parent->right) {
					x = x->parent;
					rotate_left(tree, x);
				}
				x->parent->color = 0;
				x->parent->parent->color = 1;
				rotate_right(tree, x->parent->parent);
			} else {
				if (y->color) {
					x->parent->color = 0;
					y->color = 0;
					x->parent->parent->color = 1;
					x = x->parent->parent;
				} else {
					if (x == x->parent->right) {
						x = x->parent;
						rotate_left(tree, x);
					}
					x->parent->color = 0;
					x->parent->parent->color = 1;
					rotate_right(tree, x->parent->parent);
				}
			}
		} else {
		/* case for x->parent == x->parent->parent->right */
			y = x->parent->parent->left;
			if (y == tree->nil) {
				if (x == x->parent->left) {
					x = x->parent;
					rotate_right(tree, x);
				}
				x->parent->color = 0;
				x->parent->parent->color = 1;
				rotate_left(tree, x->parent->parent);
			} else {
				if (y->color) {
					x->parent->color = 0;
					y->color = 0;
					x->parent->parent->color = 1;
					x = x->parent->parent;
				} else {
					if (x == x->parent->left) {
						x = x->parent;
						rotate_right(tree, x);
					}
					x->parent->color = 0;
					x->parent->parent->color = 1;
					rotate_left(tree, x->parent->parent);
				}
			}
		}
	}
	tree->root->color = 0;
	return newNode;
}

/**  function name	find_successor
 \   param[in]		struct red_black_tree * - pointer of the rbtree and
			struct rbt_node * - pointer of the node we want the
			the successor of
 \   return		This function returns the successor of x or NULL if no
			successor exists.
 */
struct rbt_node *find_successor(const struct red_black_tree *tree,
						struct rbt_node *x)
{
	struct rbt_node *y;

	y = x->right;
	if (y != tree->nil) {
		while (y->left != tree->nil) {
			/* returns the minium of the right subtree of x */
			y = y->left;
		}
		return y;
	} else {
		y = x->parent;
		while (x == y->right) {
			/* sentinel used instead of checking for nil */
			x = y;
			y = y->parent;
		}
		if (y == tree->root)
			return NULL;
		return y;
	}
}

/**  function name	find_predecessor
 \   param[in]		struct red_black_tree * - pointer of the rbtree and
			struct rbt_node * - pointer of the node we want the
			the predecessor of.
 \   return		This function returns the predecessor of x or NULL if no
			predecessor exists.
 */
struct rbt_node *find_predecessor(const struct red_black_tree *tree,
				  struct rbt_node *x)
{
	struct rbt_node *y;

	y = x->left;
	if (tree->nil != y) {
		while (y->right != tree->nil) {
			/* returns the maximum of the left subtree of x */
			y = y->right;
		}
		return y;
	} else {
		y = x->parent;
		while (x == y->left) {
			if (y == tree->root)
				return NULL;
			x = y;
			y = y->parent;
		}
		return y;
	}
}

/**  function name	delete_restructure
 \   param[in]		struct red_black_tree * - pointer of the rbtree and
			struct rbt_node * - pointer of the node on which
			restructuring is to be done
 \   return		none
 \   brief		Performs rotations and changes colors to restore
			red-black properties after a node is deleted
 */
void delete_restructure(struct red_black_tree *tree, struct rbt_node *x)
{
	struct rbt_node *w;

	while ((x->color == 0) && (tree->root != x)) {
		if (x == x->parent->left) {
			w = x->parent->right;
			if (w->color) {
				w->color = 0;
				x->parent->color = 1;
				rotate_left(tree, x->parent);
				w = x->parent->right;
			}
			if ((!w->right->color) && (!w->left->color)) {
				w->color = 1;
				x = x->parent;
			} else {
				if (!w->right->color) {
					w->left->color = 0;
					w->color = 1;
					rotate_right(tree, w);
					w = x->parent->right;
				}
				w->color = x->parent->color;
				x->parent->color = 0;
				w->right->color = 0;
				rotate_left(tree, x->parent);
				x = tree->root;	/* this is to exit while loop */
			}
		} else {
			w = x->parent->left;
			if (w->color) {
				w->color = 0;
				x->parent->color = 1;
				rotate_right(tree, x->parent);
				w = x->parent->left;
			}
			if ((!w->right->color) && (!w->left->color)) {
				w->color = 1;
				x = x->parent;
			} else {
				if (!w->left->color) {
					w->right->color = 0;
					w->color = 1;
					rotate_left(tree, w);
					w = x->parent->left;
				}
				w->color = x->parent->color;
				x->parent->color = 0;
				w->left->color = 0;
				rotate_right(tree, x->parent);
				x = tree->root;	/* this is to exit while loop */
			}
		}
	}
	x->color = 0;
}

/**  function name	rbtree_delete_node
 \   param[in]		struct red_black_tree * - pointer of the rbtree and
			struct rbt_node * - pointer of the node to be deleted
			from the tree
 \   return		none
 \   brief		Deletes z from tree and frees the key and data of z
			Then calls delete_restructure to restore
			red-black properties
 */
struct rbt_node *rbtree_delete_node(struct red_black_tree *tree,
				    struct rbt_node *z)
{
	struct rbt_node *y, *x;
	struct rbt_node *del_node = tree->nil;

	y = ((z->left == tree->nil)
	     || (z->right == tree->nil)) ? z : find_successor(tree, z);
	x = (y->left == tree->nil) ? y->right : y->left;

	x->parent = y->parent;
	if (y->parent == tree->nil) {
		tree->root = x;
		tree->root->parent = tree->nil;
	} else {
		if (y == y->parent->left)
			y->parent->left = x;
		else
			y->parent->right = x;
	}

	if (y != tree->nil) {
		if (y != z) {
			y->parent = z->parent;
			y->left = z->left;
			y->right = z->right;
			z->left->parent = z->right->parent = y;
			if (z->parent != tree->nil) {
				if (z == z->parent->left)
					z->parent->left = y;
				else
					z->parent->right = y;
			} else
				tree->root = y;
			if (!(y->color))
				delete_restructure(tree, x);
			del_node = z;
		} else {
			if (!(y->color))
				delete_restructure(tree, x);
			del_node = y;
		}
	}

	return del_node;

}
