/** @file
 * This file provides additional APIs for handling of RCU locks
 */

/*
 * Copyright (c) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <compiler.h>
#include <init.h>
#include <rcu_lock.h>
#include <statistics.h>
#include <lwe_malloc.h>
#include <cache.h>
#include <misc.h>

#include "statistics.h"

#define	NUM_CORES			8
#define RECLAIMEE_QUEUE_SIZE		1024
#define RECLAIMEE_QUEUE_ENTRIES 	(RECLAIMEE_QUEUE_SIZE - 1)
#define WAITING_RECLAIMEES_SIZE 	(RECLAIMEE_QUEUE_SIZE / NUM_CORES)
#define WAITING_RECLAIMEES_ENTRIES	(WAITING_RECLAIMEES_SIZE - 1)
#define MAX_WAITING_RECLAIMEES		(WAITING_RECLAIMEES_ENTRIES)

#define RCU_CPU_NOT_SYNCED	        1
#define RCU_CPU_SYNCED		        0
#define RCU_NUM_WORDS			(NUM_CORES >> 2)
#define RCU_DESYNCED_WORD		0x01010101

struct rcu_statistics_t {
	union stat64_t reclaimer_runs;
	union stat64_t reclaimed_count;
	uint32_t reclaimee_head;
};

struct rcu_reclaimee_t {
	free_func_t free_func;
	void *data;
	void *ctxt;
};

struct rcu_reclaimer_t {
	union {
		volatile uint8_t per_core[NUM_CORES];
		volatile uint32_t words[RCU_NUM_WORDS];
	} flags;
	uint32_t lock;
	struct rcu_reclaimee_t list[RECLAIMEE_QUEUE_SIZE];
	struct rcu_statistics_t *stats;
	union {
		volatile uint8_t per_core[NUM_CORES];
		volatile uint32_t words[RCU_NUM_WORDS];
	} running_cpu_map;
};

static uint32_t __rcu_update_reclaimee_head(void);
static inline uint32_t __rcu_waiting_reclaimees(uint32_t head);
static void __run_reclaimer(uint32_t head);
static void __rcu_acquire_reclaimer_lock(struct rcu_reclaimer_t *reclaimer,
					 uint32_t cpuid);
static void __rcu_desynchronize_all_cpus(void);
static void __rcu_barrier(volatile uint32_t * words);

/**	Holds the object in RCU data base for reclaimation
 */
struct rcu_reclaimer_t g_rcu_reclaimer;

/** Informs the RCU library that a new CPU want to access RCU protected DB.
 * @param[in] cpu_id CPU ID
 * note: The caller must ensure to pass a valid value of cpu_id.
 */
void
rcu_join(uint32_t cpu_id)
{
	g_rcu_reclaimer.running_cpu_map.per_core[cpu_id] = RCU_CPU_NOT_SYNCED;
}

/** Informs the RCU library that a CPU wants to stop using RCU protected DB.
 * @param[in] cpu_id CPU ID
 * note: The caller must ensure to pass a valid value of cpu_id.
 */
void rcu_leave(uint32_t cpu_id)
{
	g_rcu_reclaimer.running_cpu_map.per_core[cpu_id] = RCU_CPU_SYNCED;
}

/** RCU initializaton: An application needs to call this routine if it needs to protect some data 
 * using RCU.
 * @return Success: Pointer to initialized RCU data structure;
 *         Fail: NULL.
 */
struct rcu_reclaimer_t *
rcu_initialize(void)
{
	int i;

	for (i = 0; i < RCU_NUM_WORDS; i++) {
		g_rcu_reclaimer.flags.words[i] = 0;
		g_rcu_reclaimer.running_cpu_map.words[i] = RCU_CPU_SYNCED;
	}

	g_rcu_reclaimer.stats =
	    stats_memalign(CACHE_LINE_SIZE, sizeof(struct rcu_statistics_t));

	if (unlikely(NULL == g_rcu_reclaimer.stats))
		return NULL;

	memset(g_rcu_reclaimer.stats, 0, sizeof (struct rcu_statistics_t));

	return &g_rcu_reclaimer;
}

/** RCU reclaimer
 * @param[in] free_func Name of reclaimer function which does the work
 *	      of freeing memory or deletion of data etc.
 * @param[in] data_to_free Argument of reclaimer function.
 * @param[in] ctxt Argument of reclaimer function.
 */
void
rcu_free(free_func_t free_func, void *data_to_free, void *ctxt)
{
	uint32_t head;
	struct rcu_reclaimee_t *reclaimee;

	head = __rcu_update_reclaimee_head();
	reclaimee = &(g_rcu_reclaimer.list[head]);
	reclaimee->free_func = free_func;
	reclaimee->data = data_to_free;
	reclaimee->ctxt = ctxt;
	__run_reclaimer(head);
}

/**
 * Indicates that a particular CPU is done with all RCU read-side critical
 * sections
 * @param[in] cpu_id CPU ID which is to be synchronized.
 */
void
rcu_synchronize_cpu(uint32_t cpu_id)
{
	if (RCU_CPU_SYNCED != g_rcu_reclaimer.flags.per_core[cpu_id])
		g_rcu_reclaimer.flags.per_core[cpu_id] = RCU_CPU_SYNCED;
}

/** Increments reclaimation list head to point to next free list entry. 
 * @return Head of the list
 */
static uint32_t
__rcu_update_reclaimee_head()
{
	void *head_ptr;
	uint32_t head;

	head_ptr = &(g_rcu_reclaimer.stats->reclaimee_head);
	head = decorated_load_inc_32(head_ptr);
	head &= RECLAIMEE_QUEUE_ENTRIES;
	return head;
}

/** Returns number of entries waiting in reclaimation list. 
 * @return Number of entries in the list
 */
static inline uint32_t
__rcu_waiting_reclaimees(uint32_t head)
{
	return head & WAITING_RECLAIMEES_ENTRIES;
}

/** Reclaimer run routine. 
 *  It desynchronizes all CPUs and then waits till the time each CPU has 
 *  written its own synchronization word. At any point of time, only one
 *  CPU can be in critical section of this function.
 */
static void
__run_reclaimer(uint32_t head)
{
	struct rcu_reclaimer_t *reclaimer;
	struct rcu_reclaimee_t *reclaimee;
	uint32_t section_start, i, cpuid;

	if (__rcu_waiting_reclaimees(head) == MAX_WAITING_RECLAIMEES) {
		cpuid = get_lwe_id();
		reclaimer = &g_rcu_reclaimer;

		__rcu_acquire_reclaimer_lock(reclaimer, cpuid);
		__rcu_desynchronize_all_cpus();
		rcu_synchronize_cpu(cpuid);
		__rcu_barrier(reclaimer->flags.words);

		section_start = (head & (~WAITING_RECLAIMEES_ENTRIES));
		for (i = section_start; i <= head; i++) {
			reclaimee = &(reclaimer->list[i]);
			reclaimee->free_func(reclaimee->data, reclaimee->ctxt);
		}
		spin_unlock(&(reclaimer->lock));
	}
}

/** Acquires the reclaimer lock. Till the time caller is blocked waiting for 
 *  reclaimer lock, it synchronizes the RCU.
 */
static void
__rcu_acquire_reclaimer_lock(struct rcu_reclaimer_t *reclaimer, uint32_t cpuid)
{
	while (unlikely(spin_trylock(&(reclaimer->lock)) == 0))
		rcu_synchronize_cpu(cpuid);
}

/** Marks all CPUs as non-synchronized.  Once they are all again marked as
 *  synched, the reclaimer is clear to run.
 */
static void
__rcu_desynchronize_all_cpus()
{
	int i;

	for (i = 0; i < RCU_NUM_WORDS; i++)
		g_rcu_reclaimer.flags.words[i] =
		    g_rcu_reclaimer.running_cpu_map.words[i];
}

/** Marks the end of updater code and the beginning of reclaimer code.  It does
 * this by blocking until all pre-existing RCU read-side critical sections on
 * all CPUs have completed.  Note that synchronize_rcu() will NOT necessarily
 * wait for any subsequent RCU read-side critical sections to complete - it
 * is guaranteed to wait only for ongoing RCU read-side critical sections.
 */
static void
__rcu_barrier(volatile uint32_t * cpu_words)
{
	uint32_t sync_words[RCU_NUM_WORDS];
	uint32_t i;
	bool done;

	for (i = 0; i < RCU_NUM_WORDS; i++)
		sync_words[i] = g_rcu_reclaimer.running_cpu_map.words[i];

	do {
		done = true;
		for (i = 0; i < RCU_NUM_WORDS; i++) {
			if (sync_words[i] != 0) {
				done = false;
				sync_words[i] &= cpu_words[i];
			}
		}
	} while (done == false);
}
