/**
\file
\brief	This file provides additional APIs to create and free frame descriptors.
*/

/*
 * Copyright (c) 2009 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdarg.h>
#include <hv_misc.h>		/* for unlikely() etc. */
#include <sg_api.h>
#include <mmap.h>		/* for ptov_dpa() */
#include <log.h>

int lwe_prepare_frame(struct dpa_buff_allocator *buff_allocator,
		  struct frame_t *p_frame, struct frame_fragment_t *p_frag,
		  u32 count)
{
	u32 i;
	struct bm_buffer buf_sg;
	struct sg_entry_t *sg;
	phys_addr_t temp;

	assert(buff_allocator);
	assert(p_frame);
	assert(p_frag);
	assert(count);

	/* We do not expect this api to be called for single frames 
	 * however this code is here for completeness
	 */
	if (unlikely(count == 1)) {
		if (unlikely(p_frag[0].length > 0xFFFFF)) {
			p_frame->format = SINGLE_BUFFER_LONG;
			p_frame->offset = 0;
		} else {
			p_frame->format = SINGLE_BUFFER_SHORT;
			p_frame->offset = p_frag[0].offset;
		}

		p_frame->bpid = p_frag[0].bpid;
		p_frame->addr_hi = p_frag[0].addr_hi;
		p_frame->addr_lo = p_frag[0].addr_lo;
		p_frame->length = p_frag[0].length;

		return 0;
	}

	/*Allocate buffer for SG table */
	if (0 >= dpa_allocator_get_buff
	    (buff_allocator, (count) * sizeof (struct sg_entry_t), &buf_sg)) {
		LWE_ERROR("%s: couldn't allocate sg entry", __func__);
		return -ENOMEM;
	}

	p_frame->bpid = buf_sg.bpid;
	p_frame->addr_hi = buf_sg.hi;
	p_frame->addr_lo = buf_sg.lo;
	p_frame->length = 0;	/* Initialize it */

	temp = buf_sg.hi;
	temp = (temp << 32) | buf_sg.lo;
	sg = (struct sg_entry_t *) ptov(temp);

	for (i = 0; i < count; i++) {
		assert(p_frag[i].length);

		/*Fit the fragments in SG table */
		sg->bpid = p_frag[i].bpid;
		sg->addr_hi = p_frag[i].addr_hi;
		sg->addr_lo = p_frag[i].addr_lo;
		sg->length = p_frag[i].length;
		sg->offset = p_frag[i].offset;
		sg->reserved_offset = 0;
		sg->reserved_zero = 0;
		sg->reserved_zero2 = 0;
		sg->extension = 0;
		sg->final = 0;
		sg++;

		/*fill the p_frame with total length */
		p_frame->length += p_frag[i].length;
	}
	sg--;
	sg->final = 1;

	/*fill the p_frame format and offset */
	if (unlikely(p_frame->length > 0xFFFFF)) {
		p_frame->format = MULTI_BUFFER_LONG;
		p_frame->offset = 0;
	} else {
		p_frame->format = MULTI_BUFFER_SHORT;
		p_frame->offset = 0;	/*could have some offset later on */
	}

	return 0;
}

int lwe_prepare_fd(struct dpa_buff_allocator *buff_allocator,
	       struct qm_fd *p_fd, struct frame_t *p_frame, u32 count)
{
	u32 i;
	struct bm_buffer buf_sg;
	struct sg_entry_t *sg;
	phys_addr_t temp;

	assert(buff_allocator);
	assert(p_fd);
	assert(p_frame);
	assert(count);

	if (unlikely(count == 1)) {
		p_fd->bpid = p_frame[0].bpid;
		p_fd->addr_hi = p_frame[0].addr_hi;
		p_fd->addr_lo = p_frame[0].addr_lo;

		if (likely((p_frame->format == SINGLE_BUFFER_SHORT)
			   || (p_frame->format == MULTI_BUFFER_SHORT))) {
			p_fd->format = p_frame[0].format;
			p_fd->length20 = p_frame[0].length;
			p_fd->offset = p_frame[0].offset;
		} else {
			p_fd->_format1 = p_frame[0].format;
			p_fd->length29 = p_frame[0].length;
		}

		return 0;

	}

	/*Allocate the buffer for SG table for count frames */
	if (dpa_allocator_get_buff
	    (buff_allocator, (count) * sizeof (struct sg_entry_t),
	     &buf_sg) < 0) {
		LWE_ERROR("%s: couldn't allocate sg entry", __func__);
		return -ENOMEM;
	}

	p_fd->bpid = buf_sg.bpid;
	p_fd->addr_hi = buf_sg.hi;
	p_fd->addr_lo = buf_sg.lo;
	p_fd->_format2 = qm_fd_compound;

	temp = buf_sg.hi;
	temp = (temp << 32) | buf_sg.lo;
	sg = (struct sg_entry_t *) ptov(temp);

	/*Fit each of the frame into the SG table */
	for (i = 0; i < count; i++) {
		/*Fit the fragments in SG table */
		sg->bpid = p_frame[i].bpid;
		sg->addr_hi = p_frame[i].addr_hi;
		sg->addr_lo = p_frame[i].addr_lo;
		sg->length = p_frame[i].length;
		sg->offset = p_frame[i].offset;
		sg->reserved_offset = 0;
		sg->reserved_zero = 0;
		sg->reserved_zero2 = 0;
		sg->final = 0;
		if ((p_frame[i].format == MULTI_BUFFER_SHORT)
		    || (p_frame[i].format == MULTI_BUFFER_LONG))
			sg->extension = 1;
		else
			sg->extension = 0;
		sg++;
	}
	sg--;
	sg->final = 1;

	return 0;
}

static int free_sg_frame(const struct dpa_buff_allocator *buff_allocator,
	      uint8_t sg_buffer_bpid, phys_addr_t sg_buffer_addr,
	      phys_addr_t sg_table_addr, struct sg_entry_t *parent_sg_entry)
{
	int ret;
	struct bm_buffer buf;
	struct sg_entry_t *sg_entry;

	phys_addr_t temp_sg_buffer_addr;

	sg_entry = (struct sg_entry_t *)ptov(sg_table_addr);

	do {
		/* Processing start with new table if "E"
		 * bit is 1 for single multi-buffer frame
		 */
		if (unlikely(1 == sg_entry->extension)) {
			temp_sg_buffer_addr = sg_entry->addr_hi;
			temp_sg_buffer_addr = (temp_sg_buffer_addr << 32) |
			    sg_entry->addr_lo;

			sg_table_addr = temp_sg_buffer_addr + sg_entry->offset;

			ret = free_sg_frame(buff_allocator, sg_entry->bpid,
					temp_sg_buffer_addr, sg_table_addr,
					sg_entry);
			if (unlikely(ret < 0)) {
				LWE_ERROR("%s: free_sg_frame fail", __func__);
				return ret;
			}

			/*free up previous sg buffer */
			buf.bpid = sg_buffer_bpid;
			buf.hi = sg_buffer_addr >> 32;
			buf.lo = sg_buffer_addr & 0xFFFFFFFF;

			ret = dpa_allocator_free_buff(buff_allocator, &buf);
			if (unlikely(ret < 0)) {
				LWE_ERROR
				    ("%s: dpa_allocator_free_buff failed!",
				     __func__);
				return ret;
			}

			/* Mark the parent S/G entry if available which pointed
			 * to this S/G table as 'final'. The parent S/G entry
			 * would be NULL in case the FD pointed to this
			 * table directly (more common case).
			 */
			if (unlikely(parent_sg_entry)) {
				/*
				 * Need to turn off 'extension' bit since it
				 * takes precedence over 'final' bit.
				 */
				parent_sg_entry->extension = 0;
				parent_sg_entry->final = 1;
				/*
				 * Nullify the entry->hi and entry->lo as the
				 * sg table is freed. The following assignment
				 * is required to handle case in which
				 * subsequent BMAN rel buffer fail with -EBUSY
				 * and app calls lwe_free_fd again
				 */
				parent_sg_entry->addr_lo = 0;
				parent_sg_entry->addr_hi = 0;
			}

			return 0;
		}

		if (likely(sg_entry->addr_lo) || likely(sg_entry->addr_hi)) {
			buf.bpid = sg_entry->bpid;
			buf.hi = sg_entry->addr_hi;
			buf.lo = sg_entry->addr_lo;

			ret = dpa_allocator_free_buff(buff_allocator, &buf);
			if (unlikely(ret < 0)) {
				LWE_ERROR("%s: dpa_allocator_free_buff failed",
						__func__);
				return ret;
			}
			/*
			 * Nullify the entry->hi and entry->lo as the
			 * sg table is freed. The following assignment
			 * is required to handle case in which
			 * subsequent BMAN rel buffer fail with -EBUSY
			 * and app calls lwe_free_fd again
			 */
			sg_entry->addr_lo = 0;
			sg_entry->addr_hi = 0;
		}

		if (unlikely(1 == sg_entry->final)) {
			/*Now free up this buffer */
			buf.bpid = sg_buffer_bpid;
			buf.hi = sg_buffer_addr >> 32;
			buf.lo = sg_buffer_addr & 0xFFFFFFFF;

			ret = dpa_allocator_free_buff(buff_allocator,
						     &buf);
			if (unlikely(ret < 0)) {
				LWE_ERROR("%s: dpa_allocator_free_buff failed",
				     __func__);
				return ret;
			}

			/* Mark the parent S/G entry if available which pointed
			 * to this S/G table as 'final'. The parent S/G entry
			 * would be NULL in case the FD pointed to this
			 * table directly (more common case).
			 */
			if (unlikely(parent_sg_entry)) {
				parent_sg_entry->final = 1;
				/*
				 * Need to turn off 'extension' bit since it
				 * takes precedence over 'final' bit.
				 */
				parent_sg_entry->extension = 0;
				/*
				 * Nullify the entry->hi and entry->lo as the
				 * sg table is freed. The following assignment
				 * is required to handle case in which
				 * subsequent BMAN rel buffer fail with -EBUSY
				 * and app calls lwe_free_fd again
				 */
				parent_sg_entry->addr_lo = 0;
				parent_sg_entry->addr_hi = 0;
			}

			return 0;
		}

		sg_entry++;
	} while (likely(1));
}

/**
\brief		Drop a simple single buffer frame descriptor

\details	Free memory associated with a frame descriptor.

\param[in] 	p_fd - Frame descriptor of frame to be freed.

\return	  	0 - On success, otherwise -ve value
*/
static int free_simple_single_buf_fd(
			const struct dpa_buff_allocator *buff_allocator,
			const struct qm_fd *p_fd)
{
	struct bm_buffer buf;
	int ret;

	buf.bpid = p_fd->bpid;
	buf.hi = p_fd->addr_hi;
	buf.lo = p_fd->addr_lo;

	ret = dpa_allocator_free_buff(buff_allocator, &buf);
	if (unlikely(ret < 0)) {
		LWE_ERROR("%s: dpa_allocator_free_buff failed!", __func__);
		return ret;
	}

	return 0;
}

/**
\brief		Drop a simple multi-buffer frame descriptor

\details	Parses the scatter gather table and frees individual buffers

\param[in]	p_fd - Frame descriptor to be freed

\return		0 - On success, otherwise -ve value
 */
static int free_simple_multi_buf_fd(
			const struct dpa_buff_allocator *buff_allocator,
			const struct qm_fd *p_fd)
{
	int ret;
	uint8_t sg_buffer_bpid;
	phys_addr_t sg_buffer_addr;
	phys_addr_t sg_table_addr;

	sg_buffer_bpid = p_fd->bpid;
	sg_buffer_addr = p_fd->addr_hi;
	sg_buffer_addr = (sg_buffer_addr << 32) | p_fd->addr_lo;

	/*The first buffer pointed to by FD refers to SG table */
	sg_table_addr = p_fd->addr_hi;
	sg_table_addr = (sg_table_addr << 32) | p_fd->addr_lo;

	if (p_fd->format == qm_fd_sg)
		sg_table_addr += p_fd->offset;

	ret = free_sg_frame(buff_allocator, sg_buffer_bpid, sg_buffer_addr,
			  sg_table_addr, NULL);
	if (unlikely(ret < 0)) {
		LWE_ERROR("%s: free_sg_frame failed!", __func__);
		return ret;
	}

	return 0;
}

/**
\brief		Drop a compound frame descriptor

\details	Parse and free scatter/gather list of compound frame
		and individual buffers

\param[in]	p_fd - Frame descriptor to be freed

\return	     	0 - On success
		0 > Error code
*/
static int free_compound_fd(const struct dpa_buff_allocator *buff_allocator,
		 const struct qm_fd *p_fd)
{
	int ret;

	struct bm_buffer buf;
	struct sg_entry_t *sg_entry;

	uint8_t sg_table_bpid;
	phys_addr_t sg_table_addr;

	sg_table_bpid = p_fd->bpid;
	sg_table_addr = p_fd->addr_hi;
	sg_table_addr = (sg_table_addr << 32) | p_fd->addr_lo;

	sg_entry = (struct sg_entry_t *) ptov(sg_table_addr);

	do {
		if (unlikely(1 == sg_entry->extension)) {
			phys_addr_t frame_sg_buff_addr;

			frame_sg_buff_addr = sg_entry->addr_hi;
			frame_sg_buff_addr =
			    frame_sg_buff_addr << 32 | sg_entry->addr_lo;

			ret = free_sg_frame(buff_allocator, sg_entry->bpid,
				      frame_sg_buff_addr,
				      frame_sg_buff_addr + sg_entry->offset,
				      NULL);
			if (unlikely(ret < 0)) {
				LWE_ERROR("%s: free_sg_frame failed!",
				     __func__);
				return ret;
			}
			/*
			 * The following assignment is required to handle case
			 * in which subsequent BMAN rel buffer fail with -EBUSY
			 * and app calls lwe_free_fd again
			 */
			sg_entry->extension = 0;
			sg_entry->addr_lo = 0;
			sg_entry->addr_hi = 0;
		} else if (likely(sg_entry->addr_lo) ||
					likely(sg_entry->addr_hi)) {
			buf.bpid = sg_entry->bpid;
			buf.hi = sg_entry->addr_hi;
			buf.lo = sg_entry->addr_lo;

			ret = dpa_allocator_free_buff(buff_allocator, &buf);
			if (unlikely(ret < 0)) {
				LWE_ERROR("%s: dpa_allocator_free_buff failed",
				     __func__);
				return ret;
			}
			/*
			 * Nullify the entry->hi and entry->lo as the
			 * sg table is freed. The following assignment
			 * is required to handle case in which
			 * subsequent BMAN rel buffer fail with -EBUSY
			 * and app calls lwe_free_fd again
			 */
			sg_entry->addr_lo = 0;
			sg_entry->addr_hi = 0;
		}

		if (unlikely(1 == sg_entry->final)) {
			buf.bpid = sg_table_bpid;
			buf.hi = sg_table_addr >> 32;
			buf.lo = sg_table_addr & 0xFFFFFFFF;

			ret = dpa_allocator_free_buff(buff_allocator, &buf);
			if (unlikely(ret < 0)) {
				LWE_ERROR("%s: dpa_allocator_free_buff failed",
				     __func__);
				return ret;
			}
			return 0;
		}

		sg_entry++;
	} while (likely(1));
}

int lwe_free_fd(const struct dpa_buff_allocator *buff_allocator,
	    const struct qm_fd *p_fd)
{
	int ret;

	assert(buff_allocator);
	assert(p_fd);

	static lwe_fd_free_func_t fd_free_func[8] = {
		free_simple_single_buf_fd,
		free_compound_fd,
		free_simple_single_buf_fd,
		NULL,
		free_simple_multi_buf_fd,
		NULL,
		free_simple_multi_buf_fd
	};

	assert(likely(NULL != fd_free_func[p_fd->format]));

	ret = (fd_free_func[p_fd->format]) (buff_allocator, p_fd);

	if (unlikely(ret < 0)) {
		LWE_ERROR("%s: fd_free_func failed!", __func__);
		return ret;
	}

	return 0;
}

uint32_t get_num_frame(const struct qm_fd *p_fd)
{
	uint32_t count = 1;
	const struct sg_entry_t *sg_entry;
	phys_addr_t sg_table_addr;

	assert(p_fd);

	if (p_fd->format != qm_fd_compound)
		return count;

	sg_table_addr =	GET_ADDR36B(p_fd);
	sg_entry = (struct sg_entry_t *)ptov(sg_table_addr);

	while (0 == sg_entry->final) {
		count++;
		sg_entry++;
	}

	return count;
}

void get_frame_from_fd(const struct qm_fd *p_fd, struct frame_t *p_frame,
		       uint32_t *count)
{
	uint32_t i = 0;
	uint32_t frame_count;
	const struct sg_entry_t *sg_entry;
	phys_addr_t sg_table_addr;

	assert(p_fd);
	assert(p_frame);
	assert(count);

	frame_count = *count;
	*count = 0;

	if (p_fd->format != qm_fd_compound) {

		BPID_ADDR_CPY(&p_frame[0], p_fd);

		if (qm_fd_contig == p_fd->format || qm_fd_sg == p_fd->format) {
			p_frame[0].format = p_fd->format;
			p_frame[0].length = p_fd->length20;
			p_frame[0].offset = p_fd->offset;
		} else {
			p_frame[0].format = p_fd->_format1;
			p_frame[0].length = p_fd->length29;
		}

		*count = 1;
		return;
	}

	sg_table_addr =	GET_ADDR36B(p_fd);
	sg_entry = (struct sg_entry_t *)ptov(sg_table_addr);

	while (frame_count) {
		BPID_ADDR_CPY(&p_frame[i], sg_entry);
		p_frame[i].length = sg_entry->length;
		p_frame[i].offset = sg_entry->offset;

		if (sg_entry->extension == 1) {
			if (sg_entry->length > 0xFFFFF)
				p_frame[i].format = MULTI_BUFFER_LONG;
			else
				p_frame[i].format = MULTI_BUFFER_SHORT;
		} else {
			if (sg_entry->length > 0xFFFFF)
				p_frame[i].format = SINGLE_BUFFER_LONG;
			else
				p_frame[i].format = SINGLE_BUFFER_SHORT;
		}

		frame_count--;
		(*count)++;
		i++;

		if (sg_entry->final == 1)
			break;
		sg_entry++;
	};
}

static void append_sg_frag(struct frame_t *p_frame,
			   struct sg_entry_t *sg_entry_last,
			   struct sg_entry_t *sg_entry_new,
			   const struct frame_fragment_t *p_frag, u32 count)
{
	uint32_t i;

	for (i = 0; i < count; i++) {
		assert(p_frag[i].length);

		BPID_ADDR_CPY(sg_entry_new, &p_frag[i]);
		sg_entry_new->length = p_frag[i].length;
		sg_entry_new->offset = p_frag[i].offset;
		sg_entry_new->reserved_offset = 0;
		sg_entry_new->reserved_zero = 0;
		sg_entry_new->reserved_zero2 = 0;
		sg_entry_new->extension = 0;
		sg_entry_new->final = 0;

		/* update length of previous sg_entry and frame
		 * for new fragment
		 */
		if (sg_entry_last)
			sg_entry_last->length += p_frag[i].length;
		p_frame->length += p_frag[i].length;
	}
	sg_entry_new->final = 1;
}

int append_frame_fragment(const struct dpa_buff_allocator *buff_allocator,
			  struct frame_t *p_frame,
			  const struct frame_fragment_t *p_frag, u32 count)
{
	struct sg_entry_t *sg_entry;
	struct sg_entry_t sg_entry_last;
	struct sg_entry_t *sg_entry_new;
	struct bm_buffer buf_sg;
	struct frame_t frame_tmp;
	phys_addr_t sg_table_addr;

	assert(buff_allocator);
	assert(p_frame);
	assert(p_frag);
	assert(count);

	if (unlikely(0 >= dpa_allocator_get_buff(buff_allocator,
					(count + 1) * sizeof(struct sg_entry_t),
					&buf_sg))) {
		LWE_ERROR("%s: couldn't allocate sg entry", __func__);
		return -ENOMEM;
	}

	sg_table_addr =	GET_ADDR36B_BM_BUF(buf_sg);
	sg_entry_new = (struct sg_entry_t *)ptov(sg_table_addr);

	if ((SINGLE_BUFFER_SHORT == p_frame->format)
	    || SINGLE_BUFFER_LONG == p_frame->format) {
		BPID_ADDR_CPY(&frame_tmp, p_frame);

		BM_BUF_CPY(p_frame, buf_sg);
		/* length is already initialized for existing frag */

		BPID_ADDR_CPY(sg_entry_new, &frame_tmp);
		sg_entry_new->length = p_frame->length;
		if (SINGLE_BUFFER_SHORT == p_frame->format)
			sg_entry_new->offset = p_frame->offset;
		else
			sg_entry_new->offset = 0;

		sg_entry_new->reserved_offset = 0;
		sg_entry_new->reserved_zero = 0;
		sg_entry_new->reserved_zero2 = 0;
		sg_entry_new->extension = 0;
		sg_entry_new->final = 0;
		sg_entry_new++;

		append_sg_frag(p_frame, NULL, sg_entry_new, p_frag, count);

		if (p_frame->length > 0xFFFFF)
			p_frame->format = MULTI_BUFFER_LONG;
		else
			p_frame->format = MULTI_BUFFER_SHORT;

	} else if ((MULTI_BUFFER_SHORT == p_frame->format)
		   || MULTI_BUFFER_LONG == p_frame->format) {
		sg_table_addr =	GET_ADDR36B(p_frame);
		if (MULTI_BUFFER_SHORT == p_frame->format)
			sg_table_addr += p_frame->offset;

		sg_entry = (struct sg_entry_t *)ptov(sg_table_addr);

		do {
			if (sg_entry->extension == 1) {
				sg_table_addr = GET_ADDR36B(sg_entry);
				sg_entry =
				    (struct sg_entry_t *)ptov(sg_table_addr);

			} else if (sg_entry->final == 1) {
				sg_entry->final = 0;
				sg_entry->extension = 1;

				BPID_ADDR_CPY(&sg_entry_last, sg_entry);
				sg_entry_last.length = sg_entry->length;
				sg_entry_last.offset = sg_entry->offset;
				break;
			} else
				sg_entry++;
		} while (1);

		BM_BUF_CPY(sg_entry, buf_sg);
		/* first entry of new s/g table */
		sg_entry->length = sg_entry_last.length;

		BPID_ADDR_CPY(sg_entry_new, &sg_entry_last);
		sg_entry_new->length = sg_entry_last.length;
		sg_entry_new->offset = sg_entry_last.offset;
		sg_entry_new->reserved_offset = 0;
		sg_entry_new->reserved_zero = 0;
		sg_entry_new->reserved_zero2 = 0;
		sg_entry_new++;

		append_sg_frag(p_frame, sg_entry, sg_entry_new, p_frag, count);

		if (p_frame->length > 0xFFFFF)
			p_frame->format = MULTI_BUFFER_LONG;
	}

	return 0;
}
