/** @file
 * Stack implementation
 */

/*
 * Copyright (c) 2008 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdint.h>
#include <types.h>
#include <fsl_hcalls.h>
#include <errno.h>
#include "bitops.h"
#include "stack.h"
#include <stdlib.h>
#include <lwe_malloc.h>

struct stack_t {
	uint32_t max_height;
	uint32_t top;
	void *stack[];
};


/**Initializes the stack by allocating memory for max number of entries.
 * @param[in] max_extries : max entries to be allocated on stack.
 * @param[in] p_ecode : error code pointer. 
 * @return int : Returns the stack allocated, NULL if failed. 
 */
struct stack_t *
stack_init(uint32_t max_entries, uint32_t * p_ecode)
{
	struct stack_t *p_stack;
	int size;

	if ((!max_entries) || ((uint32_t)(-1) == max_entries)) {
		*p_ecode = EINVAL;
		return NULL;
	}

	size = sizeof (struct stack_t) + (max_entries * sizeof (void *));

	p_stack = tlmalloc(size);
	if (NULL == p_stack) {
		*p_ecode = ENOMEM;
		return NULL;
	}

	p_stack->top = ((uint32_t) (-1));
	p_stack->max_height = max_entries - 1;

	return p_stack;
}

/** Pushes into the stack. 
 * @param[in] p_stack : pointer to the stack
 * @param[in] p_item : pointer to the item to be pushed on stack
 * @return int :  Returns 0 for success, -ve value for failure 
*/
int
stack_push(struct stack_t *p_stack, void *p_item)
{
	disable_extint();

	if (p_stack->top == p_stack->max_height) {
		enable_extint();
		return -1;
	}

	(p_stack->top)++;
	p_stack->stack[p_stack->top] = p_item;

	enable_extint();

	return 0;
}

/**Pops from the stack. 
 * @param[in] p_stack : pointer to the stack
 * @return void * : pointer to the item popped
 */
void *
stack_pop(struct stack_t *p_stack)
{
	void *p_item = NULL;

	disable_extint();

	if (((uint32_t) (-1)) == p_stack->top)
		goto out;

	p_item = p_stack->stack[p_stack->top];
	p_stack->top--;

out:
	enable_extint();
	return p_item;
}

/**Frees the stack.
 * @param[in] p_stack : pointer to the stack
 * Assumes all nodes correctly returned to stack before this function is called
 */
int
stack_teardown(struct stack_t *p_stack)
{
	tlfree(p_stack);
	return 0;
}
