/** @file
 *  printf()/scanf() hooks (send/receive via byte channel) for libc
 */

/*
 * Copyright (c) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include <errno.h>
#include "hv_misc.h"
#include "fsl_hcalls.h"
#include <percpu.h>
#include "init.h"
#include "UART.h"
#include <console.h>
#include <compiler.h>
#include <ns16550.h>
#include "epapr_hcalls.h"

#define BYTE_COUNT 16

/** InitializeUART - Glue code for libc uart initilization
 *  As lwe relies on the byte channel for printf() and does not need 
 *  any uart initialization. Just return success to let libc proceed further
 */
UARTError
InitializeUART(UARTBaudRate baudrate __UNUSED)
{
	if (1 != cpu->console_ok)
		return kUARTConfigurationError;
	return kUARTNoError;
}

/** ReadUARTN - Reads a buffer from the byte channel
 *  This function is used only as a glue code for libc and should
 *  not be called by the application.
 *
 *  @param[in] bytes - Pointer to the input buffer
 *  @param[in] length - Length of the buffer
 *  @return returns 0 on success and non-zero otherwise. 
 *          See UART.h in msl-libc for various error codes
 */
UARTError
ReadUARTN(void *bytes, unsigned long length)
{
	uint32_t cnt;
	char buff[BYTE_COUNT];
	char *str = (char *) bytes;
	uint32_t ch_handle;

	if (lwe_console[cpu->console_id].driver == UART) {
		int index = lwe_console[cpu->console_id].ch_handle;
		ns16550 *ns_obj = &uart_obj[index];

		ns16550_rx(ns_obj, (uint8_t *)str);

		return kUARTNoError;
	} else {
		ch_handle = lwe_console[cpu->console_id].ch_handle;

		while (length >= BYTE_COUNT) {
			cnt = BYTE_COUNT;
			if (unlikely
				(ev_byte_channel_receive(ch_handle,
						(void *) &cnt, str)))
					return kUARTConfigurationError;
			str += cnt;
			length -= cnt;
		}

		while (likely(length)) {
			cnt = length;
			if (unlikely
				(ev_byte_channel_receive(ch_handle,
						(void *) &cnt, buff)))
					return kUARTConfigurationError;
			isr_memcpy(str, buff, cnt);
			str += cnt;
			length -= cnt;
		}

		return kUARTNoError;
	}
}

/** WriteUARTN - Dumps the given buffer on to the byte channel
 *  This function is used only as a glue code for libc and should
 *  not be called by the application.
 *
 *  @param[in] str - String to be displayed
 *  @param[in] length - Length of the string to be displayed
 *  @return returns kUARTNoError on success; asserts otherwise
 */
UARTError
WriteUARTN(const void *str, unsigned long length)
{
	unsigned int ret;
	unsigned int to_send;
	char temp_buff[2 * BYTE_COUNT];
	char *loc = (char *) str;
	uint32_t ch_handle;

	if (lwe_console[cpu->console_id].driver == UART) {
		int index = lwe_console[cpu->console_id].ch_handle;
		ns16550 *ns_obj = &uart_obj[index];

		ns16550_tx(ns_obj, (uint8_t *)str, length, CHARDEV_BLOCKING);

		return kUARTNoError;
	}

	/*Case of byte channel handled here*/
	ch_handle = lwe_console[cpu->console_id].ch_handle;

	while (length > (BYTE_COUNT - 1)) { /*Loop till less than 16 bytes left*/
		to_send = BYTE_COUNT;

		ret = ev_byte_channel_send(ch_handle,
					&to_send, loc);
		if (likely(0 == ret)) {
			loc = loc + to_send;
			length = length - to_send;
		} else if (likely(EAGAIN == ret)) {
			continue;
		} else {
			assert(0);	/*Crash here */
		}
	}

	if (0 == length)
		return kUARTNoError;

	/*Now copy into a local temp buff and send straighway*/
	isr_memcpy(temp_buff, loc, length);

	do {
		to_send = length;
		ret = ev_byte_channel_send(ch_handle,
					&to_send, loc);
		if (likely(0 == ret)) {
			loc = loc + to_send;
			length = length - to_send;
		} else if (likely(EAGAIN == ret)) {
			continue;
		} else {
			assert(0);	/*Crash here */
		}
	} while (length);

	return kUARTNoError;
}
