/** @file vmm.c
 * Virtual Address Space Manager
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stddef.h>
#include "log.h"
#include "lwe_malloc.h"

#define MIN_FREE_SIZE_LIST 0x1000
#define NUMBER_OF_LISTS 8
#define MAX_FREE_SIZE_LIST (MIN_FREE_SIZE_LIST * NUMBER_OF_LISTS)
#define MASK (~(MIN_FREE_SIZE_LIST - 1))
#define LIST_INDEX_FROM_SIZE(size) ((size & MASK)/MIN_FREE_SIZE_LIST)

#define MIN_ADDRESS_LIST 0x1000
#define MAX_ADDRESS_LIST (MIN_ADDRESS_LIST * NUMBER_OF_LISTS)
#define ADDRESS_MASK (~(MIN_FREE_SIZE_LIST - 1))
#define LIST_INDEX_FROM_ADDRESS(address) \
	((address & ADDRESS_MASK)/MIN_ADDRESS_LIST)

__PERCPU unsigned int vmem_initialized = 0;

struct global_node {
	struct global_node *g_next;
	struct global_node *g_prev;
};

typedef struct node {
	unsigned char tag;
	unsigned long address;
	unsigned int size;
	struct node *next;
	struct node *prev;
	struct global_node g_entry;
} node;

typedef node free_node;
typedef node allocated_node;

__PERCPU struct global_node global_list_head;

/* "+ 1" is required for the default list */
__PERCPU free_node free_area_lists[NUMBER_OF_LISTS + 1];
__PERCPU allocated_node allocated_area_lists[NUMBER_OF_LISTS + 1];

static allocated_node *get_allocated_list_from_address(unsigned long address)
{
	int index;

	if (address > MAX_ADDRESS_LIST)
		return &allocated_area_lists[NUMBER_OF_LISTS];

	index = LIST_INDEX_FROM_ADDRESS(address);
	return &allocated_area_lists[index];
}

static free_node *get_list_from_size(unsigned int size)
{
	int index;

	if (size > MAX_FREE_SIZE_LIST)
		return &free_area_lists[NUMBER_OF_LISTS];

	index = LIST_INDEX_FROM_SIZE(size);
	return &free_area_lists[index];
}

static free_node *get_free_area_list(unsigned int size)
{
	int index;

	if (size > MAX_FREE_SIZE_LIST) {
		if (free_area_lists[NUMBER_OF_LISTS].size >= size)
			if (free_area_lists[NUMBER_OF_LISTS].prev->size >= size)
				return &free_area_lists[NUMBER_OF_LISTS];

		return NULL;
	}

	index = LIST_INDEX_FROM_SIZE(size);
	do {
		if (free_area_lists[index].size > size) {
			/* chech whether this list contains an entry which
			 * is bigger then OR equal to the required size.
			 * This can be done by verifying the size of entrty
			 * present at tail.
			 */
			if (free_area_lists[index].prev->size >= size)
				return &free_area_lists[index];
		}

		index++;
	} while (index < (NUMBER_OF_LISTS + 1));

	return NULL;
}

static void list_add_allocated_area(allocated_node *entry)
{
	allocated_node *head;

	head = get_allocated_list_from_address(entry->address);
	head->size = head->size + entry->size;

	if (head->prev->address <= entry->address) {
		/* add allocated area at tail */
		entry->next = head;
		entry->prev = head->prev;
		head->prev->next = entry;
		head->prev = entry;
	} else if (head->next->address > entry->address) {
		/* add allocated area at head */
		entry->next = head->next;
		entry->prev = head;
		head->next->prev = entry;
		head->next = entry;
	} else {	/* add allocated area after 1st entry at head */
		entry->next = head->next->next;
		entry->prev = head->next;
		entry->next->prev = entry;
		entry->prev->next = entry;
	}
}

static void list_remove_allocated_area(allocated_node *entry)
{
	allocated_node *head;

	head = get_allocated_list_from_address(entry->address);
	head->size -= entry->size;

	entry->prev->next = entry->next;
	entry->next->prev = entry->prev;
}

static void list_add_free_area(free_node *entry)
{
	free_node *head;

	head = get_list_from_size(entry->size);
	head->size = head->size + entry->size;

	if (head->prev->size <= entry->size) {
		/* add free area at tail */
		entry->next = head;
		entry->prev = head->prev;
		head->prev->next = entry;
		head->prev = entry;
	} else if (head->next->size > entry->size) {
		/* add free area at head */
		entry->next = head->next;
		entry->prev = head;
		head->next->prev = entry;
		head->next = entry;
	} else {	/* add free area after 1st entry at head */
		entry->next = head->next->next;
		entry->prev = head->next;
		entry->next->prev = entry;
		entry->prev->next = entry;
	}
}

static void list_remove_free_area(free_node *entry)
{
	free_node *head;

	head = get_list_from_size(entry->size);
	head->size -= entry->size;

	entry->prev->next = entry->next;
	entry->next->prev = entry->prev;
}

static void list_add_global_before(struct global_node *entry,
				   struct global_node *list)
{
	entry->g_next = list;
	entry->g_prev = list->g_prev;
	entry->g_next->g_prev = entry;
	entry->g_prev->g_next = entry;
}

static void list_add_global_after(struct global_node *entry,
				  struct global_node *list)
{
	entry->g_prev = list;
	entry->g_next = list->g_next;
	entry->g_next->g_prev = entry;
	entry->g_prev->g_next = entry;
}

static void list_remove_global(struct global_node *entry)
{
	entry->g_prev->g_next = entry->g_next;
	entry->g_next->g_prev = entry->g_prev;
}

static void *allocate_free_area(free_node * nptr, unsigned int size)
{
	allocated_node *anptr = NULL;
	unsigned int free_area_available;

	free_area_available = nptr->size;

	if (nptr->size == size) {
		anptr = (allocated_node *) nptr;
		anptr->tag = 1;
		list_remove_free_area(nptr);
		list_add_allocated_area(anptr);
		return anptr;
	}

	if (free_area_available > size) {
		anptr = (allocated_node *) nptr;
		anptr = (allocated_node *) tlmalloc(sizeof(allocated_node));
		if (!anptr) {
			LWE_INFO("allocate_free_area: tlmalloc failed\n");
			return NULL;
		}

		list_remove_free_area(nptr);

		anptr->address = nptr->address;
		anptr->size = size;
		anptr->tag = 1;
		nptr->address = nptr->address + size;
		nptr->size = free_area_available - size;
		nptr->tag = 0;

		list_add_free_area(nptr);
		list_add_allocated_area(anptr);
		list_add_global_before(&anptr->g_entry, &nptr->g_entry);
	}

	return anptr;
}

static void *do_align_valloc(free_node * ptr, unsigned long align,
			     unsigned long size)
{
	allocated_node *anptr = NULL;
	free_node *fnptr = NULL;
	void *align_addr, *addr, *end_addr;

	addr = (void *)(ptr->address);
	align_addr = (void *)((unsigned long)addr & (~(align - 1)));

	if (!addr || (align_addr && (align_addr == addr))) {
		anptr = allocate_free_area(ptr, size);
	} else {
		align_addr =
		    (void *)(((unsigned long)addr + align) & (~(align - 1)));
		end_addr = (void *)ptr->address + ptr->size;
		if ((align_addr < end_addr) &&
					(align_addr + size <= end_addr)) {
			anptr =
			    (allocated_node *) tlmalloc(sizeof(allocated_node));
			if (!anptr) {
				LWE_INFO("do_align_valloc:"
					 "tlmalloc failed\n");
				return NULL;
			}

			anptr->address = (unsigned long)align_addr;
			anptr->size = size;
			anptr->tag = 1;

			/* chech if there is lower free area */
			if (align_addr + size != end_addr) {
				fnptr =
				    (free_node *) tlmalloc(sizeof(free_node));
				if (!fnptr) {
					LWE_INFO("do_align_valloc: "
						"tlmalloc failed\n");
					tlfree(anptr);
					return NULL;
				}

				fnptr->address =
				    (unsigned long)align_addr + size;
				fnptr->size = end_addr - (align_addr + size);
				fnptr->tag = 0;
			}

			list_remove_free_area(ptr);

			ptr->size = ptr->size - anptr->size;
			if (fnptr)
				ptr->size = ptr->size - fnptr->size;

			list_add_free_area(ptr);
			list_add_allocated_area(anptr);
			list_add_global_after(&anptr->g_entry, &ptr->g_entry);

			if (fnptr) {
				list_add_free_area(fnptr);
				list_add_global_after(&fnptr->g_entry,
						      &anptr->g_entry);
			}
		}
	}

	return anptr;
}

static int try_merge_upward(struct node *curr_node)
{
	struct node *prev_node;
	int ret = 0;

	if (curr_node->g_entry.g_prev != &global_list_head) {
		prev_node = (void *)(curr_node->g_entry.g_prev) -
					offsetof(node, g_entry);

		if (prev_node->tag == 0) {
			ret = 1;
			list_remove_free_area(prev_node);
			prev_node->size += curr_node->size;
			list_add_free_area(prev_node);
		}
	}

	return ret;
}

static void try_merge_downward(struct node *curr_node)
{
	struct node *next_node;

	if (curr_node->g_entry.g_next != &global_list_head) {
		next_node = (void *)(curr_node->g_entry.g_next) -
					offsetof(node, g_entry);
		if (next_node->tag == 0) {
			curr_node->size += next_node->size;
			list_remove_free_area(next_node);
			list_remove_global(&next_node->g_entry);
			tlfree(next_node);
		}
	}
}

static void initialize_free_area_lists(void)
{
	int i;

	for (i = 0; i <= NUMBER_OF_LISTS; i++) {
		free_area_lists[i].address = 0;
		free_area_lists[i].size = 0;
		free_area_lists[i].next = &free_area_lists[i];
		free_area_lists[i].prev = &free_area_lists[i];
	}
}

static void initialize_address_lists(void)
{
	int i;

	for (i = 0; i <= NUMBER_OF_LISTS; i++) {
		allocated_area_lists[i].address = 0;
		allocated_area_lists[i].size = 0;
		allocated_area_lists[i].next = &allocated_area_lists[i];
		allocated_area_lists[i].prev = &allocated_area_lists[i];
	}
}

static void initialize_global_list(struct global_node *global_list_head)
{
	global_list_head->g_next = global_list_head;
	global_list_head->g_prev = global_list_head;
}

static void vmem_init(void)
{
	initialize_global_list(&global_list_head);
	initialize_free_area_lists();
	initialize_address_lists();

	vmem_initialized = 1;
}

void *valloc(unsigned long size, unsigned long align)
{
	allocated_node *allocated_node = NULL;
	free_node *ptr, *tptr;
	free_node *head;
	int index;

	if (!vmem_initialized)
		return NULL;

	if (!size) {
		LWE_CRIT("valloc: size must be greater then zero\n");
		return NULL;
	}

	if (!align || (align & (0x1000 - 1))) {
		LWE_CRIT("valloc: Alignment must be a non-zero value and in "
			"multiple of 4k\n");
		return NULL;
	}

	head = get_free_area_list(size);
	if (!head)
		return NULL;

	do {
		ptr = head->next;
		tptr = head;
		do {
			if (ptr->size >= size)
				allocated_node = do_align_valloc(ptr, align,
								size);

			if (allocated_node)
				break;
			else {
				if (tptr == head)
					tptr = head->next;
				else if (tptr->size < ptr->size)
					tptr = ptr;

				ptr = ptr->next;
			}

		} while (ptr != head);

		if (allocated_node) {
			if (tptr != head) {
				list_remove_free_area(tptr);
				list_add_free_area(tptr);
			}

			return (void *)(allocated_node->address);
		}

		head = (void *)head + sizeof(free_node);
		index = ((unsigned long)head -
			(unsigned long)&free_area_lists) / sizeof(free_node);
	} while (index < NUMBER_OF_LISTS + 1);

	return NULL;
}

void vfree(void *ptr)
{
	allocated_node *list_head;
	allocated_node *anptr;
	int merged = 0;

	if (!vmem_initialized)
		return;

	list_head = get_allocated_list_from_address((unsigned long)ptr);

	anptr = list_head->next;

	while (anptr != list_head) {
		if (anptr->address == (unsigned long)ptr)
			break;

		anptr = anptr->next;
	}

	if (anptr->address == (unsigned long)ptr) {
		list_remove_allocated_area(anptr);
		anptr->tag = 0;
		try_merge_downward(anptr);
		merged = try_merge_upward(anptr);

		if (merged) {
			list_remove_global(&anptr->g_entry);
			tlfree(anptr);
		} else
			list_add_free_area((free_node *) anptr);
	} else
		LWE_INFO("vfree: Trying to free %p invalid address\n", ptr);
}

int valloc_add(unsigned long start, unsigned long size)
{
	if (!size) {
		LWE_CRIT("valloc_add: size must be greater then zero\n");
		return -1;
	}

	if (start >= 0xffffffff) {
		LWE_CRIT("valloc_add: invalid %#lx address\n", start);
		return -1;
	}

	free_node *fnptr = (free_node *) tlmalloc(sizeof(free_node));
	if (!fnptr) {
		LWE_CRIT("valloc_add: addition of address space"
			 "(address %#lx, size %#lx) failed, because tlmalloc "
			 "failed\n", start, size);
		return -1;
	}

	fnptr->address = start;
	fnptr->size = size;
	fnptr->tag = 0;

	struct global_node *ptr;
	struct node *gnptr;

	if (!vmem_initialized)
		vmem_init();

	ptr = global_list_head.g_next;
	while (ptr != &global_list_head) {
		gnptr = (void *)ptr - offsetof(node, g_entry);

		if (gnptr->address >= start) {
			if ((start + size) > gnptr->address) {
				LWE_CRIT("valloc_add: "
				"address space(address %#lx, size %#lx)"
				"overlaps existing address space"
				"(address %#lx, size %#x)",
					 start, size, gnptr->address,
					 gnptr->size);
				tlfree(fnptr);
				return -1;
			}

			list_add_free_area(fnptr);
			list_add_global_before(&fnptr->g_entry,
					       &gnptr->g_entry);
			return 0;
		}

		ptr = ptr->g_next;
	}

	ptr = global_list_head.g_prev;
	if (ptr != &global_list_head) {
		gnptr = (void *)ptr - offsetof(node, g_entry);
		if ((gnptr->address + gnptr->size) > start) {
			LWE_CRIT("valloc_add: "
				 "address space(address %#lx, size %#lx) "
				 "overlaps existing "
				 "address space(address %#lx, size %#x)\n",
				 start, size, gnptr->address, gnptr->size);
			tlfree(fnptr);
			return -1;
		}
	}

	list_add_free_area(fnptr);
	list_add_global_before(&fnptr->g_entry, &global_list_head);

	return 0;
}
