
/*
 * Copyright (C) 2007-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <hv_misc.h>
#include <trap_booke.h>
#include <trapframe.h>
#include <core-regs.h>
#include <bitops.h>
#include <io.h>
#include <log.h>
#include <exception_private.h>
#include <fsl_hcalls.h>
#include <mmap.h>
#include <string.h>

struct powerpc_exception {
	int vector;
	const char *name;
};

static const struct powerpc_exception powerpc_exceptions[] = {
	{ EXC_CRIT_INT, "critical input" },
	{ EXC_MCHECK, "machine check" },
	{ EXC_DSI, "data storage interrupt" },
	{ EXC_ISI, "instruction storage interrupt" },
	{ EXC_EXT_INT, "external interrupt" },
	{ EXC_ALIGN, "alignment" },
	{ EXC_PROGRAM, "program" },
	{ EXC_FPUNAVAIL,  "fpu unavailable" },
	{ EXC_SYSCALL, "system call" },
	{ EXC_AUXUNAVAIL, "aux proc unavail" },
	{ EXC_DECR, "decrementer" },
	{ EXC_FIT, "fixed-interval timer" },
	{ EXC_WDOG, "watchdog timer" },
	{ EXC_DTLB, "data tlb miss" },
	{ EXC_ITLB, "instruction tlb miss" },
	{ EXC_DEBUG, "debug" },
	{ EXC_PERFMON, "performance monitoring" },
	{ EXC_DOORBELL, "doorbell"},
	{ EXC_DOORBELLC, "doorbell critical"},
#ifdef HYPERVISOR
	{ EXC_GDOORBELL, "guest doorbell"},
	{ EXC_GDOORBELLC, "guest doorbell critical"},
	{ EXC_HCALL, "hcall"},
	{ EXC_EHPRIV, "ehpriv"},
#endif
	{ EXC_LAST, NULL }
};

static const char *trapname(int vector)
{
	const struct powerpc_exception *pe;

	for (pe = powerpc_exceptions; pe->vector != EXC_LAST; pe++) {
		if (pe->vector == vector)
			return (pe->name);
	}

	return "unknown";
}

static void traceback(trapframe_t *regs)
{
	unsigned long *sp = ((unsigned long *)regs->gpregs[1]);
	isr_printf("sp %p %lx %lx %lx %lx\n", sp, sp[0], sp[1], sp[2], sp[3]);
	sp = ((unsigned long **)regs->gpregs[1])[0];

	isr_printf("Traceback: ");

	for (int i = 1; sp != NULL; i++, sp = (unsigned long *)sp[0]) {
		if ((i % 7) == 0)
			isr_printf("\n");

		isr_printf("0x%08lx ", sp[1] - 4);
	}

	isr_printf("\n");
}

void dump_regs(trapframe_t *regs)
{
	static uint32_t dump_lock;
	register_t saved = 0;
	int lock = 0;

	if (!spin_lock_held(&dump_lock)) {
		lock = 1;
		saved = spin_lock_intsave(&dump_lock);
	}

	isr_printf("NIP 0x%08lx MSR 0x%08lx LR 0x%08lx ESR 0x%08lx EXC %d\n"
	       "CTR 0x%08lx CR 0x%08x XER 0x%08x DEAR 0x%08lx PIR %lu\n"
	       "Prev trap level %d\n",
	       regs->srr0, regs->srr1, regs->lr, mfspr(SPR_ESR), regs->exc,
	       regs->ctr, regs->cr, regs->xer, mfspr(SPR_DEAR), mfspr(SPR_PIR),
	       regs->traplevel);

	for (int i = 0; i < 32; i++) {
		isr_printf("r%02d 0x%08lx  ", i, regs->gpregs[i]);

		if ((i & 3) == 3)
			isr_printf("\n");
	}

#ifdef HYPERVISOR
	if (!(regs->srr1 & (MSR_GS | MSR_PR)))
#else
	if (!(regs->srr1 & MSR_PR))
#endif
		traceback(regs);

	if (lock)
		spin_unlock_intsave(&dump_lock, saved);
}

static void print_mcheck_exc(void)
{
	phys_addr_t addr;
	register_t mcsr;

	mcsr = mfspr(SPR_MCSR);

	isr_printf("MCSR = %#lx", mcsr);

	if (mcsr & MCSR_MAV) {
		addr = mfspr(SPR_MCAR);

		if (mcsr & MCSR_MEA)
			isr_printf("\tMCAR[EA] = %#llx\n", addr);
		else
			isr_printf("\tMCAR[PA] = %#llx\n", addr);
	}

	if (mcsr & MCSR_MCP) {
		struct hv_error mcheck_err;
		uint32_t bufsize = sizeof(struct hv_error);
		phys_addr_t mcheck_err_phys = vtop(&mcheck_err);
		uint32_t addr_hi = mcheck_err_phys >> 32;
		uint32_t addr_lo = (uint32_t)mcheck_err_phys;

		isr_printf("Machine check input signal asserted\n");
		isr_printf("Reading info from Guest event queue\n");

		int ret = fh_err_get_info(0, &bufsize, addr_hi, addr_lo, 0);
		if (ret) {
			isr_printf("Error information reading (fh_err_get_info)"
				   " from guest event queue failed\n");
			return;
		}

		isr_printf("domain: %s, error: %s, path: %s\n",
				mcheck_err.domain,
				mcheck_err.error, mcheck_err.hdev_tree_path);

		if (strcmp(mcheck_err.domain, "PAMU") == 0) {
			isr_printf("PAMU access violation avs1 = %#x,"
				   " avs2 = %#x, av_addr = %#llx\n",
				   mcheck_err.pamu.avs1, mcheck_err.pamu.avs2,
				   mcheck_err.pamu.access_violation_addr);
			isr_printf("PAMU access violation lpid = %#x,"
				   " handle = %#x\n", mcheck_err.pamu.lpid,
				    mcheck_err.pamu.liodn_handle);
		}
	}
}


void unknown_exception(trapframe_t *regs)
{
	isr_printf("\nException number: %u", regs->exc);
	isr_printf("\nException name: %s\n", trapname(regs->exc));

	switch (regs->exc) {
	case EXC_MCHECK: {
		print_mcheck_exc();
		break;
	}

	default:
		break;
	}

	dump_regs(regs);

	isr_printf("Freezing the core\n");
	while (1);
}
