/** @file
 * This file provides requisite APIs to support hardware interrupt
 */

/*
 * Copyright (c) 2008 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdbool.h>
#include <io.h>
#include <fsl_hcalls.h>
#include <libfdt.h>
#include <init.h>
#include <misc.h>
#include <log.h>
#include <errno.h>
#include <compiler.h>
#include <fsl-booke-tlb.h>
#include <mmap.h>
#include <ldt.h>
#include <hwi.h>
#include <hwi_private.h>
#include <compiler.h>
#include <epapr_hcalls.h>

volatile __PERCPU int in_irq;

struct hwi_global_s hwi_params;

__PERCPU struct hwi_percpu_s hwi_pcpu_params;

static int get_mpic_va(void)
{
	int ret = -1;
	phys_addr_t mpic_pa = 0;
	uint64_t size = 0;
	int offset = -1;

	offset = fdt_node_offset_by_compatible(devtree_ptr, 0, "epapr,hv-pic");
	if (unlikely(0 > offset)) {
		LWE_ERROR("%s: Failed for vmpic node", __func__);
		return -1;
	}
	if (unlikely(0 > get_fdt_mem_ranges(offset, &mpic_pa, &size))) {
		LWE_ERROR("%s: get_fdt_mem_ranges failed for vmpic node",
			 __func__);
		return -1;
	}

	/* set appropriate TLB entries */
	ret =
	    dev_mmap(mpic_pa, size, size, MAS2_M, TLB_MAS3_KERN,
		     &hwi_pcpu_params.mpic_va);
	if (unlikely(0 > ret)) {
		LWE_ERROR("%s: failed to create tlb entry for vmpic!",
			 __func__);
		return -1;
	}
	return 0;
}

static void default_handler(uint32_t hwi_handle, void *arg __UNUSED)
{
	LWE_PANIC("Unregisterd Interrupt - %d received,"
		  " Halting the system\n", hwi_handle);
}

static int hwi_bootcore_init(void)
{
	int i;
	int lenp;
	int offset = -1;

	/* Hardware interrupt in lagacy mode */
	offset = fdt_subnode_offset(devtree_ptr, 0, "hypervisor");
	if (unlikely(0 > offset)) {
		LWE_ERROR("%s: fdt_subnode_offset for hypervisor failed!",
			 __func__);
		return -FDT_ERR_NOTFOUND;
	}

	hwi_params.legacy = (uint32_t) fdt_get_property(devtree_ptr, offset,
							"fsl,hv-pic-legacy",
							&lenp);

	offset = fdt_node_offset_by_compatible(devtree_ptr, 0, "epapr,hv-pic");
	if (likely(0 < offset)) {
		if (NULL == fdt_node_check_compatible(devtree_ptr, offset,
						      "fsl,hv-mpic-per-cpu"))
			hwi_params.direct_eoi = true;

	} else {
		LWE_ERROR("%s: epapr,hv-pic a must property not present!",
			 __func__);
		return -FDT_ERR_NOTFOUND;
	}

	for (i = 0; i < MAX_INTERRUPTS; i++) {
		hwi_params.hwi_eoi[i] = false;
	}

	return 0;

}

static int hwi_percpu_init(void)
{
	int i;
	if (hwi_params.direct_eoi) {
		if (0 > get_mpic_va()) {
			LWE_ERROR("%s: get_mpic_va failed", __func__);
			return -1;
		}
	}

	hwi_pcpu_params.mpic_eoi = (void *)((uint32_t) hwi_pcpu_params.mpic_va +
								      MPIC_EOI);

	for (i = 0; i < MAX_INTERRUPTS; i++) {
		hwi_pcpu_params.lwe_hwi[i].handler = (void *)default_handler;
		hwi_pcpu_params.lwe_hwi[i].arg = 0;
	}

	return 0;
}

int hwi_init(int core)
{
	/* If the core is bootcore(0) do global variables initialization */
	if (core == 0) {
		if (0 > hwi_bootcore_init()) {
			LWE_ERROR("hwi_bootcore_init failed!");
			return -1;
		}
	}

	if (0 > hwi_percpu_init()) {
		LWE_ERROR("hwi_percpu_init failed!");
		return -1;
	}

	return 0;
}

int register_hwi(uint32_t hwi_handle, hwi_func handler, void *arg,
							uint32_t priority)
{
	uint32_t config;
	uint32_t dest;
	uint32_t config_prior;
	int ret = 0;

	if (hwi_handle >= MAX_INTERRUPTS) {
		LWE_ERROR("%s: interrupt source %d not supported",
			 __func__, hwi_handle);
		return -EINVAL;
	}

	if (hwi_pcpu_params.lwe_hwi[hwi_handle].handler !=
	    (void *)default_handler) {
		LWE_ERROR("%s: Handler already registered for interrupt 0x%x",
			 __func__, hwi_handle);
		return -EINVAL;
	}

	ret = ev_int_get_config(hwi_handle, &config, &config_prior,
								&dest);

	if (ret != 0) {
		LWE_ERROR("%s: hcall ev_int_get_config failed!",
			 __func__);
		goto err;
	}

	if (priority == 0 || priority > 15) {
		if (config_prior == 0)
			priority = INT_PRIORITY;
		else
			priority = config_prior;
	}

	if (hwi_params.direct_eoi) {
		if (config & MPIC_DIRECT_MASK)
			hwi_params.hwi_eoi[hwi_handle] = true;
	}

	hwi_pcpu_params.lwe_hwi[hwi_handle].handler = handler;
	hwi_pcpu_params.lwe_hwi[hwi_handle].arg = arg;

	ret = ev_int_set_config(hwi_handle, config, priority,
					lwe_instance);

	if (ret != 0) {
		LWE_ERROR
		    ("%s: hcall ev_int_set_config failed!", __func__);
		goto err1;
	}

	return ret;
err1:
	hwi_pcpu_params.lwe_hwi[hwi_handle].handler = (void *)default_handler;
	hwi_pcpu_params.lwe_hwi[hwi_handle].arg = 0;
err:
	return -ret;
}

int enable_hwi(uint32_t hwi_handle)
{
	if (hwi_handle >= MAX_INTERRUPTS) {
		LWE_ERROR("%s: interrupt source %d not supported",
			 __func__, hwi_handle);
		return -EINVAL;
	}
	if (hwi_pcpu_params.lwe_hwi[hwi_handle].handler ==
	    (void *)default_handler) {
		LWE_ERROR("%s: unregistered interrupt source 0x%x",
			 __func__, hwi_handle);
		return -EINVAL;
	}

	if (ev_int_set_mask(hwi_handle, INT_ENABLE) != 0) {
		LWE_ERROR("hwi_start: hcall ev_int_set_mask failed!");
		return -1;
	}

	return 0;
}

int disable_hwi(uint32_t hwi_handle)
{
	if (hwi_handle >= MAX_INTERRUPTS) {
		LWE_ERROR("%s: interrupt source %d not supported",
			 __func__, hwi_handle);
		return -EINVAL;
	}
	if (hwi_pcpu_params.lwe_hwi[hwi_handle].handler ==
	    (void *)default_handler) {
		LWE_ERROR("%s: unregistered interrupt source 0x%x",
			 __func__, hwi_handle);
		return -EINVAL;
	}

	if (ev_int_set_mask(hwi_handle, INT_DISABLE) != 0) {
		LWE_ERROR("hwi_stop: hcall ev_int_set_mask failed!");
		return -1;
	}

	return 0;
}

int free_hwi(uint32_t hwi_handle)
{
	uint32_t config;
	uint32_t dest;
	uint32_t priority;
	int ret = 0;

	if (hwi_handle >= MAX_INTERRUPTS) {
		LWE_ERROR("%s: interrupt source %d not supported",
			 __func__, hwi_handle);
		return -EINVAL;
	}

	if (hwi_pcpu_params.lwe_hwi[hwi_handle].handler ==
	    (void *)default_handler) {
		LWE_ERROR("%s: unregistered interrupt source 0x%x",
			 __func__, hwi_handle);
		return -EINVAL;
	}

	ret = ev_int_get_config(hwi_handle, &config, &priority,
								&dest);

	if (ret != 0) {
		LWE_ERROR("%s: hcall ev_int_get_config failed!",
			 __func__);
		return -ret;
	}

	spin_lock(&hwi_params.lock_ptr);

	ret = ev_int_set_config(hwi_handle, config, priority,
				 0);

	spin_unlock(&hwi_params.lock_ptr);

	if (ret != 0) {
		LWE_ERROR
		    ("%s: hcall ev_int_set_config failed!", __func__);
		return -ret;
	}

	hwi_pcpu_params.lwe_hwi[hwi_handle].handler = (void *)default_handler;
	hwi_pcpu_params.lwe_hwi[hwi_handle].arg = 0;

	return 0;
}

int ext_handler(void)
{
	unsigned int hwi_handle;
	in_irq++;

	if (!hwi_params.legacy) {
		hwi_handle = mfspr(SPR_EPR);
	} else {
		int ret = ev_int_iack(0, &hwi_handle);
		if (ret != 0) {
			in_irq--;
			return -ret;
		}
	}

	/* call interrupt handler */
	hwi_pcpu_params.lwe_hwi[hwi_handle].handler(hwi_handle,
						    hwi_pcpu_params.
						    lwe_hwi[hwi_handle].arg);

	if (true == hwi_params.hwi_eoi[hwi_handle]) {
		out32((uint32_t *) (hwi_pcpu_params.mpic_va), 0);
	} else {
		int ret = ev_int_eoi(hwi_handle);
		if (ret != 0) {
			in_irq--;
			return -ret;
		}
	}

	in_irq--;
	return 0;
}
