/**
\file
\brief	Exception handling registration support
*/

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <init.h>
#include <nmi.h>
#include <exception_private.h>
#include <trap_booke.h>
#include <core-regs.h>
#include <log.h>


__PERCPU struct exc_intr lwe_exc_intr[EXC_LAST] = 
			{[0 ... EXC_LAST - 1] = {NULL, NULL}};

/**
\brief	 	Machine check exception handler

\param[in] 	regs - Pointer to the trapframe structure

\note		Use of printf statement form the registered handler code is not
		supported
*/
void mcheck(struct trapframe *regs)
{
	register_t mcsr_context = mfspr(SPR_MCSR);

	in_irq++;

	if (mcsr_context & MCSR_NMI) {
		if (reg_nmi_hndlr)
			reg_nmi_hndlr();

		mtspr(SPR_MCSR, MCSR_NMI);

		in_irq--;

		return;
	}

	unknown_exception(regs);
}

/**
\brief		Exception handler for IVORs other than machine check, external
		interrupt, program interrupt, decrementer and watchdog.

\param[in] 	regs - Pointer to the trapframe structure

\note		Use of printf statement form the registered handler code is not
		supported
*/
void lwe_exception_handler(struct trapframe *regs)
{
	in_irq++;
	LWE_DEBUG("Inside %s", __func__);

	if (lwe_exc_intr[regs->exc].handler != NULL) {
		lwe_exc_intr[regs->exc].handler(regs,
				lwe_exc_intr[regs->exc].arg);
		in_irq--;
		return;
	}

	unknown_exception(regs);
}

int register_exception_handler(uint16_t vector, exc_func handler, void *arg)
{
	switch (vector) {
	case EXC_MCHECK:

	case EXC_EXT_INT:

	case EXC_DECR:

	case EXC_WDOG:

	case EXC_DOORBELL:
		LWE_ERROR("Exception handler registration not permitted");
		return -1;

	default:
		break;
	}

	if (NULL == handler) {
		LWE_ERROR("%s: NULL handler passed", __func__);
		return -1;
	}

	lwe_exc_intr[vector].handler = handler;
	lwe_exc_intr[vector].arg = arg;
	return 0;

}

