/** @file shell.c
 * LWE shell implementation
 */

/*
 * Copyright (c) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <libfdt.h>
#include <stdlib.h>
#include <hv_misc.h>
#include <fsl_hcalls.h>
#include <log.h>
#include <init.h>
#include <lwe_malloc.h>
#include <swi_private.h>
#include <hwi_private.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <shell.h>
#include <console.h>
#include <compiler.h>
#include <percpu.h>

#define MAX_CMD_BUFFER 2048
#define MAX_CMD_LEN 256
#define MAX_CMD_NAME 40
#define MAX_NUM_ARGS 20
#define MAX_HELP_STR 256
#define PROMPT_LEN 7

/** Structure of the command node
 */
struct cmd_node {
	char cmd[MAX_CMD_NAME];		/* stores command's name */
	char help[MAX_HELP_STR];	/* stores command's help string */
	shell_cmd_cb_hndlr func;	/* handler function for the command */
	struct cmd_node *next;	/* pointer to the next node in linked list*/
};

/* Stores the head/start of commands list */
static struct cmd_node *list_head;

/* Buffer to store the LWE shell prompt */
__PERCPU static char prompt[PROMPT_LEN];

/* Buffer to store the complete command entered */
__PERCPU static char cmd_data_buffer[MAX_CMD_BUFFER];

/* Maintains the index of the command entered */
__PERCPU static uint32_t index;

/** Traverse the command linked list.
 */
static void traverse_list(void)
{
	struct cmd_node *temp = NULL;
	temp = list_head;

	printf("Commands:\r\n");
	while (temp) {
		printf("  %s - %s\r\n", temp->cmd, temp->help);
		temp = temp->next;
	}
}


/** Command handler function for help command
 */
static int help_cb(int argc __UNUSED, char **args __UNUSED)
{
	traverse_list();

	return 0;
}


/** Search the user entered command in the linked list
 */
static shell_cmd_cb_hndlr cmd_search(const char *cmd)
{
	struct cmd_node *temp = list_head;

	if (list_head == NULL) {
		LWE_ERROR("Command List is empty.\r\n");
		return NULL;
	}

	while (temp != NULL) {
		if (strcmp(temp->cmd, cmd))
			temp = temp->next;
		else
			return temp->func;
	}

	return NULL;
}


/** Command list contains the command names, their help string and
 *  their respective handler functions
 *  Creates a linked list of commands
 */
static int create_list(void)
{
	struct cmd_node *tmp_node;

	tmp_node = malloc(sizeof(struct cmd_node));
	if (!tmp_node) {
		LWE_ERROR("malloc failed in %s: Cudn't allocate memory for "
			"temp command node\r\n", __func__);
		return -1;
	}

	list_head = tmp_node;
	tmp_node->next = NULL;

	strcpy(tmp_node->cmd, "help");
	strcpy(tmp_node->help, "Print command usage information");
	tmp_node->func = &help_cb;

	return 0;
}


/** Add command nodes into the linked list
 */
int add_command(char *cmd, shell_cmd_cb_hndlr hndlr_func, char *usage_str)
{
	struct cmd_node *tmp_node;
	int *tmp_head;
	tmp_head = (int *)list_head;
	shell_cmd_cb_hndlr function;

	tmp_node = malloc(sizeof(struct cmd_node));
	if (!tmp_node) {
		LWE_ERROR("malloc failed in %s: Cudn't allocate memory\r\n"
			, __func__);
		return -1;
	}
	if (list_head == NULL) {
		LWE_ERROR("%s: List is empty\r\n", __func__);
		free(tmp_node);
		return -1;
	}
	tmp_node->next = NULL;

	function = cmd_search(cmd);

	if (function) {
		LWE_ERROR("Command already exists.\r\n");
		free(tmp_node);
		return -1;
	}

	strcpy(tmp_node->cmd, cmd);
	tmp_node->func = hndlr_func;

	strcpy(tmp_node->help, usage_str);

	while (list_head->next != NULL)
		list_head = list_head->next;

	list_head->next = tmp_node;
	list_head = (struct cmd_node *)tmp_head;

	return 0;
}

/** Calls the command handler for the user entered command
 *  This function gets called from SWI context
 */
static void process_cmd(uint32_t swi_id __UNUSED, void *data)
{
	int argc = 0;
	char *token;
	char *argv[MAX_NUM_ARGS];
	shell_cmd_cb_hndlr function;
	char *cmd_buffer = data;
	int status = 0;

	token = strtok(cmd_buffer, " ");

	if (NULL == token) {
		LWE_ERROR("strtok failed for command name");
		return;
	}

	while (argc < MAX_NUM_ARGS) {
		argv[argc++] = token;

		token = strtok(NULL, " ");
		if (NULL == token)
			break;
	}

	function = cmd_search(argv[0]);

	if (function)
		function(argc, argv);
	else
		printf("Command '%s' doesn't exist\n", argv[0]);

	status = console_put((uint8_t *)prompt, strlen(prompt));
	if (status != 0)
		LWE_INFO("Sending prompt failed with status = %d", status);
}

/* ASSUMPTION -- The command till first \n in buffer
 * will be parsed and scheduled. Current code would
 * loose any other command after first \n
 */
static void parse_cmd_buff(int start_idx, int num_chars)
{
	static int last_cmd_idx;
	static int last_cmd_len;

	if (0 == last_cmd_len)
		last_cmd_idx = start_idx;	/*Store start of new command*/

	for (int i = start_idx; i < start_idx + num_chars; i++) {
		if (cmd_data_buffer[i] != '\n') {
			last_cmd_len++;
			continue;
		}

		cmd_data_buffer[i] = NULL;	/*Mark end of cmd*/

		if (0 == last_cmd_len) {
			int status = console_put((uint8_t *)prompt,
					strlen(prompt));

			if (status != 0)
				LWE_ERROR("console_put failed with "
					"status = %d", status);

			last_cmd_len = 0;
			return;

		}

		if (0 > swi_add(SWI_ID_INV,
				(void *)&cmd_data_buffer[last_cmd_idx],
				process_cmd, SWI_PRI_HIGH))
			LWE_ERROR("swi_add failed for shell cmd");

		last_cmd_len = 0;	/*Start new cmd next time*/
		return;
	}

	return;
}

/** HWI handler for the commands entered through byte channel
 *  After parsing, it registers the command as SWI
 */
static int shell_hwi_handler(void)
{
	int start_idx;
	int num_chars;

	if ((MAX_CMD_BUFFER - index) < (MAX_CMD_LEN + 1)) {
		index = 0;
	}

	start_idx = index;

	num_chars = console_gets(&(cmd_data_buffer[index]),
					MAX_CMD_BUFFER - index);
	if (num_chars == -1) {
		LWE_WARN("console_gets failed ...");
		index = 0;
		return -1;
	}

	if (num_chars > MAX_CMD_LEN) {
		LWE_WARN("cmd len: %d exceeds limit, proceeding", num_chars);
	}

	index += num_chars;
	parse_cmd_buff(start_idx, num_chars);

	return 0;
}

/** Initialises the stdout interface and prints the LWE prompt
 */
int shell_init(void)
{
	int ret = 0;
	uint32_t vcpu_num = mfspr(SPR_PIR);

	sprintf(prompt, "LWE_%d>", vcpu_num);

	/* Initialize linked list of commands and add help
	 help command into the list  */
	if (lwe_instance == 0) {
		ret = create_list();
		if (ret < 0)
			LWE_ERROR("Command database initialization and "
				"creation failed.");
#define REGISTER_CMD(cmd, cmd_handler, help_str) \
			int cmd_handler(int argc, char *argv[]); \
			add_command(cmd, cmd_handler, help_str);
#include <cmd_def.h>
#undef REGISTER_CMD
	}

	ret = attach_console_input((stdin_intr_handler_t)shell_hwi_handler);

	if (ret < 0)
		LWE_ERROR("Not able to register handler.");

	ret = console_put((uint8_t *)prompt, strlen(prompt));
	if (ret != 0)
		LWE_ERROR("Sending prompt failed.");

	return 0;
}
