/** @file
 * LWE initialization
 */

/*
 * Copyright (c) 2008 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <libfdt.h>
#include <stdlib.h>
#include <hv_misc.h>
#include <fsl_hcalls.h>
#include <percpu.h>
#include <fsl-booke-tlb.h>
#include <trapframe.h>
#include <core-regs.h>
#include <io.h>
#include <hv_dep.h>
#include <elf.h>
#include <log.h>
#include <ldt.h>
#include <init.h>
#include <misc.h>
#include <mmap.h>
#include <cache.h>
#include <barrier.h>
#include <lwe_atomic.h>
#include <malloc_wrapper.h>
#include <swi_private.h>
#include <hwi_private.h>
#include <msl_math_api_gcc.h>
#include <mmap_private.h>
#include <compiler.h>
#include <console.h>
#include "perf_mon.h"
#include <cpu_usage.h>
#include <timer_wheel_private.h>

#define BOOT_CORE	0
#define SEC_CORE	1
#define PAGE_SIZE 	0x1000
#define WIMGE_SHIFT	20
#define RWX_MASK	0x7
#define TLB1CFG_NENTRY	0x00000fff
#define DEVTREE_MAX_SIZE	(64*1024)

extern int bman_init(int secondary);
extern int qman_init(int secondary);
extern int shell_init(void);

/* generated by module_init(pme2_low_init) */
extern int m_pme2_low_init(void);
extern uint32_t _sec_start_paddr;

struct cpu g_cpu_table[MAX_CORES] =
		{
			{
				.kstack = (void *)&_percpu_bss_heap_stack_end,
				.client = 0,
			},
		};

/* Guest device tree pointer as
 * passed from the hypervisor
 */
void *devtree_ptr;

/** guest_cache_lock_on - Checks if the guest cache lock mode is 
 *  turned on in the LWE device tree
 */
uint32_t guest_cache_lock_on;

/* Secondary core marks itself in this bitamsk befire entering nap state */
unsigned long napping;
/*  core mask -- coreamsk as specifued in the device tree */
uint32_t coremask;

/* Online core mask -- Gives the number of cores on which application
 *  has been started by the boot core */
uint32_t online_core_mask = 0;
uint32_t sec_core_mask;

struct lwe_barrier sec_barrier;

/** timebase_freq - Timebase frequency as filled in by uboot 
 *  in the LWE device tree
 */
__PERCPU uint32_t timebase_freq;

/** lwe_instance - Zero based lwe instance which gets incremented
 *  by one in the order in which boot up of various LWEs happen
 */
__PERCPU uint32_t lwe_instance;

/** l1_data_cache_blk_size - Block size of L2 cache as read from the
 *  LWE device tree
 */
__PERCPU uint32_t l1_data_cache_blk_size;

/** l1_inst_cache_blk_size - Block size of L2 cache as read from the
 *  LWE device tree
 */
__PERCPU uint32_t l1_inst_cache_blk_size;

/** l2_cache_blk_size - Block size of L2 cache as read from the
 *  LWE device tree
 */
__PERCPU uint32_t l2_cache_blk_size;

uint8_t g_num_cores;

uint32_t get_tsize(uint64_t size);
static uint32_t getalignment(uint64_t addr);
static void prepare_tlb1_snapshot(cpu_t *);
uint64_t get_sizefromindex(uint32_t sizeidx);
static uint32_t min3(uint32_t a, uint32_t b, uint32_t c);

static void create_tlb1_entries(uint64_t vstart, uint64_t pstart,
		uint64_t size, uint32_t *tlb1_entry,
		uint32_t flags, uint32_t maxentry);

static void startup_error(void);

__attribute__((section(".devtree")))  __ALIGNED(CACHE_LINE_SIZE)
unsigned char devtree_area[DEVTREE_MAX_SIZE];

#define MAX_ARGC	256
#define MAX_ARGV	256

struct spintbl_entry {
	volatile entryp main;
	volatile int argc;
	volatile char *argv[MAX_ARGC];
};
struct spintbl_entry spintbl[MAX_CORES];
const void *pcpu_data_vaddr;
unsigned long cores_started;

/* Functions */
void *__real_memcpy(void *dest, const void *src, size_t n)
{
	return memcpy(dest, src, n);
}

uint32_t shutdn_dbell_cb(int irq __UNUSED, void *arg __UNUSED)
{
	swi_add(SWI_ID_SHUTDOWN, NULL, NULL, SWI_PRI_HIGH);
	return 0;
}

static void store_apps_percore_area(void)
{
	uint32_t size = (uint32_t)&_percpu_data_end - (uint32_t)&_percpu_start;
	uint32_t start = (uint32_t)&_percpu_start;

	pcpu_data_vaddr = (void *)malloc(size);
	if (!pcpu_data_vaddr) {
		LWE_PANIC("Sufficient space not available in heap to store"
			 " percpu data");
		return;
	}

	memcpy((void *)pcpu_data_vaddr, (void *)(start), size);

	return;
}

static void restore_percpu_on_self(void)
{
	uint32_t start = (uint32_t) &_percpu_start;
	unsigned long lwe_pcpu_start = (uint32_t) &_lwe_percpu_data_start;
	unsigned long lwe_pcpu_end = (uint32_t) &_lwe_percpu_data_end;
	uint32_t end = (uint32_t)&_percpu_data_end;

	memcpy((void *)start, pcpu_data_vaddr, lwe_pcpu_start - start);
	memcpy((void *)lwe_pcpu_end, pcpu_data_vaddr + lwe_pcpu_end - start,
		end - lwe_pcpu_end);
}

static int init_percore_area(uint32_t num_cores)
{
	int ret = 0;
	void *dst_vaddr;
	phys_addr_t paddr;
	uint32_t size = (uint32_t) & _percpu_end - (uint32_t) & _percpu_start;
	uint32_t start = (uint32_t) & _percpu_start;

	phys_addr_t base_paddr = vtop((void *)start);

	if (UINTMAX_MAX == base_paddr) {
		LWE_ERROR("init_percore_area:	\
			Invalid percpu start virtual address\n");
		return -1;
	}

	store_apps_percore_area();

	for (uint32_t i = 1; i < num_cores; i++) {
		paddr = base_paddr + (size * i);
		int32_t data_size = (uint32_t) &_percpu_data_end -
		    (uint32_t) &_percpu_start;

		/* if paddr is not 4k aligned, PANIC */
		if (paddr & (PAGE_SIZE - 1))
			LWE_PANIC("paddr is not 4k aligned %#llx", paddr);

		/* copy 4k at a time */
		for (uint32_t j = 0; data_size > 0; j++) {

			/* 1. map 4k using dev_mmap. */
			ret = dev_mmap(paddr + j * PAGE_SIZE,
				 PAGE_SIZE, PAGE_SIZE, TLB_MAS2_MEM | MAS2_W,
				 TLB_MAS3_KERN, &dst_vaddr);

			if (ret < 0)
				LWE_PANIC("dev_mmap failed on paddr = %#llx",
					  paddr);

			/* memcpy */
			if (data_size < PAGE_SIZE)
				memcpy(dst_vaddr,
				       (void *)(start + j * PAGE_SIZE),
				       data_size);
			else
				memcpy(dst_vaddr,
				       (void *)(start + j * PAGE_SIZE),
				       PAGE_SIZE);

			/* Unmap */
			dev_unmap(dst_vaddr);

			/* Decrement a page */
			data_size -= PAGE_SIZE;
		}
	}

	/* Clear the percpu bss */
	size = (uint32_t) &_percpu_bss_heap_stack_end -
			(uint32_t) &_percpu_bss_heap_stack_start;

	if (size == 0) {
		LWE_INFO("No percpu bss data available");
		return 0;
	}

	start = (uint32_t) &_percpu_bss_heap_stack_start;

	base_paddr = vtop((void *)start);
	for (uint32_t i = 0; i < num_cores; i++) {
		paddr = base_paddr + (size * i);
		int32_t data_size = (uint32_t) &_percpu_bss_end -
		    (uint32_t) &_percpu_bss_heap_stack_start;

		/* if paddr is not 4k aligned, PANIC */
		if (paddr & (PAGE_SIZE - 1))
			LWE_PANIC("paddr is not 4k aligned %#llx", paddr);

		/* copy 4k at a time */
		for (uint32_t j = 0; data_size > 0; j++) {

			/* 1. map 4k using dev_mmap. */
			ret = dev_mmap(paddr + j * PAGE_SIZE,
				 PAGE_SIZE, PAGE_SIZE, TLB_MAS2_MEM | MAS2_W,
				 TLB_MAS3_KERN, &dst_vaddr);

			if (ret < 0)
				LWE_PANIC("dev_mmap failed on paddr = %#llx",
					  paddr);

			/* memset */
			if (data_size < PAGE_SIZE)
				memset(dst_vaddr, 0, data_size);
			else
				memset(dst_vaddr, 0, PAGE_SIZE);

			/* Unmap */
			dev_unmap(dst_vaddr);

			/* Decrement a page */
			data_size -= PAGE_SIZE;
		}
	}

	return 0;
}

int
start_secondary_spin_table(struct boot_spin_table *table,
			   int num, cpu_t * newcpu)
{
	phys_addr_t entry = (unsigned long)&_sec_start_paddr;

	table->r3_lo = (unsigned long)newcpu;
	table->pir = num;

	out32((uint32_t *) & table->addr_hi, entry >> 32);
	out32((uint32_t *) & table->addr_lo, (uint32_t) entry);

	return 0;
}

static void release_secondary_cores(void)
{
	int node = -1;
	void *map;

	node = fdt_node_offset_by_prop_value(devtree_ptr, node, "device_type",
					     "cpu", strlen("cpu") + 1);
	if (node < 0) {
		LWE_ERROR("BROKEN: error %d (%s) reading CPU nodes, "
			  "secondary cores may not be released.\n",
			  node, fdt_strerror(node));
		return;
	}

	do {
		int len = 0;
		const char *str;
		struct cpu *cpu = NULL;

		str = fdt_getprop(devtree_ptr, node, "status", &len);
		if (!str || strcmp(str, "disabled") != 0)
			continue;

		str = fdt_getprop(devtree_ptr, node, "enable-method", &len);
		if (!str) {
			LWE_ERROR("BROKEN: Missing enable-method on"
				  "disabled cpu node");
			continue;
		}
		if (strcmp(str, "spin-table") != 0) {
			LWE_ERROR
			    ("BROKEN: Unknown enable-method \"%s\";not enabling\n",
			     str);
			continue;
		}
		const uint32_t *reg = fdt_getprop(devtree_ptr, node, "reg",
						  &len);
		if (!reg) {
			LWE_ERROR("BROKEN: Missing reg property in cpu" "node");
			continue;
		}

		const uint64_t *table =
		    fdt_getprop(devtree_ptr, node, "cpu-release-addr",
				&len);
		if (!table) {
			LWE_ERROR("BROKEN: Missing cpu-release-addr"
				  "property in cpu node");
			continue;
		}

		int ret =
		    dev_mmap((*table) & ~(PAGE_SIZE - 1), PAGE_SIZE, PAGE_SIZE,
			     TLB_MAS2_IO, TLB_MAS3_KERN, &map);
		if (ret < 0) {
			LWE_ERROR("dev_mmap failed for spin table entry");
			return;
		}

		char *table_va = map;
		table_va += *table & (PAGE_SIZE - 1);

		cpu = &(g_cpu_table[*reg]);

		cpu->kstack = (void *)((uint32_t)&_percpu_bss_heap_stack_end);

		cpu->kstack -= FRAMELEN;

		LWE_INFO("Releasing vcpu: %u", *reg);

		if (start_secondary_spin_table((void *)table_va, *reg, cpu))
			printf("BROKEN: couldn't spin up CPU%u\n", *reg);

		dev_unmap(map);

	} while ((node = fdt_node_offset_by_prop_value(devtree_ptr, node,
						       "device_type", "cpu",
						       strlen("cpu") + 1)) !=
		 -FDT_ERR_NOTFOUND);
}

/** get_valid_mask
 * Returns the mask of cores to be released from spin
 * (gets it from botoargs coremask property of present
 *  else returns the number of sec cores in device tree 
 */
static uint32_t get_valid_mask(void)
{

	uint32_t core_mask = get_sec_core_mask();
	uint32_t mask = 0;
	int ret = 0;

	core_mask |= 1;
	ret = get_bootargs_coremask(&mask);
	if (ret < 0)
		mask = core_mask;

	/* put bootcore in the mask if it doesn't have it */
	mask |= 1;
	if (~core_mask & mask)
		LWE_PANIC("bootargs - coremask attempting to use "
			  "a cpu not a part of this partition\n");

	return mask;
}

static void release_sec_core_from_nap(void)
{
	int ret = 0;
	unsigned int state;
	unsigned long nap_mask = get_sec_core_mask();
	while (1) {
		if (!nap_mask)
			break;
		for (int i = 1; i < get_num_cores(); i++) {
			if (!(atomic_read(&napping) & (1 << i)) ||
						!(nap_mask & (1 << i)))
				continue;
			ret = fh_get_core_state(-1, i, &state);
			if (ret) {
				LWE_ERROR("fh_get_core_state: FAILED: "
				       "vcpu %d status %u\n", i, ret);
				continue;
			}

			if (state == FH_VCPU_NAP) {
				nap_mask = nap_mask & ~(1 << i);
				ret = fh_exit_nap(-1, i);
				if (ret) {
					LWE_ERROR("fh_exit_nap: FAILED: "
					 "vcpu %d from boot core failed %d\n",
								      i, ret);
				}
			}
		}
	}
}

static void exit_lwe(void)
{
	/* once the core exits from main function, it removed itself from
	   online_core_mask */
	atomic_and((unsigned long *)&online_core_mask, ~(1 << lwe_instance));

	if (lwe_instance == 0)
		release_sec_core_from_nap();

	/* Synchronize all the cores at this point */
	if (barrier_sync(&sec_barrier) < 0)
		LWE_PANIC("exit :Barrier Sync failed");
}

static void common_init(int secondary)
{
	int len;

	/* Enable external and critical interrupts */
	enable_extint();
	enable_critint();

	/* Initialize the per-core heap -- tlmalloc */
	LWE_INFO("Stack begin: %p, Stack size: %u",
		 &_percpu_bss_heap_stack_end, (size_t)&_percpu_stack_size);

	tlmalloc_init((void *)&_percpu_bss_heap_stack_start,
		      (void *)((unsigned long)&_percpu_bss_heap_stack_end -
			       (unsigned long)&_percpu_stack_size - 1));

	/*Attach stdout to the console*/
	if (0 > subscribe_console())
		LWE_PANIC("subscribe_console() failed");

	/* Add available virtual addresses to the virtual address allocator */
	alloc_vspace();

	/* Initialize swi wheel */
	swi_init();

	/* Initialize hwi */
	if (hwi_init(secondary) < 0)
		LWE_PANIC("init: hwi_init failed!");

	/* Get Cache Sizes */
	len = get_cache_blk_size(L1_DATA_CACHE);
	if (len < 0) {
		LWE_WARN
		("%s: Could not find l1 data cache in device tree!!",
			__func__);
		l1_data_cache_blk_size = 0;
	} else {
		l1_data_cache_blk_size = (uint32_t)len;
	}

	len = get_cache_blk_size(L1_INST_CACHE);
	if (len < 0) {
		LWE_WARN
		("%s: Could not find l1 instruction cache in device tree!!",
			__func__);
		l1_inst_cache_blk_size = 0;
	} else {
		l1_inst_cache_blk_size = (uint32_t)len;
	}

	len = get_cache_blk_size(L2_CACHE);
	if (len < 0) {
		LWE_WARN
		("%s: Could not find l2 cache in device tree!!",
			__func__);
		l2_cache_blk_size = 0;
	} else {
		l2_cache_blk_size = (uint32_t) len;
	}

	/* Read timebase frequency */
	if (!(timebase_freq = get_tbfreq()))
		LWE_PANIC("init: Could not read timebase frequency!!");

	/* Initialize bman */
	if (bman_init(secondary))
		LWE_PANIC("init: Could not initialise Bman driver!!");

	/* Initialize bman */
	if (qman_init(secondary))
		LWE_PANIC("init: Could not initialise Qman driver!!");

	/* Initialize Performance Monitoring Block*/
	initPM();

	/*Initialise timer framework*/
	if (0 != timer_framework_init(MAX_TIMEOUT_MS,
				NUM_TIMER_OBJECTS,
				RECLAIM_QUEUE_LEN))
		LWE_PANIC("Timer framework initialisation failed");

	/*Initialise the shell*/
	if (0 != shell_init()) {
		LWE_ERROR("Shell initialisation failed");
		return;
	}
}

void lwe_spin_loop(void)
{
	int ret = 0;
loop:
	atomic_or(&napping, 1 << lwe_instance);
	ret = fh_enter_nap(-1, lwe_instance);
	if (unlikely(ret)) {
		LWE_ERROR("fh_enter_nap: FAILED: "
	       "vcpu %d from boot core failed %d\n", lwe_instance, ret);
	}

	atomic_and(&napping, ~(1 << lwe_instance));
	if (!(online_core_mask & (1 | 1 << lwe_instance))) {
		LWE_INFO("CORE EXITING SPIN LOOP\n");
		return;
	}

	while (spintbl[lwe_instance].main == NULL) {
		if (!(online_core_mask & (1 | 1 << lwe_instance))) {
			LWE_INFO("CORE EXITING\n");
			return;
		}
	}

	LWE_INFO("CORE EXITING SPIN LOOP\n");

	init_cpu_usage();
	spintbl[lwe_instance].main(spintbl[lwe_instance].argc,
					(char **)spintbl[lwe_instance].argv);

	spintbl[lwe_instance].argc = 0;
	spintbl[lwe_instance].main = NULL;

	/* once the core exits from main function, it removed itself from
	   online_core_mask */
	atomic_and((unsigned long *)&online_core_mask, ~(1 << lwe_instance));

	/* Restore apps percpu data of this core */
	restore_percpu_on_self();

	goto loop;
}

void secondary_init(void)
{
	/* Enable FPU*/
	mtmsr(mfmsr() | MSR_FP);

	/* Set the number of entries in TLB1 */
	cpu->tlb1cnt = mfspr(SPR_TLB1CFG) & TLB1CFG_NENTRY;

	/*Set the console to not OK*/
	cpu->console_ok = 0;

	/* Each core sets its pid in lwe_instance */
	lwe_instance = mfspr(SPR_PIR);

	/*Enable the logging functionality*/
	logger_init();

	LWE_INFO("Inside function %s", __func__);

	/* Initialize vtop table on secondary core */
	initialize_vptbl(lwe_instance);

	prepare_tlb1_snapshot(cpu);

	common_init(SEC_CORE);

	if (barrier_sync(&sec_barrier) < 0)
		LWE_PANIC("init :Barrier Sync failed");

	if ((coremask & (1 << lwe_instance))) {
		atomic_or((unsigned long *)&online_core_mask,
						1 << lwe_instance);
		init_cpu_usage();
		main(0, NULL);
		atomic_and((unsigned long *)&online_core_mask,
						 ~(1 << lwe_instance));
		restore_percpu_on_self();
	}

	LWE_INFO("CORE ENTERING SPIN LOOP\n");
	lwe_spin_loop();

	exit_lwe();

	while (1) ;
}

int start_cpu(uint8_t vcpu_id, entryp func, int argc, char *argv[])
{
	if (unlikely(0 != lwe_instance)) {
		LWE_ERROR("%s :INVALID PARAMETER : Only boot core is"
			  " authorized to call this function\n", __func__);
		return -EINVAL;
	}

	unsigned int state;
	int ret = fh_get_core_state(-1, vcpu_id, &state);
	if (unlikely(online_core_mask & (1 << vcpu_id) ||
					(ret && state != FH_VCPU_NAP))) {
		LWE_ERROR("%s :Core is not idle yet. Try again Later\n",
				 __func__);
		return -EAGAIN;
	}

	if (unlikely(func == NULL)) {
		LWE_ERROR("%s:Entry Point function can't be NULL\n", __func__);
		return -EINVAL;
	}

	/* Max of 255 arguments are allowed in the function */
	if (unlikely(argc > MAX_ARGC))
		argc = MAX_ARGC;

	atomic_or((unsigned long *)&online_core_mask, 1 << vcpu_id);

	spintbl[vcpu_id].argc = argc;
	for (int i = 0 ; i < argc; i++)
		strcpy((char *)spintbl[vcpu_id].argv[i], (char *)argv[i]);

	lwsync();
	spintbl[vcpu_id].main = func;

	ret = fh_exit_nap(-1, vcpu_id);
	if (unlikely(ret)) {
		LWE_ERROR("fh_exit_nap: FAILED: "
	       "vcpu %d from boot core failed %d\n", vcpu_id, ret);
	}
	return 0;
}

/** start
 * Provides the entry point, "start" which is hooked  
 * with _start of head.S.
 *
 * @param[in] devtp pointing to start offset of device tree blob
 */

void start(unsigned long devtp __UNUSED)
{
	uint32_t mask = 0;
	int node, len;
	int shutdn_irq;
	uint32_t num_cores;
	int num_consoles;

	/* Enable FPU*/
	mtmsr(mfmsr() | MSR_FP);

	/* Initialize the device tree ptr */
	devtree_ptr = (void *) &devtree_area;

	/* Each core sets its pid in lwe_instance */
	lwe_instance = mfspr(SPR_PIR);

	/* Set the number of entries in TLB1 */
	cpu->tlb1cnt = mfspr(SPR_TLB1CFG) & TLB1CFG_NENTRY;

	/*Set the console to not OK*/
	cpu->console_ok = 0;

	/*Enable the logging functionality*/
	logger_init();

	LWE_INFO("Inside function %s", __func__);

	/* Get number of cpus in the partition */
	sec_core_mask = get_sec_core_mask();
	num_cores =  __msl_count_trailing_one32(sec_core_mask + 1);
	g_num_cores = num_cores;

	LWE_INFO("Number of cores in LWE partition: %u", num_cores);

	/* Initialize vtop table on secondary core */
	initialize_vptbl(lwe_instance);

	/* Initialize bootcore's tlb1 */
	prepare_tlb1_snapshot(cpu);

	/* Initialize global heap */
	sys_malloc_init((void *)&_global_heap_start,
			(void *)(&_global_heap_end - 1));

	/* Initlaize the statistics area for stats_malloc */
	stats_malloc_init((void *)&_stats_heap_start,
			  (void *)&_stats_heap_end - 1);

	/* Initialize primary core percpu area and replicate them for
	 * other cpus 
	 */
	if (init_percore_area(num_cores)) {
		LWE_PANIC("init() error: Failed in initializing percpu area\n");
	}

	/* Initialize the spin table argv */
	for (int i = 0 ; i < MAX_CORES; i++) {
		for (int j = 0; j < MAX_ARGC; j++) {
			spintbl[i].argv[j] = malloc(MAX_ARGV * sizeof(char));
			if (!spintbl[i].argv[j])
				LWE_PANIC("Initialization of spintbl structure"
					  "failed for each core");
		}
	}

	/*Find and initialise the LWE consoles*/
	num_consoles = init_consoles();
	if (0 > num_consoles)
		LWE_PANIC("init_consoles() failed");

	LWE_INFO("Number of consoles initialised: %d", num_consoles);

	/* Check if cache locking is enabled in the sustem */
	if ((node = fdt_subnode_offset(devtree_ptr, 0, "hypervisor")) >= 0)
		if (fdt_getprop
		    (devtree_ptr, node, "fsl,hv-guest-cache-lock", &len))
			guest_cache_lock_on = 1;

	common_init(BOOT_CORE);

	/* don't grep for this fn, you won't find it - it's generated by the
	 * module_init(pme2_low_init) macro. */
	if (m_pme2_low_init())
		LWE_PANIC("init: Could not initialise PME2!!");

	mask = get_sec_core_mask() | 1;
	if (barrier_init(&sec_barrier, mask) < 0)
		LWE_PANIC("init :Barrier initialization failed");

	coremask = get_valid_mask();
	/* Bring out other cores from spinning */
	release_secondary_cores();

	atomic_or((unsigned long *)&online_core_mask, 1 << lwe_instance);
	if (barrier_sync(&sec_barrier) < 0)
		LWE_PANIC("init :Barrier Sync failed");

	shutdn_irq = get_shutdn_dbell_irq();
	if (shutdn_irq > 0) {
		if (register_hwi(shutdn_irq, shutdn_dbell_cb, NULL, 0) < 0)
			LWE_PANIC("init: register_hwi for shutdn_irq failed");

		if (enable_hwi(shutdn_irq) < 0)
			LWE_PANIC("init: enable_hwi shutdn_irq failed");
	} else
		LWE_PANIC("init: get_shutdn_dbell_irq failed");

	init_cpu_usage();
	main(0, NULL);

	exit_lwe();

	/* If all the cores have hit the barrier above, boot core would stop
	   the partition */
	LWE_INFO("All cores done. Stopping LWE partition...");
	fh_partition_stop(-1);
}

/* returns the appropriate MAS3 supervisor permission constant
 * for the r/w/x flag bits in pheader
 */
static uint8_t get_mas3_permflags(uint8_t flag)
{
	uint8_t permission = 0x0;
	if (flag & 0x1)
		permission |= MAS3_SX;
	if (flag & 0x2)
		permission |= MAS3_SW;
	if (flag & 0x4)
		permission |= MAS3_SR;

	return permission;
}

void copy_devtree(unsigned char *devtree)
{
	/* Device tree is copied into text section */
	unsigned char *dtreloc = (unsigned char *)
				((unsigned long)devtree_area
				- (unsigned long)&_start +
				(unsigned long)&_start_paddr);

	struct fdt_header *header = (struct fdt_header *)devtree;

	if (header->totalsize > DEVTREE_MAX_SIZE)
		startup_error();

	for (uint32_t i = 0; i < header->totalsize; i++)
		dtreloc[i] = devtree[i];

	for (uint32_t j = 0; j < header->totalsize; j = j + CACHE_LINE_SIZE)
		asm volatile("dcbf 0, %0" : : "r"(dtreloc + j));
}

void set_memory_tlb(uint32_t entry)
{
	uint32_t tlb1_index = entry & 0x0000ffff;
	uint32_t maxindex = (entry & 0xffff0000) >> 16;
	struct elf_header *ehdr;
	uint32_t size;

	unsigned long _begin = (unsigned long)&_start_paddr - 0x10000;

	ehdr = (struct elf_header *)_begin;
	struct program_header *ph = (struct program_header *)((uint32_t) _begin
							      + ehdr->phoff);
	int phnum = ehdr->phnum;
	for (int i = 0; i < phnum; i++) {
		size = (uint32_t) & _percpu_end - (uint32_t) & _percpu_start;
		if (ph[i].vaddr == (uint32_t) & _percpu_start) {
			create_tlb1_entries(ph[i].vaddr,
					    ph[i].paddr +
					    (size * mfspr(SPR_PIR)), size,
					    &tlb1_index, ph[i].flags, maxindex);
		} else if (ph[i].vaddr ==
			(uint32_t) &_percpu_bss_heap_stack_start) {
			uint32_t size =
				(uint32_t) &_percpu_bss_heap_stack_end -
				    (uint32_t) &_percpu_bss_heap_stack_start;
			create_tlb1_entries(ph[i].vaddr,
					    ph[i].paddr +
					    (size * mfspr(SPR_PIR)), size,
					    &tlb1_index, ph[i].flags, maxindex);

		} else
			create_tlb1_entries(ph[i].vaddr, ph[i].paddr,
					    ph[i].memsz, &tlb1_index,
					    ph[i].flags, maxindex);
	}
}

static uint32_t getalignment(uint64_t addr)
{
	uint32_t minsize = 0xfff;
	uint32_t maxsize = 0xffffffff;
	uint64_t align1, align2;
	uint64_t mask;
	uint32_t v1;

	/* Check if the address is 4k aligned */
	if (addr & minsize)
		return 0;

	mask = minsize;

	while (1) {
		align1 = mask + 1;
		align2 = align1 * 4;
		mask = (mask << 0x4) | 0xf;

		/* return 4G as the max */
		if (mask == 0xfffffffff)
			return maxsize;

		v1 = addr & mask;

		if (!v1)
			continue;

		if (!(v1 % align2))
			return align2;

		return align1;
	}
}

static uint32_t min3(uint32_t a, uint32_t b, uint32_t c)
{
	uint32_t arr[3];
	uint32_t minval = -1;
	arr[0] = a;
	arr[1] = b;
	arr[2] = c;

	for (int i = 0; i < 3; i++)
		if (arr[i] < minval)
			minval = arr[i];

	return minval;
}

static void
create_tlb1_entries(uint64_t vstart, uint64_t pstart,
		uint64_t size, uint32_t *tlb1_entry,
		uint32_t flags, uint32_t maxentry)
{
	uint32_t perm;
	uint32_t iomem;
	uint32_t align;
	uint32_t a1, a2, a3;
	uint32_t entry = *tlb1_entry;
	register_t mas0, mas1, mas2, mas3;

	perm = get_mas3_permflags(flags & RWX_MASK);
	iomem = flags >> WIMGE_SHIFT;

	if (vstart & (PAGE_SIZE - 1)
	    || pstart & (PAGE_SIZE - 1)
	    || size & (PAGE_SIZE - 1))
		return;

	while (size) {
		/* check if entry is within bounds */
		if (entry >= maxentry + 1)
			startup_error();

		a1 = getalignment(vstart);
		a2 = getalignment(pstart);
		a3 = getalignment(size);

		align = min3(a1, a2, a3);
		if (!align)
			startup_error();

		/* create a tlb entry v,p align */
		mas0 = MAS0_TLBSEL1 | MAS0_ESEL(entry);
		mas1 = MAS1_VALID | MAS1_IPROT |
		    ((get_tsize(align) << MAS1_TSIZE_SHIFT) & MAS1_TSIZE_MASK);

		mas2 = vstart | iomem;
		mas3 = pstart | perm;

		mtspr(SPR_MAS1, mas1);
		mtspr(SPR_MAS2, mas2);
		mtspr(SPR_MAS0, mas0);
		mtspr(SPR_MAS3, mas3);

		asm("isync;tlbwe;isync;msync");

		/* v,p increment with align */
		vstart = vstart + align;
		pstart = pstart + align;
		size -= align;
		entry++;
	}

	*tlb1_entry = entry;
}

uint64_t get_sizefromindex(uint32_t sizeidx)
{
	uint64_t sz = PAGE_SIZE;

	for (uint32_t i = 1; i < sizeidx; i++)
		sz = sz * 0x4;

	return sz;
}

uint32_t get_tsize(uint64_t size)
{
	int i = 0;
	uint64_t sz;
	for (i = TLB_TSIZE_4K; i <= TLB_TSIZE_4G; i++) {
		sz = get_sizefromindex(i);
		if (sz >= size)
			return i;
	}
	return 0;
}

uint32_t get_freetlb1_idx(void)
{
	uint32_t ctr;
	uint32_t tlb1count = mfspr(SPR_TLB1CFG) & TLB1CFG_NENTRY;

	/* iterate over cpu0 data structure's, tlb1 */
	for (ctr = 0; ctr < tlb1count; ctr++) {
		if (cpu->tlb1[ctr].mas1 & MAS1_VALID)
			continue;
		else
			return ctr;
	}

	return 0xff;
}

static void prepare_tlb1_snapshot(cpu_t * my_cpu)
{
	uint32_t ctr;
	uint32_t mas0;
	uint32_t _tlbsel = 1;
	uint32_t tlb1count = mfspr(SPR_TLB1CFG) & TLB1CFG_NENTRY;

	LWE_INFO("Inside function %s", __func__);

	/* iterate over cpu0 data structure's, tlb1 */
	for (ctr = 0; ctr < tlb1count; ctr++) {

		mas0 = MAS0_TLBSEL(_tlbsel) | MAS0_ESEL(ctr);
		mtspr(SPR_MAS0, mas0);
		asm volatile ("isync; tlbre; \
				isync; msync":::"memory");

		my_cpu->tlb1[ctr].mas0 = mfspr(SPR_MAS0);
		my_cpu->tlb1[ctr].mas1 = mfspr(SPR_MAS1);
		my_cpu->tlb1[ctr].mas2 = mfspr(SPR_MAS2);
		my_cpu->tlb1[ctr].mas3 = mfspr(SPR_MAS3);
		my_cpu->tlb1[ctr].mas7 = mfspr(SPR_MAS7);
	}
}

static void startup_error(void)
{
	while (1) ;
}
