/* Copyright (c) 2008-2011 Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Freescale Semiconductor nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "qman_private.h"
#include "ldt.h"

/* Global variable containing revision id (even on non-control plane systems
 * where CCSR isn't available) */
u16 qman_ip_rev;

/*****************/
/* Portal driver */
/*****************/

#define PORTAL_MAX	10
#define POOL_MAX	16
#define NUM_PORTAL_ADDR	2

/* NB: we waste an entry because idx==0 isn't valid (pool-channels have hardware
 * indexing from 1..15, as is reflected in the way SDQCR is encoded). However
 * this scheme lets us use cell-index rather than searching. */
struct __pool_channel {
	struct qm_pool_channel cfg;
	phandle ph;
};
static struct __pool_channel pools[POOL_MAX];
static u32 pools_mask;

static struct qm_portal_config configs[PORTAL_MAX];
static u8 num_portals;

static int __qm_pool_add(u32 idx, enum qm_channel channel, phandle ph)
{
	struct __pool_channel *c = &pools[idx];
	BUG_ON(!idx || (idx >= POOL_MAX));
	BUG_ON((channel < qm_channel_pool1) || (channel > qm_channel_pool15));
	if (c->ph)
		return -EBUSY;
	c->cfg.pool = QM_SDQCR_CHANNELS_POOL(idx);
	c->cfg.channel = channel;
	c->cfg.portals = 0;
	c->ph = ph;
	pools_mask |= c->cfg.pool;
	return 0;
}

static int __qm_link(struct qm_portal_config *config, phandle pool_ph)
{
	int idx = 0;
	struct __pool_channel *pool = &pools[0];
	while ((idx < POOL_MAX) && (pool->ph != pool_ph)) {
		idx++;
		pool++;
	}
	if (idx == POOL_MAX)
		return -EINVAL;
	/* Link the pool to the portal */
	pool->cfg.portals |= (1 << config->index);
	/* Link the portal to the pool */
	config->pools |= pool->cfg.pool;
	return 0;
}

static struct qm_portal_config *__qm_portal_add(
					const struct qm_portal_config *config)
{
	struct qm_portal_config *ret;
	BUG_ON((num_portals + 1) > PORTAL_MAX);
	ret = &configs[num_portals];
	*ret = *config;
	ret->portal = NULL;
	ret->index = num_portals++;
	return ret;
}

u8 qm_portal_num(void)
{
	return num_portals;
}

const struct qm_portal_config *qm_portal_config(u8 idx)
{
	if (unlikely(idx >= num_portals))
		return NULL;

	return &configs[idx];
}

u32 qm_pools(void)
{
	return pools_mask;
}

const struct qm_pool_channel *qm_pool_channel(u32 mask)
{
	u32 idx = 0, m = 1;
	struct __pool_channel *c = &pools[0];
	while ((idx < POOL_MAX) && !(mask & m)) {
		idx++;
		m <<= 1;
		c++;
	}
	if (idx == POOL_MAX)
		return NULL;
	return &c->cfg;
}

static int __init fsl_qman_pool_channel_init(int node)
{
	const phandle *ph;
	int ret, channel, index;

	if (get_fdt_int(node, "cell-index", &index) || !index || (index > 15))
		return -1;
	if (get_fdt_int(node, "fsl,qman-channel-id", &channel))
		return -1;
	if (channel != (index + qm_channel_pool1 - 1)) {
		pr_err("Mismatched 'cell-index' and 'fsl,qman-channel-id'\n");
		return -1;
	}
	ph = fdt_getprop(devtree_ptr, node, "linux,phandle", &ret);
	if (!ph || (ret != sizeof(phandle)))
		return -1;
	ret = __qm_pool_add(index, channel, *ph);
	if (!ret)
		pr_info("Qman pool channel at %d:%d\n", index, channel);
	return ret;
}

#ifdef CONFIG_FSL_QMAN_NULL_FQ_DEMUX
/* Handlers for NULL portal callbacks (ie. where the contextB field, normally
 * pointing to the corresponding FQ object, is NULL). */
static enum qman_cb_dqrr_result null_cb_dqrr(struct qman_portal *qm,
					__UNUSED struct qman_fq *fq,
					__UNUSED const struct qm_dqrr_entry *dqrr)
{
	pr_warning("Ignoring unowned DQRR frame on portal %p.\n", qm);
	return qman_cb_dqrr_consume;
}
static void null_cb_mr(struct qman_portal *qm, __UNUSED struct qman_fq *fq,
			const struct qm_mr_entry *msg)
{
	pr_warning("Ignoring unowned MR msg on portal %p, verb 0x%02x.\n",
			qm, msg->verb);
}
static const struct qman_fq_cb null_cb = {
	.dqrr = null_cb_dqrr,
	.ern = null_cb_mr,
	.dc_ern = null_cb_mr,
	.fqs = null_cb_mr
};
#endif

static int __init fsl_qman_portal_init(int node)
{
	struct qm_portal_config cfg, *pconfig;
	const phandle *ph;
	phys_addr_t portal_addr[NUM_PORTAL_ADDR];
	uint64_t portal_size[NUM_PORTAL_ADDR];
	int irq, ret, index, channel, numpools;
	u16 ip_rev = 0;

	if (fdt_node_check_compatible(devtree_ptr, node,
				"fsl,qman-portal-1.0") == 0)
		ip_rev = QMAN_REV1;
	else if (fdt_node_check_compatible(devtree_ptr, node,
				"fsl,qman-portal-1.1") == 0)
		ip_rev = QMAN_REV2;
	else if	(fdt_node_check_compatible(devtree_ptr, node,
				"fsl,qman-portal-1.2") == 0) {
		ip_rev = QMAN_REV2;
		pr_info("Treating Qman portal ver:0x0102 as 0x%04x\n", ip_rev);
	} else if (fdt_node_check_compatible(devtree_ptr, node,
				"fsl,qman-portal-2.0") == 0) {
		ip_rev = QMAN_REV2;
		pr_info("Treating Qman portal ver:0x0200 as 0x%04x\n", ip_rev);
	}
	if (!qman_ip_rev) {
		if (ip_rev)
			qman_ip_rev = ip_rev;
		else {
			pr_warning("unknown Qman version, default to rev2\n");
			qman_ip_rev = QMAN_REV2;
		}
	} else if (ip_rev && (qman_ip_rev != ip_rev))
		pr_warning("Revision=0x%04x, but portal has 0x%04x\n",
			qman_ip_rev, ip_rev);
	if (get_fdt_mem_ranges(node, portal_addr, portal_size) < 0) {
		pr_err("qm_init error, reading portal addr and size\n");
		return -1;
	}

	/* set appropriate TLB entries */
	/* Block Guide states that 16K region of QM portal should have 
	 * WIMG = x01x (coherence ON to enable stashing). 
	 * But it turns out, the 16K region should
	 * be marked as 0000 -- the LAW/PAMU needs to be marked as M=1,
	 * but if you mark the MMU entry as M=0, it saves some snoops.
	 * This seems icky, 
	 * but it's been approved by Gary W. and others as acceptable.
	 */
	ret = dev_mmap(portal_addr[0], portal_size[0], portal_size[0], 0x0,
			 MAS3_SR | MAS3_SW, &cfg.addr.addr_ce); 
	if (ret < 0) {
		pr_err("Failed to create tlb entry for qman driver\n");
		return -1;
	}
	ret = dev_mmap(portal_addr[1], portal_size[1], portal_size[1],
			 TLB_MAS2_IO, MAS3_SR | MAS3_SW, &cfg.addr.addr_ci); 
	if (ret < 0) {
		pr_err("Failed to create tlb entry for qman driver\n");
		return -1;
	}
	if (get_fdt_irq(node, &irq, NULL) < 0) {
		pr_err("Failed in getting IRQ number of");
		pr_err("qman portal %p from device tree\n", cfg.addr.addr_ce);
		return -1;
	}
	if (get_fdt_int(node, "cell-index", &index)) {
		pr_err("Failed to get 'cell-index' property\n");
		return -1;
	}
	if (get_fdt_int(node, "fsl,qman-channel-id", &channel)) {
		pr_err("Failed to get 'fsl,qman-channel-id' property\n");
		return -1;
	}
	if (channel != (index + qm_channel_swportal0)) {
		pr_err("Mismatched 'cell-index'and 'fsl,qman-channel-id'\n");
		return -1;
	}
	if (get_fdt_int(node, "fsl,hv-dma-handle", &ret))
		cfg.has_hv_dma = 1;
	else
		cfg.has_hv_dma = 0;
	cfg.irq = irq;
	cfg.channel = channel;
	cfg.pools = 0;
	/* TODO: this is ripped from the old Qman LWE driver - we should instead
	 * be looking at the 'cpu-handle' property in the device-tree node,
	 * making sure everything is kosher, etc. */

	cfg.cpu = mfspr(SPR_PIR);

	ph = fdt_getprop(devtree_ptr, node, "fsl,qman-pool-channels", &ret);
	numpools = ph ? (ret / sizeof(*ph)) : 0;

	pconfig = __qm_portal_add(&cfg);
	if (!pconfig) {
		dev_unmap(cfg.addr.addr_ce);
		dev_unmap(cfg.addr.addr_ci);
		return -ENOMEM;
	}
	pr_info("Qman portal at %p:%p (%d:%d,v%04x)\n", cfg.addr.addr_ce,
		cfg.addr.addr_ci, cfg.cpu, cfg.channel, qman_ip_rev);
	while (numpools--) {
		int tmp = __qm_link(pconfig, *(ph++));
		if (tmp)
			panic("Unrecoverable error linking pool channels");
	}
	/* If the portal is affine to a cpu and that cpu has no default affine
	 * portal, auto-initialise this one for the job. */
	if (cfg.cpu == -1)
		return 0;
	if (!qman_have_affine_portal()) {
		u32 irq_sources = 0;
#ifdef CONFIG_FSL_QMAN_HAVE_IRQ
		irq_sources = QM_PIRQ_EQCI | QM_PIRQ_EQRI | QM_PIRQ_MRI;
		/* TODO: for CGR support, we will set QM_PIRQ_CSCI too!! */
#ifdef CONFIG_FSL_QMAN_PIRQ_FAST
		irq_sources |= QM_PIRQ_DQRI;
#endif
#endif
		/* TODO: cgrs ?? */
		if (qman_create_affine_portal(pconfig,
				QMAN_PORTAL_FLAG_RSTASH |
				QMAN_PORTAL_FLAG_DSTASH,
				NULL,
#ifdef CONFIG_FSL_QMAN_NULL_FQ_DEMUX
			 	&null_cb,
#else
				NULL,
#endif
				irq_sources))
			pr_err("Qman portal auto-initialisation failed\n");
		else
			pr_info("Qman portal %d auto-initialised\n",
				pconfig->cpu);
	}
	return 0;
}

/***************/
/* Driver load */
/***************/

int qman_init(int secondary)
{
	int node;
	if (!secondary) {
		node = fdt_node_offset_by_compatible(devtree_ptr, -1,
						 "fsl,qman-pool-channel");
		for ( ; node != -1; 
			node = fdt_node_offset_by_compatible(
				devtree_ptr, node, "fsl,qman-pool-channel")) {
			int ret;
			ret = fsl_qman_pool_channel_init(node);
			if (ret)
				return ret;
		}
	}
	node = 
	    fdt_node_offset_by_compatible(devtree_ptr, -1, "fsl,qman-portal");
	for ( ; node != -1; node = fdt_node_offset_by_compatible(
				   devtree_ptr, node, "fsl,qman-portal")) {
		int ret;
		const uint32_t *cpu_phandle;
		const uint32_t *cpu_id;
		int cpu_node;

		cpu_phandle = fdt_getprop(devtree_ptr, node, "cpu-handle", NULL);
		if (!cpu_phandle)
			panic("fdt_getprop failed for cpu-handle");

		cpu_node = fdt_node_offset_by_phandle(devtree_ptr, *cpu_phandle);
		if (cpu_node < 0)
			panic("fdt_node_offset_by_phandle for cpu node");

		cpu_id = fdt_getprop(devtree_ptr, cpu_node, "reg", NULL);
		if (!cpu_id)
			panic("fdt_getprop failed for reg");

		if (*cpu_id != get_lwe_id())
			continue;

		ret = fsl_qman_portal_init(node);
		if (ret)
			return ret;
	}
#ifdef CONFIG_FSL_QMAN_FQALLOCATOR
	node = __fqalloc_init();
	if (node)
		return node;
#endif
	pr_info("Qman driver initialised (%d)\n", get_lwe_id());
	return 0;
}
