/** @file
 * Software interrupts support
 */

/*
 * Copyright (c) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <types.h>
#include <fsl_hcalls.h>
#include <errno.h>
#include "ll_struct.h"
#include "listmgr.h"
#include "misc.h"
#include "swi_private.h"
#include "lwe_malloc.h"
#include <log.h>

/*This variable stores the data required by SWI library*/
__PERCPU static struct swi_t g_swi_db;

/*Count to indicate number of SWIs to pending to be processed*/
__PERCPU volatile uint32_t g_pending_swi_count;

/**
\brief          Processes a pending SWI.

\details        Check if there is a SWI in the application's swi queue. 
		If any SWI is present, its callback handler is called if present
		otherwise SWI ID is delivered to caller as a return value.
		SWIs are dequeued as per priority order.

\param[out]      p_p_data - SWI data if callback handler not installed.

\return         SWI ID - If callback handler not installed
		SWI_ID_INV - If there is no SWI left pending to be processed

\note           Should be called periodically preferrably after every frame.
*/
uint32_t
swi_consume(void **p_p_data)
{
	list_t *list;
	struct swi_node_t *node;
	volatile struct swi_node_t *old_top;
	uint32_t i;
	uint32_t id;
	swi_callbk_hndl hndl;

	/* Walk through all the priorities in-order and all items in there.
	 * Return the SWI nodes.
	 */
	for (i = 0; likely(i < SWI_MAX_PRIORITIES); i++) {
		list = g_swi_db.swi[i];

		if (!list_is_empty(list)) {
			disable_extint();
			g_pending_swi_count--;
			node = (void *) rem_head(list);
			enable_extint();

			id = node->id;
			*p_p_data = node->p_data;
			hndl = node->hndl;

			/*Free the SWI node back to stack*/
			do {
				old_top = g_swi_db.swi_stack;
				node->ll_header.mln_succ = (node_t *)old_top;
			} while (unlikely(0 == 
				CAS((unsigned long *)&(g_swi_db.swi_stack),
				(unsigned long)old_top,
				(unsigned long)node)));

			/*If callback handle is present call it */
			if (hndl) {
				hndl(id, *p_p_data);
				continue;
			} else {
				return id;	/*Return the SWI node directly */
			}
		}
	}

	return SWI_ID_INV;	/*This means no pending SWI remains */
}

/**
\brief          Adds an SWI.

\details        Adds a node to the priority wheel list at a particular 
		priority. Allocates a node from the stack.

\param[in]      f_hndl : function to be called at given priority
\param[in] 	priority : priority to be assigned

\return         0, for SUCCESS and negative for failure.
*/
int
swi_add(uint32_t id, void *p_data, swi_callbk_hndl hndl,
	enum swi_priority_t priority)
{
	list_t *list;
	volatile struct swi_node_t *node;
	struct swi_node_t *new_top;

	if (unlikely(priority > SWI_MAX_PRIORITIES))
		return -EINVAL;

	/* Check if no SWI delivery mechanism has been specified */
	if (unlikely((SWI_ID_INV == id) && (NULL == hndl)))
		return -EINVAL;

	/* Can't have 2 SWI delivery mechanisms */
	if (unlikely((SWI_ID_INV != id) && (NULL != hndl)))
		return -EINVAL;

	do {
		node = g_swi_db.swi_stack;
		if (unlikely(NULL == node))
			return -1;

		new_top = (struct swi_node_t *)(node->ll_header.mln_succ);
	} while (unlikely(0 == CAS((unsigned long *)&(g_swi_db.swi_stack),
				(unsigned long)node, (unsigned long)new_top)));

	/* Find the priority list to insert node into */
	list = g_swi_db.swi[priority];
	assert(list != NULL);

	/* Configure the parameters that get added into the node */
	node->id = id;
	node->p_data = p_data;
	node->hndl = hndl;

	/* Node can be added anywhere in the list.
	 * We 'arbritrarily' choose the tail
	 */

	disable_extint();
	add_tail(list, (void *) node);
	g_pending_swi_count++;
	enable_extint();

	return 0;
}

/**
\brief          Initialises stack for SWI nodes.

\details        Initializing the SWI node's stack.  
		Called from swi_init.

\param[in]      f_hndl : function to be called at given priority

\return         Stack pointer for SUCCESS and NULL for failure.
*/
static struct swi_node_t *
init_swi_node_stack(uint32_t num_swi_nodes)
{
	struct swi_node_t *p_swi_node;
	uint32_t num;

	/*Allocate memory for SWI nodes */
	p_swi_node = tlmalloc(num_swi_nodes * sizeof (struct swi_node_t));
	if (unlikely(NULL == p_swi_node))
		return NULL;

	/*Initialise the stack instance to store free SWI nodes */
	g_swi_db.swi_stack = NULL;

	disable_extint();
	/*Now push SWI nodes into the stack */
	for (num = 0; likely(num < num_swi_nodes); num++) {
		p_swi_node->ll_header.mln_succ = (node_t *)g_swi_db.swi_stack;
		g_swi_db.swi_stack = p_swi_node;
		p_swi_node++;
	}
	enable_extint();

	return (struct swi_node_t *)g_swi_db.swi_stack;
}

/**
\brief          Initialises SWI feature.

\details        Allocates memory for SWis, initializes all lists.  
		Assumed to be called once at initialization time.

\return         0 for SUCCESS, negative for failure.
*/
int
swi_init(void)
{
	uint32_t i;
	list_t *list;

	/*Allocate and init linked lists for storing different priority SWIs */
	for (i = 0; likely(i < SWI_MAX_PRIORITIES); i++) {
		list = (list_t *) tlmalloc(sizeof (list_t));
		if (unlikely(!list)) {
			LWE_ERROR("tlmalloc failed for SWI list");
			return -ENOMEM;
		}
		newlist(list);
		g_swi_db.swi[i] = list;
	}

	/*Initialise stack and populate free SWI nodes in it */
	if (unlikely(NULL == init_swi_node_stack(MAX_NODES_SWI))) {
		LWE_ERROR("init_swi_node_stack failed");
		return -ENOMEM;
	}

	g_pending_swi_count = 0;

	return 0;
}
