/** @file
 * Software timers
 */

/*
 * Copyright (c) 2008 - 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <errno.h>
#include <limits.h>
#include <assert.h>
#include <stdlib.h>
#include <libfdt.h>

#include <io.h>
#include "compiler.h"
#include <timebase.h>
#include <ll_struct.h>
#include <listmgr.h>
#include <timer_wheel_private.h>
#include <log.h>

/*This variable stores the data for timer framework*/
static struct timer_framework_t timer_framework;

/**
\brief          Get the present time tick.

\details        Get the number of ticks elapsed since startup.

\return         Present value tick counter.
*/
uint32_t
get_current_tick(void)
{
	struct timer_wheel_t *my_wheel;
	uint32_t my_lwe_id;

	my_lwe_id = get_lwe_id();
	my_wheel = &(timer_framework.wheel[my_lwe_id]);

	return my_wheel->curr_tick;
}

/**
\brief          Get the period of tick in milliseconds

\details        Get the period of tick in milliseconds

\return         Tick size in milliseconds
*/
uint32_t
get_tick_size(void)
{
	return TICK_SIZE_MS;
}

/**
\brief          Get the time elapsed in milliseconds since startup

\details        Get the time elapsed in milliseconds since startup

\return         Time in milliseconds
*/
uint32_t
get_current_time(void)
{
	return (get_tick_size() * get_current_tick());
}

/**
\brief          Return difference between two new tick and old tick

\details        Takes cares of wrap around after max(uin32_t).
		Makes an assumption that wrap around has happened only once.

\param[in]      new_tick - Newer tick
\param[in]      old_tick  - Older tick

\return         Difference between newer tick and older tick.
		If positive, new_tick is greater than old_tick.
		If negative, old_tick is greater than new_tick.
*/
static int32_t
tick_diff(uint32_t new_tick, uint32_t old_tick)
{
	int32_t diff;

	/*Assumed old_tick and new_tick are not equal */
	if (likely(new_tick > old_tick)) {
		diff = new_tick - old_tick;
		if (likely(diff <= MAX_PERIOD_TICK)) {
			return diff;
		} else {
			return (diff - (MAX_PERIOD_TICK << 1));
		}
	} else if (unlikely(old_tick > new_tick)) {
		diff = old_tick - new_tick;
		if (likely(diff <= MAX_PERIOD_TICK)) {
			return (diff - (MAX_PERIOD_TICK << 1));
		} else {
			return diff;
		}
	} else {		/*Equal case */

		return 0;
	}
}

/**
\brief          Schedules a timer object in timer wheel.

\details        Based on period_tick, the timer object is added in right
		bucket of the appropriate wheel level.

\param[in]      my_wheel - Core's local wheel
\param[in]      period_tick - Period for which timer object needs to be started
\param[in]	timer_object - Timer object to schedule

\return         0 - Success
  		-1 - Failure
*/
static int32_t
schedule_timer_object(struct timer_wheel_t *my_wheel,
		      uint32_t period_tick, struct timer_object_t *timer_object)
{
	uint32_t temp;
	uint32_t insert_level;
	uint32_t next_level_shift;
	struct wheel_level_t *wheel_level;
	uint32_t wheel_offset;
	uint32_t insert_bucket;
	list_t *p_bucket_list;

	/* Determine the wheel level in which timer would go */
	/* The logic below has been written to be optimal for timers going in
	 * level-0. For each level, it check whether all the higher levels 
	 * are zero or not
	 */

	insert_level = 0;
	next_level_shift = 0;
	do {
		wheel_level = &(my_wheel->wheel_level[insert_level]);
		next_level_shift =
		    next_level_shift + bits_each_level[insert_level];

		temp = (period_tick & ~(wheel_level->mask)) >> next_level_shift;

		/*Most common case of small timers */
		if (likely(0 == temp)) {
			wheel_offset = (period_tick >> wheel_level->shift) &
			    (wheel_level->num_buckets - 1);
			break;
		}

		insert_level++;
		if (unlikely(insert_level == my_wheel->num_levels)) {
			return -1;
		}

		/*If its not level-0, then timer state must be reset_pending.. later */
		timer_object->state =
		    GET_COMBINED_TIMER_STATE(TIMER_STATE_RESET_PENDING, 0, 0);
	} while (unlikely(1));

	insert_bucket = (wheel_level->curr_bucket) + wheel_offset;
	insert_bucket = insert_bucket & (wheel_level->num_buckets - 1);

	p_bucket_list = (void *) wheel_level->buckets;
	p_bucket_list = &(p_bucket_list[insert_bucket]);

	add_tail(p_bucket_list, (void *) timer_object);

	return 0;
}

/**
\brief          Frees a timer object

\details        Adds a timer object to free list.

\param[in]      timer_object - Timer object to free

\return         void
*/
static void
free_timer_object(struct timer_object_t *timer_object)
{
	uint32_t lwe_id;
	struct timer_wheel_t *p_wheel;
	volatile struct timer_object_t *old_top;

	lwe_id = GET_LWE_ID_FROM_TIMER_ID(timer_object->timer_id);
	p_wheel = &(timer_framework.wheel[lwe_id]);

	do {
		old_top = p_wheel->free_timer_db;
		timer_object->ll_header.mln_succ = (struct node *)old_top;
	} while (unlikely(0 == CAS((unsigned long *)&(p_wheel->free_timer_db), 
			(unsigned long)old_top, (unsigned long)timer_object)));

	return;
}

static struct timer_object_t *
alloc_timer_object(struct timer_wheel_t *my_wheel)
{
	volatile struct timer_object_t *timer_object;
	struct timer_object_t *new_top;

	do {
		timer_object = my_wheel->free_timer_db;
		if (unlikely(NULL == timer_object))
			return NULL;
	
		new_top = (struct timer_object_t *)
				(timer_object->ll_header.mln_succ);
	} while (unlikely(0 == CAS((unsigned long *)&(my_wheel->free_timer_db), 
				(unsigned long)timer_object, 
				(unsigned long)new_top)));

	return (struct timer_object_t *)timer_object;
}

/**
\brief          Timer expiry dispatch function

\details        Dispatches a timer object with the timer expiry event as SWI.

\param[in]      timer_object - Timer object which has expired

\return         void
*/
static void
send_timer_expiry(struct timer_object_t *timer_object)
{
	int ecode;
	struct qm_fd expiry_fd;

	if (likely(timer_object->swi_priority < SWI_MAX_PRIORITIES)) {
		swi_add(SWI_ID_TIMER_EXPIRY, (void *) timer_object->timer_id,
			NULL, timer_object->swi_priority);
		return;
	}

	/*Else the timer expiry needs to be sent to expiry_fq */
	assert(NULL != timer_object->expiry_fq);

	expiry_fd.addr_lo = timer_object->timer_id;
	/*Rest of the fields of FD are irrelevant to be set */

	ecode = qman_enqueue(timer_object->expiry_fq, &expiry_fd, 0);
	if (unlikely(0 != ecode)) {
		LWE_ERROR("qman_enqueue_ex failed in send_timer_expiry");
		free_timer_object(timer_object);
	}

	return;
}

/**
\brief          Returns application specific data stored in timer object

\details        Checks whether the ethernet frame is allowed to be processed.

\param[in]      timer_id - Timer ID from which application data needs to be 
				extracted.

\return         Timer data (NULL is also valid value)
*/
void *
get_app_data_from_timer_id(uint32_t timer_id)
{
	uint32_t lwe_id;
	uint32_t timer_index;
	struct timer_wheel_t *p_wheel;
	struct timer_object_t *timer_object;

	lwe_id = GET_LWE_ID_FROM_TIMER_ID(timer_id);
	p_wheel = &(timer_framework.wheel[lwe_id]);

	/*Get the timer object */
	timer_index = GET_TIMER_INDEX_FROM_TIMER_ID(timer_id);

	timer_object = (void *) p_wheel->timer_objects;
	timer_object = &(timer_object[timer_index]);

	return timer_object->p_data;
}

/**
\brief          Processes a wheel level for expiring, rescheduling timers.

\details        Calculates the bucket index to be processed in wheel_level.
		In that bucket, gets all timer objects one by one and posts,
		reschedules timer objects as required by the timer object
		state.

\param[in]      my_wheel - Core's own local timer wheel
\param[in]      wheel_level  - Wheel level to process

\return         void
*/
void
proc_wheel_level(struct timer_wheel_t *my_wheel,
		 struct wheel_level_t *wheel_level)
{
	struct timer_object_t *timer_object;

	enum timer_state_t new_state;
	enum timer_state_t old_state;
	enum timer_state_t comb_state;

	uint32_t period_tick;
	uint32_t expiry_tick;
	uint32_t new_expiry_tick;

	list_t *p_proc_list;

	uint32_t curr_tick;
	int32_t value;

	curr_tick = my_wheel->curr_tick;

	p_proc_list = (void *) wheel_level->buckets;
	p_proc_list = &(p_proc_list[wheel_level->curr_bucket]);

	while ((timer_object = (void *) rem_head(p_proc_list))) {
	      proc_wheel_read_timer_state:
		expiry_tick = timer_object->expiry_tick;
		period_tick = timer_object->period_tick;

		comb_state = timer_object->state;
		old_state = GET_TIMER_STATE_ID(comb_state);

		if (unlikely(TIMER_STATE_STOPPED == old_state)) {
			timer_object->state =
			    GET_COMBINED_TIMER_STATE(TIMER_STATE_FREE, 0, 0);
			free_timer_object(timer_object);
			continue;
		}

		if (likely(TIMER_STATE_RESET_PENDING == old_state)) {
			value = tick_diff(expiry_tick, curr_tick);
			if (likely(value > 0)) {
				new_state =
				    GET_COMBINED_TIMER_STATE
				    (TIMER_STATE_SCHEDULED, 0, 0);

				if (unlikely
				    (0 ==
				     CAS((void *) &(timer_object->state),
					 comb_state, new_state))) {
					goto proc_wheel_read_timer_state;
				}

				schedule_timer_object(my_wheel, value,
						      timer_object);

				continue;
			}
		}

		/* Else the timer is expired, change to TIMER_STATE_EXPIRED */

		/*Needed for case of periodic timer */
		new_expiry_tick = my_wheel->curr_tick + period_tick;
		if (unlikely(0 == CAS((void *) &(timer_object->expiry_tick),
				      expiry_tick, new_expiry_tick))) {
			/*reset_timer have been executed in between */
			goto proc_wheel_read_timer_state;
		}

		new_state = GET_COMBINED_TIMER_STATE(TIMER_STATE_EXPIRED, 1, 0);
		if (unlikely
		    (0 ==
		     CAS((void *) &(timer_object->state), comb_state,
			 new_state))) {
			/*reset/stop_timer has been executed in between */
			goto proc_wheel_read_timer_state;
		}

		/*Now send the timer expiry */
		send_timer_expiry(timer_object);
	}			/*end while */

}

/**
\brief          Process the core local timer wheel on receving hardware timer 
		tick.

\details        On receipt of hardware timer tick, process level-0 of wheel and 
		further wheel levels if the lower level has wrapped around.

\param[in]      swi_id - Unused parameter. Passed by SWI framework in this 
		callback.
\param[in]      data - Unused parameter. Passed by SWI framework in this

\return         void
*/
static void
proc_timer_wheel(uint32_t swi_id __UNUSED, void *data __UNUSED)
{
	struct timer_wheel_t *my_wheel;
	struct wheel_level_t *wheel_level;

	struct reclaim_queue_t *reclaim_queue;
	struct timer_object_t *timer_object;

	uint32_t my_lwe_id;
	uint32_t level;
	uint32_t count;

	my_lwe_id = get_lwe_id();
	my_wheel = &(timer_framework.wheel[my_lwe_id]);

	(my_wheel->curr_tick)++;

	level = 0;
	do {
		wheel_level = &(my_wheel->wheel_level[level]);

		/*Incr curr_bucket so that timer objects are not added to this list now */
		wheel_level->curr_bucket = (wheel_level->curr_bucket + 1) &
		    (wheel_level->num_buckets - 1);

		proc_wheel_level(my_wheel, wheel_level);

		if (likely(wheel_level->curr_bucket !=
			   (wheel_level->num_buckets - 1))) {
			break;
		}

		level++;
	} while (likely(level < my_wheel->num_levels));

	/*Now process fast reclaim queues */
	for (count = 0; likely(count < MAX_NUM_LWE_INSTANCES); count++) {
		reclaim_queue = (my_wheel->reclaim_queues)[count];

		while ((timer_object =
			reclaim_queue->queue[reclaim_queue->read_index])) {
			/* Freeup the queue slot */
			reclaim_queue->queue[reclaim_queue->read_index] = NULL;

			assert(likely
			       (TIMER_STATE_STOPPED == timer_object->state));

			/*Delete the object from bucket and free up the object here */
			remove((void *) timer_object);

			timer_object->state =
			    GET_COMBINED_TIMER_STATE(TIMER_STATE_FREE, 0, 0);

			free_timer_object(timer_object);

			reclaim_queue->read_index++;
			if (unlikely
			    (reclaim_queue->read_index ==
			     my_wheel->reclaim_queue_len)) {
				reclaim_queue->read_index = 0;
			}
		}
	}
}

/**
\brief          ISR handler for decrementer expiry

\details        This is called for every decrementer interrupt after 
		every tick milliseconds

\return         void
*/
void
proc_decr_expiry(void)
{
	mtspr(SPR_TSR, TSR_DIS);

	/*Sending a SWI to process lowest level wheel */
	/*p_data agrument in swi_add denotes wheel level to be processed */
	swi_add(SWI_ID_INV, NULL, proc_timer_wheel, SWI_PRI_HIGH);

	return;
}

/**
\brief          Starts a hardware ticker (decrementer) for TICK_MS period.

\details        Reads device tree for timebase frequency and then programs
		decrementer to give periodic ticks every TICK_MS milliseconds.

\return         Current time in milliseconds since system startup
*/
static uint32_t
start_hw_ticker(void)
{
	uint64_t tb;

	uint32_t tb_inc_per_ms;
	uint32_t tb_inc_per_tick;
	uint32_t tb_inc_remaining;

	uint32_t curr_ms;
	uint32_t curr_tick;

	register_t reg;

	/* Timebase is synchronized across all cores */

	/* Ticker is started only at boundary of a tick. Since the ticker 
	   period is same across cores, all cores will get ticker interrupt 
	   at the same time
	 */
	LWE_INFO("TB frequency = %u", timebase_freq);

	tb_inc_per_ms = timebase_freq / 1000;	/* Fractional part..? later */

	tb_inc_per_tick = (tb_inc_per_ms * TICK_SIZE_MS);

	reg = disable_critint_save();

	/*Calculate how many ticks are remaining till next tick boundary */
	/*Disable preemption later */
	tb = get_timebase();
	curr_ms = tb / tb_inc_per_ms;
	curr_tick = tb / tb_inc_per_tick;

	tb_inc_remaining = ((curr_tick + 1) * tb_inc_per_tick) - tb;

	/* Disable the decrementer if already enabled. Do we need this. later */
	mtspr(SPR_TCR, mfspr(SPR_TCR) & (~TCR_ARE) & (~TCR_DIE));

	/*Initial load value equal to next tick boundary */
	mtspr(SPR_DEC, tb_inc_remaining);

	/*Reload value equal to tick size */
	mtspr(SPR_DECAR, tb_inc_per_tick);

	mtspr(SPR_TCR, mfspr(SPR_TCR) | TCR_ARE | TCR_DIE);

	restore_critint(reg);

	return curr_ms;
}

/**
\brief          Initialises a timer object with default values.

\details        Initialises a timer object with default values.

\param[in]      timer_object - Timer object to be initialised
\param[in]      lwe_id  - LWE instance to which this object belongs to.
\param[in]      index  - Location of object in array of timer objects

\return         void
*/
static void
timer_object_init(struct timer_object_t *timer_object, uint32_t lwe_id,
		  uint32_t index)
{
	timer_object->timer_index = index;
	timer_object->lwe_id = lwe_id;
	timer_object->state = TIMER_STATE_FREE;
}

/**
\brief          Allocates memory for wheel levels.

\details        Allocates memory for wheel levels.

\param[in]      my_wheel - Core local timer wheel

\return         0 - SUCCESS
  		-1 - Failure
*/
static int
alloc_wheel_levels(struct timer_wheel_t *my_wheel)
{
	uint32_t count;
	uint32_t level;
	struct wheel_level_t *wheel_level;
	list_t *p_bucket_list;

	/*Allocate memory for each wheel level */
	for (level = 0; level < my_wheel->num_levels; level++) {
		wheel_level = &(my_wheel->wheel_level[level]);

		/* Allocate the memory for buckets */
		wheel_level->buckets =
		    malloc((sizeof (list_t) * wheel_level->num_buckets));

		if (unlikely(NULL == wheel_level->buckets)) {
			LWE_ERROR("malloc fail for level: %u", level);
			return -1;
		}

		/*Init bucket linked lists in all wheel levels */
		p_bucket_list = (void *) wheel_level->buckets;
		for (count = 0; count < wheel_level->num_buckets; count++) {
			newlist(&(p_bucket_list[count]));
		}
	}

	return 0;
}

/**
\brief          Initialises required number of wheel levels

\details        Determines how many wheel levels are required for supporting
		timer of duration *max_timeout_tick.

\param[in]      my_wheel - Core's local timer wheel
\param[in]      *max_timeout_tick  - Maximum timeout tick to be supported

\return         Non-zero - Number of levels initialised.
  		0 - Failure
*/
static uint32_t
init_wheel_levels(struct timer_wheel_t *my_wheel, uint32_t * max_timeout_tick)
{
	uint32_t count;
	uint32_t acc_ticks;
	uint32_t total_ticks;
	uint32_t highest_level;
	uint32_t num_levels;	/*Number of wheel-levels needed */
	uint32_t half_buckets;
	uint32_t pruned_ticks;
	uint32_t num_buckets;
	uint32_t lower_level_bits;

	struct wheel_level_t *wheel_level;

	/*Determine how many levels of wheel are needed */
	acc_ticks = 1;
	num_levels = 0;
	do {
		num_buckets = 1 << bits_each_level[num_levels];
		total_ticks = acc_ticks * num_buckets;
		num_levels++;

		if (likely(total_ticks >= *max_timeout_tick)) {
			break;
		}

		acc_ticks = total_ticks;
	} while (likely(num_levels < MAX_WHEEL_LEVEL));

	/*Timer framework cannot serve this much big timeout */
	if (unlikely(total_ticks < *max_timeout_tick)) {
		LWE_ERROR("Too big timeout value requested");
		return 0;
	}

	LWE_INFO("Number of wheel levels required: %u", num_levels);

	/*Now prune the extra buckets at highest level.. save memory */
	highest_level = num_levels - 1;
	while (1) {
		num_buckets = 1 << bits_each_level[highest_level];
		half_buckets = num_buckets >> 1;

		pruned_ticks = half_buckets * acc_ticks;
		total_ticks = total_ticks - pruned_ticks;

		if (total_ticks < *max_timeout_tick) {
			*max_timeout_tick = total_ticks + pruned_ticks;
			break;
		}

		bits_each_level[highest_level]--;

		/*Break if total ticks now is equal to what we need to support*/
		if (total_ticks == *max_timeout_tick)
			break;
	}

	count = 0;
	lower_level_bits = 0;
	do {
		wheel_level = &(my_wheel->wheel_level[count]);

		num_buckets = (1 << bits_each_level[count]);

		wheel_level->curr_bucket = 0;
		wheel_level->num_buckets = num_buckets;

		wheel_level->shift = lower_level_bits;
		wheel_level->mask = (num_buckets - 1) << lower_level_bits;

		LWE_INFO("Level-%u, Buckets: %u, Shift: %u, Mask: 0x%08x",
			 count, num_buckets, wheel_level->shift,
			 wheel_level->mask);

		lower_level_bits = lower_level_bits + bits_each_level[count];
		count++;
	} while (likely(count < num_levels));

	LWE_INFO("Max timeout ticks adjusted to: %u", *max_timeout_tick);

	my_wheel->num_levels = num_levels;

	if (unlikely(0 > alloc_wheel_levels(my_wheel))) {
		LWE_ERROR("alloc_wheel_levels failed");
		return 0;
	}

	return num_levels;
}

/**
\brief          Initialises the timer framework

\details        Called by application once at the time of system startup.

\param[in]      max_timeout_ms - Maximum timeout requried in milliseconds
\param[in]	num_timer_objects - Maximum  number of timers to be started
\param[in]	reclaim_queue_len - Length of timer object reclaimation queue

\return         0 - Success
  		-1 - Failure

\note		This routine needs to be called from all the cores in LWE
		partition which need to use timer wheel services.
*/
int
timer_framework_init(uint32_t max_timeout_ms,
		     uint32_t num_timer_objects, uint32_t reclaim_queue_len)
{
	uint32_t my_lwe_id;
	uint32_t max_timeout_tick;
	uint32_t num_levels;
	uint32_t count;
	uint32_t size;

	struct timer_object_t *timer_object;
	struct wheel_level_t *wheel_level;
	struct timer_wheel_t *my_wheel;

	my_lwe_id = get_lwe_id();
	my_wheel = &(timer_framework.wheel[my_lwe_id]);

	if (unlikely(num_timer_objects > MAX_NUM_TIMER_OBJECTS)) {
		return -1;
	}

	max_timeout_tick = GET_TICK_FROM_MS(max_timeout_ms);
	if (unlikely((max_timeout_tick > (uint32_t) MAX_PERIOD_TICK) ||
		     (0UL == max_timeout_tick))) {
		return -1;
	}

	/*Determine how many levels of wheel we need */
	num_levels = init_wheel_levels(my_wheel, &max_timeout_tick);
	if (unlikely(0 == num_levels)) {
		return -1;
	}

	my_wheel->max_timeout_tick = max_timeout_tick;

	/* Allocate memory for timer objects */
	/* Align it to multiple of 64 bytes .. later */
	my_wheel->timer_objects =
	    malloc((sizeof (struct timer_object_t) * num_timer_objects));

	if (unlikely(NULL == my_wheel->timer_objects)) {
		LWE_ERROR("malloc failed for timer objects");
		return -1;
	}

	/* Initialise free timer pool */
	my_wheel->free_timer_db = NULL;

	/*Initialise and release timer objects in timer pool */
	timer_object = (struct timer_object_t *)(my_wheel->timer_objects);
	for (count = 0; count < num_timer_objects; count++) {
		timer_object_init(timer_object, my_lwe_id, count);

		timer_object->ll_header.mln_succ = 
			(struct node *)(my_wheel->free_timer_db);
		my_wheel->free_timer_db = timer_object;

		timer_object++;
	}

	/*Allocate memory for reclaimation queue */
	size = (sizeof (struct reclaim_queue_t) +
	     (sizeof (void *) * reclaim_queue_len));
	my_wheel->reclaim_queue_len = reclaim_queue_len;

	for (count = 0; count < MAX_NUM_LWE_INSTANCES; count++) {
		(my_wheel->reclaim_queues)[count] = malloc(size);
		if (NULL == (my_wheel->reclaim_queues)[count]) {
			LWE_ERROR("malloc failed for reclaim queue");
			free(my_wheel->timer_objects);
			return -1;
		}

		/*Freeing all slots in reclaim queue */
		memset((my_wheel->reclaim_queues)[count], NULL, size);

		/* As such not needed, just being explicit */
		((my_wheel->reclaim_queues)[count])->read_index = 0;
		((my_wheel->reclaim_queues)[count])->write_index = 0;
	}

	/* At startup, the curr_bucket in each wheel level is at end */
	for (count = 0; count < num_levels; count++) {
		wheel_level = &(my_wheel->wheel_level[count]);
		wheel_level->curr_bucket = wheel_level->num_buckets - 1;
	}

	return 0;
}

/**
\brief		Timer framework start function

\details	Starts the tick processing in timer framework on local core

\return		On success 0, on failure negative ecode.
*/
int timer_framework_start(void)
{
	struct timer_wheel_t *my_wheel;
	uint32_t curr_time_ms;

	my_wheel = &(timer_framework.wheel[lwe_instance]);

	/* Start the hardware timer and get current time */
	curr_time_ms = start_hw_ticker();
	my_wheel->curr_tick = (curr_time_ms / TICK_SIZE_MS);
	return 0;
}

/**
\brief		Timer framework stop function

\details	Stops the tick processing in timer framework on local core

\return		On success 0, on failure negative ecode.
*/
int timer_framework_stop(void)
{
	register_t reg;

	reg = disable_critint_save();

	mtspr(SPR_TCR, mfspr(SPR_TCR) & (~TCR_ARE) & (~TCR_DIE));
	mtspr(SPR_DEC, 0);
	mtspr(SPR_TSR, TSR_DIS);

	restore_critint(reg);

	return 0;
}


/**
\brief          Timer start function

\details        Starts a software timer

\param[in]      period_ms - Period for which timer to be started
\param[in]      is_periodic - Whether timer is periodic or not
\param[in]      expiry_fq - FQ ID on which to receive timer expiry 
\param[in]      swi_priority - Priority of SWI for timer expiry
\param[in]      callbk - Timer expiry callback function
\param[in]      p_data - Application data stored with timer context 

\return         On success, timer ID, on failure INV_TIMER_ID.
*/
uint32_t
start_timer(uint32_t period_ms,
	    bool is_periodic,
	    struct qman_fq * expiry_fq,
	    enum swi_priority_t swi_priority,
	    timer_callbk_t callbk, void *p_data)
{
	int32_t ret;
	uint32_t my_lwe_id;
	uint32_t period_tick;

	struct timer_object_t *timer_object;
	struct timer_wheel_t *my_wheel;

	/*Too small timer period given*/
	assert(period_ms >= TICK_SIZE_MS);

	/*Timer expiry event delivery mechanism not specified */
	assert((NULL != callbk) || (NULL != expiry_fq));

	/*Timer expiry event can be delivered in only one way */
	assert((NULL == callbk) || (NULL == expiry_fq));

	if (NULL != callbk)
		assert(swi_priority < SWI_MAX_PRIORITIES);

	my_lwe_id = get_lwe_id();
	my_wheel = &(timer_framework.wheel[my_lwe_id]);

	timer_object = alloc_timer_object(my_wheel);
	if (unlikely(NULL == timer_object))
		return INV_TIMER_ID;

	period_tick = GET_TICK_FROM_MS(period_ms);

	timer_object->is_periodic = is_periodic;
	timer_object->wheel_id = my_lwe_id;

	if (expiry_fq) {
		timer_object->expiry_fq = expiry_fq;
		timer_object->swi_priority = SWI_MAX_PRIORITIES;
	}
	else {
		timer_object->callbk = callbk;
		timer_object->swi_priority = swi_priority;
	}

	timer_object->p_data = p_data;
	timer_object->period_tick = period_tick;
	timer_object->expiry_tick = my_wheel->curr_tick + period_tick;
	timer_object->state =
	    GET_COMBINED_TIMER_STATE(TIMER_STATE_SCHEDULED, 0, 0);

	ret = schedule_timer_object(my_wheel, period_tick, timer_object);
	if (unlikely(ret < 0)) {
		/*Attempt to start timer of very big duration */
		LWE_ERROR("schedule_timer_object failed");
		free_timer_object(timer_object);
		return INV_TIMER_ID;
	}

	/*This forces all stores to complete in memory before timer ID is returned */
	smp_mbar();

	return timer_object->timer_id;
}

static void
add_to_reclaim_queue(struct timer_object_t *timer_object)
{
	struct reclaim_queue_t *p_reclaim_queue;
	struct timer_wheel_t *p_wheel;
	uint32_t write_index;
	uint32_t lwe_id;

	/*Get the LWE ID in which timer object has been scheduled */
	lwe_id = timer_object->wheel_id;

	p_wheel = &(timer_framework.wheel[lwe_id]);

	/* Get hold of relcaim queue and add the timer object to it */
	p_reclaim_queue = (p_wheel->reclaim_queues)[lwe_id];

	write_index = p_reclaim_queue->write_index;
	if (unlikely(NULL != p_reclaim_queue->queue[write_index])) {
		LWE_WARN("No free slot available in fast reclaim queue");
		return;
	}

	p_reclaim_queue->queue[write_index] = timer_object;

	write_index++;
	if (unlikely(write_index == p_wheel->reclaim_queue_len))
		write_index = 0;

	p_reclaim_queue->write_index = write_index;
	
	return;
}

/**
\brief          Stops a timer

\details        Stops a software timer.

\param[in]      timer_id - Timer ID to be stopped

\return         0 - Success
  		-1 - Failure
*/
int
stop_timer(uint32_t timer_id)
{
	uint32_t lwe_id;
	uint32_t timer_index;
	uint32_t inflight;
	struct timer_wheel_t *p_wheel;
	struct timer_object_t *timer_object;
	enum timer_state_t comb_state;
	enum timer_state_t state_id;

	lwe_id = GET_LWE_ID_FROM_TIMER_ID(timer_id);
	p_wheel = &(timer_framework.wheel[lwe_id]);

	/*Get the timer object */
	timer_index = GET_TIMER_INDEX_FROM_TIMER_ID(timer_id);

	timer_object = (void *) p_wheel->timer_objects;
	timer_object = &(timer_object[timer_index]);

	comb_state = timer_object->state;
	state_id = GET_TIMER_STATE_ID(comb_state);

	if (unlikely((TIMER_STATE_FREE == state_id) ||
		     (TIMER_STATE_STOPPED == state_id))) {
		LWE_ERROR("stop_timer for a timer in wrong state");
		return -1;
	}

	inflight = GET_TIMER_INFLIGHT_STATUS(comb_state);

	timer_object->state =
	    GET_COMBINED_TIMER_STATE(TIMER_STATE_STOPPED, inflight, 0);

	/* If the timer being stopped is quite far away, 
	 * add it to fast reclaim queue if its not inflight. 
	 */

	if (unlikely(1 == inflight))
		return 0;

	if ((p_wheel->curr_tick - timer_object->expiry_tick) < 4)
		return 0;

	add_to_reclaim_queue(timer_object);

	return 0;
}

/**
\brief          Clone a timer object

\details        Copy a timer object from old to new.

\param[in]      p_new - Timer object to clone into.
\param[in]      p_old - Timer object to clone from.

\return         void
*/
static void
clone_timer_object(struct timer_object_t *p_new, struct timer_object_t *p_old)
{
	p_new->expiry_fq = p_old->expiry_fq;
	p_new->is_periodic = p_old->is_periodic;
	p_new->swi_priority = p_old->swi_priority;
	p_new->callbk = p_old->callbk;
	p_new->p_data = p_old->p_data;
}

/**
\brief          Reset a timer

\details        Resets a software timer.

\param[in]      timer_id - Timer ID to reset.
\param[in]      new_period_ms  - New period for which timer needs to be restarted

\return         Old or new timer ID on success.
  		INV_TIMER_ID on failure
*/
uint32_t
reset_timer(uint32_t timer_id, uint32_t new_period_ms)
{
	int32_t ret;
	uint32_t lwe_id;
	uint32_t my_lwe_id;
	uint32_t curr_tick;
	uint32_t timer_index;

	struct timer_wheel_t *p_wheel;
	struct timer_wheel_t *my_wheel;

	struct timer_object_t *timer_object;
	struct timer_object_t *p_new_timer_object;

	enum timer_state_t comb_state;
	enum timer_state_t old_state;
	enum timer_state_t new_state;

	uint32_t new_period_tick;
	uint32_t new_expiry_tick;
	uint32_t old_expiry_tick;

	uint32_t state_num;
	uint32_t inflight;
	int32_t value;

	lwe_id = GET_LWE_ID_FROM_TIMER_ID(timer_id);
	p_wheel = &(timer_framework.wheel[lwe_id]);

	my_lwe_id = get_lwe_id();
	my_wheel = &(timer_framework.wheel[my_lwe_id]);

	/*Get the timer object */
	timer_index = GET_TIMER_INDEX_FROM_TIMER_ID(timer_id);

	timer_object = (void *) p_wheel->timer_objects;
	timer_object = &(timer_object[timer_index]);

	comb_state = timer_object->state;
	old_state = GET_TIMER_STATE_ID(comb_state);

	if (unlikely((TIMER_STATE_FREE == old_state) ||
		     (TIMER_STATE_STOPPED == old_state))) {
		return INV_TIMER_ID;
	}

	if (likely(0 == new_period_ms)) {	/* Use the older value for timeout period */
		new_period_tick = timer_object->period_tick;
	} else if (unlikely(new_period_ms < TICK_SIZE_MS)) {	/*Less than tick size */
		return INV_TIMER_ID;
	} else {
		/*Round of new_period_ms to multiple of TICK_SIZE_MS */
		new_period_tick = GET_TICK_FROM_MS(new_period_ms);
		if (unlikely(new_period_tick > my_wheel->max_timeout_tick)) {
			return INV_TIMER_ID;
		}
	}

	curr_tick = my_wheel->curr_tick;

	old_expiry_tick = timer_object->expiry_tick;
	new_expiry_tick = curr_tick + new_period_tick;	/*check later */

	value = tick_diff(new_expiry_tick, curr_tick) -
	    tick_diff(old_expiry_tick, curr_tick);
	if (likely(value >= 0)) {
		timer_object->period_tick = new_period_tick;
		timer_object->expiry_tick =
		    p_wheel->curr_tick + new_period_tick;

		state_num = GET_TIMER_STATE_NUM(old_state);
		state_num++;
		inflight = GET_TIMER_INFLIGHT_STATUS(old_state);
		new_state = TIMER_STATE_RESET_PENDING;
		comb_state =
		    GET_COMBINED_TIMER_STATE(new_state, inflight, state_num);

		smp_mbar();	/*This is required to force that 
				   period_tick and expiry_tick updates are seen
				   by all cores in system before state change */
		timer_object->state = comb_state;

		smp_mbar();

		return timer_id;
	}

	/*Else new_expiry_time_ms is less than old_expiry_time_ms */
	/*Start a new timer and stop old timer */

	p_new_timer_object = alloc_timer_object(my_wheel);
	if (unlikely(NULL == p_new_timer_object))
		return INV_TIMER_ID;

	clone_timer_object(p_new_timer_object, timer_object);

	p_new_timer_object->wheel_id = my_lwe_id;
	p_new_timer_object->period_tick = new_period_tick;
	p_new_timer_object->expiry_tick = new_expiry_tick;
	p_new_timer_object->state =
	    GET_COMBINED_TIMER_STATE(TIMER_STATE_SCHEDULED, 0, 0);

	ret = schedule_timer_object(my_wheel, new_period_tick,
				    p_new_timer_object);
	if (unlikely(ret < 0)) {
		/*Attempt to start timer of very big duration */
		LWE_ERROR("schedule_timer_object failed");
		free_timer_object(p_new_timer_object);
		return INV_TIMER_ID;
	}

	if (unlikely(0 != stop_timer(timer_id))) 
		LWE_ERROR("Error stopping timer ID: %u in reset_timer",
			  timer_id);

	return (p_new_timer_object->timer_id);
}

/**
\brief          Timer expiry process function

\details        To be called by application on receiving timer expiry event.

\param[in]      timer_id - Timer ID which has expired,

\return         void
*/
void
proc_timer_expiry(uint32_t timer_id)
{
	int32_t value;
	uint32_t lwe_id;
	uint32_t my_lwe_id;
	uint32_t timer_index;
	struct timer_wheel_t *p_wheel;
	struct timer_wheel_t *my_wheel;
	struct timer_object_t *timer_object;
	enum timer_state_t state_id;

	my_lwe_id = get_lwe_id();
	my_wheel = &(timer_framework.wheel[my_lwe_id]);

	/*Get the timer object */
	lwe_id = GET_LWE_ID_FROM_TIMER_ID(timer_id);
	timer_index = GET_TIMER_INDEX_FROM_TIMER_ID(timer_id);

	p_wheel = &(timer_framework.wheel[lwe_id]);
	timer_object = (void *) p_wheel->timer_objects;
	timer_object = &(timer_object[timer_index]);

	state_id = GET_TIMER_STATE_ID(timer_object->state);

      proc_expiry_check_timer_state:
	assert(likely(TIMER_STATE_FREE != state_id));

	if (unlikely(TIMER_STATE_STOPPED == state_id)) {
		goto proc_expiry_free_timer;
	}

	if (unlikely(TIMER_STATE_RESET_PENDING == state_id)) {
		goto proc_expiry_restart_timer;
	}

      proc_expiry_call_callbk:
	/* Now call the application callback function */
	timer_object->callbk(timer_id, timer_object->p_data);

	/* Since app may reset or stop the same timer, this may change state */
	state_id = GET_TIMER_STATE_ID(timer_object->state);
	if (unlikely(TIMER_STATE_EXPIRED != state_id)) {
		goto proc_expiry_check_timer_state;
	}

	/*If timer is not periodic, free the timer object */
	if (false == timer_object->is_periodic) {
	      proc_expiry_free_timer:
		timer_object->state =
		    GET_COMBINED_TIMER_STATE(TIMER_STATE_FREE, 0, 0);
		free_timer_object(timer_object);
		return;
	}

      proc_expiry_restart_timer:
	value = tick_diff(timer_object->expiry_tick, my_wheel->curr_tick);
	if (likely(value > 0)) {
		timer_object->wheel_id = my_lwe_id;
		timer_object->state =
		    GET_COMBINED_TIMER_STATE(TIMER_STATE_SCHEDULED, 0, 0);

		schedule_timer_object(my_wheel, value, timer_object);

		return;
	}

	timer_object->expiry_tick = timer_object->expiry_tick +
	    timer_object->period_tick;

	goto proc_expiry_call_callbk;
}

int shell_cmd_sts(int argc, char *argv[])
{
	uint32_t timer_id;
	struct timer_wheel_t *p_wheel;
	struct timer_wheel_t *p_sched_wheel;
	struct timer_object_t *timer_object;

	uint32_t lwe_id;
	uint32_t sched_wheel_id;
	uint32_t timer_index;

	if (argc < 2) {
		printf("usage: sts <timer_id>\n");
		return -1;
	}

	errno = 0;
	timer_id = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == timer_id)
			|| (0 != errno && 0 == timer_id)) {
		printf("strtoul failed converting number of bytes.\n");
		return -1;
	}

	lwe_id = GET_LWE_ID_FROM_TIMER_ID(timer_id);
	p_wheel = &(timer_framework.wheel[lwe_id]);

	/*Get the timer object */
	timer_index = GET_TIMER_INDEX_FROM_TIMER_ID(timer_id);

	timer_object = (void *) p_wheel->timer_objects;
	timer_object = &(timer_object[timer_index]);

	sched_wheel_id = timer_object->wheel_id;
	p_sched_wheel = &(timer_framework.wheel[sched_wheel_id]);

	printf("\tTimer ID: %#x\n", timer_id);
	printf("\tOwner LWE: %u\n", lwe_id);
	printf("\tScheduling wheel: %u\n", sched_wheel_id);
	printf("\tCurrent tick: %u\n", p_sched_wheel->curr_tick);
	printf("\tExpiry tick: %u\n", timer_object->expiry_tick);

	printf("\tPeriodic: %d\n", timer_object->is_periodic);
	if (timer_object->is_periodic)
		printf("\tPeriodicity: %u ticks\n", timer_object->period_tick);

	printf("\tTimer state: ");
	switch (timer_object->state) {
		case TIMER_STATE_FREE: {
			printf("FREE\n");
			break;
		}
		case TIMER_STATE_STOPPED: {
			printf("STOPPED\n");
			break;
		}

		case TIMER_STATE_SCHEDULED: {
			printf("SCHEDULED\n");
			break;
		}

		case TIMER_STATE_RESET_PENDING: {
			printf("RESET_PENDING\n");
			break;
		}

		case TIMER_STATE_EXPIRED: {
			printf("EXPIRED\n");
			break;
		}
	}

	return 0;
}
