/** @file
 * This file provides requisite APIs to support watchdog
 */

/*
 * Copyright (c) 2009-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include "io.h"
#include "ldt.h"
#include <stdint.h>
#include <types.h>
#include <fsl_hcalls.h>
#include "init.h"
#include "wdog.h"
#include "bitops.h"
#include "trapframe.h"
#include "wdog_private.h"
#include <compiler.h>

__PERCPU wdoghndlr_t reg_wdog_hndlr = NULL;

/**
 * Calculating integer log to the base 2 of a  given 64 bit value
 * @param[in] val -- 64 bit value whose log has to be calculated
 * @return -- log to the base 2 of val. (Result is rounded of to it's ceiling)
 */
static inline int
int_log2(uint64_t val)
{
	val = val - 1;

	if ((val >> 32) & 0xffffffff)
		return (64 - count_msb_zeroes(val >> 32));
	else
		return (32 - count_msb_zeroes(val));
}

/**
 * To find timebase bit which would toggle from 0 to 1 for a given timeout
 * @param[in] -- timeout in milliseconds
 * @return -- bit of tbu and tbl whose toggle will result in specified timeout
 */
static inline int
get_timebase_bit(uint32_t timeout_ms)
{
	uint32_t tb_freq = get_tbfreq();
	uint64_t tb_ticks = (uint64_t) tb_freq * (timeout_ms/1000);

	return 64 - int_log2(tb_ticks);
}

/** 
 * Start watch dog timer with the given timeout values in milliseconds
 * @param[in] handler --  Handler called when the watchdog expires.
 * @param[in] wrc - Watchdog reset control
 * @param[in] timeout -- expiry value in milliseconds
 * @return -- Success: 0; Fail: -EINVAL
 */
int
wdog_init(wdoghndlr_t hndlr, uint32_t wrc, uint32_t timeout_ms)
{
	int tb_bit = get_timebase_bit(timeout_ms);
	register_t reg;

	if (unlikely(NULL == hndlr))
		return -EINVAL;

	reg = disable_critint_save();

	reg_wdog_hndlr = hndlr;

	mtspr(SPR_TSR, TSR_ENW | TSR_WIS);

	if (wrc)
		write_tcr(TCR_WRC_MASK, wrc);

	write_tcr(TCR_WP_MASK, TCR_WP_SET(tb_bit));
	write_tcr(TCR_WIE, TCR_WIE);

	restore_critint(reg);

	return 0;
}

/** 
 *  Stop watchdog timer
 */
void
wdog_stop(void)
{
	write_tcr(TCR_WP_MASK | TCR_WIE, TCR_WP_SET(0));
}

/** 
 * wdog_handler - watchdog exception handler
 *  @param[in] regs -- Pointer to the trapframe structure
 */
void
wdog_handler(struct trapframe *regs __UNUSED)
{
	in_irq++;

	if (reg_wdog_hndlr == NULL)
		while (1)
			;

	reg_wdog_hndlr();

	in_irq--;
}
