/* Copyright (c) 2009-2010 Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Freescale Semiconductor nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "init.h"
#include "misc.h"
#include "compat.h"

#define NUM_POINTERS	20
#define NUM_TESTS	200000 /* undef to loop forever */
#define SIZE_LOW	8
#define SIZE_HIGH	1024
#define NUM_CPUS	7

#undef CONFIG_VERBOSE

static inline void do_lfsr(u32 *mix)
{
	*mix = (*mix >> 1) ^ (-(*mix & 1u) & 0xd0000001u);
}

/* Returns negative for free, positive for allocate. Should be deterministic
 * when num_ptrs is zero or NUM_POINTERS, otherwise it should be biased towards
 * maintaining num_ptrs "near" NUM_POINTERS/2. */
static int choose_alloc_or_free(u32 *mix, unsigned int num_ptrs)
{
	unsigned int x;
	/* So;  0 <= num_ptrs < NUM_POINTERS
	 * Pick a number x in [0..NUM_POINTERS-1], and then;
	 *   if x < num_ptrs, we free
	 *   if x >= num_ptrs, we alloc
	 */
	do_lfsr(mix);
	x = *mix % NUM_POINTERS;
	if (x < num_ptrs)
		return (int)x - num_ptrs;
	return ((int)x + 1 - num_ptrs);
}

#define MAX_ALIGN_POWER		8	/* 2 ^ 8 == 256 */
static void choose_size(u32 *mix, u32 *sz, u32 *align)
{
	do_lfsr(mix);
	*sz = SIZE_LOW + (*mix % (SIZE_HIGH - SIZE_LOW + 1));
	do_lfsr(mix);
	*align = 1 << ((*mix % MAX_ALIGN_POWER) + 1);
}

static int choose_idx(u32 *mix, unsigned int num_ptrs)
{
	do_lfsr(mix);
	return (*mix % num_ptrs);
}

int main(__always_unused int argc, __always_unused char *argv[])
{
	void *ptrs[NUM_POINTERS];
	unsigned int num_ptrs = 0, loop = 0;
	int is_alloc = 0;
	u32 mix = 0xdeadbeef;

	if (get_lwe_id() >= NUM_CPUS)
		return 0;
	loop = 50 + get_lwe_id();
	while (--loop)
		do_lfsr(&mix);
	pr_info("\"alloc\" starting\n");
#ifdef NUM_TESTS
	while (loop++ < NUM_TESTS) {
#else
	while (1) {
#endif
		if (!is_alloc) {
			is_alloc = choose_alloc_or_free(&mix, num_ptrs);
#ifdef CONFIG_VERBOSE
			pr_info("Choosing is_alloc=%d, total=%d\n", is_alloc,
					num_ptrs);
#endif
		}
		if (is_alloc > 0) {
			u32 sz, align;
			choose_size(&mix, &sz, &align);
			ptrs[num_ptrs] = memalign(align, sz);
			BUG_ON(!ptrs[num_ptrs]);
#ifdef CONFIG_VERBOSE
			pr_info("Alloc %p, sz/a %d/%d, total=%d (mix 0x%08x)\n",
				ptrs[num_ptrs], sz, align, num_ptrs + 1, mix);
#endif
			num_ptrs++;
			is_alloc--;
		} else {
			unsigned int idx = choose_idx(&mix, num_ptrs);
#ifdef CONFIG_VERBOSE
			pr_info("Free idx %d (mix 0x%08x)\n", idx, mix);
			pr_info("Free %p\n", ptrs[idx]);
#endif
			free(ptrs[idx]);
			num_ptrs--;
#ifdef CONFIG_VERBOSE
			pr_info("Free'd, total=%d\n", num_ptrs);
#endif
			if (idx < num_ptrs)
				memcpy(&ptrs[idx], &ptrs[idx + 1],
					(num_ptrs - idx) * sizeof(void*));
			is_alloc++;
		}
	}
	while (num_ptrs--) {
#ifdef CONFIG_VERBOSE
		pr_info("Free %p, total=%d\n", ptrs[num_ptrs], num_ptrs);
#endif
		free(ptrs[num_ptrs]);
	}
	pr_info("\"alloc\" complete\n");
	return 0;
}

