/* Copyright (c) 2009-2010 Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Freescale Semiconductor nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "init.h"
#include "misc.h"
#include "timebase.h"
#include "compat.h"
#include "fsl_qman.h" /* for qman_poll() */
#include "fsl_bman.h"

#define my_mfspr(reg) \
({ \
	register_t ret; \
	asm volatile("mfspr %0, %1" : "=r" (ret) : "i" (reg) : "memory"); \
	ret; \
})
static inline uint64_t
my_get_timebase(void)
{
	uint32_t hi, lo, chk;

	/*
	 * To make sure that there is no carry over
	 * between checking of TBU and TBL
	 */
	do {
		hi = my_mfspr(SPR_ATBU);
		lo = my_mfspr(SPR_ATBL);
		chk = my_mfspr(SPR_ATBU);
	} while (unlikely(hi != chk));

	return (uint64_t) hi << 32 | (uint64_t) lo;
}

/*************/
/* constants */
/*************/

#define MAX_CPUS_TO_TEST	7
#define NUM_IGNORE		18
#define NUM_CPUS		7
#define NUM_BUFS_PER_RELEASE	1
#define TEST_BPID(n)		(30 + (n))
#undef MODEL_CHKPT
#undef DO_CHECKSUMS

/***************/
/* global vars */
/***************/

#ifdef DO_CHECKSUMS
/* When releasing, this is our checksum */
static __PERCPU u32 rsum;
/* When acquiring, this is our checksum */
static __PERCPU u32 asum;
#endif

struct waker {
	volatile int kick;
} __cacheline_aligned;
static struct waker waker[NUM_CPUS];

static __PERCPU u64 rel_capture[2] ____cacheline_aligned;
static __PERCPU u64 acq_capture[2] ____cacheline_aligned;
static __PERCPU u32 test_buffers;
static __PERCPU u32 test_start;
static __PERCPU u32 rel_jam, acq_jam;

/* This is our test array (only used by cpu 0) */
struct test {
	int num_cpus;
	int stockpiling;
	u32 num_buffers;
};
static const struct test tests[] = {
	{ 1, 0, 4096},
	{ 2, 0, 4096},
	{ 3, 0, 4096},
	{ 4, 0, 4096},
	{ 5, 0, 4096},
	{ 6, 0, 4096},
	{ 7, 0, 4096},
	{ 1, 1, 4096},
	{ 2, 1, 4096},
	{ 3, 1, 4096},
	{ 4, 1, 4096},
	{ 5, 1, 4096},
	{ 6, 1, 4096},
	{ 7, 1, 4096},
	{ 0, 0, 0}
};

static int elapse(u64 *a)
{
	return (int)(a[1] - a[0]);
}

static int avg(u64 *a)
{
	int e = elapse(a);
	int d = test_start;
	return (e + (d / 2)) / d;
}

/********/
/* test */
/********/

#ifdef DO_CHECKSUMS
#define fill_in(b) \
do { \
	rsum += base; \
	b.hi = 0xfaba; \
	b.lo = base++; \
} while (0)
#define read_in(b) \
do { \
	asum += b.lo; /* should we check buf->hi==BASE_HI? */ \
} while (0)
#else
#define fill_in(b) \
do { \
	b.lo = base++; \
} while (0)
#define read_in(b)
#endif

static void do_releases(struct bman_pool *pool)
{
	struct bm_buffer bufs[NUM_BUFS_PER_RELEASE];
	unsigned int loop = test_buffers;
	u32 base = 0xdeadbeef;
#ifdef DO_CHECKSUMS
	rsum = 0;
#endif
	dcbt_rw(rel_capture);
	while (loop) {
		int err;
		if (loop == test_start)
			rel_capture[0] = my_get_timebase();
		/* Loops (and gcc) seem to generate more cycles than the
		 * interface we're trying to benchmark! Take matters into one's
		 * own hands... */
#if NUM_BUFS_PER_RELEASE > 0
		fill_in(bufs[0]);
#endif
#if NUM_BUFS_PER_RELEASE > 1
		fill_in(bufs[1]);
#endif
#if NUM_BUFS_PER_RELEASE > 2
		fill_in(bufs[2]);
#endif
#if NUM_BUFS_PER_RELEASE > 3
		fill_in(bufs[3]);
#endif
#if NUM_BUFS_PER_RELEASE > 4
		fill_in(bufs[4]);
#endif
#if NUM_BUFS_PER_RELEASE > 5
		fill_in(bufs[5]);
#endif
#if NUM_BUFS_PER_RELEASE > 6
		fill_in(bufs[6]);
#endif
#if NUM_BUFS_PER_RELEASE > 7
		fill_in(bufs[7]);
#endif
retry:
		err = bman_release(pool, &bufs[0], NUM_BUFS_PER_RELEASE, 0);
		if (unlikely(err)) {
			rel_jam++;
			barrier();
			goto retry;
		}
		loop--;
	}
	rel_capture[1] = my_get_timebase();
}

static void do_acquires(struct bman_pool *pool)
{
	struct bm_buffer bufs[NUM_BUFS_PER_RELEASE];
	unsigned int loop = test_buffers;
#ifdef DO_CHECKSUMS
	asum = 0;
#endif
	dcbt_rw(acq_capture);
	while (loop) {
		int err;
		if (loop == test_start)
			acq_capture[0] = my_get_timebase();
retry:
		err = bman_acquire(pool, &bufs[0], NUM_BUFS_PER_RELEASE, 0);
		if (unlikely(err != NUM_BUFS_PER_RELEASE)) {
			BUG_ON(err >= 0);
			acq_jam++;
			barrier();
			goto retry;
		}
#if NUM_BUFS_PER_RELEASE > 0
		read_in(bufs[0]);
#endif
#if NUM_BUFS_PER_RELEASE > 1
		read_in(bufs[1]);
#endif
#if NUM_BUFS_PER_RELEASE > 2
		read_in(bufs[2]);
#endif
#if NUM_BUFS_PER_RELEASE > 3
		read_in(bufs[3]);
#endif
#if NUM_BUFS_PER_RELEASE > 4
		read_in(bufs[4]);
#endif
#if NUM_BUFS_PER_RELEASE > 5
		read_in(bufs[5]);
#endif
#if NUM_BUFS_PER_RELEASE > 6
		read_in(bufs[6]);
#endif
#if NUM_BUFS_PER_RELEASE > 7
		read_in(bufs[7]);
#endif
		loop--;
	}
	acq_capture[1] = my_get_timebase();
}

#ifdef MODEL_CHKPT
static volatile int done_print;
static spinlock_t bringup_lock = SPIN_LOCK_UNLOCKED;
#endif
static unsigned long cpus_ready = 0;

/* At the beginning of each test, the cpus all call this routine. The secondary
 * cores signal their readyness for the next test by incrementing "ready", the
 * primary core polls until "ready" hits the desired total. (Ie. except for
 * one-off writes from secondary cores, "ready" stays cache-local to cpu 0.)
 * After incrementing "ready", the secondary cores wait for the go-ahead by
 * polling on the percpu "waker" signal, which cpu 0 sets once everyone is
 * ready. (Ie. except for one-off writes from cpu 0, the "waker" values stay
 * cache-local to their respective cpus.)
 *
 * The idea behind all this is to avoid bursts of coherency noise when tests are
 * starting up or finishing, as these would contaminate the results. This
 * signalling involves some coherency action, but only for the one-off writes,
 * the important thing is that the polling should stay cache-local. */
static void sync_cpus(int whoami)
{
	unsigned int dumbloop;
	atomic_inc(&cpus_ready);
	if (whoami) {
		/* secondary cpus wait for the "go" */
		while (!waker[whoami].kick)
			;
		waker[whoami].kick = 0;
	} else {
		/* cpu 0 waits for the others to be ready */
		while (atomic_read(&cpus_ready) < NUM_CPUS)
			;
		atomic_set(&cpus_ready, 0);
#ifndef MODEL_CHKPT
		dumbloop = 0;
		while (dumbloop < 0x100000)
			dumbloop++;
#endif
		dumbloop = 1;
		while (dumbloop < NUM_CPUS)
			waker[dumbloop++].kick = 1;
	}
}

int main(__always_unused int argc, __always_unused char *argv[])
{
	struct bman_pool_params params;
	int whoami = get_lwe_id();
	struct bman_pool *pool = NULL; /* unnecessary, but gcc doesn't see */
	const struct test *test = &tests[0];

#ifdef MODEL_CHKPT
	/* Do a dance so that we can checkpoint when we see "blastman starting",
	 * know that no cpu has yet started testing, and that post-checkpoint we
	 * do no more console activity on any cpu. */
	spin_lock(&bringup_lock);
	if (!done_print) {
		pr_info("%d: blastman starting\n", get_lwe_id());
		done_print = 1;
	}
	spin_unlock(&bringup_lock);
#endif

	while (test->num_cpus) {
		int doIrun;
		if (MAX_CPUS_TO_TEST < test->num_cpus)
			goto next_test;
		doIrun = (whoami < test->num_cpus) ? 1 : 0;

		test_buffers = test->num_buffers;
		test_start = test->num_buffers - NUM_IGNORE;
		rel_jam = acq_jam = 0;

		if (doIrun) {
			params.bpid = TEST_BPID(whoami);
			params.flags = 0;
			if (test->stockpiling)
				params.flags |= BMAN_POOL_FLAG_STOCKPILE;
			pool = bman_new_pool(&params);
			if (!pool)
				panic("bman_new_pool() failed\n");
		}

		sync_cpus(whoami);

		if (doIrun) {
			do_releases(pool);
			do_acquires(pool);
		}

		sync_cpus(whoami);

		if (doIrun) {
			bman_free_pool(pool);
			pr_info("test {%d,%c,%d} rel %d, acq %d, jam (%d:%d)\n",
				test->num_cpus, test->stockpiling ? 'Y' : 'x',
				test->num_buffers, avg(rel_capture),
				avg(acq_capture), rel_jam, acq_jam);
		}
next_test:
		test++;
	}

	return 0;
}

