/* Copyright (c) 2010 Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Freescale Semiconductor nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "init.h"
#include "misc.h"
#include "compat.h"
#include "fsl_qman.h"

/*************/
/* constants */
/*************/

/*************************************/
/* Predeclarations (eg. for fq_base) */
/*************************************/

static enum qman_cb_dqrr_result cb_dqrr(struct qman_portal *,
					struct qman_fq *,
					const struct qm_dqrr_entry *);
static void cb_ern(struct qman_portal *, struct qman_fq *,
			const struct qm_mr_entry *);
static void cb_dc_ern(struct qman_portal *, struct qman_fq *,
			const struct qm_mr_entry *);
static void cb_fqs(struct qman_portal *, struct qman_fq *,
			const struct qm_mr_entry *);
static void cgr_cb(struct qman_portal *, struct qman_cgr *, int);

/***************/
/* global vars */
/***************/

static __PERCPU struct qm_fd fd;
static __PERCPU struct qm_fd fd_ern;
static __PERCPU struct qm_fd fd_dq;
static __PERCPU struct qm_fd fd_dq_ern;
static __PERCPU struct qman_fq fq_base = {
	.cb = {
		.dqrr = cb_dqrr,
		.ern = cb_ern,
		.dc_ern = cb_dc_ern,
		.fqs = cb_fqs
	}
};
static __PERCPU int retire_complete;
static __PERCPU int sdqcr_complete;
static __PERCPU struct qman_cgr cgr;
static __PERCPU int ern_occured;
static __PERCPU int congestion_state;

/**********************/
/* internal functions */
/**********************/

/* Helpers for initialising and "incrementing" a frame descriptor */
static void fd_init(struct qm_fd *__fd)
{
	__fd->addr_hi = 0xba;		/* high 16-bits */
	__fd->addr_lo = 0xdeadbeef;	/* low 32-bits */
	__fd->format = qm_fd_contig_big;
	__fd->length29 = 0x00000009;
	__fd->cmd = 0xfeedf00d;
}

static void fd_inc(struct qm_fd *__fd)
{
	__fd->addr_lo++;
	__fd->addr_hi--;
	__fd->cmd++;
}

/* The only part of the 'fd' we can't memcmp() is the ppid */
static int fd_cmp(const struct qm_fd *a, const struct qm_fd *b)
{
	int r = a->addr_hi - b->addr_hi;
	if (!r)
		r = a->addr_lo - b->addr_lo;
	if (!r)
		r = a->format - b->format;
	if (!r)
		r = a->cmd - b->cmd;
	return r;
}

/********/
/* test */
/********/

static void do_enqueues(struct qman_fq *fq, u32 length, u32 ern)
{
	struct qm_fd *lfd = &fd;

	if (ern)
		lfd = &fd_ern;

	lfd->length29 = length;
	pr_info("Enqueue: %08x\n", lfd->length29);
	if (qman_enqueue(fq, lfd, QM_EQCR_VERB_COLOUR_RED))
		panic("qman_enqueue() failed\n");
	fd_inc(lfd);
}

int main(int argc __UNUSED, char *argv[] __UNUSED)
{
	struct qm_mcc_initfq initfq;
	int whoami = get_lwe_id();
	int res;
	unsigned int flags;
	struct qman_fq *fq = &fq_base;
	struct qm_mcc_initcgr cgr_state;

	pr_info("cgr_test_high starting\n");

	fd_init(&fd);
	fd_init(&fd_dq);
	fd_init(&fd_dq_ern);
	fd_init(&fd_ern);

	/* create a CGR object */
	/* initialise for notification and setup to tail drop */
	cgr.cb = cgr_cb; /* set callback */
	cgr.cgrid = whoami;
	memset(&cgr_state, 0x0, sizeof(struct qm_mcc_initcgr));
	cgr_state.we_mask = QM_CGR_WE_CSCN_EN | QM_CGR_WE_CS_THRES |
			QM_CGR_WE_CSTD_EN | QM_CGR_WE_MODE;
	cgr_state.cgr.mode = 0;
	cgr_state.cgr.cscn_en = 1; /* turn on congestion notifications */
	cgr_state.cgr.cstd_en = 1;
	cgr_state.cgr.cs_thres.TA = 8;
	cgr_state.cgr.cs_thres.Tn = 0;
	pr_info("Creating CGR 0x%d\n", cgr.cgrid);
	res = qman_create_cgr(&cgr, QMAN_CGR_FLAG_USE_INIT, &cgr_state);
	if (res)
		panic("qman_create_cgr failed\n");
	/* Initialise (parked) FQ */
	if (qman_create_fq(0, QMAN_FQ_FLAG_DYNAMIC_FQID, fq))
		panic("qman_create_fq() failed\n");
	/* enable congestion via cgr */
	memset(&initfq, 0, sizeof(initfq));
	initfq.we_mask = QM_INITFQ_WE_FQCTRL | QM_INITFQ_WE_CGID;
	initfq.fqd.fq_ctrl = QM_FQCTRL_CGE;
	initfq.fqd.cgid = (u8)whoami;
	if (qman_init_fq(fq, QMAN_INITFQ_FLAG_LOCAL, &initfq))
		panic("qman_init_fq() failed\n");
	/* enqueue a frame which will cause congestion
	 * this first frame will not be rejected, further one will */
	pr_info("Wait for congestion...\n");
	do_enqueues(fq, 9, 0);
	wait_event(nothing, congestion_state);
	pr_info("Congestion entered\n");
	/* this one should be rejected */
	pr_info("Enqueue frame which should cause rejection\n");
	do_enqueues(fq, 1, 1);
	wait_event(nothing, ern_occured);
	ern_occured = 0;
	pr_info("ERN occured\n");
	pr_info("scheduled dequeue (till-empty)\n");
	if (qman_schedule_fq(fq))
		panic("qman_schedule_fq() failed\n");
	wait_event(nothing, sdqcr_complete);
	sdqcr_complete = 0;
	pr_info("wait for exit from congestion\n");
	wait_event(nothing, !congestion_state);
	/* Retire and OOS the FQ */
	res = qman_retire_fq(fq, &flags);
	if (res < 0) {
		pr_err("qman_retire_fq() failed 0x%x\n", res);
		panic("qman_retire_fq() failed\n");
	}
	wait_event(nothing, retire_complete);
	retire_complete = 0;

	if (flags & QMAN_FQ_STATE_BLOCKOOS)
		panic("leaking frames\n");
	res = qman_oos_fq(fq);
	if (res) {
		pr_err("qman_oos_rq() failed 0x%x\n", res);
		panic("qman_oos_fq() failed\n");
	}
	qman_destroy_fq(fq, 0);
	pr_info("Delete CGR\n");
	res = qman_delete_cgr(&cgr);
	if (res)
		panic("Error deleting cgr\n");

	/* Test WRED */
	pr_info("cgr WRED starting\n");
	fd_init(&fd);
	fd_init(&fd_dq);
	fd_init(&fd_dq_ern);
	fd_init(&fd_ern);

	/* create a CGR object */
	/* initialise for notification and setup to tail drop */
	cgr.cb = cgr_cb; /* set callback */
	cgr.cgrid = whoami;
	memset(&cgr_state, 0x0, sizeof(struct qm_mcc_initcgr));
	cgr_state.we_mask = QM_CGR_WE_WR_PARM_R | QM_CGR_WE_WR_EN_R |
	                        QM_CGR_WE_MODE;
	/* MinTH = 7483.51 */
	/* MaxTH = 15872 */
	cgr_state.cgr.mode = 0;
	cgr_state.cgr.wr_parm_r.MA = 62;
	cgr_state.cgr.wr_parm_r.Mn = 8;
	cgr_state.cgr.wr_parm_r.SA = 125;
	cgr_state.cgr.wr_parm_r.Sn = 13;
	cgr_state.cgr.wr_parm_r.Pn = 31;
	cgr_state.cgr.wr_en_r = 1;

	pr_info("Creating CGR 0x%d\n", cgr.cgrid);
	res = qman_create_cgr(&cgr, QMAN_CGR_FLAG_USE_INIT, &cgr_state);
	if (res)
		panic("qman_create_cgr failed\n");
	/* Initialise (parked) FQ */
	if (qman_create_fq(0, QMAN_FQ_FLAG_DYNAMIC_FQID, fq))
		panic("qman_create_fq() failed\n");
	/* enable congestion via cgr */
	memset(&initfq, 0, sizeof(initfq));
	initfq.we_mask = QM_INITFQ_WE_FQCTRL | QM_INITFQ_WE_CGID;
	initfq.fqd.fq_ctrl = QM_FQCTRL_CGE;
	initfq.fqd.cgid = (u8)whoami;
	if (qman_init_fq(fq, QMAN_INITFQ_FLAG_LOCAL, &initfq))
		panic("qman_init_fq() failed\n");
	/* enqueue a frame which will cause congestion
	 * this first frame will not be rejected, further one will */
	do_enqueues(fq, 16000, 0);
	/* this one should be rejected */
	pr_info("Enqueue frame which should cause rejection\n");
	do_enqueues(fq, 16000, 1);
	wait_event(nothing, ern_occured);
	pr_info("ERN occured\n");
	pr_info("scheduled dequeue (till-empty)\n");
	if (qman_schedule_fq(fq))
		panic("qman_schedule_fq() failed\n");
	wait_event(nothing, sdqcr_complete);
	pr_info("sdqcr_complete set\n");
	/* Retire and OOS the FQ */
	res = qman_retire_fq(fq, &flags);
	if (res < 0) {
		pr_err("qman_retire_fq() failed 0x%x\n", res);
		panic("qman_retire_fq() failed\n");
	}
	wait_event(nothing, retire_complete);
	retire_complete = 0;

	if (flags & QMAN_FQ_STATE_BLOCKOOS)
	        panic("leaking frames\n");
	res = qman_oos_fq(fq);
	if (res) {
		pr_err("qman_oos_rq() failed 0x%x\n", res);
		panic("qman_oos_fq() failed\n");
	}
	qman_destroy_fq(fq, 0);

        pr_info("Delete CGR\n");
        res = qman_delete_cgr(&cgr);
        if (res)
                panic("Error deleting cgr\n");


	pr_info("cgr test finished\n");
	return 0;
}

static enum qman_cb_dqrr_result cb_dqrr(struct qman_portal *p __UNUSED,
					struct qman_fq *fq __UNUSED,
					const struct qm_dqrr_entry *dq)
{
	pr_info("Dequeue: length %08x\n", dq->fd.length29);
	if (fd_cmp(&fd_dq, &dq->fd)) {
		pr_err("BADNESS: dequeued frame doesn't match;\n");
		pr_err("Got:\n");
		hexdump(&dq->fd, sizeof(dq->fd));
		pr_err("Expected:\n");
		hexdump(&fd_dq, sizeof(fd_dq));
		BUG();
	}
	fd_inc(&fd_dq);
	if (!(dq->stat & QM_DQRR_STAT_UNSCHEDULED) && !fd_cmp(&fd_dq, &fd)) {
		pr_info("Setting sdqcr_complete\n");
		sdqcr_complete = 1;
	}
	return qman_cb_dqrr_consume;
}

static void cb_ern(struct qman_portal *p __UNUSED, struct qman_fq *fq __UNUSED,
			const struct qm_mr_entry *msg)
{
	pr_info("cb_ern: rc= 0x%x, fd.length29 %08x\n",
		(u32)msg->ern.rc, msg->ern.fd.length29);
	if (fd_cmp(&fd_dq_ern, &msg->ern.fd)) {
		pr_err("BADNESS: dequeued frame doesn't match;\n");
		pr_err("Got:\n");
		hexdump(&msg->ern.fd, sizeof(msg->ern.fd));
		pr_err("Expected:\n");
		hexdump(&fd_dq_ern, sizeof(fd_dq_ern));
		BUG();
	}
	fd_inc(&fd_dq_ern);
	ern_occured = 1;
}

static void cb_dc_ern(struct qman_portal *p __UNUSED,
			struct qman_fq *fq __UNUSED,
			const struct qm_mr_entry *msg __UNUSED)
{
	panic("cb_dc_ern() unimplemented");
}

static void cb_fqs(struct qman_portal *p __UNUSED, struct qman_fq *fq __UNUSED,
			const struct qm_mr_entry *msg)
{
	u8 verb = (msg->verb & QM_MR_VERB_TYPE_MASK);
	if ((verb != QM_MR_VERB_FQRN) && (verb != QM_MR_VERB_FQRNI))
		panic("unexpected FQS message");
	pr_info("Retirement message received\n");
	retire_complete = 1;
}

static void cgr_cb(struct qman_portal *qm __UNUSED,
		struct qman_cgr *cgr, int congested)
{
	pr_info("CGR CB (0x%x) portal %d, congestion is 0x%x\n",
		cgr->cgrid, cgr->chan - qm_channel_swportal0,
		congested);
	congestion_state = congested;
}
