/* Copyright (c) 2008-2010 Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Freescale Semiconductor nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "init.h"
#include "misc.h"
#include "compat.h"
#include "fsl_pme.h"

/* Define this to get success-path hexdumps */
#undef CONFIG_VERBOSE

MODULE_AUTHOR("Geoff Thorpe");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("FSL PME2 (p4080) high-level self-test");

/* Default Flow Context State */
static u8 __fl_ctx_exp[] = {
	0x88, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xe0,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00
};

struct ctrl_token {
	struct pme_ctx_ctrl_token token;
	int done;
	struct qm_fd result_fd;
};

static void ctrl_cb(__always_unused struct pme_ctx *ctx, const struct qm_fd *fd,
				struct pme_ctx_ctrl_token *token)
{
	struct ctrl_token *t = (void *)token;
	memcpy(&t->result_fd, fd, sizeof(*fd));
	t->done = 1;
}

int main(__always_unused int argc, __always_unused char *argv[])
{
	struct pme_flow flow;
	struct qm_fqd_stashing stashing;
	struct ctrl_token t = {
		.token = {
			.cb = ctrl_cb
		}
	};
	struct pme_ctx *ctx;
	int ret;
	enum pme_status status;

	pr_info("PME2: high-level test starting\n");
	pme_sw_flow_init(&flow);
	ctx = memalign(sizeof(struct pme_ctx), sizeof(struct pme_ctx));
        BUG_ON(!ctx);

	ret = pme_ctx_init(ctx, PME_CTX_FLAG_LOCAL, 0, 4, 4, 0, NULL);
	BUG_ON(ret);
	/* enable the context */
	pme_ctx_enable(ctx);
	/* update flow context with defaults from pme_sw_flow_init() */
	t.done = 0;
	ret = pme_ctx_ctrl_update_flow(ctx, PME_CMD_FCW_ALL, &flow,
					&t.token);
	BUG_ON(ret);
	wait_event(NULL, t.done == 1);
	/* Check for error */
	status = pme_fd_res_status(&t.result_fd);
	if (status) {
		pr_info("Update Flow failed %d\r\n", status);
		BUG_ON(1);
	}
#ifdef CONFIG_VERBOSE
	pr_info("Update Flow Passed\r\n");
#endif

	/* read back flow settings */
	t.done = 0;
	ret = pme_ctx_ctrl_read_flow(ctx, PME_CMD_FCW_ALL, &flow, &t.token);
	BUG_ON(ret);
	wait_event(NULL, t.done == 1);
	/* Check for error */
	status = pme_fd_res_status(&t.result_fd);
	if (status) {
		pr_info("Read Flow failed %d\r\n", status);
		BUG_ON(1);
	}
#ifdef CONFIG_VERBOSE
	pr_info("Read Flow Passed\r\n");
#endif

	if (memcmp(&flow, __fl_ctx_exp, sizeof(flow))!= 0) {
		pr_info("Default Flow Context Read FAIL\n");
		pr_info("Expected:\n");
		hexdump(__fl_ctx_exp, sizeof(__fl_ctx_exp));
		pr_info("Received:\n");
		hexdump(&flow, sizeof(flow));
		BUG_ON(1);
	} else
		pr_info("Default Flow Context Read OK\n");
	/* start a NOP */
	t.done = 0;
	ret = pme_ctx_ctrl_nop(ctx, 0, &t.token);
	BUG_ON(ret);
	wait_event(NULL, t.done == 1);
	/* Check for error */
	status = pme_fd_res_status(&t.result_fd);
	if (status) {
		pr_info("NOP failed %d\r\n", status);
		BUG_ON(1);
	}
#ifdef CONFIG_VERBOSE
	pr_info("NOP Passed\r\n");
#endif
	/* start an update to add residue to the context */
	flow.ren = 1;
	t.done = 0;
	ret = pme_ctx_ctrl_update_flow(ctx, PME_CMD_FCW_RES, &flow, &t.token);
	BUG_ON(ret);
	wait_event(NULL, t.done == 1);
	/* Check for error */
	status = pme_fd_res_status(&t.result_fd);
	if (status) {
		pr_info("Update Flow failed %d\r\n", status);
		BUG_ON(1);
	}
#ifdef CONFIG_VERBOSE
	pr_info("Update Flow Passed\r\n");
#endif
	/* start a disable */
	t.done = 0;
	ret = pme_ctx_disable(ctx, 0, &t.token);
	if (ret == 1) {
		pr_info("Waiting for disable to complete\n");
		wait_event(NULL, t.done == 1);
	} else if (ret)
		BUG();
	/* do some reconfiguration */
	ret = pme_ctx_reconfigure_tx(ctx, 63, 7);
	BUG_ON(ret);
	stashing.exclusive = 0;
	stashing.annotation_cl = 0;
	stashing.data_cl = 0;
	stashing.context_cl = 0;
	ret = pme_ctx_reconfigure_rx(ctx, 7, 0, &stashing);
	BUG_ON(ret);
	/* reenable */
	ret = pme_ctx_enable(ctx);
	BUG_ON(ret);
	/* read back flow settings */
	t.done = 0;
	ret = pme_ctx_ctrl_read_flow(ctx, PME_CMD_FCW_RES, &flow, &t.token);
	BUG_ON(ret);
	wait_event(NULL, t.done == 1);
	/* Check for error */
	status = pme_fd_res_status(&t.result_fd);
	if (status) {
		pr_info("Read Flow failed %d\r\n", status);
		BUG_ON(1);
	}
#ifdef CONFIG_VERBOSE
	pr_info("Read Flow Passed\r\n");
#endif
	/* NOP */
	t.done = 0;
	ret = pme_ctx_ctrl_nop(ctx, 0, &t.token);
	BUG_ON(ret);
	wait_event(NULL, t.done == 1);
	/* Check for error */
	status = pme_fd_res_status(&t.result_fd);
	if (status) {
		pr_info("NOP failed %d\r\n", status);
		BUG_ON(1);
	}
#ifdef CONFIG_VERBOSE
	pr_info("NOP Passed\r\n");
#endif
	/* Disable, and done */
	t.done = 0;
	ret = pme_ctx_disable(ctx, 0, &t.token);
	if (ret == 1) {
		pr_info("Waiting for disable to complete\n");
		wait_event(NULL, t.done == 1);
	} else if (ret)
		BUG();
	pme_ctx_finish(ctx);
	free(ctx);
	pr_info("PME2: high-level test done\n");
	return 0;
}

