/* Copyright (c) 2008-2010 Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Freescale Semiconductor nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "init.h"
#include "misc.h"
#include "compat.h"
#include "fsl_qman.h"
#include "fsl_bman.h" /* for bman_poll() */

/*************/
/* constants */
/*************/

#define CGR_ID		27
#define POOL_ID		12
#define FQ_FLAGS	QMAN_FQ_FLAG_DYNAMIC_FQID
#define NUM_ENQUEUES	10
#define NUM_PARTIAL	4
#define PORTAL_SDQCR	(QM_SDQCR_SOURCE_CHANNELS | \
			QM_SDQCR_TYPE_PRIO_QOS | \
			QM_SDQCR_TOKEN_SET(0x98) | \
			QM_SDQCR_CHANNELS_DEDICATED | \
			QM_SDQCR_CHANNELS_POOL(POOL_ID))
#define PORTAL_OPAQUE	(void *)0xf00dbeef
#define PORTAL_FLAGS	QMAN_PORTAL_FLAG_IRQ

/*************************************/
/* Predeclarations (eg. for fq_base) */
/*************************************/

static enum qman_cb_dqrr_result cb_dqrr(struct qman_portal *,
					struct qman_fq *,
					const struct qm_dqrr_entry *);
static void cb_ern(struct qman_portal *, struct qman_fq *,
			const struct qm_mr_entry *);
static void cb_dc_ern(struct qman_portal *, struct qman_fq *,
			const struct qm_mr_entry *);
static void cb_fqs(struct qman_portal *, struct qman_fq *,
			const struct qm_mr_entry *);

/***************/
/* global vars */
/***************/

static __PERCPU struct qm_fd fd;
static __PERCPU struct qm_fd fd_dq;
static __PERCPU struct qman_fq fq_base = {
	.cb = {
		.dqrr = cb_dqrr,
		.ern = cb_ern,
		.dc_ern = cb_dc_ern,
		.fqs = cb_fqs
	}
};
static __PERCPU int retire_complete;
static __PERCPU int sdqcr_complete;

/**********************/
/* internal functions */
/**********************/

/* Helpers for initialising and "incrementing" a frame descriptor */
static void fd_init(struct qm_fd *__fd)
{
	__fd->addr_hi = 0xba;		/* high 16-bits */
	__fd->addr_lo = 0xdeadbeef;	/* low 32-bits */
	__fd->format = qm_fd_contig_big;
	__fd->length29 = 0x0000ffff;
	__fd->cmd = 0xfeedf00d;
}

static void fd_inc(struct qm_fd *__fd)
{
	__fd->addr_lo++;
	__fd->addr_hi--;
	__fd->length29--;
	__fd->cmd++;
}

/* The only part of the 'fd' we can't memcmp() is the ppid */
static int fd_cmp(const struct qm_fd *a, const struct qm_fd *b)
{
	int r = a->addr_hi - b->addr_hi;
	if (!r)
		r = a->addr_lo - b->addr_lo;
	if (!r)
		r = a->format - b->format;
	if (!r)
		r = a->opaque - b->opaque;
	if (!r)
		r = a->cmd - b->cmd;
	return r;
}

/********/
/* test */
/********/

static void delay_1000(void)
{
	u64 now, then = mfatb();
	do {
		qman_poll();
		now = mfatb();
	} while (now < (then + 1000));
}

static void do_enqueues(struct qman_fq *fq)
{
	unsigned int loop;
	int retries;
	for (loop = 0; loop < NUM_ENQUEUES; loop++) {
		retries = 5;
		pr_info("Enqueue: %08x\n", fd.addr_lo);
		while (qman_enqueue(fq, &fd, 0)) {
			if (!(retries--))
				panic("qman_enqueue() failed");
			delay_1000();
		}
		fd_inc(&fd);
	}
	/* Enqueues are posted, so we need to wait till Qman has consumed them
	 * before dequeuing, otherwise the test assumptions may be broken. */
	retries = 5;
	while (!qman_eqcr_is_empty()) {
		if (!(retries--))
			panic("qman_enqueue() not syncing");
		delay_1000();
	}
}

static void do_vdqcr(struct qman_fq *fq, u32 vdqcr)
{
	u32 flags;
	int retries = 5;
	while (qman_volatile_dequeue(fq, 0, vdqcr)) {
		if (!(retries--))
			panic("qman_volatile_dequeue() failed");
		delay_1000();
	}
	retries = 5;
	while (1) {
		qman_fq_state(fq, NULL, &flags);
		if (!(flags & QMAN_FQ_STATE_VDQCR))
			/* complete */
			return;
		if (!(retries--))
			panic("qman_volatile_dequeue() not syncing");
		delay_1000();
	}
}

int main(__always_unused int argc, __always_unused char *argv[])
{
	struct qman_cgrs cgrs;
	int res;
	unsigned int flags;
	struct qman_fq *fq = &fq_base;

	pr_info("qman_test_high starting\n");
	fd_init(&fd);
	fd_init(&fd_dq);
	qman_cgrs_init(&cgrs);
	qman_cgrs_set(&cgrs, CGR_ID);

	pr_info("high-level test, start ccmode\n");

	/* Initialise (parked) FQ */
	if (qman_create_fq(0, FQ_FLAGS, fq))
		panic("qman_create_fq() failed\n");
	if (qman_init_fq(fq, QMAN_INITFQ_FLAG_LOCAL, NULL))
		panic("qman_init_fq() failed\n");

#ifndef CONFIG_FSL_QMAN_TEST_HIGH_BUG_VDQCR
	/* Do enqueues + VDQCR, twice. (Parked FQ) */
	do_enqueues(fq);
	pr_info("VDQCR (till-empty);\n");
	do_vdqcr(fq, QM_VDQCR_NUMFRAMES_TILLEMPTY);

	do_enqueues(fq);
	pr_info("VDQCR (%d of %d);\n", NUM_PARTIAL, NUM_ENQUEUES);
	do_vdqcr(fq, QM_VDQCR_NUMFRAMES_SET(NUM_PARTIAL));
	pr_info("VDQCR (%d of %d);\n", NUM_ENQUEUES - NUM_PARTIAL,
					NUM_ENQUEUES);
	do_vdqcr(fq, QM_VDQCR_NUMFRAMES_SET(NUM_ENQUEUES - NUM_PARTIAL));
#endif

	do_enqueues(fq);
	pr_info("scheduled dequeue (till-empty)\n");
	if (qman_schedule_fq(fq))
		panic("qman_schedule_fq() failed\n");
	wait_event(nothing, sdqcr_complete);

	/* Retire and OOS the FQ */
	res = qman_retire_fq(fq, &flags);
	if (res < 0)
		panic("qman_retire_fq() failed\n");
	wait_event(nothing, retire_complete);

	if (flags & QMAN_FQ_STATE_BLOCKOOS)
		panic("leaking frames\n");
	if (qman_oos_fq(fq))
		panic("qman_oos_fq() failed\n");
	qman_destroy_fq(fq, 0);
	pr_info("qman_test_high finished\n");
	return 0;
}

static enum qman_cb_dqrr_result cb_dqrr(__always_unused struct qman_portal *p,
					__always_unused struct qman_fq *fq,
					const struct qm_dqrr_entry *dq)
{
	pr_info("Dequeue: %08x (%s)\n", dq->fd.addr_lo,
		(dq->stat & QM_DQRR_STAT_UNSCHEDULED) ? "VDQCR" : "SDQCR");
	if (fd_cmp(&fd_dq, &dq->fd)) {
		pr_err("BADNESS: dequeued frame doesn't match;\n");
		pr_err("Got:\n");
		hexdump(&dq->fd, sizeof(dq->fd));
		pr_err("Expected:\n");
		hexdump(&fd_dq, sizeof(fd_dq));
		BUG();
	}
	fd_inc(&fd_dq);
	if (!(dq->stat & QM_DQRR_STAT_UNSCHEDULED) && !fd_cmp(&fd_dq, &fd))
		sdqcr_complete = 1;
	return qman_cb_dqrr_consume;
}

static void cb_ern(__always_unused struct qman_portal *p,
		__always_unused struct qman_fq *fq,
		__always_unused const struct qm_mr_entry *msg)
{
	panic("cb_ern() unimplemented");
}

static void cb_dc_ern(__always_unused struct qman_portal *p,
		__always_unused struct qman_fq *fq,
		__always_unused const struct qm_mr_entry *msg)
{
	panic("cb_dc_ern() unimplemented");
}

static void cb_fqs(__always_unused struct qman_portal *p,
		__always_unused struct qman_fq *fq,
		const struct qm_mr_entry *msg)
{
	u8 verb = (msg->verb & QM_MR_VERB_TYPE_MASK);
	if ((verb != QM_MR_VERB_FQRN) && (verb != QM_MR_VERB_FQRNI))
		panic("unexpected FQS message");
	pr_info("Retirement message received\n");
	retire_complete = 1;
}

