/* Copyright (c) 2010 Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Freescale Semiconductor nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "init.h"
#include "misc.h"
#include "compat.h"
#include "fsl_qman.h"

/*************/
/* constants */
/*************/

#define NUM_ENQUEUES	10
#define NUM_PARTIAL	4

/*************************************/
/* Predeclarations (eg. for fq_base) */
/*************************************/

static enum qman_cb_dqrr_result cb_dqrr(struct qman_portal *,
					struct qman_fq *,
					const struct qm_dqrr_entry *);
static void cb_ern(struct qman_portal *, struct qman_fq *,
			const struct qm_mr_entry *);
static void cb_dc_ern(struct qman_portal *, struct qman_fq *,
			const struct qm_mr_entry *);
static void cb_fqs(struct qman_portal *, struct qman_fq *,
			const struct qm_mr_entry *);

/***************/
/* global vars */
/***************/

static __PERCPU struct qm_fd fd;
static __PERCPU struct qm_fd fd_dq;
static __PERCPU struct qman_fq fq_base = {
	.cb = {
		.dqrr = cb_dqrr,
		.ern = cb_ern,
		.dc_ern = cb_dc_ern,
		.fqs = cb_fqs
	}
};

/**********************/
/* internal functions */
/**********************/

/* Helpers for initialising and "incrementing" a frame descriptor */
static void fd_init(struct qm_fd *__fd)
{
	__fd->addr_hi = 0xba;		/* high 16-bits */
	__fd->addr_lo = 0xdeadbeef;	/* low 32-bits */
	__fd->format = qm_fd_contig_big;
	__fd->length29 = 0x0000ffff;
	__fd->cmd = 0xfeedf00d;
}

static void fd_inc(struct qm_fd *__fd)
{
	__fd->addr_lo++;
	__fd->addr_hi--;
	__fd->length29--;
	__fd->cmd++;
}

/* The only part of the 'fd' we can't memcmp() is the ppid */
static int fd_cmp(const struct qm_fd *a, const struct qm_fd *b)
{
	int r = a->addr_hi - b->addr_hi;
	if (!r)
		r = a->addr_lo - b->addr_lo;
	if (!r)
		r = a->format - b->format;
	if (!r)
		r = a->opaque - b->opaque;
	if (!r)
		r = a->cmd - b->cmd;
	return r;
}

/********/
/* test */
/********/

static void delay_1000(void)
{
	u64 now, then = mfatb();
	do {
		qman_poll();
		now = mfatb();
	} while (now < (then + 1000));
}

static void do_enqueues(struct qman_fq *fq)
{
	unsigned int loop;
	int retries;
	for (loop = 0; loop < NUM_ENQUEUES; loop++) {
		retries = 5;
		pr_info("Enqueue: %08x\n", fd.addr_lo);
		while (qman_enqueue(fq, &fd, 0)) {
			if (!(retries--))
				panic("qman_enqueue() failed");
			delay_1000();
		}
		fd_inc(&fd);
	}
	/* Enqueues are posted, so we need to wait till Qman has consumed them
	 * before dequeuing, otherwise the test assumptions may be broken. */
	retries = 5;
	while (!qman_eqcr_is_empty()) {
		if (!(retries--))
			panic("qman_enqueue() not syncing");
		delay_1000();
	}
}

static void do_vdqcr(struct qman_fq *fq, u32 vdqcr)
{
	u32 flags;
	int retries = 5;
	while (qman_volatile_dequeue(fq, 0, vdqcr)) {
		if (!(retries--))
			panic("qman_volatile_dequeue() failed");
		delay_1000();
	}
	retries = 5;
	while (1) {
		qman_fq_state(fq, NULL, &flags);
		if (!(flags & QMAN_FQ_STATE_VDQCR))
			/* complete */
			return;
		if (!(retries--))
			panic("qman_volatile_dequeue() not syncing");
		delay_1000();
	}
}

int main(__always_unused int argc, __always_unused char *argv[])
{
	u32 flags;
	enum qman_fq_state state;
	struct qman_fq *fq = &fq_base;
	struct qm_mcc_initfq opts;

	pr_info("vdqcr starting\n");
	fd_init(&fd);
	fd_init(&fd_dq);

	/* Initialise (parked) FQ */
	pr_info("  Initialising FQs...\n");
	if (qman_create_fq(0, QMAN_FQ_FLAG_DYNAMIC_FQID |
				QMAN_FQ_FLAG_TO_DCPORTAL, fq))
		panic("qman_create_fq() failed\n");
	opts.we_mask = QM_INITFQ_WE_DESTWQ | QM_INITFQ_WE_CONTEXTA |
					QM_INITFQ_WE_CONTEXTB;
	/* The "dest" values are just for illustration, they aren't used in any
	 * way because we don't schedule the FQ. */
	opts.fqd.dest.channel = qm_channel_pme;
	opts.fqd.dest.wq = 3;
	opts.fqd.context_a.opaque = 0x87654321abbaf00dULL;
	opts.fqd.context_b = 0xdeadbeef;
	if (qman_init_fq(fq, 0, &opts))
		panic("qman_init_fq() failed\n");

	pr_info("  Enqueuing %d frames...\n", NUM_ENQUEUES);
	do_enqueues(fq);

	pr_info("  Volatile dequeue %d frames...\n", NUM_PARTIAL);
	do_vdqcr(fq, QM_VDQCR_EXACT | QM_VDQCR_NUMFRAMES_SET(NUM_PARTIAL));

	pr_info("  Retire FQ...\n");
	if (qman_retire_fq(fq, &flags))
		panic("qman_retire_fq() failed");
	if (!(flags & QMAN_FQ_STATE_NE))
		panic("FQ is empty?");
	if (!(flags & QMAN_FQ_STATE_BLOCKOOS))
		panic("FQ can OOS already?");

	pr_info("  Volatile dequeue until empty...\n");
	do_vdqcr(fq, QM_VDQCR_NUMFRAMES_TILLEMPTY);

	pr_info("  Checking state...\n");
	qman_fq_state(fq, &state, &flags);
	if (state != qman_fq_state_retired)
		panic("not in the retired state");
	if (flags & QMAN_FQ_STATE_BLOCKOOS)
		panic("FQ can't OOS");

	pr_info("  OOS FQ...\n");
	if (qman_oos_fq(fq))
		panic("qman_oos_fq() failed\n");
	qman_destroy_fq(fq, 0);
	pr_info("vdqcr finished\n");
	return 0;
}

static enum qman_cb_dqrr_result cb_dqrr(__always_unused struct qman_portal *p,
					__always_unused struct qman_fq *fq,
					const struct qm_dqrr_entry *dq)
{
	pr_info("Dequeue: %08x (%s)\n", dq->fd.addr_lo,
		(dq->stat & QM_DQRR_STAT_UNSCHEDULED) ? "VDQCR" : "SDQCR");
	if (fd_cmp(&fd_dq, &dq->fd)) {
		pr_err("BADNESS: dequeued frame doesn't match;\n");
		pr_err("Got:\n");
		hexdump(&dq->fd, sizeof(dq->fd));
		pr_err("Expected:\n");
		hexdump(&fd_dq, sizeof(fd_dq));
		BUG();
	}
	fd_inc(&fd_dq);
	return qman_cb_dqrr_consume;
}

static void cb_ern(__always_unused struct qman_portal *p,
		__always_unused struct qman_fq *fq,
		__always_unused const struct qm_mr_entry *msg)
{
	panic("cb_ern() unimplemented");
}

static void cb_dc_ern(__always_unused struct qman_portal *p,
		__always_unused struct qman_fq *fq,
		__always_unused const struct qm_mr_entry *msg)
{
	panic("cb_dc_ern() unimplemented");
}

static void cb_fqs(__always_unused struct qman_portal *p,
		__always_unused struct qman_fq *fq,
		const struct qm_mr_entry *msg)
{
	u8 verb = (msg->verb & QM_MR_VERB_TYPE_MASK);
	if (verb != QM_MR_VERB_FQRNI)
		panic("unexpected FQS message");
}

